//     Underscore.js 1.4.4
//     http://underscorejs.org
//     (c) 2009-2013 Jeremy Ashkenas, DocumentCloud Inc.
//     Underscore may be freely distributed under the MIT license.

(function() {

  // Baseline setup
  // --------------

  // Establish the root object, `window` in the browser, or `global` on the server.
  var root = this;

  // Save the previous value of the `_` variable.
  var previousUnderscore = root._;

  // Establish the object that gets returned to break out of a loop iteration.
  var breaker = {};

  // Save bytes in the minified (but not gzipped) version:
  var ArrayProto = Array.prototype, ObjProto = Object.prototype, FuncProto = Function.prototype;

  // Create quick reference variables for speed access to core prototypes.
  var push             = ArrayProto.push,
      slice            = ArrayProto.slice,
      concat           = ArrayProto.concat,
      toString         = ObjProto.toString,
      hasOwnProperty   = ObjProto.hasOwnProperty;

  // All **ECMAScript 5** native function implementations that we hope to use
  // are declared here.
  var
    nativeForEach      = ArrayProto.forEach,
    nativeMap          = ArrayProto.map,
    nativeReduce       = ArrayProto.reduce,
    nativeReduceRight  = ArrayProto.reduceRight,
    nativeFilter       = ArrayProto.filter,
    nativeEvery        = ArrayProto.every,
    nativeSome         = ArrayProto.some,
    nativeIndexOf      = ArrayProto.indexOf,
    nativeLastIndexOf  = ArrayProto.lastIndexOf,
    nativeIsArray      = Array.isArray,
    nativeKeys         = Object.keys,
    nativeBind         = FuncProto.bind;

  // Create a safe reference to the Underscore object for use below.
  var _ = function(obj) {
    if (obj instanceof _) return obj;
    if (!(this instanceof _)) return new _(obj);
    this._wrapped = obj;
  };

  // Export the Underscore object for **Node.js**, with
  // backwards-compatibility for the old `require()` API. If we're in
  // the browser, add `_` as a global object via a string identifier,
  // for Closure Compiler "advanced" mode.
  if (typeof exports !== 'undefined') {
    if (typeof module !== 'undefined' && module.exports) {
      exports = module.exports = _;
    }
    exports._ = _;
  } else {
    root._ = _;
  }

  // Current version.
  _.VERSION = '1.4.4';

  // Collection Functions
  // --------------------

  // The cornerstone, an `each` implementation, aka `forEach`.
  // Handles objects with the built-in `forEach`, arrays, and raw objects.
  // Delegates to **ECMAScript 5**'s native `forEach` if available.
  var each = _.each = _.forEach = function(obj, iterator, context) {
    if (obj == null) return;
    if (nativeForEach && obj.forEach === nativeForEach) {
      obj.forEach(iterator, context);
    } else if (obj.length === +obj.length) {
      for (var i = 0, l = obj.length; i < l; i++) {
        if (iterator.call(context, obj[i], i, obj) === breaker) return;
      }
    } else {
      for (var key in obj) {
        if (_.has(obj, key)) {
          if (iterator.call(context, obj[key], key, obj) === breaker) return;
        }
      }
    }
  };

  // Return the results of applying the iterator to each element.
  // Delegates to **ECMAScript 5**'s native `map` if available.
  _.map = _.collect = function(obj, iterator, context) {
    var results = [];
    if (obj == null) return results;
    if (nativeMap && obj.map === nativeMap) return obj.map(iterator, context);
    each(obj, function(value, index, list) {
      results[results.length] = iterator.call(context, value, index, list);
    });
    return results;
  };

  var reduceError = 'Reduce of empty array with no initial value';

  // **Reduce** builds up a single result from a list of values, aka `inject`,
  // or `foldl`. Delegates to **ECMAScript 5**'s native `reduce` if available.
  _.reduce = _.foldl = _.inject = function(obj, iterator, memo, context) {
    var initial = arguments.length > 2;
    if (obj == null) obj = [];
    if (nativeReduce && obj.reduce === nativeReduce) {
      if (context) iterator = _.bind(iterator, context);
      return initial ? obj.reduce(iterator, memo) : obj.reduce(iterator);
    }
    each(obj, function(value, index, list) {
      if (!initial) {
        memo = value;
        initial = true;
      } else {
        memo = iterator.call(context, memo, value, index, list);
      }
    });
    if (!initial) throw new TypeError(reduceError);
    return memo;
  };

  // The right-associative version of reduce, also known as `foldr`.
  // Delegates to **ECMAScript 5**'s native `reduceRight` if available.
  _.reduceRight = _.foldr = function(obj, iterator, memo, context) {
    var initial = arguments.length > 2;
    if (obj == null) obj = [];
    if (nativeReduceRight && obj.reduceRight === nativeReduceRight) {
      if (context) iterator = _.bind(iterator, context);
      return initial ? obj.reduceRight(iterator, memo) : obj.reduceRight(iterator);
    }
    var length = obj.length;
    if (length !== +length) {
      var keys = _.keys(obj);
      length = keys.length;
    }
    each(obj, function(value, index, list) {
      index = keys ? keys[--length] : --length;
      if (!initial) {
        memo = obj[index];
        initial = true;
      } else {
        memo = iterator.call(context, memo, obj[index], index, list);
      }
    });
    if (!initial) throw new TypeError(reduceError);
    return memo;
  };

  // Return the first value which passes a truth test. Aliased as `detect`.
  _.find = _.detect = function(obj, iterator, context) {
    var result;
    any(obj, function(value, index, list) {
      if (iterator.call(context, value, index, list)) {
        result = value;
        return true;
      }
    });
    return result;
  };

  // Return all the elements that pass a truth test.
  // Delegates to **ECMAScript 5**'s native `filter` if available.
  // Aliased as `select`.
  _.filter = _.select = function(obj, iterator, context) {
    var results = [];
    if (obj == null) return results;
    if (nativeFilter && obj.filter === nativeFilter) return obj.filter(iterator, context);
    each(obj, function(value, index, list) {
      if (iterator.call(context, value, index, list)) results[results.length] = value;
    });
    return results;
  };

  // Return all the elements for which a truth test fails.
  _.reject = function(obj, iterator, context) {
    return _.filter(obj, function(value, index, list) {
      return !iterator.call(context, value, index, list);
    }, context);
  };

  // Determine whether all of the elements match a truth test.
  // Delegates to **ECMAScript 5**'s native `every` if available.
  // Aliased as `all`.
  _.every = _.all = function(obj, iterator, context) {
    iterator || (iterator = _.identity);
    var result = true;
    if (obj == null) return result;
    if (nativeEvery && obj.every === nativeEvery) return obj.every(iterator, context);
    each(obj, function(value, index, list) {
      if (!(result = result && iterator.call(context, value, index, list))) return breaker;
    });
    return !!result;
  };

  // Determine if at least one element in the object matches a truth test.
  // Delegates to **ECMAScript 5**'s native `some` if available.
  // Aliased as `any`.
  var any = _.some = _.any = function(obj, iterator, context) {
    iterator || (iterator = _.identity);
    var result = false;
    if (obj == null) return result;
    if (nativeSome && obj.some === nativeSome) return obj.some(iterator, context);
    each(obj, function(value, index, list) {
      if (result || (result = iterator.call(context, value, index, list))) return breaker;
    });
    return !!result;
  };

  // Determine if the array or object contains a given value (using `===`).
  // Aliased as `include`.
  _.contains = _.include = function(obj, target) {
    if (obj == null) return false;
    if (nativeIndexOf && obj.indexOf === nativeIndexOf) return obj.indexOf(target) != -1;
    return any(obj, function(value) {
      return value === target;
    });
  };

  // Invoke a method (with arguments) on every item in a collection.
  _.invoke = function(obj, method) {
    var args = slice.call(arguments, 2);
    var isFunc = _.isFunction(method);
    return _.map(obj, function(value) {
      return (isFunc ? method : value[method]).apply(value, args);
    });
  };

  // Convenience version of a common use case of `map`: fetching a property.
  _.pluck = function(obj, key) {
    return _.map(obj, function(value){ return value[key]; });
  };

  // Convenience version of a common use case of `filter`: selecting only objects
  // containing specific `key:value` pairs.
  _.where = function(obj, attrs, first) {
    if (_.isEmpty(attrs)) return first ? null : [];
    return _[first ? 'find' : 'filter'](obj, function(value) {
      for (var key in attrs) {
        if (attrs[key] !== value[key]) return false;
      }
      return true;
    });
  };

  // Convenience version of a common use case of `find`: getting the first object
  // containing specific `key:value` pairs.
  _.findWhere = function(obj, attrs) {
    return _.where(obj, attrs, true);
  };

  // Return the maximum element or (element-based computation).
  // Can't optimize arrays of integers longer than 65,535 elements.
  // See: https://bugs.webkit.org/show_bug.cgi?id=80797
  _.max = function(obj, iterator, context) {
    if (!iterator && _.isArray(obj) && obj[0] === +obj[0] && obj.length < 65535) {
      return Math.max.apply(Math, obj);
    }
    if (!iterator && _.isEmpty(obj)) return -Infinity;
    var result = {computed : -Infinity, value: -Infinity};
    each(obj, function(value, index, list) {
      var computed = iterator ? iterator.call(context, value, index, list) : value;
      computed >= result.computed && (result = {value : value, computed : computed});
    });
    return result.value;
  };

  // Return the minimum element (or element-based computation).
  _.min = function(obj, iterator, context) {
    if (!iterator && _.isArray(obj) && obj[0] === +obj[0] && obj.length < 65535) {
      return Math.min.apply(Math, obj);
    }
    if (!iterator && _.isEmpty(obj)) return Infinity;
    var result = {computed : Infinity, value: Infinity};
    each(obj, function(value, index, list) {
      var computed = iterator ? iterator.call(context, value, index, list) : value;
      computed < result.computed && (result = {value : value, computed : computed});
    });
    return result.value;
  };

  // Shuffle an array.
  _.shuffle = function(obj) {
    var rand;
    var index = 0;
    var shuffled = [];
    each(obj, function(value) {
      rand = _.random(index++);
      shuffled[index - 1] = shuffled[rand];
      shuffled[rand] = value;
    });
    return shuffled;
  };

  // An internal function to generate lookup iterators.
  var lookupIterator = function(value) {
    return _.isFunction(value) ? value : function(obj){ return obj[value]; };
  };

  // Sort the object's values by a criterion produced by an iterator.
  _.sortBy = function(obj, value, context) {
    var iterator = lookupIterator(value);
    return _.pluck(_.map(obj, function(value, index, list) {
      return {
        value : value,
        index : index,
        criteria : iterator.call(context, value, index, list)
      };
    }).sort(function(left, right) {
      var a = left.criteria;
      var b = right.criteria;
      if (a !== b) {
        if (a > b || a === void 0) return 1;
        if (a < b || b === void 0) return -1;
      }
      return left.index < right.index ? -1 : 1;
    }), 'value');
  };

  // An internal function used for aggregate "group by" operations.
  var group = function(obj, value, context, behavior) {
    var result = {};
    var iterator = lookupIterator(value || _.identity);
    each(obj, function(value, index) {
      var key = iterator.call(context, value, index, obj);
      behavior(result, key, value);
    });
    return result;
  };

  // Groups the object's values by a criterion. Pass either a string attribute
  // to group by, or a function that returns the criterion.
  _.groupBy = function(obj, value, context) {
    return group(obj, value, context, function(result, key, value) {
      (_.has(result, key) ? result[key] : (result[key] = [])).push(value);
    });
  };

  // Counts instances of an object that group by a certain criterion. Pass
  // either a string attribute to count by, or a function that returns the
  // criterion.
  _.countBy = function(obj, value, context) {
    return group(obj, value, context, function(result, key) {
      if (!_.has(result, key)) result[key] = 0;
      result[key]++;
    });
  };

  // Use a comparator function to figure out the smallest index at which
  // an object should be inserted so as to maintain order. Uses binary search.
  _.sortedIndex = function(array, obj, iterator, context) {
    iterator = iterator == null ? _.identity : lookupIterator(iterator);
    var value = iterator.call(context, obj);
    var low = 0, high = array.length;
    while (low < high) {
      var mid = (low + high) >>> 1;
      iterator.call(context, array[mid]) < value ? low = mid + 1 : high = mid;
    }
    return low;
  };

  // Safely convert anything iterable into a real, live array.
  _.toArray = function(obj) {
    if (!obj) return [];
    if (_.isArray(obj)) return slice.call(obj);
    if (obj.length === +obj.length) return _.map(obj, _.identity);
    return _.values(obj);
  };

  // Return the number of elements in an object.
  _.size = function(obj) {
    if (obj == null) return 0;
    return (obj.length === +obj.length) ? obj.length : _.keys(obj).length;
  };

  // Array Functions
  // ---------------

  // Get the first element of an array. Passing **n** will return the first N
  // values in the array. Aliased as `head` and `take`. The **guard** check
  // allows it to work with `_.map`.
  _.first = _.head = _.take = function(array, n, guard) {
    if (array == null) return void 0;
    return (n != null) && !guard ? slice.call(array, 0, n) : array[0];
  };

  // Returns everything but the last entry of the array. Especially useful on
  // the arguments object. Passing **n** will return all the values in
  // the array, excluding the last N. The **guard** check allows it to work with
  // `_.map`.
  _.initial = function(array, n, guard) {
    return slice.call(array, 0, array.length - ((n == null) || guard ? 1 : n));
  };

  // Get the last element of an array. Passing **n** will return the last N
  // values in the array. The **guard** check allows it to work with `_.map`.
  _.last = function(array, n, guard) {
    if (array == null) return void 0;
    if ((n != null) && !guard) {
      return slice.call(array, Math.max(array.length - n, 0));
    } else {
      return array[array.length - 1];
    }
  };

  // Returns everything but the first entry of the array. Aliased as `tail` and `drop`.
  // Especially useful on the arguments object. Passing an **n** will return
  // the rest N values in the array. The **guard**
  // check allows it to work with `_.map`.
  _.rest = _.tail = _.drop = function(array, n, guard) {
    return slice.call(array, (n == null) || guard ? 1 : n);
  };

  // Trim out all falsy values from an array.
  _.compact = function(array) {
    return _.filter(array, _.identity);
  };

  // Internal implementation of a recursive `flatten` function.
  var flatten = function(input, shallow, output) {
    each(input, function(value) {
      if (_.isArray(value)) {
        shallow ? push.apply(output, value) : flatten(value, shallow, output);
      } else {
        output.push(value);
      }
    });
    return output;
  };

  // Return a completely flattened version of an array.
  _.flatten = function(array, shallow) {
    return flatten(array, shallow, []);
  };

  // Return a version of the array that does not contain the specified value(s).
  _.without = function(array) {
    return _.difference(array, slice.call(arguments, 1));
  };

  // Produce a duplicate-free version of the array. If the array has already
  // been sorted, you have the option of using a faster algorithm.
  // Aliased as `unique`.
  _.uniq = _.unique = function(array, isSorted, iterator, context) {
    if (_.isFunction(isSorted)) {
      context = iterator;
      iterator = isSorted;
      isSorted = false;
    }
    var initial = iterator ? _.map(array, iterator, context) : array;
    var results = [];
    var seen = [];
    each(initial, function(value, index) {
      if (isSorted ? (!index || seen[seen.length - 1] !== value) : !_.contains(seen, value)) {
        seen.push(value);
        results.push(array[index]);
      }
    });
    return results;
  };

  // Produce an array that contains the union: each distinct element from all of
  // the passed-in arrays.
  _.union = function() {
    return _.uniq(concat.apply(ArrayProto, arguments));
  };

  // Produce an array that contains every item shared between all the
  // passed-in arrays.
  _.intersection = function(array) {
    var rest = slice.call(arguments, 1);
    return _.filter(_.uniq(array), function(item) {
      return _.every(rest, function(other) {
        return _.indexOf(other, item) >= 0;
      });
    });
  };

  // Take the difference between one array and a number of other arrays.
  // Only the elements present in just the first array will remain.
  _.difference = function(array) {
    var rest = concat.apply(ArrayProto, slice.call(arguments, 1));
    return _.filter(array, function(value){ return !_.contains(rest, value); });
  };

  // Zip together multiple lists into a single array -- elements that share
  // an index go together.
  _.zip = function() {
    var args = slice.call(arguments);
    var length = _.max(_.pluck(args, 'length'));
    var results = new Array(length);
    for (var i = 0; i < length; i++) {
      results[i] = _.pluck(args, "" + i);
    }
    return results;
  };

  // Converts lists into objects. Pass either a single array of `[key, value]`
  // pairs, or two parallel arrays of the same length -- one of keys, and one of
  // the corresponding values.
  _.object = function(list, values) {
    if (list == null) return {};
    var result = {};
    for (var i = 0, l = list.length; i < l; i++) {
      if (values) {
        result[list[i]] = values[i];
      } else {
        result[list[i][0]] = list[i][1];
      }
    }
    return result;
  };

  // If the browser doesn't supply us with indexOf (I'm looking at you, **MSIE**),
  // we need this function. Return the position of the first occurrence of an
  // item in an array, or -1 if the item is not included in the array.
  // Delegates to **ECMAScript 5**'s native `indexOf` if available.
  // If the array is large and already in sort order, pass `true`
  // for **isSorted** to use binary search.
  _.indexOf = function(array, item, isSorted) {
    if (array == null) return -1;
    var i = 0, l = array.length;
    if (isSorted) {
      if (typeof isSorted == 'number') {
        i = (isSorted < 0 ? Math.max(0, l + isSorted) : isSorted);
      } else {
        i = _.sortedIndex(array, item);
        return array[i] === item ? i : -1;
      }
    }
    if (nativeIndexOf && array.indexOf === nativeIndexOf) return array.indexOf(item, isSorted);
    for (; i < l; i++) if (array[i] === item) return i;
    return -1;
  };

  // Delegates to **ECMAScript 5**'s native `lastIndexOf` if available.
  _.lastIndexOf = function(array, item, from) {
    if (array == null) return -1;
    var hasIndex = from != null;
    if (nativeLastIndexOf && array.lastIndexOf === nativeLastIndexOf) {
      return hasIndex ? array.lastIndexOf(item, from) : array.lastIndexOf(item);
    }
    var i = (hasIndex ? from : array.length);
    while (i--) if (array[i] === item) return i;
    return -1;
  };

  // Generate an integer Array containing an arithmetic progression. A port of
  // the native Python `range()` function. See
  // [the Python documentation](http://docs.python.org/library/functions.html#range).
  _.range = function(start, stop, step) {
    if (arguments.length <= 1) {
      stop = start || 0;
      start = 0;
    }
    step = arguments[2] || 1;

    var len = Math.max(Math.ceil((stop - start) / step), 0);
    var idx = 0;
    var range = new Array(len);

    while(idx < len) {
      range[idx++] = start;
      start += step;
    }

    return range;
  };

  // Function (ahem) Functions
  // ------------------

  // Create a function bound to a given object (assigning `this`, and arguments,
  // optionally). Delegates to **ECMAScript 5**'s native `Function.bind` if
  // available.
  _.bind = function(func, context) {
    if (func.bind === nativeBind && nativeBind) return nativeBind.apply(func, slice.call(arguments, 1));
    var args = slice.call(arguments, 2);
    return function() {
      return func.apply(context, args.concat(slice.call(arguments)));
    };
  };

  // Partially apply a function by creating a version that has had some of its
  // arguments pre-filled, without changing its dynamic `this` context.
  _.partial = function(func) {
    var args = slice.call(arguments, 1);
    return function() {
      return func.apply(this, args.concat(slice.call(arguments)));
    };
  };

  // Bind all of an object's methods to that object. Useful for ensuring that
  // all callbacks defined on an object belong to it.
  _.bindAll = function(obj) {
    var funcs = slice.call(arguments, 1);
    if (funcs.length === 0) funcs = _.functions(obj);
    each(funcs, function(f) { obj[f] = _.bind(obj[f], obj); });
    return obj;
  };

  // Memoize an expensive function by storing its results.
  _.memoize = function(func, hasher) {
    var memo = {};
    hasher || (hasher = _.identity);
    return function() {
      var key = hasher.apply(this, arguments);
      return _.has(memo, key) ? memo[key] : (memo[key] = func.apply(this, arguments));
    };
  };

  // Delays a function for the given number of milliseconds, and then calls
  // it with the arguments supplied.
  _.delay = function(func, wait) {
    var args = slice.call(arguments, 2);
    return setTimeout(function(){ return func.apply(null, args); }, wait);
  };

  // Defers a function, scheduling it to run after the current call stack has
  // cleared.
  _.defer = function(func) {
    return _.delay.apply(_, [func, 1].concat(slice.call(arguments, 1)));
  };

  // Returns a function, that, when invoked, will only be triggered at most once
  // during a given window of time.
  _.throttle = function(func, wait) {
    var context, args, timeout, result;
    var previous = 0;
    var later = function() {
      previous = new Date;
      timeout = null;
      result = func.apply(context, args);
    };
    return function() {
      var now = new Date;
      var remaining = wait - (now - previous);
      context = this;
      args = arguments;
      if (remaining <= 0) {
        clearTimeout(timeout);
        timeout = null;
        previous = now;
        result = func.apply(context, args);
      } else if (!timeout) {
        timeout = setTimeout(later, remaining);
      }
      return result;
    };
  };

  // Returns a function, that, as long as it continues to be invoked, will not
  // be triggered. The function will be called after it stops being called for
  // N milliseconds. If `immediate` is passed, trigger the function on the
  // leading edge, instead of the trailing.
  _.debounce = function(func, wait, immediate) {
    var timeout, result;
    return function() {
      var context = this, args = arguments;
      var later = function() {
        timeout = null;
        if (!immediate) result = func.apply(context, args);
      };
      var callNow = immediate && !timeout;
      clearTimeout(timeout);
      timeout = setTimeout(later, wait);
      if (callNow) result = func.apply(context, args);
      return result;
    };
  };

  // Returns a function that will be executed at most one time, no matter how
  // often you call it. Useful for lazy initialization.
  _.once = function(func) {
    var ran = false, memo;
    return function() {
      if (ran) return memo;
      ran = true;
      memo = func.apply(this, arguments);
      func = null;
      return memo;
    };
  };

  // Returns the first function passed as an argument to the second,
  // allowing you to adjust arguments, run code before and after, and
  // conditionally execute the original function.
  _.wrap = function(func, wrapper) {
    return function() {
      var args = [func];
      push.apply(args, arguments);
      return wrapper.apply(this, args);
    };
  };

  // Returns a function that is the composition of a list of functions, each
  // consuming the return value of the function that follows.
  _.compose = function() {
    var funcs = arguments;
    return function() {
      var args = arguments;
      for (var i = funcs.length - 1; i >= 0; i--) {
        args = [funcs[i].apply(this, args)];
      }
      return args[0];
    };
  };

  // Returns a function that will only be executed after being called N times.
  _.after = function(times, func) {
    if (times <= 0) return func();
    return function() {
      if (--times < 1) {
        return func.apply(this, arguments);
      }
    };
  };

  // Object Functions
  // ----------------

  // Retrieve the names of an object's properties.
  // Delegates to **ECMAScript 5**'s native `Object.keys`
  _.keys = nativeKeys || function(obj) {
    if (obj !== Object(obj)) throw new TypeError('Invalid object');
    var keys = [];
    for (var key in obj) if (_.has(obj, key)) keys[keys.length] = key;
    return keys;
  };

  // Retrieve the values of an object's properties.
  _.values = function(obj) {
    var values = [];
    for (var key in obj) if (_.has(obj, key)) values.push(obj[key]);
    return values;
  };

  // Convert an object into a list of `[key, value]` pairs.
  _.pairs = function(obj) {
    var pairs = [];
    for (var key in obj) if (_.has(obj, key)) pairs.push([key, obj[key]]);
    return pairs;
  };

  // Invert the keys and values of an object. The values must be serializable.
  _.invert = function(obj) {
    var result = {};
    for (var key in obj) if (_.has(obj, key)) result[obj[key]] = key;
    return result;
  };

  // Return a sorted list of the function names available on the object.
  // Aliased as `methods`
  _.functions = _.methods = function(obj) {
    var names = [];
    for (var key in obj) {
      if (_.isFunction(obj[key])) names.push(key);
    }
    return names.sort();
  };

  // Extend a given object with all the properties in passed-in object(s).
  _.extend = function(obj) {
    each(slice.call(arguments, 1), function(source) {
      if (source) {
        for (var prop in source) {
          obj[prop] = source[prop];
        }
      }
    });
    return obj;
  };

  // Return a copy of the object only containing the whitelisted properties.
  _.pick = function(obj) {
    var copy = {};
    var keys = concat.apply(ArrayProto, slice.call(arguments, 1));
    each(keys, function(key) {
      if (key in obj) copy[key] = obj[key];
    });
    return copy;
  };

   // Return a copy of the object without the blacklisted properties.
  _.omit = function(obj) {
    var copy = {};
    var keys = concat.apply(ArrayProto, slice.call(arguments, 1));
    for (var key in obj) {
      if (!_.contains(keys, key)) copy[key] = obj[key];
    }
    return copy;
  };

  // Fill in a given object with default properties.
  _.defaults = function(obj) {
    each(slice.call(arguments, 1), function(source) {
      if (source) {
        for (var prop in source) {
          if (obj[prop] == null) obj[prop] = source[prop];
        }
      }
    });
    return obj;
  };

  // Create a (shallow-cloned) duplicate of an object.
  _.clone = function(obj) {
    if (!_.isObject(obj)) return obj;
    return _.isArray(obj) ? obj.slice() : _.extend({}, obj);
  };

  // Invokes interceptor with the obj, and then returns obj.
  // The primary purpose of this method is to "tap into" a method chain, in
  // order to perform operations on intermediate results within the chain.
  _.tap = function(obj, interceptor) {
    interceptor(obj);
    return obj;
  };

  // Internal recursive comparison function for `isEqual`.
  var eq = function(a, b, aStack, bStack) {
    // Identical objects are equal. `0 === -0`, but they aren't identical.
    // See the Harmony `egal` proposal: http://wiki.ecmascript.org/doku.php?id=harmony:egal.
    if (a === b) return a !== 0 || 1 / a == 1 / b;
    // A strict comparison is necessary because `null == undefined`.
    if (a == null || b == null) return a === b;
    // Unwrap any wrapped objects.
    if (a instanceof _) a = a._wrapped;
    if (b instanceof _) b = b._wrapped;
    // Compare `[[Class]]` names.
    var className = toString.call(a);
    if (className != toString.call(b)) return false;
    switch (className) {
      // Strings, numbers, dates, and booleans are compared by value.
      case '[object String]':
        // Primitives and their corresponding object wrappers are equivalent; thus, `"5"` is
        // equivalent to `new String("5")`.
        return a == String(b);
      case '[object Number]':
        // `NaN`s are equivalent, but non-reflexive. An `egal` comparison is performed for
        // other numeric values.
        return a != +a ? b != +b : (a == 0 ? 1 / a == 1 / b : a == +b);
      case '[object Date]':
      case '[object Boolean]':
        // Coerce dates and booleans to numeric primitive values. Dates are compared by their
        // millisecond representations. Note that invalid dates with millisecond representations
        // of `NaN` are not equivalent.
        return +a == +b;
      // RegExps are compared by their source patterns and flags.
      case '[object RegExp]':
        return a.source == b.source &&
               a.global == b.global &&
               a.multiline == b.multiline &&
               a.ignoreCase == b.ignoreCase;
    }
    if (typeof a != 'object' || typeof b != 'object') return false;
    // Assume equality for cyclic structures. The algorithm for detecting cyclic
    // structures is adapted from ES 5.1 section 15.12.3, abstract operation `JO`.
    var length = aStack.length;
    while (length--) {
      // Linear search. Performance is inversely proportional to the number of
      // unique nested structures.
      if (aStack[length] == a) return bStack[length] == b;
    }
    // Add the first object to the stack of traversed objects.
    aStack.push(a);
    bStack.push(b);
    var size = 0, result = true;
    // Recursively compare objects and arrays.
    if (className == '[object Array]') {
      // Compare array lengths to determine if a deep comparison is necessary.
      size = a.length;
      result = size == b.length;
      if (result) {
        // Deep compare the contents, ignoring non-numeric properties.
        while (size--) {
          if (!(result = eq(a[size], b[size], aStack, bStack))) break;
        }
      }
    } else {
      // Objects with different constructors are not equivalent, but `Object`s
      // from different frames are.
      var aCtor = a.constructor, bCtor = b.constructor;
      if (aCtor !== bCtor && !(_.isFunction(aCtor) && (aCtor instanceof aCtor) &&
                               _.isFunction(bCtor) && (bCtor instanceof bCtor))) {
        return false;
      }
      // Deep compare objects.
      for (var key in a) {
        if (_.has(a, key)) {
          // Count the expected number of properties.
          size++;
          // Deep compare each member.
          if (!(result = _.has(b, key) && eq(a[key], b[key], aStack, bStack))) break;
        }
      }
      // Ensure that both objects contain the same number of properties.
      if (result) {
        for (key in b) {
          if (_.has(b, key) && !(size--)) break;
        }
        result = !size;
      }
    }
    // Remove the first object from the stack of traversed objects.
    aStack.pop();
    bStack.pop();
    return result;
  };

  // Perform a deep comparison to check if two objects are equal.
  _.isEqual = function(a, b) {
    return eq(a, b, [], []);
  };

  // Is a given array, string, or object empty?
  // An "empty" object has no enumerable own-properties.
  _.isEmpty = function(obj) {
    if (obj == null) return true;
    if (_.isArray(obj) || _.isString(obj)) return obj.length === 0;
    for (var key in obj) if (_.has(obj, key)) return false;
    return true;
  };

  // Is a given value a DOM element?
  _.isElement = function(obj) {
    return !!(obj && obj.nodeType === 1);
  };

  // Is a given value an array?
  // Delegates to ECMA5's native Array.isArray
  _.isArray = nativeIsArray || function(obj) {
    return toString.call(obj) == '[object Array]';
  };

  // Is a given variable an object?
  _.isObject = function(obj) {
    return obj === Object(obj);
  };

  // Add some isType methods: isArguments, isFunction, isString, isNumber, isDate, isRegExp.
  each(['Arguments', 'Function', 'String', 'Number', 'Date', 'RegExp'], function(name) {
    _['is' + name] = function(obj) {
      return toString.call(obj) == '[object ' + name + ']';
    };
  });

  // Define a fallback version of the method in browsers (ahem, IE), where
  // there isn't any inspectable "Arguments" type.
  if (!_.isArguments(arguments)) {
    _.isArguments = function(obj) {
      return !!(obj && _.has(obj, 'callee'));
    };
  }

  // Optimize `isFunction` if appropriate.
  if (typeof (/./) !== 'function') {
    _.isFunction = function(obj) {
      return typeof obj === 'function';
    };
  }

  // Is a given object a finite number?
  _.isFinite = function(obj) {
    return isFinite(obj) && !isNaN(parseFloat(obj));
  };

  // Is the given value `NaN`? (NaN is the only number which does not equal itself).
  _.isNaN = function(obj) {
    return _.isNumber(obj) && obj != +obj;
  };

  // Is a given value a boolean?
  _.isBoolean = function(obj) {
    return obj === true || obj === false || toString.call(obj) == '[object Boolean]';
  };

  // Is a given value equal to null?
  _.isNull = function(obj) {
    return obj === null;
  };

  // Is a given variable undefined?
  _.isUndefined = function(obj) {
    return obj === void 0;
  };

  // Shortcut function for checking if an object has a given property directly
  // on itself (in other words, not on a prototype).
  _.has = function(obj, key) {
    return hasOwnProperty.call(obj, key);
  };

  // Utility Functions
  // -----------------

  // Run Underscore.js in *noConflict* mode, returning the `_` variable to its
  // previous owner. Returns a reference to the Underscore object.
  _.noConflict = function() {
    root._ = previousUnderscore;
    return this;
  };

  // Keep the identity function around for default iterators.
  _.identity = function(value) {
    return value;
  };

  // Run a function **n** times.
  _.times = function(n, iterator, context) {
    var accum = Array(n);
    for (var i = 0; i < n; i++) accum[i] = iterator.call(context, i);
    return accum;
  };

  // Return a random integer between min and max (inclusive).
  _.random = function(min, max) {
    if (max == null) {
      max = min;
      min = 0;
    }
    return min + Math.floor(Math.random() * (max - min + 1));
  };

  // List of HTML entities for escaping.
  var entityMap = {
    escape: {
      '&': '&amp;',
      '<': '&lt;',
      '>': '&gt;',
      '"': '&quot;',
      "'": '&#x27;',
      '/': '&#x2F;'
    }
  };
  entityMap.unescape = _.invert(entityMap.escape);

  // Regexes containing the keys and values listed immediately above.
  var entityRegexes = {
    escape:   new RegExp('[' + _.keys(entityMap.escape).join('') + ']', 'g'),
    unescape: new RegExp('(' + _.keys(entityMap.unescape).join('|') + ')', 'g')
  };

  // Functions for escaping and unescaping strings to/from HTML interpolation.
  _.each(['escape', 'unescape'], function(method) {
    _[method] = function(string) {
      if (string == null) return '';
      return ('' + string).replace(entityRegexes[method], function(match) {
        return entityMap[method][match];
      });
    };
  });

  // If the value of the named property is a function then invoke it;
  // otherwise, return it.
  _.result = function(object, property) {
    if (object == null) return null;
    var value = object[property];
    return _.isFunction(value) ? value.call(object) : value;
  };

  // Add your own custom functions to the Underscore object.
  _.mixin = function(obj) {
    each(_.functions(obj), function(name){
      var func = _[name] = obj[name];
      _.prototype[name] = function() {
        var args = [this._wrapped];
        push.apply(args, arguments);
        return result.call(this, func.apply(_, args));
      };
    });
  };

  // Generate a unique integer id (unique within the entire client session).
  // Useful for temporary DOM ids.
  var idCounter = 0;
  _.uniqueId = function(prefix) {
    var id = ++idCounter + '';
    return prefix ? prefix + id : id;
  };

  // By default, Underscore uses ERB-style template delimiters, change the
  // following template settings to use alternative delimiters.
  _.templateSettings = {
    evaluate    : /<%([\s\S]+?)%>/g,
    interpolate : /<%=([\s\S]+?)%>/g,
    escape      : /<%-([\s\S]+?)%>/g
  };

  // When customizing `templateSettings`, if you don't want to define an
  // interpolation, evaluation or escaping regex, we need one that is
  // guaranteed not to match.
  var noMatch = /(.)^/;

  // Certain characters need to be escaped so that they can be put into a
  // string literal.
  var escapes = {
    "'":      "'",
    '\\':     '\\',
    '\r':     'r',
    '\n':     'n',
    '\t':     't',
    '\u2028': 'u2028',
    '\u2029': 'u2029'
  };

  var escaper = /\\|'|\r|\n|\t|\u2028|\u2029/g;

  // JavaScript micro-templating, similar to John Resig's implementation.
  // Underscore templating handles arbitrary delimiters, preserves whitespace,
  // and correctly escapes quotes within interpolated code.
  _.template = function(text, data, settings) {
    var render;
    settings = _.defaults({}, settings, _.templateSettings);

    // Combine delimiters into one regular expression via alternation.
    var matcher = new RegExp([
      (settings.escape || noMatch).source,
      (settings.interpolate || noMatch).source,
      (settings.evaluate || noMatch).source
    ].join('|') + '|$', 'g');

    // Compile the template source, escaping string literals appropriately.
    var index = 0;
    var source = "__p+='";
    text.replace(matcher, function(match, escape, interpolate, evaluate, offset) {
      source += text.slice(index, offset)
        .replace(escaper, function(match) { return '\\' + escapes[match]; });

      if (escape) {
        source += "'+\n((__t=(" + escape + "))==null?'':_.escape(__t))+\n'";
      }
      if (interpolate) {
        source += "'+\n((__t=(" + interpolate + "))==null?'':__t)+\n'";
      }
      if (evaluate) {
        source += "';\n" + evaluate + "\n__p+='";
      }
      index = offset + match.length;
      return match;
    });
    source += "';\n";

    // If a variable is not specified, place data values in local scope.
    if (!settings.variable) source = 'with(obj||{}){\n' + source + '}\n';

    source = "var __t,__p='',__j=Array.prototype.join," +
      "print=function(){__p+=__j.call(arguments,'');};\n" +
      source + "return __p;\n";

    try {
      render = new Function(settings.variable || 'obj', '_', source);
    } catch (e) {
      e.source = source;
      throw e;
    }

    if (data) return render(data, _);
    var template = function(data) {
      return render.call(this, data, _);
    };

    // Provide the compiled function source as a convenience for precompilation.
    template.source = 'function(' + (settings.variable || 'obj') + '){\n' + source + '}';

    return template;
  };

  // Add a "chain" function, which will delegate to the wrapper.
  _.chain = function(obj) {
    return _(obj).chain();
  };

  // OOP
  // ---------------
  // If Underscore is called as a function, it returns a wrapped object that
  // can be used OO-style. This wrapper holds altered versions of all the
  // underscore functions. Wrapped objects may be chained.

  // Helper function to continue chaining intermediate results.
  var result = function(obj) {
    return this._chain ? _(obj).chain() : obj;
  };

  // Add all of the Underscore functions to the wrapper object.
  _.mixin(_);

  // Add all mutator Array functions to the wrapper.
  each(['pop', 'push', 'reverse', 'shift', 'sort', 'splice', 'unshift'], function(name) {
    var method = ArrayProto[name];
    _.prototype[name] = function() {
      var obj = this._wrapped;
      method.apply(obj, arguments);
      if ((name == 'shift' || name == 'splice') && obj.length === 0) delete obj[0];
      return result.call(this, obj);
    };
  });

  // Add all accessor Array functions to the wrapper.
  each(['concat', 'join', 'slice'], function(name) {
    var method = ArrayProto[name];
    _.prototype[name] = function() {
      return result.call(this, method.apply(this._wrapped, arguments));
    };
  });

  _.extend(_.prototype, {

    // Start chaining a wrapped Underscore object.
    chain: function() {
      this._chain = true;
      return this;
    },

    // Extracts the result from a wrapped and chained object.
    value: function() {
      return this._wrapped;
    }

  });

  // AMD define happens at the end for compatibility with AMD loaders
  // that don't enforce next-turn semantics on modules.
  if (typeof define === 'function' && define.amd) {
    define('underscore', function() {
      return _;
    });
  }

}).call(this);
;/**
 *
 * Color picker
 * Author: Stefan Petre www.eyecon.ro
 * 
 * Dual licensed under the MIT and GPL licenses
 * 
 */
(function ($) {
	var ColorPicker = function () {
		var
			ids = {},
			inAction,
			charMin = 65,
			visible,
			tpl = '<div class="colorpicker"><div class="colorpicker_color"><div><div></div></div></div><div class="colorpicker_hue"><div></div></div><div class="colorpicker_new_color"></div><div class="colorpicker_current_color"></div><div class="colorpicker_hex"><input type="text" maxlength="6" size="6" /></div><div class="colorpicker_rgb_r colorpicker_field"><input type="text" maxlength="3" size="3" /><span></span></div><div class="colorpicker_rgb_g colorpicker_field"><input type="text" maxlength="3" size="3" /><span></span></div><div class="colorpicker_rgb_b colorpicker_field"><input type="text" maxlength="3" size="3" /><span></span></div><div class="colorpicker_hsb_h colorpicker_field"><input type="text" maxlength="3" size="3" /><span></span></div><div class="colorpicker_hsb_s colorpicker_field"><input type="text" maxlength="3" size="3" /><span></span></div><div class="colorpicker_hsb_b colorpicker_field"><input type="text" maxlength="3" size="3" /><span></span></div><div class="colorpicker_submit"></div></div>',
			defaults = {
				eventName: 'click',
				onShow: function () {},
				onBeforeShow: function(){},
				onHide: function () {},
				onChange: function () {},
				onSubmit: function () {},
				color: 'ff0000',
				livePreview: true,
				flat: false
			},
			fillRGBFields = function  (hsb, cal) {
				var rgb = HSBToRGB(hsb);
				$(cal).data('colorpicker').fields
					.eq(1).val(rgb.r).end()
					.eq(2).val(rgb.g).end()
					.eq(3).val(rgb.b).end();
			},
			fillHSBFields = function  (hsb, cal) {
				$(cal).data('colorpicker').fields
					.eq(4).val(hsb.h).end()
					.eq(5).val(hsb.s).end()
					.eq(6).val(hsb.b).end();
			},
			fillHexFields = function (hsb, cal) {
				$(cal).data('colorpicker').fields
					.eq(0).val(HSBToHex(hsb)).end();
			},
			setSelector = function (hsb, cal) {
				$(cal).data('colorpicker').selector.css('backgroundColor', '#' + HSBToHex({h: hsb.h, s: 100, b: 100}));
				$(cal).data('colorpicker').selectorIndic.css({
					left: parseInt(150 * hsb.s/100, 10),
					top: parseInt(150 * (100-hsb.b)/100, 10)
				});
			},
			setHue = function (hsb, cal) {
				$(cal).data('colorpicker').hue.css('top', parseInt(150 - 150 * hsb.h/360, 10));
			},
			setCurrentColor = function (hsb, cal) {
				$(cal).data('colorpicker').currentColor.css('backgroundColor', '#' + HSBToHex(hsb));
			},
			setNewColor = function (hsb, cal) {
				$(cal).data('colorpicker').newColor.css('backgroundColor', '#' + HSBToHex(hsb));
			},
			keyDown = function (ev) {
				var pressedKey = ev.charCode || ev.keyCode || -1;
				if ((pressedKey > charMin && pressedKey <= 90) || pressedKey == 32) {
					return false;
				}
				var cal = $(this).parent().parent();
				if (cal.data('colorpicker').livePreview === true) {
					change.apply(this);
				}
			},
			change = function (ev) {
				var cal = $(this).parent().parent(), col;
				if (this.parentNode.className.indexOf('_hex') > 0) {
					cal.data('colorpicker').color = col = HexToHSB(fixHex(this.value));
				} else if (this.parentNode.className.indexOf('_hsb') > 0) {
					cal.data('colorpicker').color = col = fixHSB({
						h: parseInt(cal.data('colorpicker').fields.eq(4).val(), 10),
						s: parseInt(cal.data('colorpicker').fields.eq(5).val(), 10),
						b: parseInt(cal.data('colorpicker').fields.eq(6).val(), 10)
					});
				} else {
					cal.data('colorpicker').color = col = RGBToHSB(fixRGB({
						r: parseInt(cal.data('colorpicker').fields.eq(1).val(), 10),
						g: parseInt(cal.data('colorpicker').fields.eq(2).val(), 10),
						b: parseInt(cal.data('colorpicker').fields.eq(3).val(), 10)
					}));
				}
				if (ev) {
					fillRGBFields(col, cal.get(0));
					fillHexFields(col, cal.get(0));
					fillHSBFields(col, cal.get(0));
				}
				setSelector(col, cal.get(0));
				setHue(col, cal.get(0));
				setNewColor(col, cal.get(0));
				cal.data('colorpicker').onChange.apply(cal, [col, HSBToHex(col), HSBToRGB(col)]);
			},
			blur = function (ev) {
				var cal = $(this).parent().parent();
				cal.data('colorpicker').fields.parent().removeClass('colorpicker_focus');
			},
			focus = function () {
				charMin = this.parentNode.className.indexOf('_hex') > 0 ? 70 : 65;
				$(this).parent().parent().data('colorpicker').fields.parent().removeClass('colorpicker_focus');
				$(this).parent().addClass('colorpicker_focus');
			},
			downIncrement = function (ev) {
				var field = $(this).parent().find('input').focus();
				var current = {
					el: $(this).parent().addClass('colorpicker_slider'),
					max: this.parentNode.className.indexOf('_hsb_h') > 0 ? 360 : (this.parentNode.className.indexOf('_hsb') > 0 ? 100 : 255),
					y: ev.pageY,
					field: field,
					val: parseInt(field.val(), 10),
					preview: $(this).parent().parent().data('colorpicker').livePreview					
				};
				$(document).bind('mouseup', current, upIncrement);
				$(document).bind('mousemove', current, moveIncrement);
			},
			moveIncrement = function (ev) {
				ev.data.field.val(Math.max(0, Math.min(ev.data.max, parseInt(ev.data.val + ev.pageY - ev.data.y, 10))));
				if (ev.data.preview) {
					change.apply(ev.data.field.get(0), [true]);
				}
				return false;
			},
			upIncrement = function (ev) {
				change.apply(ev.data.field.get(0), [true]);
				ev.data.el.removeClass('colorpicker_slider').find('input').focus();
				$(document).unbind('mouseup', upIncrement);
				$(document).unbind('mousemove', moveIncrement);
				return false;
			},
			downHue = function (ev) {
				var current = {
					cal: $(this).parent(),
					y: $(this).offset().top
				};
				current.preview = current.cal.data('colorpicker').livePreview;
				$(document).bind('mouseup', current, upHue);
				$(document).bind('mousemove', current, moveHue);
			},
			moveHue = function (ev) {
				change.apply(
					ev.data.cal.data('colorpicker')
						.fields
						.eq(4)
						.val(parseInt(360*(150 - Math.max(0,Math.min(150,(ev.pageY - ev.data.y))))/150, 10))
						.get(0),
					[ev.data.preview]
				);
				return false;
			},
			upHue = function (ev) {
				fillRGBFields(ev.data.cal.data('colorpicker').color, ev.data.cal.get(0));
				fillHexFields(ev.data.cal.data('colorpicker').color, ev.data.cal.get(0));
				$(document).unbind('mouseup', upHue);
				$(document).unbind('mousemove', moveHue);
				return false;
			},
			downSelector = function (ev) {
				var current = {
					cal: $(this).parent(),
					pos: $(this).offset()
				};
				current.preview = current.cal.data('colorpicker').livePreview;
				$(document).bind('mouseup', current, upSelector);
				$(document).bind('mousemove', current, moveSelector);
			},
			moveSelector = function (ev) {
				change.apply(
					ev.data.cal.data('colorpicker')
						.fields
						.eq(6)
						.val(parseInt(100*(150 - Math.max(0,Math.min(150,(ev.pageY - ev.data.pos.top))))/150, 10))
						.end()
						.eq(5)
						.val(parseInt(100*(Math.max(0,Math.min(150,(ev.pageX - ev.data.pos.left))))/150, 10))
						.get(0),
					[ev.data.preview]
				);
				return false;
			},
			upSelector = function (ev) {
				fillRGBFields(ev.data.cal.data('colorpicker').color, ev.data.cal.get(0));
				fillHexFields(ev.data.cal.data('colorpicker').color, ev.data.cal.get(0));
				$(document).unbind('mouseup', upSelector);
				$(document).unbind('mousemove', moveSelector);
				return false;
			},
			enterSubmit = function (ev) {
				$(this).addClass('colorpicker_focus');
			},
			leaveSubmit = function (ev) {
				$(this).removeClass('colorpicker_focus');
			},
			clickSubmit = function (ev) {
				var cal = $(this).parent();
				var col = cal.data('colorpicker').color;
				cal.data('colorpicker').origColor = col;
				setCurrentColor(col, cal.get(0));
				cal.data('colorpicker').onSubmit(col, HSBToHex(col), HSBToRGB(col), cal.data('colorpicker').el);
			},
			show = function (ev) {
				var cal = $('#' + $(this).data('colorpickerId'));
				cal.data('colorpicker').onBeforeShow.apply(this, [cal.get(0)]);
				var pos = $(this).offset();
				var viewPort = getViewport();
				var top = pos.top + this.offsetHeight;
				var left = pos.left;
				if (top + 176 > viewPort.t + viewPort.h) {
					top -= this.offsetHeight + 176;
				}
				if (left + 356 > viewPort.l + viewPort.w) {
					left -= 356;
				}
				cal.css({left: left + 'px', top: top + 'px'});
				if (cal.data('colorpicker').onShow.apply(this, [cal.get(0)]) != false) {
					cal.show();
				}
				$(document).bind('mousedown', {cal: cal}, hide);
				return false;
			},
			hide = function (ev) {
				if (!isChildOf(ev.data.cal.get(0), ev.target, ev.data.cal.get(0))) {
					if (ev.data.cal.data('colorpicker').onHide.apply(this, [ev.data.cal.get(0)]) != false) {
						ev.data.cal.hide();
					}
					$(document).unbind('mousedown', hide);
				}
			},
			isChildOf = function(parentEl, el, container) {
				if (parentEl == el) {
					return true;
				}
				if (parentEl.contains) {
					return parentEl.contains(el);
				}
				if ( parentEl.compareDocumentPosition ) {
					return !!(parentEl.compareDocumentPosition(el) & 16);
				}
				var prEl = el.parentNode;
				while(prEl && prEl != container) {
					if (prEl == parentEl)
						return true;
					prEl = prEl.parentNode;
				}
				return false;
			},
			getViewport = function () {
				var m = document.compatMode == 'CSS1Compat';
				return {
					l : window.pageXOffset || (m ? document.documentElement.scrollLeft : document.body.scrollLeft),
					t : window.pageYOffset || (m ? document.documentElement.scrollTop : document.body.scrollTop),
					w : window.innerWidth || (m ? document.documentElement.clientWidth : document.body.clientWidth),
					h : window.innerHeight || (m ? document.documentElement.clientHeight : document.body.clientHeight)
				};
			},
			fixHSB = function (hsb) {
				return {
					h: Math.min(360, Math.max(0, hsb.h)),
					s: Math.min(100, Math.max(0, hsb.s)),
					b: Math.min(100, Math.max(0, hsb.b))
				};
			}, 
			fixRGB = function (rgb) {
				return {
					r: Math.min(255, Math.max(0, rgb.r)),
					g: Math.min(255, Math.max(0, rgb.g)),
					b: Math.min(255, Math.max(0, rgb.b))
				};
			},
			fixHex = function (hex) {
				var len = 6 - hex.length;
				if (len > 0) {
					var o = [];
					for (var i=0; i<len; i++) {
						o.push('0');
					}
					o.push(hex);
					hex = o.join('');
				}
				return hex;
			}, 
			HexToRGB = function (hex) {
				var hex = parseInt(((hex.indexOf('#') > -1) ? hex.substring(1) : hex), 16);
				return {r: hex >> 16, g: (hex & 0x00FF00) >> 8, b: (hex & 0x0000FF)};
			},
			HexToHSB = function (hex) {
				return RGBToHSB(HexToRGB(hex));
			},
			RGBToHSB = function (rgb) {
				var hsb = {
					h: 0,
					s: 0,
					b: 0
				};
				var min = Math.min(rgb.r, rgb.g, rgb.b);
				var max = Math.max(rgb.r, rgb.g, rgb.b);
				var delta = max - min;
				hsb.b = max;
				if (max != 0) {
					
				}
				hsb.s = max != 0 ? 255 * delta / max : 0;
				if (hsb.s != 0) {
					if (rgb.r == max) {
						hsb.h = (rgb.g - rgb.b) / delta;
					} else if (rgb.g == max) {
						hsb.h = 2 + (rgb.b - rgb.r) / delta;
					} else {
						hsb.h = 4 + (rgb.r - rgb.g) / delta;
					}
				} else {
					hsb.h = -1;
				}
				hsb.h *= 60;
				if (hsb.h < 0) {
					hsb.h += 360;
				}
				hsb.s *= 100/255;
				hsb.b *= 100/255;
				return hsb;
			},
			HSBToRGB = function (hsb) {
				var rgb = {};
				var h = Math.round(hsb.h);
				var s = Math.round(hsb.s*255/100);
				var v = Math.round(hsb.b*255/100);
				if(s == 0) {
					rgb.r = rgb.g = rgb.b = v;
				} else {
					var t1 = v;
					var t2 = (255-s)*v/255;
					var t3 = (t1-t2)*(h%60)/60;
					if(h==360) h = 0;
					if(h<60) {rgb.r=t1;	rgb.b=t2; rgb.g=t2+t3}
					else if(h<120) {rgb.g=t1; rgb.b=t2;	rgb.r=t1-t3}
					else if(h<180) {rgb.g=t1; rgb.r=t2;	rgb.b=t2+t3}
					else if(h<240) {rgb.b=t1; rgb.r=t2;	rgb.g=t1-t3}
					else if(h<300) {rgb.b=t1; rgb.g=t2;	rgb.r=t2+t3}
					else if(h<360) {rgb.r=t1; rgb.g=t2;	rgb.b=t1-t3}
					else {rgb.r=0; rgb.g=0;	rgb.b=0}
				}
				return {r:Math.round(rgb.r), g:Math.round(rgb.g), b:Math.round(rgb.b)};
			},
			RGBToHex = function (rgb) {
				var hex = [
					rgb.r.toString(16),
					rgb.g.toString(16),
					rgb.b.toString(16)
				];
				$.each(hex, function (nr, val) {
					if (val.length == 1) {
						hex[nr] = '0' + val;
					}
				});
				return hex.join('');
			},
			HSBToHex = function (hsb) {
				return RGBToHex(HSBToRGB(hsb));
			},
			restoreOriginal = function () {
				var cal = $(this).parent();
				var col = cal.data('colorpicker').origColor;
				cal.data('colorpicker').color = col;
				fillRGBFields(col, cal.get(0));
				fillHexFields(col, cal.get(0));
				fillHSBFields(col, cal.get(0));
				setSelector(col, cal.get(0));
				setHue(col, cal.get(0));
				setNewColor(col, cal.get(0));
			};
		return {
			init: function (opt) {
				opt = $.extend({}, defaults, opt||{});
				if (typeof opt.color == 'string') {
					opt.color = HexToHSB(opt.color);
				} else if (opt.color.r != undefined && opt.color.g != undefined && opt.color.b != undefined) {
					opt.color = RGBToHSB(opt.color);
				} else if (opt.color.h != undefined && opt.color.s != undefined && opt.color.b != undefined) {
					opt.color = fixHSB(opt.color);
				} else {
					return this;
				}
				return this.each(function () {
					if (!$(this).data('colorpickerId')) {
						var options = $.extend({}, opt);
						options.origColor = opt.color;
						var id = 'collorpicker_' + parseInt(Math.random() * 1000);
						$(this).data('colorpickerId', id);
						var cal = $(tpl).attr('id', id);
						if (options.flat) {
							cal.appendTo(this).show();
						} else {
							cal.appendTo(document.body);
						}
						options.fields = cal
											.find('input')
												.bind('keyup', keyDown)
												.bind('change', change)
												.bind('blur', blur)
												.bind('focus', focus);
						cal
							.find('span').bind('mousedown', downIncrement).end()
							.find('>div.colorpicker_current_color').bind('click', restoreOriginal);
						options.selector = cal.find('div.colorpicker_color').bind('mousedown', downSelector);
						options.selectorIndic = options.selector.find('div div');
						options.el = this;
						options.hue = cal.find('div.colorpicker_hue div');
						cal.find('div.colorpicker_hue').bind('mousedown', downHue);
						options.newColor = cal.find('div.colorpicker_new_color');
						options.currentColor = cal.find('div.colorpicker_current_color');
						cal.data('colorpicker', options);
						cal.find('div.colorpicker_submit')
							.bind('mouseenter', enterSubmit)
							.bind('mouseleave', leaveSubmit)
							.bind('click', clickSubmit);
						fillRGBFields(options.color, cal.get(0));
						fillHSBFields(options.color, cal.get(0));
						fillHexFields(options.color, cal.get(0));
						setHue(options.color, cal.get(0));
						setSelector(options.color, cal.get(0));
						setCurrentColor(options.color, cal.get(0));
						setNewColor(options.color, cal.get(0));
						if (options.flat) {
							cal.css({
								position: 'relative',
								display: 'block'
							});
						} else {
							$(this).bind(options.eventName, show);
						}
					}
				});
			},
			showPicker: function() {
				return this.each( function () {
					if ($(this).data('colorpickerId')) {
						show.apply(this);
					}
				});
			},
			hidePicker: function() {
				return this.each( function () {
					if ($(this).data('colorpickerId')) {
						$('#' + $(this).data('colorpickerId')).hide();
					}
				});
			},
			setColor: function(col) {
				if (typeof col == 'string') {
					col = HexToHSB(col);
				} else if (col.r != undefined && col.g != undefined && col.b != undefined) {
					col = RGBToHSB(col);
				} else if (col.h != undefined && col.s != undefined && col.b != undefined) {
					col = fixHSB(col);
				} else {
					return this;
				}
				return this.each(function(){
					if ($(this).data('colorpickerId')) {
						var cal = $('#' + $(this).data('colorpickerId'));
						cal.data('colorpicker').color = col;
						cal.data('colorpicker').origColor = col;
						fillRGBFields(col, cal.get(0));
						fillHSBFields(col, cal.get(0));
						fillHexFields(col, cal.get(0));
						setHue(col, cal.get(0));
						setSelector(col, cal.get(0));
						setCurrentColor(col, cal.get(0));
						setNewColor(col, cal.get(0));
					}
				});
			}
		};
	}();
	$.fn.extend({
		ColorPicker: ColorPicker.init,
		ColorPickerHide: ColorPicker.hidePicker,
		ColorPickerShow: ColorPicker.showPicker,
		ColorPickerSetColor: ColorPicker.setColor
	});
})(jQuery);/*!
 * Lazy Load - jQuery plugin for lazy loading images
 *
 * Copyright (c) 2007-2015 Mika Tuupola
 *
 * Licensed under the MIT license:
 *   http://www.opensource.org/licenses/mit-license.php
 *
 * Project home:
 *   http://www.appelsiini.net/projects/lazyload
 *
 * Version:  1.9.7
 *
 */

(function($, window, document, undefined) {
    var $window = $(window);

    $.fn.lazyload = function(options) {
        var elements = this;
        var $container;
        var settings = {
            threshold       : 0,
            failure_limit   : 0,
            event           : "scroll",
            effect          : "show",
            container       : window,
            data_attribute  : "original",
            skip_invisible  : false,
            appear          : null,
            load            : null,
            placeholder     : "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsQAAA7EAZUrDhsAAAANSURBVBhXYzh8+PB/AAffA0nNPuCLAAAAAElFTkSuQmCC"
        };

        function update() {
            var counter = 0;

            elements.each(function() {
                var $this = $(this);
                if (settings.skip_invisible && !$this.is(":visible")) {
                    return;
                }
                if ($.abovethetop(this, settings) ||
                    $.leftofbegin(this, settings)) {
                        /* Nothing. */
                } else if (!$.belowthefold(this, settings) &&
                    !$.rightoffold(this, settings)) {
                        $this.trigger("appear");
                        /* if we found an image we'll load, reset the counter */
                        counter = 0;
                } else {
                    if (++counter > settings.failure_limit) {
                        return false;
                    }
                }
            });

        }

        if(options) {
            /* Maintain BC for a couple of versions. */
            if (undefined !== options.failurelimit) {
                options.failure_limit = options.failurelimit;
                delete options.failurelimit;
            }
            if (undefined !== options.effectspeed) {
                options.effect_speed = options.effectspeed;
                delete options.effectspeed;
            }

            $.extend(settings, options);
        }

        /* Cache container as jQuery as object. */
        $container = (settings.container === undefined ||
                      settings.container === window) ? $window : $(settings.container);

        /* Fire one scroll event per scroll. Not one scroll event per image. */
        if (0 === settings.event.indexOf("scroll")) {
            $container.bind(settings.event, function() {
                return update();
            });
        }

        this.each(function() {
            var self = this;
            var $self = $(self);

            self.loaded = false;

            /* If no src attribute given use data:uri. */
            if ($self.attr("src") === undefined || $self.attr("src") === false) {
                if ($self.is("img")) {
                    $self.attr("src", settings.placeholder);
                }
            }

            /* When appear is triggered load original image. */
            $self.one("appear", function() {
                if (!this.loaded) {
                    if (settings.appear) {
                        var elements_left = elements.length;
                        settings.appear.call(self, elements_left, settings);
                    }
                    $("<img />")
                        .bind("load", function() {

                            var original = $self.attr("data-" + settings.data_attribute);
                            $self.hide();
                            if ($self.is("img")) {
                                $self.attr("src", original);
                            } else {
                                $self.css("background-image", "url('" + original + "')");
                            }
                            $self[settings.effect](settings.effect_speed);

                            self.loaded = true;

                            /* Remove image from array so it is not looped next time. */
                            var temp = $.grep(elements, function(element) {
                                return !element.loaded;
                            });
                            elements = $(temp);

                            if (settings.load) {
                                var elements_left = elements.length;
                                settings.load.call(self, elements_left, settings);
                            }
                        })
                        .attr("src", $self.attr("data-" + settings.data_attribute));
                }
            });

            /* When wanted event is triggered load original image */
            /* by triggering appear.                              */
            if (0 !== settings.event.indexOf("scroll")) {
                $self.bind(settings.event, function() {
                    if (!self.loaded) {
                        $self.trigger("appear");
                    }
                });
            }
        });

        /* Check if something appears when window is resized. */
        $window.bind("resize", function() {
            update();
        });

        /* With IOS5 force loading images when navigating with back button. */
        /* Non optimal workaround. */
        if ((/(?:iphone|ipod|ipad).*os 5/gi).test(navigator.appVersion)) {
            $window.bind("pageshow", function(event) {
                if (event.originalEvent && event.originalEvent.persisted) {
                    elements.each(function() {
                        $(this).trigger("appear");
                    });
                }
            });
        }

        /* Force initial check if images should appear. */
        $(document).ready(function() {
            update();
        });

        return this;
    };

    /* Convenience methods in jQuery namespace.           */
    /* Use as  $.belowthefold(element, {threshold : 100, container : window}) */

    $.belowthefold = function(element, settings) {
        var fold;

        if (settings.container === undefined || settings.container === window) {
            fold = (window.innerHeight ? window.innerHeight : $window.height()) + $window.scrollTop();
        } else {
            fold = $(settings.container).offset().top + $(settings.container).height();
        }

        return fold <= $(element).offset().top - settings.threshold;
    };

    $.rightoffold = function(element, settings) {
        var fold;

        if (settings.container === undefined || settings.container === window) {
            fold = $window.width() + $window.scrollLeft();
        } else {
            fold = $(settings.container).offset().left + $(settings.container).width();
        }

        return fold <= $(element).offset().left - settings.threshold;
    };

    $.abovethetop = function(element, settings) {
        var fold;

        if (settings.container === undefined || settings.container === window) {
            fold = $window.scrollTop();
        } else {
            fold = $(settings.container).offset().top;
        }

        return fold >= $(element).offset().top + settings.threshold  + $(element).height();
    };

    $.leftofbegin = function(element, settings) {
        var fold;

        if (settings.container === undefined || settings.container === window) {
            fold = $window.scrollLeft();
        } else {
            fold = $(settings.container).offset().left;
        }

        return fold >= $(element).offset().left + settings.threshold + $(element).width();
    };

    $.inviewport = function(element, settings) {
         return !$.rightoffold(element, settings) && !$.leftofbegin(element, settings) &&
                !$.belowthefold(element, settings) && !$.abovethetop(element, settings);
     };

    /* Custom selectors for your convenience.   */
    /* Use as $("img:below-the-fold").something() or */
    /* $("img").filter(":below-the-fold").something() which is faster */

    $.extend($.expr[":"], {
        "below-the-fold" : function(a) { return $.belowthefold(a, {threshold : 0}); },
        "above-the-top"  : function(a) { return !$.belowthefold(a, {threshold : 0}); },
        "right-of-screen": function(a) { return $.rightoffold(a, {threshold : 0}); },
        "left-of-screen" : function(a) { return !$.rightoffold(a, {threshold : 0}); },
        "in-viewport"    : function(a) { return $.inviewport(a, {threshold : 0}); },
        /* Maintain BC for couple of versions. */
        "above-the-fold" : function(a) { return !$.belowthefold(a, {threshold : 0}); },
        "right-of-fold"  : function(a) { return $.rightoffold(a, {threshold : 0}); },
        "left-of-fold"   : function(a) { return !$.rightoffold(a, {threshold : 0}); }
    });

})(jQuery, window, document);
;/**
 * humane.js
 * Humanized Messages for Notifications
 * @author Marc Harter (@wavded)
 * @example
 *   humane.log('hello world');
 * See more usage examples at: http://wavded.github.com/humane-js/
 */

;!function (name, context, definition) {
   if (typeof module !== 'undefined') module.exports = definition(name, context)
   else if (typeof define === 'function' && typeof define.amd  === 'object') define(definition)
   else context[name] = definition(name, context)
}('humane', this, function (name, context) {
   var win = window
   var doc = document

   var ENV = {
      on: function (el, type, cb) {
         'addEventListener' in win ? el.addEventListener(type,cb,false) : el.attachEvent('on'+type,cb)
      },
      off: function (el, type, cb) {
         'removeEventListener' in win ? el.removeEventListener(type,cb,false) : el.detachEvent('on'+type,cb)
      },
      bind: function (fn, ctx) {
         return function () { fn.apply(ctx,arguments) }
      },
      isArray: Array.isArray || function (obj) { return Object.prototype.toString.call(obj) === '[object Array]' },
      config: function (preferred, fallback) {
         return preferred != null ? preferred : fallback
      },
      transSupport: false,
      useFilter: /msie [678]/i.test(navigator.userAgent), // sniff, sniff
      _checkTransition: function () {
         var el = doc.createElement('div')
         var vendors = { webkit: 'webkit', Moz: '', O: 'o', ms: 'MS' }

         for (var vendor in vendors)
            if (vendor + 'Transition' in el.style) {
               this.vendorPrefix = vendors[vendor]
               this.transSupport = true
            }
      }
   }
   ENV._checkTransition()

   var Humane = function (o) {
      o || (o = {})
      this.queue = []
      this.baseCls = o.baseCls || 'humane'
      this.addnCls = o.addnCls || ''
      this.timeout = 'timeout' in o ? o.timeout : 2500
      this.waitForMove = o.waitForMove || false
      this.clickToClose = o.clickToClose || false
      this.forceNew = o.forceNew || false
      this.container = o.container || doc.body

      try { this._setupEl() } // attempt to setup elements
      catch (e) {
        ENV.on(win,'load',ENV.bind(this._setupEl, this)) // dom wasn't ready, wait till ready
      }
   }

   Humane.prototype = {
      constructor: Humane,
      _setupEl: function () {
         var el = doc.createElement('div')
         el.style.display = 'none'
         this.container.appendChild(el)
         this.el = el
         this.removeEvent = ENV.bind(this.remove,this)
         this.transEvent = ENV.bind(this._afterAnimation,this)
         this._run()
      },
      _afterTimeout: function () {
         if (!ENV.config(this.currentMsg.waitForMove,this.waitForMove)) this.remove()

         else if (!this.removeEventsSet) {
            ENV.on(doc.body,'mousemove',this.removeEvent)
            ENV.on(doc.body,'click',this.removeEvent)
            ENV.on(doc.body,'keypress',this.removeEvent)
            ENV.on(doc.body,'touchstart',this.removeEvent)
            this.removeEventsSet = true
         }
      },
      _run: function () {
         if (this._animating || !this.queue.length || !this.el) return

         this._animating = true
         if (this.currentTimer) {
            clearTimeout(this.currentTimer)
            this.currentTimer = null
         }

         var msg = this.queue.shift()
         var clickToClose = ENV.config(msg.clickToClose,this.clickToClose)

         if (clickToClose) {
            ENV.on(this.el,'click',this.removeEvent)
            ENV.on(this.el,'touchstart',this.removeEvent)
         }

         var timeout = ENV.config(msg.timeout,this.timeout)

         if (timeout > 0)
            this.currentTimer = setTimeout(ENV.bind(this._afterTimeout,this), timeout)

         if (ENV.isArray(msg.html)) msg.html = '<ul><li>'+msg.html.join('<li>')+'</ul>'

         this.el.innerHTML = msg.html
         this.currentMsg = msg
         this.el.className = this.baseCls
         if (ENV.transSupport) {
            this.el.style.display = 'block'
            setTimeout(ENV.bind(this._showMsg,this),50)
         } else {
            this._showMsg()
         }

      },
      _setOpacity: function (opacity) {
         if (ENV.useFilter)
            this.el.filters.item('DXImageTransform.Microsoft.Alpha').Opacity = opacity*100
         else
            this.el.style.opacity = String(opacity)
      },
      _showMsg: function () {
         var addnCls = ENV.config(this.currentMsg.addnCls,this.addnCls)
         if (ENV.transSupport) {
            this.el.className = this.baseCls+' '+addnCls+' '+this.baseCls+'-animate'
         }
         else {
            var opacity = 0
            this.el.className = this.baseCls+' '+addnCls+' '+this.baseCls+'-js-animate'
            this._setOpacity(0) // reset value so hover states work
            this.el.style.display = 'block'

            var self = this
            var interval = setInterval(function(){
               if (opacity < 1) {
                  opacity += 0.1
                  if (opacity > 1) opacity = 1
                  self._setOpacity(opacity)
               }
               else clearInterval(interval)
            }, 30)
         }
      },
      _hideMsg: function () {
         var addnCls = ENV.config(this.currentMsg.addnCls,this.addnCls)
         if (ENV.transSupport) {
            this.el.className = this.baseCls+' '+addnCls
            ENV.on(this.el,ENV.vendorPrefix ? ENV.vendorPrefix+'TransitionEnd' : 'transitionend',this.transEvent)
         }
         else {
            var opacity = 1
            var self = this
            var interval = setInterval(function(){
               if(opacity > 0) {
                  opacity -= 0.1
                  if (opacity < 0) opacity = 0
                  self._setOpacity(opacity);
               }
               else {
                  self.el.className = self.baseCls+' '+addnCls
                  clearInterval(interval)
                  self._afterAnimation()
               }
            }, 30)
         }
      },
      _afterAnimation: function () {
         if (ENV.transSupport) ENV.off(this.el,ENV.vendorPrefix ? ENV.vendorPrefix+'TransitionEnd' : 'transitionend',this.transEvent)

         if (this.currentMsg.cb) this.currentMsg.cb()
         this.el.style.display = 'none'

         this._animating = false
         this._run()
      },
      remove: function (e) {
         var cb = typeof e == 'function' ? e : null

         ENV.off(doc.body,'mousemove',this.removeEvent)
         ENV.off(doc.body,'click',this.removeEvent)
         ENV.off(doc.body,'keypress',this.removeEvent)
         ENV.off(doc.body,'touchstart',this.removeEvent)
         ENV.off(this.el,'click',this.removeEvent)
         ENV.off(this.el,'touchstart',this.removeEvent)
         this.removeEventsSet = false

         if (cb) this.currentMsg.cb = cb
         if (this._animating) this._hideMsg()
         else if (cb) cb()
      },
      log: function (html, o, cb, defaults) {
         var msg = {}
         if (defaults)
           for (var opt in defaults)
               msg[opt] = defaults[opt]

         if (typeof o == 'function') cb = o
         else if (o)
            for (var opt in o) msg[opt] = o[opt]

         msg.html = html
         if (cb) msg.cb = cb
         this.queue.push(msg)
         this._run()
         return this
      },
      spawn: function (defaults) {
         var self = this
         return function (html, o, cb) {
            self.log.call(self,html,o,cb,defaults)
            return self
         }
      },
      create: function (o) { return new Humane(o) }
   }
   return new Humane()
})
;/*
 * JavaScript Load Image
 * https://github.com/blueimp/JavaScript-Load-Image
 *
 * Copyright 2011, Sebastian Tschan
 * https://blueimp.net
 *
 * Licensed under the MIT license:
 * https://opensource.org/licenses/MIT
 */

/* global define, URL, webkitURL, FileReader */

;(function ($) {
  'use strict'

  // Loads an image for a given File object.
  // Invokes the callback with an img or optional canvas
  // element (if supported by the browser) as parameter:
  function loadImage (file, callback, options) {
    var img = document.createElement('img')
    var url
    img.onerror = function (event) {
      return loadImage.onerror(img, event, file, callback, options)
    }
    img.onload = function (event) {
      return loadImage.onload(img, event, file, callback, options)
    }
    if (typeof file === 'string') {
      loadImage.fetchBlob(
        file,
        function (blob) {
          if (blob) {
            file = blob
            url = loadImage.createObjectURL(file)
          } else {
            url = file
            if (options && options.crossOrigin) {
              img.crossOrigin = options.crossOrigin
            }
          }
          img.src = url
        },
        options
      )
      return img
    } else if (
      loadImage.isInstanceOf('Blob', file) ||
      // Files are also Blob instances, but some browsers
      // (Firefox 3.6) support the File API but not Blobs:
      loadImage.isInstanceOf('File', file)
    ) {
      url = img._objectURL = loadImage.createObjectURL(file)
      if (url) {
        img.src = url
        return img
      }
      return loadImage.readFile(file, function (e) {
        var target = e.target
        if (target && target.result) {
          img.src = target.result
        } else if (callback) {
          callback(e)
        }
      })
    }
  }
  // The check for URL.revokeObjectURL fixes an issue with Opera 12,
  // which provides URL.createObjectURL but doesn't properly implement it:
  var urlAPI =
    ($.createObjectURL && $) ||
    ($.URL && URL.revokeObjectURL && URL) ||
    ($.webkitURL && webkitURL)

  function revokeHelper (img, options) {
    if (img._objectURL && !(options && options.noRevoke)) {
      loadImage.revokeObjectURL(img._objectURL)
      delete img._objectURL
    }
  }

  // If the callback given to this function returns a blob, it is used as image
  // source instead of the original url and overrides the file argument used in
  // the onload and onerror event callbacks:
  loadImage.fetchBlob = function (url, callback, options) {
    callback()
  }

  loadImage.isInstanceOf = function (type, obj) {
    // Cross-frame instanceof check
    return Object.prototype.toString.call(obj) === '[object ' + type + ']'
  }

  loadImage.transform = function (img, options, callback, file, data) {
    callback(img, data)
  }

  loadImage.onerror = function (img, event, file, callback, options) {
    revokeHelper(img, options)
    if (callback) {
      callback.call(img, event)
    }
  }

  loadImage.onload = function (img, event, file, callback, options) {
    revokeHelper(img, options)
    if (callback) {
      loadImage.transform(img, options, callback, file, {})
    }
  }

  loadImage.createObjectURL = function (file) {
    return urlAPI ? urlAPI.createObjectURL(file) : false
  }

  loadImage.revokeObjectURL = function (url) {
    return urlAPI ? urlAPI.revokeObjectURL(url) : false
  }

  // Loads a given File object via FileReader interface,
  // invokes the callback with the event object (load or error).
  // The result can be read via event.target.result:
  loadImage.readFile = function (file, callback, method) {
    if ($.FileReader) {
      var fileReader = new FileReader()
      fileReader.onload = fileReader.onerror = callback
      method = method || 'readAsDataURL'
      if (fileReader[method]) {
        fileReader[method](file)
        return fileReader
      }
    }
    return false
  }

  if (typeof define === 'function' && define.amd) {
    define(function () {
      return loadImage
    })
  } else if (typeof module === 'object' && module.exports) {
    module.exports = loadImage
  } else {
    $.loadImage = loadImage
  }
})((typeof window !== 'undefined' && window) || this)
;/*
 * jQuery Iframe Transport Plugin 1.7
 * https://github.com/blueimp/jQuery-File-Upload
 *
 * Copyright 2011, Sebastian Tschan
 * https://blueimp.net
 *
 * Licensed under the MIT license:
 * http://www.opensource.org/licenses/MIT
 */

/*jslint unparam: true, nomen: true */
/*global define, window, document */

(function (factory) {
    'use strict';
    if (typeof define === 'function' && define.amd) {
        // Register as an anonymous AMD module:
        define(['jquery'], factory);
    } else {
        // Browser globals:
        factory(window.jQuery);
    }
}(function ($) {
    'use strict';

    // Helper variable to create unique names for the transport iframes:
    var counter = 0;

    // The iframe transport accepts three additional options:
    // options.fileInput: a jQuery collection of file input fields
    // options.paramName: the parameter name for the file form data,
    //  overrides the name property of the file input field(s),
    //  can be a string or an array of strings.
    // options.formData: an array of objects with name and value properties,
    //  equivalent to the return data of .serializeArray(), e.g.:
    //  [{name: 'a', value: 1}, {name: 'b', value: 2}]
    $.ajaxTransport('iframe', function (options) {
        if (options.async) {
            var form,
                iframe,
                addParamChar;
            return {
                send: function (_, completeCallback) {
                    form = $('<form style="display:none;"></form>');
                    form.attr('accept-charset', options.formAcceptCharset);
                    addParamChar = /\?/.test(options.url) ? '&' : '?';
                    // XDomainRequest only supports GET and POST:
                    if (options.type === 'DELETE') {
                        options.url = options.url + addParamChar + '_method=DELETE';
                        options.type = 'POST';
                    } else if (options.type === 'PUT') {
                        options.url = options.url + addParamChar + '_method=PUT';
                        options.type = 'POST';
                    } else if (options.type === 'PATCH') {
                        options.url = options.url + addParamChar + '_method=PATCH';
                        options.type = 'POST';
                    }
                    // javascript:false as initial iframe src
                    // prevents warning popups on HTTPS in IE6.
                    // IE versions below IE8 cannot set the name property of
                    // elements that have already been added to the DOM,
                    // so we set the name along with the iframe HTML markup:
                    counter += 1;
                    iframe = $(
                        '<iframe src="javascript:false;" name="iframe-transport-' +
                            counter + '"></iframe>'
                    ).bind('load', function () {
                        var fileInputClones,
                            paramNames = $.isArray(options.paramName) ?
                                    options.paramName : [options.paramName];
                        iframe
                            .unbind('load')
                            .bind('load', function () {
                                var response;
                                // Wrap in a try/catch block to catch exceptions thrown
                                // when trying to access cross-domain iframe contents:
                                try {
                                    response = iframe.contents();
                                    // Google Chrome and Firefox do not throw an
                                    // exception when calling iframe.contents() on
                                    // cross-domain requests, so we unify the response:
                                    if (!response.length || !response[0].firstChild) {
                                        throw new Error();
                                    }
                                } catch (e) {
                                    response = undefined;
                                }
                                // The complete callback returns the
                                // iframe content document as response object:
                                completeCallback(
                                    200,
                                    'success',
                                    {'iframe': response}
                                );
                                // Fix for IE endless progress bar activity bug
                                // (happens on form submits to iframe targets):
                                $('<iframe src="javascript:false;"></iframe>')
                                    .appendTo(form);
                                window.setTimeout(function () {
                                    // Removing the form in a setTimeout call
                                    // allows Chrome's developer tools to display
                                    // the response result
                                    form.remove();
                                }, 0);
                            });
                        form
                            .prop('target', iframe.prop('name'))
                            .prop('action', options.url)
                            .prop('method', options.type);
                        if (options.formData) {
                            $.each(options.formData, function (index, field) {
                                $('<input type="hidden"/>')
                                    .prop('name', field.name)
                                    .val(field.value)
                                    .appendTo(form);
                            });
                        }
                        if (options.fileInput && options.fileInput.length &&
                                options.type === 'POST') {
                            fileInputClones = options.fileInput.clone();
                            // Insert a clone for each file input field:
                            options.fileInput.after(function (index) {
                                return fileInputClones[index];
                            });
                            if (options.paramName) {
                                options.fileInput.each(function (index) {
                                    $(this).prop(
                                        'name',
                                        paramNames[index] || options.paramName
                                    );
                                });
                            }
                            // Appending the file input fields to the hidden form
                            // removes them from their original location:
                            form
                                .append(options.fileInput)
                                .prop('enctype', 'multipart/form-data')
                                // enctype must be set as encoding for IE:
                                .prop('encoding', 'multipart/form-data');
                        }
                        form.submit();
                        // Insert the file input fields at their original location
                        // by replacing the clones with the originals:
                        if (fileInputClones && fileInputClones.length) {
                            options.fileInput.each(function (index, input) {
                                var clone = $(fileInputClones[index]);
                                $(input).prop('name', clone.prop('name'));
                                clone.replaceWith(input);
                            });
                        }
                    });
                    form.append(iframe).appendTo(document.body);
                },
                abort: function () {
                    if (iframe) {
                        // javascript:false as iframe src aborts the request
                        // and prevents warning popups on HTTPS in IE6.
                        // concat is used to avoid the "Script URL" JSLint error:
                        iframe
                            .unbind('load')
                            .prop('src', 'javascript'.concat(':false;'));
                    }
                    if (form) {
                        form.remove();
                    }
                }
            };
        }
    });

    // The iframe transport returns the iframe content document as response.
    // The following adds converters from iframe to text, json, html, xml
    // and script.
    // Please note that the Content-Type for JSON responses has to be text/plain
    // or text/html, if the browser doesn't include application/json in the
    // Accept header, else IE will show a download dialog.
    // The Content-Type for XML responses on the other hand has to be always
    // application/xml or text/xml, so IE properly parses the XML response.
    // See also
    // https://github.com/blueimp/jQuery-File-Upload/wiki/Setup#content-type-negotiation
    $.ajaxSetup({
        converters: {
            'iframe text': function (iframe) {
                return iframe && $(iframe[0].body).text();
            },
            'iframe json': function (iframe) {
                return iframe && $.parseJSON($(iframe[0].body).text());
            },
            'iframe html': function (iframe) {
                return iframe && $(iframe[0].body).html();
            },
            'iframe xml': function (iframe) {
                var xmlDoc = iframe && iframe[0];
                return xmlDoc && $.isXMLDoc(xmlDoc) ? xmlDoc :
                        $.parseXML((xmlDoc.XMLDocument && xmlDoc.XMLDocument.xml) ||
                            $(xmlDoc.body).html());
            },
            'iframe script': function (iframe) {
                return iframe && $.globalEval($(iframe[0].body).text());
            }
        }
    });

}));
;/*
 * jQuery File Upload Plugin 5.31.6
 * https://github.com/blueimp/jQuery-File-Upload
 *
 * Copyright 2010, Sebastian Tschan
 * https://blueimp.net
 *
 * Licensed under the MIT license:
 * http://www.opensource.org/licenses/MIT
 */

/*jslint nomen: true, unparam: true, regexp: true */
/*global define, window, document, location, File, Blob, FormData */

(function (factory) {
    'use strict';
    if (typeof define === 'function' && define.amd) {
        // Register as an anonymous AMD module:
        define([
            'jquery',
            'jquery.ui.widget'
        ], factory);
    } else {
        // Browser globals:
        factory(window.jQuery);
    }
}(function ($) {
    'use strict';

    // The FileReader API is not actually used, but works as feature detection,
    // as e.g. Safari supports XHR file uploads via the FormData API,
    // but not non-multipart XHR file uploads:
    $.support.xhrFileUpload = !!(window.XMLHttpRequestUpload && window.FileReader);
    $.support.xhrFormDataFileUpload = !!window.FormData;

    // Detect support for Blob slicing (required for chunked uploads):
    $.support.blobSlice = window.Blob && (Blob.prototype.slice ||
        Blob.prototype.webkitSlice || Blob.prototype.mozSlice);

    // The fileupload widget listens for change events on file input fields defined
    // via fileInput setting and paste or drop events of the given dropZone.
    // In addition to the default jQuery Widget methods, the fileupload widget
    // exposes the "add" and "send" methods, to add or directly send files using
    // the fileupload API.
    // By default, files added via file input selection, paste, drag & drop or
    // "add" method are uploaded immediately, but it is possible to override
    // the "add" callback option to queue file uploads.
    $.widget('blueimp.fileupload', {

        options: {
            // The drop target element(s), by the default the complete document.
            // Set to null to disable drag & drop support:
            dropZone: $(document),
            // The paste target element(s), by the default the complete document.
            // Set to null to disable paste support:
            pasteZone: $(document),
            // The file input field(s), that are listened to for change events.
            // If undefined, it is set to the file input fields inside
            // of the widget element on plugin initialization.
            // Set to null to disable the change listener.
            fileInput: undefined,
            // By default, the file input field is replaced with a clone after
            // each input field change event. This is required for iframe transport
            // queues and allows change events to be fired for the same file
            // selection, but can be disabled by setting the following option to false:
            replaceFileInput: true,
            // The parameter name for the file form data (the request argument name).
            // If undefined or empty, the name property of the file input field is
            // used, or "files[]" if the file input name property is also empty,
            // can be a string or an array of strings:
            paramName: undefined,
            // By default, each file of a selection is uploaded using an individual
            // request for XHR type uploads. Set to false to upload file
            // selections in one request each:
            singleFileUploads: true,
            // To limit the number of files uploaded with one XHR request,
            // set the following option to an integer greater than 0:
            limitMultiFileUploads: undefined,
            // Set the following option to true to issue all file upload requests
            // in a sequential order:
            sequentialUploads: false,
            // To limit the number of concurrent uploads,
            // set the following option to an integer greater than 0:
            limitConcurrentUploads: undefined,
            // Set the following option to true to force iframe transport uploads:
            forceIframeTransport: false,
            // Set the following option to the location of a redirect url on the
            // origin server, for cross-domain iframe transport uploads:
            redirect: undefined,
            // The parameter name for the redirect url, sent as part of the form
            // data and set to 'redirect' if this option is empty:
            redirectParamName: undefined,
            // Set the following option to the location of a postMessage window,
            // to enable postMessage transport uploads:
            postMessage: undefined,
            // By default, XHR file uploads are sent as multipart/form-data.
            // The iframe transport is always using multipart/form-data.
            // Set to false to enable non-multipart XHR uploads:
            multipart: true,
            // To upload large files in smaller chunks, set the following option
            // to a preferred maximum chunk size. If set to 0, null or undefined,
            // or the browser does not support the required Blob API, files will
            // be uploaded as a whole.
            maxChunkSize: undefined,
            // When a non-multipart upload or a chunked multipart upload has been
            // aborted, this option can be used to resume the upload by setting
            // it to the size of the already uploaded bytes. This option is most
            // useful when modifying the options object inside of the "add" or
            // "send" callbacks, as the options are cloned for each file upload.
            uploadedBytes: undefined,
            // By default, failed (abort or error) file uploads are removed from the
            // global progress calculation. Set the following option to false to
            // prevent recalculating the global progress data:
            recalculateProgress: true,
            // Interval in milliseconds to calculate and trigger progress events:
            progressInterval: 100,
            // Interval in milliseconds to calculate progress bitrate:
            bitrateInterval: 500,
            // By default, uploads are started automatically when adding files:
            autoUpload: true,

            // Error and info messages:
            messages: {
                uploadedBytes: 'Uploaded bytes exceed file size'
            },

            // Translation function, gets the message key to be translated
            // and an object with context specific data as arguments:
            i18n: function (message, context) {
                message = this.messages[message] || message.toString();
                if (context) {
                    $.each(context, function (key, value) {
                        message = message.replace('{' + key + '}', value);
                    });
                }
                return message;
            },

            // Additional form data to be sent along with the file uploads can be set
            // using this option, which accepts an array of objects with name and
            // value properties, a function returning such an array, a FormData
            // object (for XHR file uploads), or a simple object.
            // The form of the first fileInput is given as parameter to the function:
            formData: function (form) {
                return form.serializeArray();
            },

            // The add callback is invoked as soon as files are added to the fileupload
            // widget (via file input selection, drag & drop, paste or add API call).
            // If the singleFileUploads option is enabled, this callback will be
            // called once for each file in the selection for XHR file uploads, else
            // once for each file selection.
            //
            // The upload starts when the submit method is invoked on the data parameter.
            // The data object contains a files property holding the added files
            // and allows you to override plugin options as well as define ajax settings.
            //
            // Listeners for this callback can also be bound the following way:
            // .bind('fileuploadadd', func);
            //
            // data.submit() returns a Promise object and allows to attach additional
            // handlers using jQuery's Deferred callbacks:
            // data.submit().done(func).fail(func).always(func);
            add: function (e, data) {
                if (data.autoUpload || (data.autoUpload !== false &&
                        $(this).fileupload('option', 'autoUpload'))) {
                    data.process().done(function () {
                        data.submit();
                    });
                }
            },

            // Other callbacks:

            // Callback for the submit event of each file upload:
            // submit: function (e, data) {}, // .bind('fileuploadsubmit', func);

            // Callback for the start of each file upload request:
            // send: function (e, data) {}, // .bind('fileuploadsend', func);

            // Callback for successful uploads:
            // done: function (e, data) {}, // .bind('fileuploaddone', func);

            // Callback for failed (abort or error) uploads:
            // fail: function (e, data) {}, // .bind('fileuploadfail', func);

            // Callback for completed (success, abort or error) requests:
            // always: function (e, data) {}, // .bind('fileuploadalways', func);

            // Callback for upload progress events:
            // progress: function (e, data) {}, // .bind('fileuploadprogress', func);

            // Callback for global upload progress events:
            // progressall: function (e, data) {}, // .bind('fileuploadprogressall', func);

            // Callback for uploads start, equivalent to the global ajaxStart event:
            // start: function (e) {}, // .bind('fileuploadstart', func);

            // Callback for uploads stop, equivalent to the global ajaxStop event:
            // stop: function (e) {}, // .bind('fileuploadstop', func);

            // Callback for change events of the fileInput(s):
            // change: function (e, data) {}, // .bind('fileuploadchange', func);

            // Callback for paste events to the pasteZone(s):
            // paste: function (e, data) {}, // .bind('fileuploadpaste', func);

            // Callback for drop events of the dropZone(s):
            // drop: function (e, data) {}, // .bind('fileuploaddrop', func);

            // Callback for dragover events of the dropZone(s):
            // dragover: function (e) {}, // .bind('fileuploaddragover', func);

            // Callback for the start of each chunk upload request:
            // chunksend: function (e, data) {}, // .bind('fileuploadchunksend', func);

            // Callback for successful chunk uploads:
            // chunkdone: function (e, data) {}, // .bind('fileuploadchunkdone', func);

            // Callback for failed (abort or error) chunk uploads:
            // chunkfail: function (e, data) {}, // .bind('fileuploadchunkfail', func);

            // Callback for completed (success, abort or error) chunk upload requests:
            // chunkalways: function (e, data) {}, // .bind('fileuploadchunkalways', func);

            // The plugin options are used as settings object for the ajax calls.
            // The following are jQuery ajax settings required for the file uploads:
            processData: false,
            contentType: false,
            cache: false
        },

        // A list of options that require reinitializing event listeners and/or
        // special initialization code:
        _specialOptions: [
            'fileInput',
            'dropZone',
            'pasteZone',
            'multipart',
            'forceIframeTransport'
        ],

        _blobSlice: $.support.blobSlice && function () {
            var slice = this.slice || this.webkitSlice || this.mozSlice;
            return slice.apply(this, arguments);
        },

        _BitrateTimer: function () {
            this.timestamp = ((Date.now) ? Date.now() : (new Date()).getTime());
            this.loaded = 0;
            this.bitrate = 0;
            this.getBitrate = function (now, loaded, interval) {
                var timeDiff = now - this.timestamp;
                if (!this.bitrate || !interval || timeDiff > interval) {
                    this.bitrate = (loaded - this.loaded) * (1000 / timeDiff) * 8;
                    this.loaded = loaded;
                    this.timestamp = now;
                }
                return this.bitrate;
            };
        },

        _isXHRUpload: function (options) {
            return !options.forceIframeTransport &&
                ((!options.multipart && $.support.xhrFileUpload) ||
                $.support.xhrFormDataFileUpload);
        },

        _getFormData: function (options) {
            var formData;
            if (typeof options.formData === 'function') {
                return options.formData(options.form);
            }
            if ($.isArray(options.formData)) {
                return options.formData;
            }
            if ($.type(options.formData) === 'object') {
                formData = [];
                $.each(options.formData, function (name, value) {
                    formData.push({name: name, value: value});
                });
                return formData;
            }
            return [];
        },

        _getTotal: function (files) {
            var total = 0;
            $.each(files, function (index, file) {
                total += file.size || 1;
            });
            return total;
        },

        _initProgressObject: function (obj) {
            var progress = {
                loaded: 0,
                total: 0,
                bitrate: 0
            };
            if (obj._progress) {
                $.extend(obj._progress, progress);
            } else {
                obj._progress = progress;
            }
        },

        _initResponseObject: function (obj) {
            var prop;
            if (obj._response) {
                for (prop in obj._response) {
                    if (obj._response.hasOwnProperty(prop)) {
                        delete obj._response[prop];
                    }
                }
            } else {
                obj._response = {};
            }
        },

        _onProgress: function (e, data) {
            if (e.lengthComputable) {
                var now = ((Date.now) ? Date.now() : (new Date()).getTime()),
                    loaded;
                if (data._time && data.progressInterval &&
                        (now - data._time < data.progressInterval) &&
                        e.loaded !== e.total) {
                    return;
                }
                data._time = now;
                loaded = Math.floor(
                    e.loaded / e.total * (data.chunkSize || data._progress.total)
                ) + (data.uploadedBytes || 0);
                // Add the difference from the previously loaded state
                // to the global loaded counter:
                this._progress.loaded += (loaded - data._progress.loaded);
                this._progress.bitrate = this._bitrateTimer.getBitrate(
                    now,
                    this._progress.loaded,
                    data.bitrateInterval
                );
                data._progress.loaded = data.loaded = loaded;
                data._progress.bitrate = data.bitrate = data._bitrateTimer.getBitrate(
                    now,
                    loaded,
                    data.bitrateInterval
                );
                // Trigger a custom progress event with a total data property set
                // to the file size(s) of the current upload and a loaded data
                // property calculated accordingly:
                this._trigger('progress', e, data);
                // Trigger a global progress event for all current file uploads,
                // including ajax calls queued for sequential file uploads:
                this._trigger('progressall', e, this._progress);
            }
        },

        _initProgressListener: function (options) {
            var that = this,
                xhr = options.xhr ? options.xhr() : $.ajaxSettings.xhr();
            // Accesss to the native XHR object is required to add event listeners
            // for the upload progress event:
            if (xhr.upload) {
                $(xhr.upload).bind('progress', function (e) {
                    var oe = e.originalEvent;
                    // Make sure the progress event properties get copied over:
                    e.lengthComputable = oe.lengthComputable;
                    e.loaded = oe.loaded;
                    e.total = oe.total;
                    that._onProgress(e, options);
                });
                options.xhr = function () {
                    return xhr;
                };
            }
        },

        _isInstanceOf: function (type, obj) {
            // Cross-frame instanceof check
            return Object.prototype.toString.call(obj) === '[object ' + type + ']';
        },

        _initXHRData: function (options) {
            var that = this,
                formData,
                file = options.files[0],
                // Ignore non-multipart setting if not supported:
                multipart = options.multipart || !$.support.xhrFileUpload,
                paramName = options.paramName[0];
            options.headers = options.headers || {};
            if (options.contentRange) {
                options.headers['Content-Range'] = options.contentRange;
            }
            if (!multipart || options.blob || !this._isInstanceOf('File', file)) {
                options.headers['Content-Disposition'] = 'attachment; filename="' +
                    encodeURI(file.name) + '"';
            }
            if (!multipart) {
                options.contentType = file.type;
                options.data = options.blob || file;
            } else if ($.support.xhrFormDataFileUpload) {
                if (options.postMessage) {
                    // window.postMessage does not allow sending FormData
                    // objects, so we just add the File/Blob objects to
                    // the formData array and let the postMessage window
                    // create the FormData object out of this array:
                    formData = this._getFormData(options);
                    if (options.blob) {
                        formData.push({
                            name: paramName,
                            value: options.blob
                        });
                    } else {
                        $.each(options.files, function (index, file) {
                            formData.push({
                                name: options.paramName[index] || paramName,
                                value: file
                            });
                        });
                    }
                } else {
                    if (that._isInstanceOf('FormData', options.formData)) {
                        formData = options.formData;
                    } else {
                        formData = new FormData();
                        $.each(this._getFormData(options), function (index, field) {
                            formData.append(field.name, field.value);
                        });
                    }
                    if (options.blob) {
                        formData.append(paramName, options.blob, file.name);
                    } else {
                        $.each(options.files, function (index, file) {
                            // This check allows the tests to run with
                            // dummy objects:
                            if (that._isInstanceOf('File', file) ||
                                    that._isInstanceOf('Blob', file)) {
                                formData.append(
                                    options.paramName[index] || paramName,
                                    file,
                                    file.name
                                );
                            }
                        });
                    }
                }
                options.data = formData;
            }
            // Blob reference is not needed anymore, free memory:
            options.blob = null;
        },

        _initIframeSettings: function (options) {
            var targetHost = $('<a></a>').prop('href', options.url).prop('host');
            // Setting the dataType to iframe enables the iframe transport:
            options.dataType = 'iframe ' + (options.dataType || '');
            // The iframe transport accepts a serialized array as form data:
            options.formData = this._getFormData(options);
            // Add redirect url to form data on cross-domain uploads:
            if (options.redirect && targetHost && targetHost !== location.host) {
                options.formData.push({
                    name: options.redirectParamName || 'redirect',
                    value: options.redirect
                });
            }
        },

        _initDataSettings: function (options) {
            if (this._isXHRUpload(options)) {
                if (!this._chunkedUpload(options, true)) {
                    if (!options.data) {
                        this._initXHRData(options);
                    }
                    this._initProgressListener(options);
                }
                if (options.postMessage) {
                    // Setting the dataType to postmessage enables the
                    // postMessage transport:
                    options.dataType = 'postmessage ' + (options.dataType || '');
                }
            } else {
                this._initIframeSettings(options);
            }
        },

        _getParamName: function (options) {
            var fileInput = $(options.fileInput),
                paramName = options.paramName;
            if (!paramName) {
                paramName = [];
                fileInput.each(function () {
                    var input = $(this),
                        name = input.prop('name') || 'files[]',
                        i = (input.prop('files') || [1]).length;
                    while (i) {
                        paramName.push(name);
                        i -= 1;
                    }
                });
                if (!paramName.length) {
                    paramName = [fileInput.prop('name') || 'files[]'];
                }
            } else if (!$.isArray(paramName)) {
                paramName = [paramName];
            }
            return paramName;
        },

        _initFormSettings: function (options) {
            // Retrieve missing options from the input field and the
            // associated form, if available:
            if (!options.form || !options.form.length) {
                options.form = $(options.fileInput.prop('form'));
                // If the given file input doesn't have an associated form,
                // use the default widget file input's form:
                if (!options.form.length) {
                    options.form = $(this.options.fileInput.prop('form'));
                }
            }
            options.paramName = this._getParamName(options);
            if (!options.url) {
                options.url = options.form.prop('action') || location.href;
            }
            // The HTTP request method must be "POST" or "PUT":
            options.type = (options.type || options.form.prop('method') || '')
                .toUpperCase();
            if (options.type !== 'POST' && options.type !== 'PUT' &&
                    options.type !== 'PATCH') {
                options.type = 'POST';
            }
            if (!options.formAcceptCharset) {
                options.formAcceptCharset = options.form.attr('accept-charset');
            }
        },

        _getAJAXSettings: function (data) {
            var options = $.extend({}, this.options, data);
            this._initFormSettings(options);
            this._initDataSettings(options);
            return options;
        },

        // jQuery 1.6 doesn't provide .state(),
        // while jQuery 1.8+ removed .isRejected() and .isResolved():
        _getDeferredState: function (deferred) {
            if (deferred.state) {
                return deferred.state();
            }
            if (deferred.isResolved()) {
                return 'resolved';
            }
            if (deferred.isRejected()) {
                return 'rejected';
            }
            return 'pending';
        },

        // Maps jqXHR callbacks to the equivalent
        // methods of the given Promise object:
        _enhancePromise: function (promise) {
            promise.success = promise.done;
            promise.error = promise.fail;
            promise.complete = promise.always;
            return promise;
        },

        // Creates and returns a Promise object enhanced with
        // the jqXHR methods abort, success, error and complete:
        _getXHRPromise: function (resolveOrReject, context, args) {
            var dfd = $.Deferred(),
                promise = dfd.promise();
            context = context || this.options.context || promise;
            if (resolveOrReject === true) {
                dfd.resolveWith(context, args);
            } else if (resolveOrReject === false) {
                dfd.rejectWith(context, args);
            }
            promise.abort = dfd.promise;
            return this._enhancePromise(promise);
        },

        // Adds convenience methods to the data callback argument:
        _addConvenienceMethods: function (e, data) {
            var that = this,
                getPromise = function (data) {
                    return $.Deferred().resolveWith(that, [data]).promise();
                };
            data.process = function (resolveFunc, rejectFunc) {
                if (resolveFunc || rejectFunc) {
                    data._processQueue = this._processQueue =
                        (this._processQueue || getPromise(this))
                            .pipe(resolveFunc, rejectFunc);
                }
                return this._processQueue || getPromise(this);
            };
            data.submit = function () {
                if (this.state() !== 'pending') {
                    data.jqXHR = this.jqXHR =
                        (that._trigger('submit', e, this) !== false) &&
                        that._onSend(e, this);
                }
                return this.jqXHR || that._getXHRPromise();
            };
            data.abort = function () {
                if (this.jqXHR) {
                    return this.jqXHR.abort();
                }
                return that._getXHRPromise();
            };
            data.state = function () {
                if (this.jqXHR) {
                    return that._getDeferredState(this.jqXHR);
                }
                if (this._processQueue) {
                    return that._getDeferredState(this._processQueue);
                }
            };
            data.progress = function () {
                return this._progress;
            };
            data.response = function () {
                return this._response;
            };
        },

        // Parses the Range header from the server response
        // and returns the uploaded bytes:
        _getUploadedBytes: function (jqXHR) {
            var range = jqXHR.getResponseHeader('Range'),
                parts = range && range.split('-'),
                upperBytesPos = parts && parts.length > 1 &&
                    parseInt(parts[1], 10);
            return upperBytesPos && upperBytesPos + 1;
        },

        // Uploads a file in multiple, sequential requests
        // by splitting the file up in multiple blob chunks.
        // If the second parameter is true, only tests if the file
        // should be uploaded in chunks, but does not invoke any
        // upload requests:
        _chunkedUpload: function (options, testOnly) {
            options.uploadedBytes = options.uploadedBytes || 0;
            var that = this,
                file = options.files[0],
                fs = file.size,
                ub = options.uploadedBytes,
                mcs = options.maxChunkSize || fs,
                slice = this._blobSlice,
                dfd = $.Deferred(),
                promise = dfd.promise(),
                jqXHR,
                upload;
            if (!(this._isXHRUpload(options) && slice && (ub || mcs < fs)) ||
                    options.data) {
                return false;
            }
            if (testOnly) {
                return true;
            }
            if (ub >= fs) {
                file.error = options.i18n('uploadedBytes');
                return this._getXHRPromise(
                    false,
                    options.context,
                    [null, 'error', file.error]
                );
            }
            // The chunk upload method:
            upload = function () {
                // Clone the options object for each chunk upload:
                var o = $.extend({}, options),
                    currentLoaded = o._progress.loaded;
                o.blob = slice.call(
                    file,
                    ub,
                    ub + mcs,
                    file.type
                );
                // Store the current chunk size, as the blob itself
                // will be dereferenced after data processing:
                o.chunkSize = o.blob.size;
                // Expose the chunk bytes position range:
                o.contentRange = 'bytes ' + ub + '-' +
                    (ub + o.chunkSize - 1) + '/' + fs;
                // Process the upload data (the blob and potential form data):
                that._initXHRData(o);
                // Add progress listeners for this chunk upload:
                that._initProgressListener(o);
                jqXHR = ((that._trigger('chunksend', null, o) !== false && $.ajax(o)) ||
                        that._getXHRPromise(false, o.context))
                    .done(function (result, textStatus, jqXHR) {
                        ub = that._getUploadedBytes(jqXHR) ||
                            (ub + o.chunkSize);
                        // Create a progress event if no final progress event
                        // with loaded equaling total has been triggered
                        // for this chunk:
                        if (currentLoaded + o.chunkSize - o._progress.loaded) {
                            that._onProgress($.Event('progress', {
                                lengthComputable: true,
                                loaded: ub - o.uploadedBytes,
                                total: ub - o.uploadedBytes
                            }), o);
                        }
                        options.uploadedBytes = o.uploadedBytes = ub;
                        o.result = result;
                        o.textStatus = textStatus;
                        o.jqXHR = jqXHR;
                        that._trigger('chunkdone', null, o);
                        that._trigger('chunkalways', null, o);
                        if (ub < fs) {
                            // File upload not yet complete,
                            // continue with the next chunk:
                            upload();
                        } else {
                            dfd.resolveWith(
                                o.context,
                                [result, textStatus, jqXHR]
                            );
                        }
                    })
                    .fail(function (jqXHR, textStatus, errorThrown) {
                        o.jqXHR = jqXHR;
                        o.textStatus = textStatus;
                        o.errorThrown = errorThrown;
                        that._trigger('chunkfail', null, o);
                        that._trigger('chunkalways', null, o);
                        dfd.rejectWith(
                            o.context,
                            [jqXHR, textStatus, errorThrown]
                        );
                    });
            };
            this._enhancePromise(promise);
            promise.abort = function () {
                return jqXHR.abort();
            };
            upload();
            return promise;
        },

        _beforeSend: function (e, data) {
            if (this._active === 0) {
                // the start callback is triggered when an upload starts
                // and no other uploads are currently running,
                // equivalent to the global ajaxStart event:
                this._trigger('start');
                // Set timer for global bitrate progress calculation:
                this._bitrateTimer = new this._BitrateTimer();
                // Reset the global progress values:
                this._progress.loaded = this._progress.total = 0;
                this._progress.bitrate = 0;
            }
            // Make sure the container objects for the .response() and
            // .progress() methods on the data object are available
            // and reset to their initial state:
            this._initResponseObject(data);
            this._initProgressObject(data);
            data._progress.loaded = data.loaded = data.uploadedBytes || 0;
            data._progress.total = data.total = this._getTotal(data.files) || 1;
            data._progress.bitrate = data.bitrate = 0;
            this._active += 1;
            // Initialize the global progress values:
            this._progress.loaded += data.loaded;
            this._progress.total += data.total;
        },

        _onDone: function (result, textStatus, jqXHR, options) {
            var total = options._progress.total,
                response = options._response;
            if (options._progress.loaded < total) {
                // Create a progress event if no final progress event
                // with loaded equaling total has been triggered:
                this._onProgress($.Event('progress', {
                    lengthComputable: true,
                    loaded: total,
                    total: total
                }), options);
            }
            response.result = options.result = result;
            response.textStatus = options.textStatus = textStatus;
            response.jqXHR = options.jqXHR = jqXHR;
            this._trigger('done', null, options);
        },

        _onFail: function (jqXHR, textStatus, errorThrown, options) {
            var response = options._response;
            if (options.recalculateProgress) {
                // Remove the failed (error or abort) file upload from
                // the global progress calculation:
                this._progress.loaded -= options._progress.loaded;
                this._progress.total -= options._progress.total;
            }
            response.jqXHR = options.jqXHR = jqXHR;
            response.textStatus = options.textStatus = textStatus;
            response.errorThrown = options.errorThrown = errorThrown;
            this._trigger('fail', null, options);
        },

        _onAlways: function (jqXHRorResult, textStatus, jqXHRorError, options) {
            // jqXHRorResult, textStatus and jqXHRorError are added to the
            // options object via done and fail callbacks
            this._trigger('always', null, options);
        },

        _onSend: function (e, data) {
            if (!data.submit) {
                this._addConvenienceMethods(e, data);
            }
            var that = this,
                jqXHR,
                aborted,
                slot,
                pipe,
                options = that._getAJAXSettings(data),
                send = function () {
                    that._sending += 1;
                    // Set timer for bitrate progress calculation:
                    options._bitrateTimer = new that._BitrateTimer();
                    jqXHR = jqXHR || (
                        ((aborted || that._trigger('send', e, options) === false) &&
                        that._getXHRPromise(false, options.context, aborted)) ||
                        that._chunkedUpload(options) || $.ajax(options)
                    ).done(function (result, textStatus, jqXHR) {
                        that._onDone(result, textStatus, jqXHR, options);
                    }).fail(function (jqXHR, textStatus, errorThrown) {
                        that._onFail(jqXHR, textStatus, errorThrown, options);
                    }).always(function (jqXHRorResult, textStatus, jqXHRorError) {
                        that._onAlways(
                            jqXHRorResult,
                            textStatus,
                            jqXHRorError,
                            options
                        );
                        that._sending -= 1;
                        that._active -= 1;
                        if (options.limitConcurrentUploads &&
                                options.limitConcurrentUploads > that._sending) {
                            // Start the next queued upload,
                            // that has not been aborted:
                            var nextSlot = that._slots.shift();
                            while (nextSlot) {
                                if (that._getDeferredState(nextSlot) === 'pending') {
                                    nextSlot.resolve();
                                    break;
                                }
                                nextSlot = that._slots.shift();
                            }
                        }
                        if (that._active === 0) {
                            // The stop callback is triggered when all uploads have
                            // been completed, equivalent to the global ajaxStop event:
                            that._trigger('stop');
                        }
                    });
                    return jqXHR;
                };
            this._beforeSend(e, options);
            if (this.options.sequentialUploads ||
                    (this.options.limitConcurrentUploads &&
                    this.options.limitConcurrentUploads <= this._sending)) {
                if (this.options.limitConcurrentUploads > 1) {
                    slot = $.Deferred();
                    this._slots.push(slot);
                    pipe = slot.pipe(send);
                } else {
                    this._sequence = this._sequence.pipe(send, send);
                    pipe = this._sequence;
                }
                // Return the piped Promise object, enhanced with an abort method,
                // which is delegated to the jqXHR object of the current upload,
                // and jqXHR callbacks mapped to the equivalent Promise methods:
                pipe.abort = function () {
                    aborted = [undefined, 'abort', 'abort'];
                    if (!jqXHR) {
                        if (slot) {
                            slot.rejectWith(options.context, aborted);
                        }
                        return send();
                    }
                    return jqXHR.abort();
                };
                return this._enhancePromise(pipe);
            }
            return send();
        },

        _onAdd: function (e, data) {
            var that = this,
                result = true,
                options = $.extend({}, this.options, data),
                limit = options.limitMultiFileUploads,
                paramName = this._getParamName(options),
                paramNameSet,
                paramNameSlice,
                fileSet,
                i;
            if (!(options.singleFileUploads || limit) ||
                    !this._isXHRUpload(options)) {
                fileSet = [data.files];
                paramNameSet = [paramName];
            } else if (!options.singleFileUploads && limit) {
                fileSet = [];
                paramNameSet = [];
                for (i = 0; i < data.files.length; i += limit) {
                    fileSet.push(data.files.slice(i, i + limit));
                    paramNameSlice = paramName.slice(i, i + limit);
                    if (!paramNameSlice.length) {
                        paramNameSlice = paramName;
                    }
                    paramNameSet.push(paramNameSlice);
                }
            } else {
                paramNameSet = paramName;
            }
            data.originalFiles = data.files;
            $.each(fileSet || data.files, function (index, element) {
                var newData = $.extend({}, data);
                newData.files = fileSet ? element : [element];
                newData.paramName = paramNameSet[index];
                that._initResponseObject(newData);
                that._initProgressObject(newData);
                that._addConvenienceMethods(e, newData);
                result = that._trigger('add', e, newData);
                return result;
            });
            return result;
        },

        _replaceFileInput: function (input) {
            var inputClone = input.clone(true);
            $('<form></form>').append(inputClone)[0].reset();
            // Detaching allows to insert the fileInput on another form
            // without loosing the file input value:
            input.after(inputClone).detach();
            // Avoid memory leaks with the detached file input:
            $.cleanData(input.unbind('remove'));
            // Replace the original file input element in the fileInput
            // elements set with the clone, which has been copied including
            // event handlers:
            this.options.fileInput = this.options.fileInput.map(function (i, el) {
                if (el === input[0]) {
                    return inputClone[0];
                }
                return el;
            });
            // If the widget has been initialized on the file input itself,
            // override this.element with the file input clone:
            if (input[0] === this.element[0]) {
                this.element = inputClone;
            }
        },

        _handleFileTreeEntry: function (entry, path) {
            var that = this,
                dfd = $.Deferred(),
                errorHandler = function (e) {
                    if (e && !e.entry) {
                        e.entry = entry;
                    }
                    // Since $.when returns immediately if one
                    // Deferred is rejected, we use resolve instead.
                    // This allows valid files and invalid items
                    // to be returned together in one set:
                    dfd.resolve([e]);
                },
                dirReader;
            path = path || '';
            if (entry.isFile) {
                if (entry._file) {
                    // Workaround for Chrome bug #149735
                    entry._file.relativePath = path;
                    dfd.resolve(entry._file);
                } else {
                    entry.file(function (file) {
                        file.relativePath = path;
                        dfd.resolve(file);
                    }, errorHandler);
                }
            } else if (entry.isDirectory) {
                dirReader = entry.createReader();
                dirReader.readEntries(function (entries) {
                    that._handleFileTreeEntries(
                        entries,
                        path + entry.name + '/'
                    ).done(function (files) {
                        dfd.resolve(files);
                    }).fail(errorHandler);
                }, errorHandler);
            } else {
                // Return an empy list for file system items
                // other than files or directories:
                dfd.resolve([]);
            }
            return dfd.promise();
        },

        _handleFileTreeEntries: function (entries, path) {
            var that = this;
            return $.when.apply(
                $,
                $.map(entries, function (entry) {
                    return that._handleFileTreeEntry(entry, path);
                })
            ).pipe(function () {
                return Array.prototype.concat.apply(
                    [],
                    arguments
                );
            });
        },

        _getDroppedFiles: function (dataTransfer) {
            dataTransfer = dataTransfer || {};
            var items = dataTransfer.items;
            if (items && items.length && (items[0].webkitGetAsEntry ||
                    items[0].getAsEntry)) {
                return this._handleFileTreeEntries(
                    $.map(items, function (item) {
                        var entry;
                        if (item.webkitGetAsEntry) {
                            entry = item.webkitGetAsEntry();
                            if (entry) {
                                // Workaround for Chrome bug #149735:
                                entry._file = item.getAsFile();
                            }
                            return entry;
                        }
                        return item.getAsEntry();
                    })
                );
            }
            return $.Deferred().resolve(
                $.makeArray(dataTransfer.files)
            ).promise();
        },

        _getSingleFileInputFiles: function (fileInput) {
            fileInput = $(fileInput);
            var entries = fileInput.prop('webkitEntries') ||
                    fileInput.prop('entries'),
                files,
                value;
            if (entries && entries.length) {
                return this._handleFileTreeEntries(entries);
            }
            files = $.makeArray(fileInput.prop('files'));
            if (!files.length) {
                value = fileInput.prop('value');
                if (!value) {
                    return $.Deferred().resolve([]).promise();
                }
                // If the files property is not available, the browser does not
                // support the File API and we add a pseudo File object with
                // the input value as name with path information removed:
                files = [{name: value.replace(/^.*\\/, '')}];
            } else if (files[0].name === undefined && files[0].fileName) {
                // File normalization for Safari 4 and Firefox 3:
                $.each(files, function (index, file) {
                    file.name = file.fileName;
                    file.size = file.fileSize;
                });
            }
            return $.Deferred().resolve(files).promise();
        },

        _getFileInputFiles: function (fileInput) {
            if (!(fileInput instanceof $) || fileInput.length === 1) {
                return this._getSingleFileInputFiles(fileInput);
            }
            return $.when.apply(
                $,
                $.map(fileInput, this._getSingleFileInputFiles)
            ).pipe(function () {
                return Array.prototype.concat.apply(
                    [],
                    arguments
                );
            });
        },

        _onChange: function (e) {
            var that = this,
                data = {
                    fileInput: $(e.target),
                    form: $(e.target.form)
                };
            this._getFileInputFiles(data.fileInput).always(function (files) {
                data.files = files;
                if (that.options.replaceFileInput) {
                    that._replaceFileInput(data.fileInput);
                }
                if (that._trigger('change', e, data) !== false) {
                    that._onAdd(e, data);
                }
            });
        },

        _onPaste: function (e) {
            var items = e.originalEvent && e.originalEvent.clipboardData &&
                    e.originalEvent.clipboardData.items,
                data = {files: []};
            if (items && items.length) {
                $.each(items, function (index, item) {
                    var file = item.getAsFile && item.getAsFile();
                    if (file) {
                        data.files.push(file);
                    }
                });
                if (this._trigger('paste', e, data) === false ||
                        this._onAdd(e, data) === false) {
                    return false;
                }
            }
        },

        _onDrop: function (e) {
            e.dataTransfer = e.originalEvent && e.originalEvent.dataTransfer;
            var that = this,
                dataTransfer = e.dataTransfer,
                data = {};
            if (dataTransfer && dataTransfer.files && dataTransfer.files.length) {
                e.preventDefault();
                this._getDroppedFiles(dataTransfer).always(function (files) {
                    data.files = files;
                    if (that._trigger('drop', e, data) !== false) {
                        that._onAdd(e, data);
                    }
                });
            }
        },

        _onDragOver: function (e) {
            e.dataTransfer = e.originalEvent && e.originalEvent.dataTransfer;
            var dataTransfer = e.dataTransfer;
            if (dataTransfer) {
                if (this._trigger('dragover', e) === false) {
                    return false;
                }
                if ($.inArray('Files', dataTransfer.types) !== -1) {
                    dataTransfer.dropEffect = 'copy';
                    e.preventDefault();
                }
            }
        },

        _initEventHandlers: function () {
            if (this._isXHRUpload(this.options)) {
                this._on(this.options.dropZone, {
                    dragover: this._onDragOver,
                    drop: this._onDrop
                });
                this._on(this.options.pasteZone, {
                    paste: this._onPaste
                });
            }
            this._on(this.options.fileInput, {
                change: this._onChange
            });
        },

        _destroyEventHandlers: function () {
            this._off(this.options.dropZone, 'dragover drop');
            this._off(this.options.pasteZone, 'paste');
            this._off(this.options.fileInput, 'change');
        },

        _setOption: function (key, value) {
            var reinit = $.inArray(key, this._specialOptions) !== -1;
            if (reinit) {
                this._destroyEventHandlers();
            }
            this._super(key, value);
            if (reinit) {
                this._initSpecialOptions();
                this._initEventHandlers();
            }
        },

        _initSpecialOptions: function () {
            var options = this.options;
            if (options.fileInput === undefined) {
                options.fileInput = this.element.is('input[type="file"]') ?
                        this.element : this.element.find('input[type="file"]');
            } else if (!(options.fileInput instanceof $)) {
                options.fileInput = $(options.fileInput);
            }
            if (!(options.dropZone instanceof $)) {
                options.dropZone = $(options.dropZone);
            }
            if (!(options.pasteZone instanceof $)) {
                options.pasteZone = $(options.pasteZone);
            }
        },

        _getRegExp: function (str) {
            var parts = str.split('/'),
                modifiers = parts.pop();
            parts.shift();
            return new RegExp(parts.join('/'), modifiers);
        },

        _isRegExpOption: function (key, value) {
            return key !== 'url' && $.type(value) === 'string' &&
                /^\/.*\/[igm]{0,3}$/.test(value);
        },

        _initDataAttributes: function () {
            var that = this,
                options = this.options;
            // Initialize options set via HTML5 data-attributes:
            $.each(
                $(this.element[0].cloneNode(false)).data(),
                function (key, value) {
                    if (that._isRegExpOption(key, value)) {
                        value = that._getRegExp(value);
                    }
                    options[key] = value;
                }
            );
        },

        _create: function () {
            this._initDataAttributes();
            this._initSpecialOptions();
            this._slots = [];
            this._sequence = this._getXHRPromise(true);
            this._sending = this._active = 0;
            this._initProgressObject(this);
            this._initEventHandlers();
        },

        // This method is exposed to the widget API and allows to query
        // the number of active uploads:
        active: function () {
            return this._active;
        },

        // This method is exposed to the widget API and allows to query
        // the widget upload progress.
        // It returns an object with loaded, total and bitrate properties
        // for the running uploads:
        progress: function () {
            return this._progress;
        },

        // This method is exposed to the widget API and allows adding files
        // using the fileupload API. The data parameter accepts an object which
        // must have a files property and can contain additional options:
        // .fileupload('add', {files: filesList});
        add: function (data) {
            var that = this;
            if (!data || this.options.disabled) {
                return;
            }
            if (data.fileInput && !data.files) {
                this._getFileInputFiles(data.fileInput).always(function (files) {
                    data.files = files;
                    that._onAdd(null, data);
                });
            } else {
                data.files = $.makeArray(data.files);
                this._onAdd(null, data);
            }
        },

        // This method is exposed to the widget API and allows sending files
        // using the fileupload API. The data parameter accepts an object which
        // must have a files or fileInput property and can contain additional options:
        // .fileupload('send', {files: filesList});
        // The method returns a Promise object for the file upload call.
        send: function (data) {
            if (data && !this.options.disabled) {
                if (data.fileInput && !data.files) {
                    var that = this,
                        dfd = $.Deferred(),
                        promise = dfd.promise(),
                        jqXHR,
                        aborted;
                    promise.abort = function () {
                        aborted = true;
                        if (jqXHR) {
                            return jqXHR.abort();
                        }
                        dfd.reject(null, 'abort', 'abort');
                        return promise;
                    };
                    this._getFileInputFiles(data.fileInput).always(
                        function (files) {
                            if (aborted) {
                                return;
                            }
                            data.files = files;
                            jqXHR = that._onSend(null, data).then(
                                function (result, textStatus, jqXHR) {
                                    dfd.resolve(result, textStatus, jqXHR);
                                },
                                function (jqXHR, textStatus, errorThrown) {
                                    dfd.reject(jqXHR, textStatus, errorThrown);
                                }
                            );
                        }
                    );
                    return this._enhancePromise(promise);
                }
                data.files = $.makeArray(data.files);
                if (data.files.length) {
                    return this._onSend(null, data);
                }
            }
            return this._getXHRPromise(false, data && data.context);
        }

    });

}));
;(function(window, $) {
    'use strict';

    // Builds a 4 characters unique id
    var uniqId = function(prefix) {
        prefix = prefix || '';

        return prefix + ("0000" + (Math.random() * Math.pow(36, 4) << 0).toString(36)).substr(-4);
    };

    // Builds query parameters from a geoname object
    var RequestDataBuilder = function(geo, datas) {
        this.datas = datas || {};

        if (!geo instanceof GeotoCompleter) {
           throw 'You must provide an instance of GeotoCompleter';
        }

        if (geo.getOption('sort')) {
            this.datas.sort = geo.getOption('sort');
        }
        if (geo.getOption('client-ip')) {
            this.datas.sortParams = geo.getOption('sortParams');
        }
        if (geo.getOption('country')) {
            this.datas.country = geo.getOption('country');
        }
        if (geo.getOption('name')) {
            this.datas.name = geo.getOption('name');
        }
        if (geo.getOption('limit')) {
            this.datas.limit = geo.getOption('limit');
        }
    };

    RequestDataBuilder.prototype.getRequestDatas = function() {
        return this.datas;
    };

    // Handles request to the remote Geonames Server
    var RequestManager = function(server) {
        var _request = false;
        var _endpoint = server;

        this.search = function(resource, datas, errorCallback, parseresults) {
            _request = $.ajax({
                type: "GET",
                dataType: "jsonp",
                jsonpCallback: "parseresults",
                url: _endpoint + resource,
                beforeSend: function() {
                    if (_request && typeof _request.abort === 'function') {
                        _request.abort();
                    }
                },
                error: errorCallback,
                data: datas
            })
            .done(parseresults)
            .always(function() {
                _request = false;
            });
        };
    };

    var GeotoCompleter = function(el, serverEndpoint, options) {
        if (typeof $.ui === 'undefined') {
            throw 'jQuery UI must be loaded';
        }

        if($(el).data('geocompleter')) {
            return $(el).data('geocompleter');
        }

        var _serverEndpoint = serverEndpoint.substr(serverEndpoint.length - 1) === '/' ? serverEndpoint : serverEndpoint + '/';

        this._requestManager = new RequestManager(_serverEndpoint);

        this.$el = $(el);

        this._opts = $.extend({
            "name": null,
            "sort": null,
            "client-ip": null,
            "country": null,
            "limit": null,
            "init-input": true
        }, options);

        this.$input = null;

        this.$el.data('geocompleter', this);
    };

    GeotoCompleter.prototype.getOption = function(name) {
        if (!(name in this._opts)) {
            return null;
        }

        return this._opts[name];
    };

    GeotoCompleter.prototype.setOption = function(name, value) {
        if (!(name in this._opts)) {
            return this;
        }

        this._opts[name] = value;

        return this;
    };

    GeotoCompleter.prototype.getAutocompleter = function() {
        return this.$input;
    };

    GeotoCompleter.prototype.destroy = function() {
        if (this.$input) {
            this.$input.remove();
            this.$el.show();
            this.$el.val('');
            this.$el.data('geocompleter', null);
        }
    };

    GeotoCompleter.prototype.init = function() {
        if (null !== this.$input) {
            return;
        }

        var self = this;

        var updateGeonameField = function(value) {
            self.$el.val(value);
        };

        var updateCityField = function(value) {
            self.$input.val(value);
        };

        var resetGeonameField = function() {
            self.$el.val('');
        };

        var resetCityField = function() {
            self.$input.val('');
        };

        var isGeonameFieldSetted = function() {
            return self.$el.val() !== '';
        };

        var highlight = function (s, t) {
            var matcher = new RegExp("("+$.ui.autocomplete.escapeRegex(t)+")", "ig" );
            return s.replace(matcher, "<span class='ui-state-highlight'>$1</span>");
        };

        // Creates city input
        this.$input = $('<input />')
                .attr('name', uniqId(this.$el.attr('name')))
                .attr('id', uniqId(this.$el.attr('id')))
                .attr('type', 'text')
                .attr('class', this.$el.attr('class'))
                .addClass("geocompleter-input");

        // Prevents form submission when pressing ENTER
        this.$input.keypress(function(event) {
            var code = (event.keyCode ? event.keyCode : event.which);
            if(code === $.ui.keyCode.ENTER ) {
                event.preventDefault();
                return false;
            }
        });

        // On any keyup except (esc, up, down, enter) fields are desynchronised, reset geonames field
        this.$input.keyup(function(event) {
            var code = (event.keyCode ? event.keyCode : event.which);
            var unBindKeys = [
                $.ui.keyCode.ESCAPE,
                $.ui.keyCode.UP,
                $.ui.keyCode.DOWN,
                $.ui.keyCode.LEFT,
                $.ui.keyCode.RIGHT,
                $.ui.keyCode.ENTER
            ];

            if (-1 === $.inArray(code, unBindKeys)){
                resetGeonameField();
            }
        });

        this.$el.hide();
        this.$el.after(this.$input);

        // Overrides prototype to render values without autoescape, useful to highlight values
        $.ui.autocomplete.prototype._renderItem = function( ul, item) {
            return $( "<li></li>" )
              .data( "item.autocomplete", item )
              .append( $( "<a></a>" ).html( item.label ) )
              .appendTo( ul );
        };

        // Saves response content
        var responseContent;

        // Builds a jquery autocompleter
        this.$input.autocomplete({
            create: function(event) {
                if(self.$el.val() !== "" && self.getOption("init-input")) {
                    self._requestManager.search(
                        "city/" + parseInt(self.$el.val(), 10),
                        {},
                        function(jqXhr, status, error) {
                            return;
                        }, function (data) {
                            var country = data.country.name || "";
                            self.$input.val(data.name + ("" !== country ? "," + country : ""));
                        }
                    );
                }
            },
            source: function(request, response) {
                var name, country, terms = '';

                terms = request.term.split(',');

                if (terms.length === 2) {
                    country = terms.pop();
                }

                name = terms.pop();

                self.setOption('name', $.trim(name));
                self.setOption('country', $.trim(country));

                var requestDataBuilder = new RequestDataBuilder(self);

                self._requestManager.search(
                    "city",
                    requestDataBuilder.getRequestDatas(),
                    function(jqXhr, status, error) {
                        if (jqXhr.status !== 0 && jqXhr.statusText !== 'abort') {
                            response([]);
                            self.$input.trigger('geotocompleter.request.error', [jqXhr, status, error]);
                        }
                    }, function(data) {
                        response($.map(data || [], function(item) {
                            var country = country ? country : name;
                            var labelName = highlight(item.name, name);
                            var labelCountry = highlight((item.country ? item.country.name || '' : ''), country);
                            var labelRegion = highlight((item.region ? item.region.name || '' : ''), name);

                            return {
                                label:  labelName + ("" !== labelCountry ? ", " + labelCountry : "") + ("" !== labelRegion ? " <span class='region'>" + labelRegion + "</span>" : ""),
                                value: item.name + (item.country ? ", " + item.country.name : ''),
                                geonameid: item.geonameid
                            };
                        }));
                    }
                );
            },
            messages: {
                noResults: '',
                results: function() {}
            },
            response: function (event, ui) {
                responseContent = [];
                if (ui.content) {
                    responseContent = ui.content;
                    // Sets geoname id if values are re synchronized
                    if (ui.content.length > 0) {
                        var items = $.grep(ui.content, function(item) {
                            return item.value === self.$input.val() ? item : null;
                        });

                        if (items.length > 0) {
                            updateGeonameField(items[0].geonameid);
                        }
                    }
                }
            },
            select: function(event, ui) {
                if (ui.item) {
                    updateGeonameField(ui.item.geonameid);
                }
            },
            focus: function (event, ui) {
                var code = (event.keyCode ? event.keyCode : event.which);
                // Updates geoname ID only if key up and key down are pressed
                if (ui.item && -1 !== $.inArray(code, [$.ui.keyCode.DOWN, $.ui.keyCode.UP])) {
                    updateGeonameField(ui.item.geonameid);
                }
            },
            close : function (event, ui) {
                var ev = event.originalEvent;

                if ("undefined" === typeof ev) {
                    return false;
                }

                var code = (ev.keyCode ? ev.keyCode : ev.which);
                // If esc key is pressed or user leaves the input
                if ((ev.type === "keydown" && code === $.ui.keyCode.ESCAPE) || ev.type === "blur") {
                    if (isGeonameFieldSetted() && responseContent.length > 0) {
                        var geonameId = self.$el.val();
                        // Update city input according to the setted geonameId
                        var responseValues = $.grep(responseContent, function(item) {
                            return item.geonameid === geonameId ? item : null;
                        });

                        if (responseValues.length > 0) {
                            self.$input.val(responseValues[0].value);
                        }

                        return false;
                    }

                    // Resets both field as nothing is no more sychronized
                    resetGeonameField();
                    resetCityField();
                }
            }
        }).autocomplete("widget").addClass("geocompleter-menu");

        var onInit = self.getOption('onInit');

        // On Initialization callback
        if (onInit !== null && typeof onInit === 'function') {
            onInit(this.$el, this.$input);
        }
    };

    var methods = {
        init: function(options) {
            var settings = $.extend({
                server: ''
            }, options);

            if ('' === settings.server) {
                throw '"server" must be set';
            }

            return this.each(function() {
                var geocompleter = new GeotoCompleter(this, settings.server, settings);
                geocompleter.init();
            });
        },
        destroy: function() {
            return this.each(function() {
                var geocompleter = $(this).data('geocompleter');
                if (geocompleter) {
                    geocompleter.destroy();
                }
            });
        },
        autocompleter: function() {
            var args = arguments;
            return this.each(function() {
                var geocompleter = $(this).data('geocompleter');
                if (args[0] === "on" && typeof args[1] === "string" && typeof args[2] === "function") {
                    // Bind addition events
                    geocompleter.getAutocompleter().on(args[1], args[2]);
                } else {
                    $.fn.autocomplete.apply(geocompleter.getAutocompleter(),args);
                }
            });
        }
    };

    $.fn.geocompleter = function(method) {
        if (methods[method]) {
            return methods[method].apply(this, Array.prototype.slice.call(arguments, 1));
        } else if (typeof method === 'object' || !method) {
            return methods.init.apply(this, arguments);
        } else {
            $.error('Method ' + method + ' does not exist on jQuery.geocompleter');
        }
    };
})(window, jQuery);
;var publicationModule = (function () {

    var ajaxState = {
        query: null,
        isRunning: false
    };

    var curPage;
    var $answers = $('#answers');

    // refresh current view
    $answers.on('click', '.feed_reload', function (event) {
        event.preventDefault();
        fetchPublications(curPage)
    });

    // navigate to a specific feed
    $answers.on('click', '.ajax_answers', function (event) {
        event.preventDefault();
        var $this = $(this);
        var append = $this.hasClass('append');
        var noScroll = $this.hasClass('no_scroll');

        _fetchRemote($(event.currentTarget).attr('href'), {})
            .then(function (data) {
                if (!append) {
                    $answers.empty();
                    if (!noScroll) {
                        $answers.scrollTop(0);
                    }
                    $answers.append(data);

                    $answers.find("img.lazyload").lazyload({
                        container: $answers
                    });
                }
                else {
                    $('.see_more.loading', $answers).remove();
                    $answers.append(data);

                    $answers.find("img.lazyload").lazyload({
                        container: $answers
                    });

                    if (!noScroll) {
                        $answers.animate({
                            'scrollTop': ($answers.scrollTop() + $answers.innerHeight() - 80)
                        });
                    }
                }
                afterSearch();
            });

    });

    // subscribe_rss
    $answers.on('click', '.subscribe_rss', function (event) {
        event.preventDefault();
        var $this = $(this);

        if (typeof(renew) === 'undefined')
            renew = 'false';
        else
            renew = renew ? 'true' : 'false';

        var buttons = {};
        buttons[language.renewRss] = function () {
            $this.trigger({
                type: 'click',
                renew: true
            });
        };
        buttons[language.fermer] = function () {
            $('#DIALOG').empty().dialog('destroy');
        };

        event.stopPropagation();

        $.ajax({
            type: "GET",
            url: $this.attr('href') + (event.renew === true ? '?renew=true' : ''),
            dataType: 'json',
            success: function (data) {
                if (data.texte !== false && data.titre !== false) {
                    if ($("#DIALOG").data("ui-dialog")) {
                        $("#DIALOG").dialog('destroy');
                    }
                    $("#DIALOG").attr('title', data.titre)
                        .empty()
                        .append(data.texte)
                        .dialog({
                            autoOpen: false,
                            closeOnEscape: true,
                            resizable: false,
                            draggable: false,
                            modal: true,
                            buttons: buttons,
                            width: 650,
                            height: 250,
                            overlay: {
                                backgroundColor: '#000',
                                opacity: 0.7
                            }
                        }).dialog('open');

                }
            }
        });
    });

    // edit a feed
    $answers.on('click', '.feed .entry a.feed_edit', function () {
        var $this = $(this);
        $.ajax({
            type: "GET",
            url: $this.attr('href'),
            dataType: 'html',
            success: function (data) {
                return _createPublicationModal(data);
            }
        });
        return false;
    });

    // remove a feed
    $answers.on('click', '.feed .entry a.feed_delete', function () {
        if (!confirm('etes vous sur de vouloir supprimer cette entree ?'))
            return false;
        var $this = $(this);
        $.ajax({
            type: "POST",
            url: $this.attr('href'),
            dataType: 'json',
            success: function (data) {
                if (data.error === false) {
                    var $entry = $this.closest('.entry');
                    $entry.animate({
                        height: 0,
                        opacity: 0
                    }, function () {
                        $entry.remove();
                    });
                }
                else
                    alert(data.message);
            }
        });
        return false;
    });


    $answers.on('mouseover', '.feed .entry', function () {
        $(this).addClass('hover');
    });

    $answers.on('mouseout', '.feed .entry', function () {
        $(this).removeClass('hover');
    });

    $answers.on('click', '.see_more a', function (event) {
        $see_more = $(this).closest('.see_more');
        $see_more.addClass('loading');
    });


    var _fetchRemote = function (url, data) {
        var page = 0;
        if (data.page === undefined) {
            page = data.page;
        }

        return ajaxState.query = $.ajax({
            type: "GET",
            url: url,
            dataType: 'html',
            data: data,
            beforeSend: function () {
                if (ajaxState.isRunning && ajaxState.query.abort)
                    answAjax.abort();
                if (page === 0)
                    clearAnswers();
                ajaxState.isRunning = true;
                $answers.addClass('loading');
            },
            error: function () {
                ajaxState.isRunning = false;
                $answers.removeClass('loading');
            },
            timeout: function () {
                ajaxState.isRunning = false;
                $answers.removeClass('loading');
            },
            success: function (data) {
                ajaxState.isRunning = false;
            }
        });
    };

    var _createPublicationModal = function (data) {

        var buttons = {};
        buttons[language.valider] = function () {
            var dialog = p4.Dialog.get(1);
            var error = false;
            var $form = $('form.main_form', dialog.getDomElement());

            $('.required_text', $form).each(function (i, el) {
                if ($.trim($(el).val()) === '') {
                    $(el).addClass('error');
                    error = true;
                }
            });

            if (error) {
                alert(language.feed_require_fields);
            }

            if ($('input[name="feed_id"]', $form).val() === '') {
                alert(language.feed_require_feed);
                error = true;
            }

            if (error) {
                return false;
            }

            $.ajax({
                type: 'POST',
                url: $form.attr('action'),
                data: $form.serializeArray(),
                dataType: 'json',
                beforeSend: function () {
                    $('button', dialog.getDomElement()).prop('disabled', true);
                },
                error: function () {
                    $('button', dialog.getDomElement()).prop('disabled', false);
                },
                timeout: function () {
                    $('button', dialog.getDomElement()).prop('disabled', false);
                },
                success: function (data) {
                    $('button', dialog.getDomElement()).prop('disabled', false);
                    if (data.error === true) {
                        alert(data.message);
                        return;
                    }

                    if ($('form.main_form', dialog.getDomElement()).hasClass('entry_update')) {
                        var id = $('form input[name="entry_id"]', dialog.getDomElement()).val();
                        var container = $('#entry_' + id);

                        container.replaceWith(data.datas);

                        container.hide().fadeIn();

                        $answers.find("img.lazyload").lazyload({
                            container: $answers
                        });
                    }

                    p4.Dialog.Close(1);
                }
            });
            p4.Dialog.Close(1);
        };

        var dialog = p4.Dialog.Create({
            size: 'Full',
            closeOnEscape: true,
            closeButton: true,
            buttons: buttons
        });

        dialog.setContent(data);

        var $feeds_item = $('.feeds .feed', dialog.getDomElement());
        var $form = $('form.main_form', dialog.getDomElement());

        $feeds_item.bind('click', function () {
            $feeds_item.removeClass('selected');
            $(this).addClass('selected');
            $('input[name="feed_id"]', $form).val($('input', this).val());
        }).hover(function () {
            $(this).addClass('hover');
        }, function () {
            $(this).removeClass('hover');
        });

        $form.bind('submit', function () {
            return false;
        });

        return;
    };

    var fetchPublications = function (page) {
        curPage = page;
        return _fetchRemote('../prod/feeds/', {
            page: page
        })
            .then(function (data) {
                $('.next_publi_link', $answers).remove();

                $answers.append(data);

                $answers.find("img.lazyload").lazyload({
                    container: $answers
                });

                afterSearch();
                if (page > 0) {
                    $answers.stop().animate({
                        scrollTop: $answers.scrollTop() + $answers.height()
                    }, 700);
                }
                return;
            });
    };


    var publishRecords = function (type, value) {
        var options = {
            lst: '',
            ssel: '',
            act: ''
        };

        switch (type) {
            case "IMGT":
            case "CHIM":
                options.lst = value;
                break;

            case "STORY":
                options.story = value;
                break;
            case "SSTT":
                options.ssel = value;
                break;
        }

        $.post("../prod/feeds/requestavailable/"
            , options
            , function (data) {

                return _createPublicationModal(data);
            });

        return;
    };

    return {
        fetchPublications: function (page) {
            return fetchPublications(page);
        },
        publishRecords: function (type, value) {
            return publishRecords(type, value)
        }
    };
})();
;/*!
 * jQuery Form Plugin
 * version: 2.49 (18-OCT-2010)
 * @requires jQuery v1.3.2 or later
 *
 * Examples and documentation at: http://malsup.com/jquery/form/
 * Dual licensed under the MIT and GPL licenses:
 *   http://www.opensource.org/licenses/mit-license.php
 *   http://www.gnu.org/licenses/gpl.html
 */
;(function($) {

/*
	Usage Note:
	-----------
	Do not use both ajaxSubmit and ajaxForm on the same form.  These
	functions are intended to be exclusive.  Use ajaxSubmit if you want
	to bind your own submit handler to the form.  For example,

	$(document).ready(function() {
		$('#myForm').bind('submit', function(e) {
			e.preventDefault(); // <-- important
			$(this).ajaxSubmit({
				target: '#output'
			});
		});
	});

	Use ajaxForm when you want the plugin to manage all the event binding
	for you.  For example,

	$(document).ready(function() {
		$('#myForm').ajaxForm({
			target: '#output'
		});
	});

	When using ajaxForm, the ajaxSubmit function will be invoked for you
	at the appropriate time.
*/

/**
 * ajaxSubmit() provides a mechanism for immediately submitting
 * an HTML form using AJAX.
 */
$.fn.ajaxSubmit = function(options) {
	// fast fail if nothing selected (http://dev.jquery.com/ticket/2752)
	if (!this.length) {
		log('ajaxSubmit: skipping submit process - no element selected');
		return this;
	}

	if (typeof options == 'function') {
		options = { success: options };
	}

	var url = $.trim(this.attr('action'));
	if (url) {
		// clean url (don't include hash vaue)
		url = (url.match(/^([^#]+)/)||[])[1];
	}
	url = url || window.location.href || '';

	options = $.extend(true, {
		url:  url,
		type: this.attr('method') || 'GET',
		iframeSrc: /^https/i.test(window.location.href || '') ? 'javascript:false' : 'about:blank'
	}, options);

	// hook for manipulating the form data before it is extracted;
	// convenient for use with rich editors like tinyMCE or FCKEditor
	var veto = {};
	this.trigger('form-pre-serialize', [this, options, veto]);
	if (veto.veto) {
		log('ajaxSubmit: submit vetoed via form-pre-serialize trigger');
		return this;
	}

	// provide opportunity to alter form data before it is serialized
	if (options.beforeSerialize && options.beforeSerialize(this, options) === false) {
		log('ajaxSubmit: submit aborted via beforeSerialize callback');
		return this;
	}

	var n,v,a = this.formToArray(options.semantic);
	if (options.data) {
		options.extraData = options.data;
		for (n in options.data) {
			if(options.data[n] instanceof Array) {
				for (var k in options.data[n]) {
					a.push( { name: n, value: options.data[n][k] } );
				}
			}
			else {
				v = options.data[n];
				v = $.isFunction(v) ? v() : v; // if value is fn, invoke it
				a.push( { name: n, value: v } );
			}
		}
	}

	// give pre-submit callback an opportunity to abort the submit
	if (options.beforeSubmit && options.beforeSubmit(a, this, options) === false) {
		log('ajaxSubmit: submit aborted via beforeSubmit callback');
		return this;
	}

	// fire vetoable 'validate' event
	this.trigger('form-submit-validate', [a, this, options, veto]);
	if (veto.veto) {
		log('ajaxSubmit: submit vetoed via form-submit-validate trigger');
		return this;
	}

	var q = $.param(a);

	if (options.type.toUpperCase() == 'GET') {
		options.url += (options.url.indexOf('?') >= 0 ? '&' : '?') + q;
		options.data = null;  // data is null for 'get'
	}
	else {
		options.data = q; // data is the query string for 'post'
	}

	var $form = this, callbacks = [];
	if (options.resetForm) {
		callbacks.push(function() { $form.resetForm(); });
	}
	if (options.clearForm) {
		callbacks.push(function() { $form.clearForm(); });
	}

	// perform a load on the target only if dataType is not provided
	if (!options.dataType && options.target) {
		var oldSuccess = options.success || function(){};
		callbacks.push(function(data) {
			var fn = options.replaceTarget ? 'replaceWith' : 'html';
			$(options.target)[fn](data).each(oldSuccess, arguments);
		});
	}
	else if (options.success) {
		callbacks.push(options.success);
	}

	options.success = function(data, status, xhr) { // jQuery 1.4+ passes xhr as 3rd arg
		var context = options.context || options;   // jQuery 1.4+ supports scope context 
		for (var i=0, max=callbacks.length; i < max; i++) {
			callbacks[i].apply(context, [data, status, xhr || $form, $form]);
		}
	};

	// are there files to upload?
	var fileInputs = $('input:file', this).length > 0;
	var mp = 'multipart/form-data';
	var multipart = ($form.attr('enctype') == mp || $form.attr('encoding') == mp);

	// options.iframe allows user to force iframe mode
	// 06-NOV-09: now defaulting to iframe mode if file input is detected
   if (options.iframe !== false && (fileInputs || options.iframe || multipart)) {
	   // hack to fix Safari hang (thanks to Tim Molendijk for this)
	   // see:  http://groups.google.com/group/jquery-dev/browse_thread/thread/36395b7ab510dd5d
	   if (options.closeKeepAlive) {
		   $.get(options.closeKeepAlive, fileUpload);
		}
	   else {
		   fileUpload();
		}
   }
   else {
	   $.ajax(options);
   }

	// fire 'notify' event
	this.trigger('form-submit-notify', [this, options]);
	return this;


	// private function for handling file uploads (hat tip to YAHOO!)
	function fileUpload() {
		var form = $form[0];

		if ($(':input[name=submit],:input[id=submit]', form).length) {
			// if there is an input with a name or id of 'submit' then we won't be
			// able to invoke the submit fn on the form (at least not x-browser)
			alert('Error: Form elements must not have name or id of "submit".');
			return;
		}
		
		var s = $.extend(true, {}, $.ajaxSettings, options);
		s.context = s.context || s;
		var id = 'jqFormIO' + (new Date().getTime()), fn = '_'+id;
		window[fn] = function() {
			var f = $io.data('form-plugin-onload');
			if (f) {
				f();
				window[fn] = undefined;
				try { delete window[fn]; } catch(e){}
			}
		}
		var $io = $('<iframe id="' + id + '" name="' + id + '" src="'+ s.iframeSrc +'" onload="window[\'_\'+this.id]()" />');
		var io = $io[0];

		$io.css({ position: 'absolute', top: '-1000px', left: '-1000px' });

		var xhr = { // mock object
			aborted: 0,
			responseText: null,
			responseXML: null,
			status: 0,
			statusText: 'n/a',
			getAllResponseHeaders: function() {},
			getResponseHeader: function() {},
			setRequestHeader: function() {},
			abort: function() {
				this.aborted = 1;
				$io.attr('src', s.iframeSrc); // abort op in progress
			}
		};

		var g = s.global;
		// trigger ajax global events so that activity/block indicators work like normal
		if (g && ! $.active++) {
			$.event.trigger("ajaxStart");
		}
		if (g) {
			$.event.trigger("ajaxSend", [xhr, s]);
		}

		if (s.beforeSend && s.beforeSend.call(s.context, xhr, s) === false) {
			if (s.global) { 
				$.active--;
			}
			return;
		}
		if (xhr.aborted) {
			return;
		}

		var cbInvoked = false;
		var timedOut = 0;

		// add submitting element to data if we know it
		var sub = form.clk;
		if (sub) {
			var n = sub.name;
			if (n && !sub.disabled) {
				s.extraData = s.extraData || {};
				s.extraData[n] = sub.value;
				if (sub.type == "image") {
					s.extraData[n+'.x'] = form.clk_x;
					s.extraData[n+'.y'] = form.clk_y;
				}
			}
		}

		// take a breath so that pending repaints get some cpu time before the upload starts
		function doSubmit() {
			// make sure form attrs are set
			var t = $form.attr('target'), a = $form.attr('action');

			// update form attrs in IE friendly way
			form.setAttribute('target',id);
			if (form.getAttribute('method') != 'POST') {
				form.setAttribute('method', 'POST');
			}
			if (form.getAttribute('action') != s.url) {
				form.setAttribute('action', s.url);
			}

			// ie borks in some cases when setting encoding
			if (! s.skipEncodingOverride) {
				$form.attr({
					encoding: 'multipart/form-data',
					enctype:  'multipart/form-data'
				});
			}

			// support timout
			if (s.timeout) {
				setTimeout(function() { timedOut = true; cb(); }, s.timeout);
			}

			// add "extra" data to form if provided in options
			var extraInputs = [];
			try {
				if (s.extraData) {
					for (var n in s.extraData) {
						extraInputs.push(
							$('<input type="hidden" name="'+n+'" value="'+s.extraData[n]+'" />')
								.appendTo(form)[0]);
					}
				}

				// add iframe to doc and submit the form
				$io.appendTo('body');
				$io.data('form-plugin-onload', cb);
				form.submit();
			}
			finally {
				// reset attrs and remove "extra" input elements
				form.setAttribute('action',a);
				if(t) {
					form.setAttribute('target', t);
				} else {
					$form.removeAttr('target');
				}
				$(extraInputs).remove();
			}
		}

		if (s.forceSync) {
			doSubmit();
		}
		else {
			setTimeout(doSubmit, 10); // this lets dom updates render
		}
	
		var data, doc, domCheckCount = 50;

		function cb() {
			if (cbInvoked) {
				return;
			}

			$io.removeData('form-plugin-onload');
			
			var ok = true;
			try {
				if (timedOut) {
					throw 'timeout';
				}
				// extract the server response from the iframe
				doc = io.contentWindow ? io.contentWindow.document : io.contentDocument ? io.contentDocument : io.document;
				
				var isXml = s.dataType == 'xml' || doc.XMLDocument || $.isXMLDoc(doc);
				log('isXml='+isXml);
				if (!isXml && window.opera && (doc.body == null || doc.body.innerHTML == '')) {
					if (--domCheckCount) {
						// in some browsers (Opera) the iframe DOM is not always traversable when
						// the onload callback fires, so we loop a bit to accommodate
						log('requeing onLoad callback, DOM not available');
						setTimeout(cb, 250);
						return;
					}
					// let this fall through because server response could be an empty document
					//log('Could not access iframe DOM after mutiple tries.');
					//throw 'DOMException: not available';
				}

				//log('response detected');
				cbInvoked = true;
				xhr.responseText = doc.documentElement ? doc.documentElement.innerHTML : null; 
				xhr.responseXML = doc.XMLDocument ? doc.XMLDocument : doc;
				xhr.getResponseHeader = function(header){
					var headers = {'content-type': s.dataType};
					return headers[header];
				};

				var scr = /(json|script)/.test(s.dataType);
				if (scr || s.textarea) {
					// see if user embedded response in textarea
					var ta = doc.getElementsByTagName('textarea')[0];
					if (ta) {
						xhr.responseText = ta.value;
					}
					else if (scr) {
						// account for browsers injecting pre around json response
						var pre = doc.getElementsByTagName('pre')[0];
						var b = doc.getElementsByTagName('body')[0];
						if (pre) {
							xhr.responseText = pre.innerHTML;
						}
						else if (b) {
							xhr.responseText = b.innerHTML;
						}
					}			  
				}
				else if (s.dataType == 'xml' && !xhr.responseXML && xhr.responseText != null) {
					xhr.responseXML = toXml(xhr.responseText);
				}
				data = $.httpData(xhr, s.dataType);
			}
			catch(e){
				log('error caught:',e);
				ok = false;
				xhr.error = e;
				$.handleError(s, xhr, 'error', e);
			}

			// ordering of these callbacks/triggers is odd, but that's how $.ajax does it
			if (ok) {
				s.success.call(s.context, data, 'success', xhr);
				if (g) {
					$.event.trigger("ajaxSuccess", [xhr, s]);
				}
			}
			if (g) {
				$.event.trigger("ajaxComplete", [xhr, s]);
			}
			if (g && ! --$.active) {
				$.event.trigger("ajaxStop");
			}
			if (s.complete) {
				s.complete.call(s.context, xhr, ok ? 'success' : 'error');
			}

			// clean up
			setTimeout(function() {
				$io.removeData('form-plugin-onload');
				$io.remove();
				xhr.responseXML = null;
			}, 100);
		}

		function toXml(s, doc) {
			if (window.ActiveXObject) {
				doc = new ActiveXObject('Microsoft.XMLDOM');
				doc.async = 'false';
				doc.loadXML(s);
			}
			else {
				doc = (new DOMParser()).parseFromString(s, 'text/xml');
			}
			return (doc && doc.documentElement && doc.documentElement.tagName != 'parsererror') ? doc : null;
		}
	}
};

/**
 * ajaxForm() provides a mechanism for fully automating form submission.
 *
 * The advantages of using this method instead of ajaxSubmit() are:
 *
 * 1: This method will include coordinates for <input type="image" /> elements (if the element
 *	is used to submit the form).
 * 2. This method will include the submit element's name/value data (for the element that was
 *	used to submit the form).
 * 3. This method binds the submit() method to the form for you.
 *
 * The options argument for ajaxForm works exactly as it does for ajaxSubmit.  ajaxForm merely
 * passes the options argument along after properly binding events for submit elements and
 * the form itself.
 */
$.fn.ajaxForm = function(options) {
	// in jQuery 1.3+ we can fix mistakes with the ready state
	if (this.length === 0) {
		var o = { s: this.selector, c: this.context };
		if (!$.isReady && o.s) {
			log('DOM not ready, queuing ajaxForm');
			$(function() {
				$(o.s,o.c).ajaxForm(options);
			});
			return this;
		}
		// is your DOM ready?  http://docs.jquery.com/Tutorials:Introducing_$(document).ready()
		log('terminating; zero elements found by selector' + ($.isReady ? '' : ' (DOM not ready)'));
		return this;
	}
	
	return this.ajaxFormUnbind().bind('submit.form-plugin', function(e) {
		if (!e.isDefaultPrevented()) { // if event has been canceled, don't proceed
			e.preventDefault();
			$(this).ajaxSubmit(options);
		}
	}).bind('click.form-plugin', function(e) {
		var target = e.target;
		var $el = $(target);
		if (!($el.is(":submit,input:image"))) {
			// is this a child element of the submit el?  (ex: a span within a button)
			var t = $el.closest(':submit');
			if (t.length == 0) {
				return;
			}
			target = t[0];
		}
		var form = this;
		form.clk = target;
		if (target.type == 'image') {
			if (e.offsetX != undefined) {
				form.clk_x = e.offsetX;
				form.clk_y = e.offsetY;
			} else if (typeof $.fn.offset == 'function') { // try to use dimensions plugin
				var offset = $el.offset();
				form.clk_x = e.pageX - offset.left;
				form.clk_y = e.pageY - offset.top;
			} else {
				form.clk_x = e.pageX - target.offsetLeft;
				form.clk_y = e.pageY - target.offsetTop;
			}
		}
		// clear form vars
		setTimeout(function() { form.clk = form.clk_x = form.clk_y = null; }, 100);
	});
};

// ajaxFormUnbind unbinds the event handlers that were bound by ajaxForm
$.fn.ajaxFormUnbind = function() {
	return this.unbind('submit.form-plugin click.form-plugin');
};

/**
 * formToArray() gathers form element data into an array of objects that can
 * be passed to any of the following ajax functions: $.get, $.post, or load.
 * Each object in the array has both a 'name' and 'value' property.  An example of
 * an array for a simple login form might be:
 *
 * [ { name: 'username', value: 'jresig' }, { name: 'password', value: 'secret' } ]
 *
 * It is this array that is passed to pre-submit callback functions provided to the
 * ajaxSubmit() and ajaxForm() methods.
 */
$.fn.formToArray = function(semantic) {
	var a = [];
	if (this.length === 0) {
		return a;
	}

	var form = this[0];
	var els = semantic ? form.getElementsByTagName('*') : form.elements;
	if (!els) {
		return a;
	}
	
	var i,j,n,v,el,max,jmax;
	for(i=0, max=els.length; i < max; i++) {
		el = els[i];
		n = el.name;
		if (!n) {
			continue;
		}

		if (semantic && form.clk && el.type == "image") {
			// handle image inputs on the fly when semantic == true
			if(!el.disabled && form.clk == el) {
				a.push({name: n, value: $(el).val()});
				a.push({name: n+'.x', value: form.clk_x}, {name: n+'.y', value: form.clk_y});
			}
			continue;
		}

		v = $.fieldValue(el, true);
		if (v && v.constructor == Array) {
			for(j=0, jmax=v.length; j < jmax; j++) {
				a.push({name: n, value: v[j]});
			}
		}
		else if (v !== null && typeof v != 'undefined') {
			a.push({name: n, value: v});
		}
	}

	if (!semantic && form.clk) {
		// input type=='image' are not found in elements array! handle it here
		var $input = $(form.clk), input = $input[0];
		n = input.name;
		if (n && !input.disabled && input.type == 'image') {
			a.push({name: n, value: $input.val()});
			a.push({name: n+'.x', value: form.clk_x}, {name: n+'.y', value: form.clk_y});
		}
	}
	return a;
};

/**
 * Serializes form data into a 'submittable' string. This method will return a string
 * in the format: name1=value1&amp;name2=value2
 */
$.fn.formSerialize = function(semantic) {
	//hand off to jQuery.param for proper encoding
	return $.param(this.formToArray(semantic));
};

/**
 * Serializes all field elements in the jQuery object into a query string.
 * This method will return a string in the format: name1=value1&amp;name2=value2
 */
$.fn.fieldSerialize = function(successful) {
	var a = [];
	this.each(function() {
		var n = this.name;
		if (!n) {
			return;
		}
		var v = $.fieldValue(this, successful);
		if (v && v.constructor == Array) {
			for (var i=0,max=v.length; i < max; i++) {
				a.push({name: n, value: v[i]});
			}
		}
		else if (v !== null && typeof v != 'undefined') {
			a.push({name: this.name, value: v});
		}
	});
	//hand off to jQuery.param for proper encoding
	return $.param(a);
};

/**
 * Returns the value(s) of the element in the matched set.  For example, consider the following form:
 *
 *  <form><fieldset>
 *	  <input name="A" type="text" />
 *	  <input name="A" type="text" />
 *	  <input name="B" type="checkbox" value="B1" />
 *	  <input name="B" type="checkbox" value="B2"/>
 *	  <input name="C" type="radio" value="C1" />
 *	  <input name="C" type="radio" value="C2" />
 *  </fieldset></form>
 *
 *  var v = $(':text').fieldValue();
 *  // if no values are entered into the text inputs
 *  v == ['','']
 *  // if values entered into the text inputs are 'foo' and 'bar'
 *  v == ['foo','bar']
 *
 *  var v = $(':checkbox').fieldValue();
 *  // if neither checkbox is checked
 *  v === undefined
 *  // if both checkboxes are checked
 *  v == ['B1', 'B2']
 *
 *  var v = $(':radio').fieldValue();
 *  // if neither radio is checked
 *  v === undefined
 *  // if first radio is checked
 *  v == ['C1']
 *
 * The successful argument controls whether or not the field element must be 'successful'
 * (per http://www.w3.org/TR/html4/interact/forms.html#successful-controls).
 * The default value of the successful argument is true.  If this value is false the value(s)
 * for each element is returned.
 *
 * Note: This method *always* returns an array.  If no valid value can be determined the
 *	   array will be empty, otherwise it will contain one or more values.
 */
$.fn.fieldValue = function(successful) {
	for (var val=[], i=0, max=this.length; i < max; i++) {
		var el = this[i];
		var v = $.fieldValue(el, successful);
		if (v === null || typeof v == 'undefined' || (v.constructor == Array && !v.length)) {
			continue;
		}
		v.constructor == Array ? $.merge(val, v) : val.push(v);
	}
	return val;
};

/**
 * Returns the value of the field element.
 */
$.fieldValue = function(el, successful) {
	var n = el.name, t = el.type, tag = el.tagName.toLowerCase();
	if (successful === undefined) {
		successful = true;
	}

	if (successful && (!n || el.disabled || t == 'reset' || t == 'button' ||
		(t == 'checkbox' || t == 'radio') && !el.checked ||
		(t == 'submit' || t == 'image') && el.form && el.form.clk != el ||
		tag == 'select' && el.selectedIndex == -1)) {
			return null;
	}

	if (tag == 'select') {
		var index = el.selectedIndex;
		if (index < 0) {
			return null;
		}
		var a = [], ops = el.options;
		var one = (t == 'select-one');
		var max = (one ? index+1 : ops.length);
		for(var i=(one ? index : 0); i < max; i++) {
			var op = ops[i];
			if (op.selected) {
				var v = op.value;
				if (!v) { // extra pain for IE...
					v = (op.attributes && op.attributes['value'] && !(op.attributes['value'].specified)) ? op.text : op.value;
				}
				if (one) {
					return v;
				}
				a.push(v);
			}
		}
		return a;
	}
	return $(el).val();
};

/**
 * Clears the form data.  Takes the following actions on the form's input fields:
 *  - input text fields will have their 'value' property set to the empty string
 *  - select elements will have their 'selectedIndex' property set to -1
 *  - checkbox and radio inputs will have their 'checked' property set to false
 *  - inputs of type submit, button, reset, and hidden will *not* be effected
 *  - button elements will *not* be effected
 */
$.fn.clearForm = function() {
	return this.each(function() {
		$('input,select,textarea', this).clearFields();
	});
};

/**
 * Clears the selected form elements.
 */
$.fn.clearFields = $.fn.clearInputs = function() {
	return this.each(function() {
		var t = this.type, tag = this.tagName.toLowerCase();
		if (t == 'text' || t == 'password' || tag == 'textarea') {
			this.value = '';
		}
		else if (t == 'checkbox' || t == 'radio') {
			this.checked = false;
		}
		else if (tag == 'select') {
			this.selectedIndex = -1;
		}
	});
};

/**
 * Resets the form data.  Causes all form elements to be reset to their original value.
 */
$.fn.resetForm = function() {
	return this.each(function() {
		// guard against an input with the name of 'reset'
		// note that IE reports the reset function as an 'object'
		if (typeof this.reset == 'function' || (typeof this.reset == 'object' && !this.reset.nodeType)) {
			this.reset();
		}
	});
};

/**
 * Enables or disables any matching elements.
 */
$.fn.enable = function(b) {
	if (b === undefined) {
		b = true;
	}
	return this.each(function() {
		this.disabled = !b;
	});
};

/**
 * Checks/unchecks any matching checkboxes or radio buttons and
 * selects/deselects and matching option elements.
 */
$.fn.selected = function(select) {
	if (select === undefined) {
		select = true;
	}
	return this.each(function() {
		var t = this.type;
		if (t == 'checkbox' || t == 'radio') {
			this.checked = select;
		}
		else if (this.tagName.toLowerCase() == 'option') {
			var $sel = $(this).parent('select');
			if (select && $sel[0] && $sel[0].type == 'select-one') {
				// deselect all other options
				$sel.find('option').selected(false);
			}
			this.selected = select;
		}
	});
};

// helper fn for console logging
// set $.fn.ajaxSubmit.debug to true to enable debug logging
function log() {
	if ($.fn.ajaxSubmit.debug) {
		var msg = '[jquery.form] ' + Array.prototype.join.call(arguments,'');
		if (window.console && window.console.log) {
			window.console.log(msg);
		}
		else if (window.opera && window.opera.postError) {
			window.opera.postError(msg);
		}
	}
};

})(jQuery);
;/*
 * Selection Object
 *
 *
 */

(function (window) {

    var Selectable = function ($container, options) {

        var defaults = {
                allow_multiple: false,
                selector: '',
                callbackSelection: null,
                selectStart: null,
                selectStop: null,
                limit: null
            },
            options = (typeof options == 'object') ? options : {};

        var $this = this;

        if ($container.data('selectionnable')) {
            /* this container is already selectionnable */
            if (window.console) {
                console.error('Trying to apply new selection to existing one');
            }

            return $container.data('selectionnable');
        }

        this.$container = $container;
        this.options = jQuery.extend(defaults, options);
        this.datas = new Array();

        this.$container.data('selectionnable', this);
        this.$container.addClass('selectionnable');
        this.$container
            .on('click', this.options.selector, function(event) {
                event.preventDefault();
                if (typeof $this.options.selectStart === 'function') {
                    $this.options.selectStart(jQuery.extend(jQuery.Event('selectStart'), event), $this);
                }

                var $that = jQuery(this);

                var k = get_value($that, $this);

                if (is_shift_key(event) && jQuery('.last_selected', this.$container).filter($this.options.selector).length != 0) {
                    var lst = jQuery($this.options.selector, this.$container);

                    var index1 = jQuery.inArray(jQuery('.last_selected', this.$container).filter($this.options.selector)[0], lst);
                    var index2 = jQuery.inArray($that[0], lst);

                    if (index2 < index1) {
                        var tmp = index1;
                        index1 = (index2 - 1) < 0 ? index2 : (index2 - 1);
                        index2 = tmp;
                    }

                    var stopped = false;

                    if (index2 != -1 && index1 != -1) {
                        var exp = $this.options.selector + ':gt(' + index1 + '):lt(' + (index2 - index1) + ')';

                        $.each(jQuery(exp, this.$container), function (i, n) {
                            if (!jQuery(n).hasClass('selected') && stopped === false) {
                                if (!$this.hasReachLimit()) {
                                    var k = get_value(jQuery(n), $this);
                                    $this.push(k);
                                    jQuery(n).addClass('selected');
                                }
                                else {
                                    alert(language.max_record_selected);
                                    stopped = true;
                                }
                            }
                        });
                    }

                    if ($this.has(k) === false && stopped === false) {
                        if (!$this.hasReachLimit()) {
                            $this.push(k);
                            $that.addClass('selected');
                        }
                        else {
                            alert(language.max_record_selected);
                        }
                    }
                }
                else {
                    if (!is_ctrl_key(event)) {
                        $this.empty().push(k);
                        jQuery('.selected', this.$container).filter($this.options.selector).removeClass('selected');
                        $that.addClass('selected');
                    }
                    else {
                        if ($this.has(k) === true) {
                            $this.remove(k);
                            $that.removeClass('selected');
                        }
                        else {
                            if (!$this.hasReachLimit()) {
                                $this.push(k);
                                $that.addClass('selected');
                            }
                            else {
                                alert(language.max_record_selected);
                            }
                        }
                    }
                }

                jQuery('.last_selected', this.$container).removeClass('last_selected');
                $that.addClass('last_selected');


                if (typeof $this.options.selectStop === 'function') {
                    $this.options.selectStop(jQuery.extend(jQuery.Event('selectStop'), event), $this);
                }

            });

        return;
    };

    function get_value(element, Selectable) {
        if (typeof Selectable.options.callbackSelection === 'function') {
            return Selectable.options.callbackSelection(jQuery(element));
        }
        else {
            return jQuery('input[name="id"]', jQuery(element)).val();
        }
    }

    function is_ctrl_key(event) {
        if (event.altKey)
            return true;
        if (event.ctrlKey)
            return true;
        if (event.metaKey)	// apple key opera
            return true;
        if (event.keyCode == '17')	// apple key opera
            return true;
        if (event.keyCode == '224')	// apple key mozilla
            return true;
        if (event.keyCode == '91')	// apple key safari
            return true;

        return false;
    }

    function is_shift_key(event) {
        if (event.shiftKey)
            return true;
        return false;
    }


    Selectable.prototype = {
        push: function (element) {
            if (this.options.allow_multiple === true || !this.has(element)) {
                this.datas.push(element);
            }

            return this;
        },
        hasReachLimit: function () {
            if (this.options.limit !== null && this.options.limit <= this.datas.length) {
                return true;
            }
            return false;
        },
        remove: function (element) {
            this.datas = jQuery.grep(this.datas, function (n) {
                return(n !== element);
            });

            return this;
        },
        has: function (element) {

            return jQuery.inArray(element, this.datas) >= 0;
        },
        get: function () {

            return this.datas;
        },
        empty: function () {
            var $this = this;
            this.datas = new Array();

            jQuery(this.options.selector, this.$container).filter('.selected:visible').removeClass('selected');

            if (typeof $this.options.selectStop === 'function') {
                $this.options.selectStop(jQuery.Event('selectStop'), $this);
            }

            return this;
        },
        length: function () {

            return this.datas.length;
        },
        size: function () {

            return this.datas.length;
        },
        serialize: function (separator) {

            separator = separator || ';';

            return this.datas.join(separator);
        },
        selectAll: function () {
            this.select('*');

            return this;
        },
        select: function (selector) {
            var $this = this,
                stopped = false;

            jQuery(this.options.selector, this.$container).filter(selector).not('.selected').filter(':visible').each(function () {
                if (!$this.hasReachLimit()) {
                    $this.push(get_value(this, $this));
                    $(this).addClass('selected');
                }
                else {
                    if (stopped === false) {
                        alert(language.max_record_selected);
                    }
                    stopped = true;
                }
            });


            if (typeof $this.options.selectStop === 'function') {
                $this.options.selectStop(jQuery.Event('selectStop'), $this);
            }

            return this;
        }
    };

    window.Selectable = Selectable;
})(window);
;(function (window) {

    function checkVocabId(VocabularyId) {
        if (typeof VocabularyId === 'undefined')
            VocabularyId = null;

        if (VocabularyId === '')
            VocabularyId = null;

        return VocabularyId;
    }

    var recordFieldValue = function (meta_id, value, VocabularyId) {

        VocabularyId = checkVocabId(VocabularyId);

        this.datas = {
            meta_id: meta_id,
            value: value,
            VocabularyId: VocabularyId
        };

        var $this = this;
    };

    recordFieldValue.prototype = {
        getValue: function () {
            return this.datas.value;
        },
        getMetaId: function () {
            return this.datas.meta_id;
        },
        getVocabularyId: function () {
            return this.datas.VocabularyId;
        },
        setValue: function (value, VocabularyId) {

            this.datas.value = value;
            this.datas.VocabularyId = checkVocabId(VocabularyId);
            return this;
        },
        remove: function () {
            this.datas.value = '';
            this.datas.VocabularyId = null;

            return this;
        }
    };

    var databoxField = function (name, label, meta_struct_id, options) {

        var defaults = {
                multi: false,
                required: false,
                readonly: false,
                maxLength: null,
                minLength: null,
                type: 'string',
                separator: null,
                vocabularyControl: null,
                vocabularyRestricted: false
            },
            options = (typeof options == 'object') ? options : {};

        if (isNaN(meta_struct_id)) {
            throw 'meta_struct_id should be a number';
        }

        this.name = name;
        this.label = label;
        this.meta_struct_id = meta_struct_id;
        this.options = jQuery.extend(defaults, options);

        if (this.options.multi === true && this.options.separator === null) {
            this.options.separator = ';';
        }

    };

    databoxField.prototype = {
        getMetaStructId: function () {
            return this.meta_struct_id;
        },
        getName: function () {
            return this.name;
        },
        getLabel: function () {
            return this.label;
        },
        isMulti: function () {
            return this.options.multi;
        },
        isRequired: function () {
            return this.options.required;
        },
        isReadonly: function () {
            return this.options.readonly;
        },
        getMaxLength: function () {
            return this.options.maxLength;
        },
        getMinLength: function () {
            return this.options.minLength;
        },
        getType: function () {
            return this.options.type;
        },
        getSeparator: function () {
            return this.options.separator;
        }
    };

    var recordField = function (databoxField, arrayValues) {

        this.databoxField = databoxField;
        this.options = {
            dirty: false
        };
        this.datas = new Array();

        if (arrayValues instanceof Array) {
            if (arrayValues.length > 1 && !databoxField.isMulti())
                throw 'You can not add multiple values to a non multi field ' + databoxField.getName();

            var first = true;

            for (v in arrayValues) {
                if (typeof arrayValues[v] !== 'object') {
                    if (window.console) {
                        console.error('Trying to add a non-recordFieldValue to the field...');
                    }

                    continue;
                }

                if (isNaN(arrayValues[v].getMetaId())) {
                    if (window.console) {
                        console.error('Trying to add a recordFieldValue without metaId...');
                    }

                    continue;
                }

                if (!first && this.options.multi === false) {
                    if (window.console) {
                        console.error('Trying to add multi values in a non-multi field');
                    }
                }

                if (window.console) {
                    console.log('adding a value : ', arrayValues[v]);
                }

                this.datas.push(arrayValues[v]);
                first = false;
            }
        }

        var $this = this;
    }
    recordField.prototype = {
        getName: function () {
            return this.databoxField.getName();
        },
        getMetaStructId: function () {
            return this.databoxField.getMetaStructId();
        },
        isMulti: function () {
            return this.databoxField.isMulti();
        },
        isRequired: function () {
            return this.databoxField.isRequired();
        },
        isDirty: function () {
            return this.options.dirty;
        },
        addValue: function (value, merge, VocabularyId) {

            VocabularyId = checkVocabId(VocabularyId);

            merge = !!merge;

            if (this.databoxField.isReadonly()) {
                if (window.console) {
                    console.error('Unable to set a value to a readonly field');
                }

                return;
            }

            if (window.console) {
                console.log('adding value ', value, ' vocId : ', VocabularyId, '  ; merge is ', merge);
            }

            if (this.isMulti()) {
                if (!this.hasValue(value, VocabularyId)) {
                    if (window.console) {
                        console.log('adding new multi value ', value);
                    }
                    if( value === '') {
                        return;
                    }
                    this.datas.push(new recordFieldValue(null, value, VocabularyId));
                    this.options.dirty = true;
                }
                else {
                    if (window.console) {
                        console.log('already have ', value);
                    }
                }
            }
            else {
                if (merge === true && this.isEmpty() === false && VocabularyId === null) {
                    if (window.console) {
                        console.log('Merging value ', value);
                    }

                    this.datas[0].setValue(this.datas[0].getValue() + ' ' + value, VocabularyId);

                    this.options.dirty = true;
                }
                else {
                    if (merge === true && this.isEmpty() === false && VocabularyId !== null) {
                        if (window.console) {
                            console.error('Cannot merge vocabularies');
                        }
                        this.datas[0].setValue(value, VocabularyId);
                    }
                    else {

                        if (!this.hasValue(value, VocabularyId)) {
                            if (this.datas.length === 0) {
                                if (window.console) {
                                    console.log('Adding new value ', value);
                                }
                                this.datas.push(new recordFieldValue(null, value, VocabularyId));
                            }
                            else {
                                if (window.console) {
                                    console.log('Updating value ', value);
                                }
                                this.datas[0].setValue(value, VocabularyId);
                            }
                            this.options.dirty = true;
                        }
                    }
                }
            }

            return this;
        },
        hasValue: function (value, VocabularyId) {

            if (typeof value === 'undefined') {
                if (window.console) {
                    console.error('Trying to check the presence of an undefined value');
                }
            }

            VocabularyId = checkVocabId(VocabularyId);

            for (d in this.datas) {
                if (VocabularyId !== null) {
                    if (this.datas[d].getVocabularyId() === VocabularyId) {
                        if (window.console) {
                            console.log('already got the vocab ID');
                        }
                        return true;
                    }
                }
                else if (this.datas[d].getVocabularyId() === null && this.datas[d].getValue() == value) {
                    if (window.console) {
                        console.log('already got this value');
                    }
                    return true;
                }
            }
            return false;
        },
        removeValue: function (value, vocabularyId) {

            if (this.databoxField.isReadonly()) {
                if (window.console) {
                    console.error('Unable to set a value to a readonly field');
                }

                return;
            }

            vocabularyId = checkVocabId(vocabularyId);

            if (window.console) {
                console.log('Try to remove value ', value, vocabularyId, this.datas);
            }

            for (d in this.datas) {
                if (window.console) {
                    console.log('loopin... ', this.datas[d].getValue());
                }
                if (this.datas[d].getVocabularyId() !== null) {
                    if (this.datas[d].getVocabularyId() == vocabularyId) {
                        if (window.console) {
                            console.log('Found within the vocab ! removing... ');
                        }
                        this.datas[d].remove();
                        this.options.dirty = true;
                    }
                }
                else if (this.datas[d].getValue() == value) {
                    if (window.console) {
                        console.log('Found ! removing... ');
                    }
                    this.datas[d].remove();
                    this.options.dirty = true;
                }
            }
            return this;
        },
        isEmpty: function () {
            var empty = true;

            for (d in this.datas) {
                if (this.datas[d].getValue() !== '')
                    empty = false;
            }
            return empty;
        },
        empty: function () {

            if (this.databoxField.isReadonly()) {
                if (window.console) {
                    console.error('Unable to set a value to a readonly field');
                }

                return;
            }

            for (d in this.datas) {
                this.datas[d].remove();
                this.options.dirty = true;
            }
            return this;
        },
        getValue: function () {

            if (this.isMulti())
                throw 'This field is multi, I can not give you a single value';

            if (this.isEmpty())
                return null;

            return this.datas[0];
        },
        getValues: function () {

            if (!this.isMulti()) {
                throw 'This field is not multi, I can not give you multiple values';
            }

            if (this.isEmpty())
                return new Array();

            var arrayValues = [];

            for (d in this.datas) {
                if (this.datas[d].getValue() === '')
                    continue;

                arrayValues.push(this.datas[d]);
            }

            return arrayValues;
        },
        sort: function (algo) {
            this.datas.sort(algo);

            return this;
        },
        getSerializedValues: function () {

            var arrayValues = [];
            var values = this.getValues();

            for (v in values) {
                arrayValues.push(values[v].getValue());
            }

            return arrayValues.join(' ; ');
        },
        replaceValue: function (search, replace) {

            if (this.databoxField.isReadonly()) {
                if (window.console) {
                    console.error('Unable to set a value to a readonly field');
                }

                return 0;
            }

            var n = 0;

            for (d in this.datas) {
                if (this.datas[d].getVocabularyId() !== null) {
                    continue;
                }

                var value = this.datas[d].getValue();
                var replacedValue = value.replace(search, replace);

                if (value === replacedValue) {
                    continue;
                }

                n++;

                this.removeValue(value);

                if (!this.hasValue(replacedValue)) {
                    this.addValue(replacedValue);
                }

                this.options.dirty = true;
            }

            return n;
        },
        exportDatas: function () {

            var returnValue = new Array();

            for (d in this.datas) {
                var temp = {
                    meta_id: this.datas[d].getMetaId() ? this.datas[d].getMetaId() : '',
                    meta_struct_id: this.getMetaStructId(),
                    value: this.datas[d].getValue()
                };

                if (this.datas[d].getVocabularyId()) {
                    temp.vocabularyId = this.datas[d].getVocabularyId();
                }
                returnValue.push(temp);
            }

            return returnValue;
        }
    };

    window.p4 = window.p4 || {};

    window.p4.databoxField = databoxField;
    window.p4.recordFieldValue = recordFieldValue;
    window.p4.recordField = recordField;

})(window);
;var p4 = p4 || {};

(function (window, p4, $) {

    var Lists = function () {

    };

    var List = function (id) {

        if (parseInt(id) <= 0) {
            throw 'Invalid list id';
        }

        this.id = id;
    };

    Lists.prototype = {
        create: function (name, callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/',
                dataType: 'json',
                data: {name: name},
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            var list = new List(data.list_id);
                            callback(list);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });

        },
        get: function (callback, type) {

            var $this = this;
            type = typeof type === 'undefined' ? 'json' : type;

            $.ajax({
                type: 'GET',
                url: '/prod/lists/all/',
                dataType: type,
                data: {},
                success: function (data) {
                    if (type == 'json') {
                        if (data.success) {
                            humane.info(data.message);

                            if (typeof callback === 'function') {
                                callback(data.result);
                            }
                        }
                        else {
                            humane.error(data.message);
                        }
                    }
                    else {
                        if (typeof callback === 'function') {
                            callback(data);
                        }
                    }
                }
            });
        }

    }

    List.prototype = {
        addUsers: function (arrayUsers, callback) {

            if (!arrayUsers instanceof Array) {
                throw 'addUsers takes array as argument';
            }

            var $this = this;
            var data = {usr_ids: $(arrayUsers).toArray()};

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/' + $this.id + '/add/',
                dataType: 'json',
                data: data,
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this, data);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        addUser: function (usr_id, callback) {
            this.addUsers([usr_id], callback);
        },
        remove: function (callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/' + this.id + '/delete/',
                dataType: 'json',
                data: {},
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        update: function (name, callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/' + this.id + '/update/',
                dataType: 'json',
                data: { name: name },
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        removeUser: function (usr_id, callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/' + this.id + '/remove/' + usr_id + '/',
                dataType: 'json',
                data: {},
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this, data);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        shareWith: function (usr_id, role, callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/' + this.id + '/share/' + usr_id + '/',
                dataType: 'json',
                data: {role: role},
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        unshareWith: function (callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/lists/list/' + this.id + '/unshare/' + usr_id + '/',
                dataType: 'json',
                data: {},
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        get: function (callback) {

            var $this = this;

            $.ajax({
                type: 'GET',
                url: '/prod/lists/list/' + this.id + '/',
                dataType: 'json',
                data: {},
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);

                        if (typeof callback === 'function') {
                            callback($this, data);
                        }
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        }
    }

    p4.Lists = new Lists();
    document.List = List;

})(document, p4, jQuery);
;(function () {
    $(document).ready(function () {
        humane.info = humane.spawn({addnCls: 'humane-libnotify-info', timeout: 1000});
        humane.error = humane.spawn({addnCls: 'humane-libnotify-error', timeout: 1000});

        $('body').on('click', 'a.dialog', function (event) {
            $('.context-menu').hide();
            var $this = $(this), size = 'Medium';

            if ($this.hasClass('small-dialog')) {
                size = 'Small';
            } else if ($this.hasClass('full-dialog')) {
                size = 'Full';
            }

            var options = {
                size: size,
                loading: true,
                title: $this.attr('title'),
                closeOnEscape: true
            };

            $dialog = p4.Dialog.Create(options);

            $.ajax({
                type: "GET",
                url: $this.attr('href'),
                dataType: 'html',
                success: function (data) {
                    $dialog.setContent(data);
                    return;
                }
            });

            return false;
        });
    });
}());
;;
(function (window) {

    var Feedback = function ($container, context) {
        this.container = $($container);

        this.Context = context;

        this.selection = new Selectable(
            $('.user_content .badges', this.container),
            {
                selector: '.badge'
            }
        );

        var $this = this;

        this.container.on('click', '.content .options .select-all', function (event) {
            $this.selection.selectAll();
        });

        this.container.on('click', '.content .options .unselect-all', function (event) {
            $this.selection.empty();
        });

        $('.UserTips', this.container).tooltip();

        this.container.on('click', 'a.user_adder', function (event) {
            var $this = $(this);

            $.ajax({
                type: "GET",
                url: $this.attr('href'),
                dataType: 'html',
                beforeSend: function () {
                    var options = {
                        size: 'Medium',
                        title: $this.html()
                    };
                    p4.Dialog.Create(options, 2).getDomElement().addClass('loading');
                },
                success: function (data) {
                    p4.Dialog.get(2).getDomElement().removeClass('loading').empty().append(data);
                    return;
                },
                error: function () {
                    p4.Dialog.get(2).Close();
                    return;
                },
                timeout: function () {
                    p4.Dialog.get(2).Close();
                    return;
                }
            });

            return false;
        });

        this.container.on('click', '.recommended_users', function (event) {
            var usr_id = $('input[name="usr_id"]', $(this)).val();

            $this.loadUser(usr_id, $this.selectUser);

            return false;
        });

        this.container.on('click', '.recommended_users_list', function (event) {

            var content = $('#push_user_recommendations').html();

            var options = {
                size: 'Small',
                title: $(this).attr('title')
            };

            $dialog = p4.Dialog.Create(options, 2);
            $dialog.setContent(content);

            $dialog.getDomElement().find('a.adder').bind('click', function () {

                $(this).addClass('added');

                var usr_id = $(this).closest('tr').find('input[name="usr_id"]').val();

                $this.loadUser(usr_id, $this.selectUser);

                return false;
            });

            $dialog.getDomElement().find('a.adder').each(function (i, el) {

                var usr_id = $(this).closest('tr').find('input[name="usr_id"]').val();

                if ($('.badge_' + usr_id, $this.container).length > 0) {
                    $(this).addClass('added');
                }

            });


            return false;
        });

        //this.container.on('submit', '#PushBox form[name="FeedBackForm"]', function (event) {
        $('#PushBox form[name="FeedBackForm"]').bind('submit', function () {

            var $this = $(this);

            $.ajax({
                type: $this.attr('method'),
                url: $this.attr('action'),
                dataType: 'json',
                data: $this.serializeArray(),
                beforeSend: function () {

                },
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);
                        p4.Dialog.Close(1);
                        p4.WorkZone.refresh();
                    }
                    else {
                        humane.error(data.message);
                    }
                    return;
                },
                error: function () {

                    return;
                },
                timeout: function () {

                    return;
                }
            });

            return false;
        });

        $('.FeedbackSend', this.container).bind('click', function () {
            if ($('.badges .badge', $container).length === 0) {
                alert(language.FeedBackNoUsersSelected);
                return;
            }

            var buttons = {};

            buttons[language.send] = function () {
                if ($.trim($('input[name="name"]', $dialog.getDomElement()).val()) === '') {
                    var options = {
                        size: 'Alert',
                        closeButton: true,
                        title: language.warning
                    }
                    var $dialogAlert = p4.Dialog.Create(options, 3);
                    $dialogAlert.setContent(language.FeedBackNameMandatory);

                    return false;
                }

                $dialog.Close();

                $('input[name="name"]', $FeedBackForm).val($('input[name="name"]', $dialog.getDomElement()).val());
                $('input[name="duration"]', $FeedBackForm).val($('select[name="duration"]', $dialog.getDomElement()).val());
                $('textarea[name="message"]', $FeedBackForm).val($('textarea[name="message"]', $dialog.getDomElement()).val());
                $('input[name="recept"]', $FeedBackForm).prop('checked', $('input[name="recept"]', $dialog.getDomElement()).prop('checked'));
                $('input[name="force_authentication"]', $FeedBackForm).prop('checked', $('input[name="force_authentication"]', $dialog.getDomElement()).prop('checked'));

                $FeedBackForm.trigger('submit');
            };

            var options = {
                size: 'Medium',
                buttons: buttons,
                loading: true,
                title: language.send,
                closeOnEscape: true,
                cancelButton: true
            };

            var $dialog = p4.Dialog.Create(options, 2);

            var $FeedBackForm = $('form[name="FeedBackForm"]', $container);

            var html = _.template($("#feedback_sendform_tpl").html());

            $dialog.setContent(html);

            $('input[name="name"]', $dialog.getDomElement()).val($('input[name="name"]', $FeedBackForm).val());
            $('textarea[name="message"]', $dialog.getDomElement()).val($('textarea[name="message"]', $FeedBackForm).val());
            $('.' + $this.Context, $dialog.getDomElement()).show();

            $('form', $dialog.getDomElement()).submit(function () {
                return false;
            });
        });

        $('.user_content .badges', this.container).disableSelection();


        // toggle download feature for users
        this.container.on('click', '.user_content .badges .badge .toggle', function (event) {
            var $this = $(this);

            $this.toggleClass('status_off status_on');

            $this.find('input').val($this.hasClass('status_on') ? '1' : '0');

            return false;
        });

        // toggle feature state of selected users
        this.container.on('click', '.general_togglers .general_toggler', function (event) {
            var feature = $(this).attr('feature');

            var $badges = $('.user_content .badge.selected', this.container);

            var toggles = $('.status_off.toggle_' + feature, $badges);

            if (toggles.length === 0) {
                var toggles = $('.status_on.toggle_' + feature, $badges);
            }
            if (toggles.length === 0) {
                humane.info('No user selected');
            }

            toggles.trigger('click');
            return false;
        });

        this.container.on('click', '.user_content .badges .badge .deleter', function (event) {
            var $elem = $(this).closest('.badge');
            $elem.fadeOut(function () {
                $elem.remove();
            });
            return false;
        });

        this.container.on('click', '.list_manager', function (event) {
            $('#PushBox').hide();
            $('#ListManager').show();
            return false;
        });

        this.container.on('click', 'a.list_loader', function (event) {
            var url = $(this).attr('href');

            var callbackList = function (list) {
                for (var i in list.entries) {
                    this.selectUser(list.entries[i].User);
                }
            };

            $this.loadList(url, callbackList);

            return false;
        });

        $('form.list_saver', this.container).bind('submit', function () {
            var $form = $(this);
            var $input = $('input[name="name"]', $form);

            var users = p4.Feedback.getUsers();

            if (users.length === 0) {
                humane.error('No users');
                return false;
            }

            p4.Lists.create($input.val(), function (list) {
                $input.val('');
                list.addUsers(users);
            });

            return false;
        });

        $('input[name="users-search"]', this.container).autocomplete({
            minLength: 2,
            source: function (request, response) {
                $.ajax({
                    url: '/prod/push/search-user/',
                    dataType: "json",
                    data: {
                        query: request.term
                    },
                    success: function (data) {
                        response(data);
                    }
                });
            },
            focus: function (event, ui) {
                $('input[name="users-search"]').val(ui.item.label);
            },
            select: function (event, ui) {
                if (ui.item.type === 'USER') {
                    $this.selectUser(ui.item);
                } else if (ui.item.type === 'LIST') {
                    for (var e in ui.item.entries) {
                        $this.selectUser(ui.item.entries[e].User);
                    }
                }
                return false;
            }
        })
            .data("ui-autocomplete")._renderItem = function (ul, item) {
            var html = "";

            if (item.type === 'USER') {
                html = _.template($("#list_user_tpl").html(), {

                    item: item
                });
            } else if (item.type === 'LIST') {
                html = _.template($("#list_list_tpl").html(), {
                    item: item
                });
            }

            return  $(html).data("ui-autocomplete-item", item).appendTo(ul);
        };

        return this;
    };

    Feedback.prototype = {
        selectUser: function (user) {
            if (typeof user !== 'object') {
                if (window.console) {
                    console.log('trying to select a user with wrong datas');
                }
            }
            if ($('.badge_' + user.usr_id, this.container).length > 0) {
                humane.info('User already selected');
                return;
            }

            var html = _.template($("#" + this.Context.toLowerCase() + "_badge_tpl").html(), {
                user: user
            });

            p4.Feedback.appendBadge(html);
        },
        loadUser: function (usr_id, callback) {
            var $this = this;

            $.ajax({
                type: 'GET',
                url: '/prod/push/user/' + usr_id + '/',
                dataType: 'json',
                data: {
                    usr_id: usr_id
                },
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback.call($this, data);
                    }
                }
            });
        },
        loadList: function (url, callback) {
            var $this = this;

            $.ajax({
                type: 'GET',
                url: url,
                dataType: 'json',
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback.call($this, data);
                    }
                }
            });
        },
        appendBadge: function (badge) {
            $('.user_content .badges', this.container).append(badge);
        },
        addUser: function ($form, callback) {

            var $this = this;

            $.ajax({
                type: 'POST',
                url: '/prod/push/add-user/',
                dataType: 'json',
                data: $form.serializeArray(),
                success: function (data) {
                    if (data.success) {
                        humane.info(data.message);
                        $this.selectUser(data.user);
                        callback();
                    }
                    else {
                        humane.error(data.message);
                    }
                }
            });
        },
        getSelection: function () {
            return this.selection;
        },
        getUsers: function () {
            return $('.user_content .badge', this.container).map(function () {
                return $('input[name="id"]', $(this)).val();
            });
        }
    };


    var ListManager = function ($container) {

        this.list = null;
        this.container = $container;

        $container.on('click', '.back_link', function() {
            $('#PushBox').show();
            $('#ListManager').hide();
            return false;
        });

        $container.on('click', 'a.list_sharer', function() {

            var $this = $(this),
                options = {
                    size: 'Small',
                    closeButton: true,
                    title: $this.attr('title')
                },
                $dialog = p4.Dialog.Create(options, 2);

            $dialog.load($this.attr('href'), 'GET');

            return false;
        });

        $container.on('click', 'a.user_adder', function() {

            var $this = $(this);

            $.ajax({
                type: "GET",
                url: $this.attr('href'),
                dataType: 'html',
                beforeSend: function () {
                    var options = {
                        size: 'Medium',
                        title: $this.html()
                    };
                    p4.Dialog.Create(options, 2).getDomElement().addClass('loading');
                },
                success: function (data) {
                    p4.Dialog.get(2).getDomElement().removeClass('loading').empty().append(data);
                    return;
                },
                error: function () {
                    p4.Dialog.get(2).Close();
                    return;
                },
                timeout: function () {
                    p4.Dialog.get(2).Close();
                    return;
                }
            });

            return false;
        });


        var initLeft = function () {
            $('a.list_refresh', $container).bind('click', function (event) {

                var callback = function (datas) {
                    $('.all-lists', $container).removeClass('loading').append(datas);
                    initLeft();
                };

                $('.all-lists', $container).empty().addClass('loading');

                p4.Lists.get(callback, 'html');

                return false;
            });

            $('a.list_adder', $container).bind('click', function (event) {

                var makeDialog = function (box) {

                    var buttons = {};

                    buttons[language.valider] = function () {

                        var callbackOK = function () {
                            $('a.list_refresh', $container).trigger('click');
                            p4.Dialog.get(2).Close();
                        };

                        var name = $('input[name="name"]', p4.Dialog.get(2).getDomElement()).val();

                        if ($.trim(name) === '') {
                            alert(language.listNameCannotBeEmpty);
                            return;
                        }

                        p4.Lists.create(name, callbackOK);
                    };

                    var options = {
                        cancelButton: true,
                        buttons: buttons,
                        size: '700x170'
                    };

                    p4.Dialog.Create(options, 2).setContent(box);
                };

                var html = _.template($("#list_editor_dialog_add_tpl").html());

                makeDialog(html);

                return false;
            });

            $('li.list a.list_link', $container).bind('click', function (event) {

                var $this = $(this);

                $this.closest('.lists').find('.list.selected').removeClass('selected');
                $this.parent('li.list').addClass('selected');

                $.ajax({
                    type: 'GET',
                    url: $this.attr('href'),
                    dataType: 'html',
                    success: function (data) {
                        $('.editor', $container).removeClass('loading').append(data);
                        initRight();
                    },
                    beforeSend: function () {
                        $('.editor', $container).empty().addClass('loading');
                    }
                });

                return false;
            });
        };

        var initRight = function () {

            var $container = $('#ListManager .editor');

            $('form[name="list-editor-search"]', $container).bind('submit', function () {

                var $this = $(this);
                var dest = $('.list-editor-results', $container);

                $.ajax({
                    url: $this.attr('action'),
                    type: $this.attr('method'),
                    dataType: "html",
                    data: $this.serializeArray(),
                    beforeSend: function () {
                        dest.empty().addClass('loading');
                    },
                    success: function (datas) {

                        dest.empty().removeClass('loading').append(datas);
                    }
                });
                return false;
            });

            $('form[name="list-editor-search"] select, form[name="list-editor-search"] input[name="ListUser"]', $container).bind('change', function () {
                $(this).closest('form').trigger('submit');
            });

            $('.EditToggle', $container).bind('click', function () {
                $('.content.readonly, .content.readwrite', $('#ListManager')).toggle();
                return false;
            });
            $('.Refresher', $container).bind('click', function () {
                $('#ListManager ul.lists .list.selected a').trigger('click');
                return false;
            });

            $('form[name="SaveName"]', $container).bind('submit', function () {
                var $this = $(this);

                $.ajax({
                    type: $this.attr('method'),
                    url: $this.attr('action'),
                    dataType: 'json',
                    data: $this.serializeArray(),
                    beforeSend: function () {

                    },
                    success: function (data) {
                        if (data.success) {
                            humane.info(data.message);
                            $('#ListManager .lists .list_refresh').trigger('click');
                        }
                        else {
                            humane.error(data.message);
                        }
                        return;
                    },
                    error: function () {

                        return;
                    },
                    timeout: function () {

                        return;
                    }
                });

                return false;
            });


            $('button.deleter', $container).bind('click', function (event) {

                var list_id = $(this).data("list-id");

                var makeDialog = function (box) {

                    var buttons = {};

                    buttons[language.valider] = function () {

                        var callbackOK = function () {
                            $('#ListManager .all-lists a.list_refresh').trigger('click');
                            p4.Dialog.get(2).Close();
                        };

                        var List = new document.List(list_id);
                        List.remove(callbackOK);
                    };

                    var options = {
                        cancelButton: true,
                        buttons: buttons,
                        size: 'Alert'
                    };

                    p4.Dialog.Create(options, 2).setContent(box);
                };

                var html = _.template($("#list_editor_dialog_delete_tpl").html());

                makeDialog(html);

                return false;
            });
        };

        initLeft();

        $('.badges a.deleter', this.container).on('click', function () {
            var badge = $(this).closest('.badge');

            var usr_id = badge.find('input[name="id"]').val();


            var callback = function (list, datas) {
                $('.counter.current, .list.selected .counter', $('#ListManager')).each(function () {
                    $(this).text(parseInt($(this).text()) - 1);
                });

                badge.remove();
            };

            p4.ListManager.getList().removeUser(usr_id, callback);

            return false;
        });

    };

    ListManager.prototype = {
        workOn: function (list_id) {
            this.list = new document.List(list_id);
        },
        getList: function () {
            return this.list;
        },
        appendBadge: function (datas) {
            $('#ListManager .badges').append(datas);
        }
    };


    window.Feedback = Feedback;
    window.ListManager = ListManager;

}(window));
;var p4 = p4 || {};

(function (p4, window) {

    p4.Results = {
        'Selection': new Selectable($('#answers'), {
            selector: '.IMGT',
            limit: 800,
            selectStart: function (event, selection) {
                $('#answercontextwrap table:visible').hide();
            },
            selectStop: function (event, selection) {
                viewNbSelect();
            },
            callbackSelection: function (element) {
                var elements = $(element).attr('id').split('_');

                return elements.slice(elements.length - 2, elements.length).join('_');
            }
        })
    };

    return;
}(p4, window));
;document.getElementById('loader_bar').style.width = '30%';

var p4 = p4 || {};

var baskAjax, baskAjaxrunning;
baskAjaxrunning = false;
var answAjax, answAjaxrunning;
answAjaxrunning = false;
var searchAjax, searchAjaxRunning;
searchAjaxRunning = false;

var language = {};
var bodySize = {
    x: 0,
    y: 0
};

var selectedFacets = {};
var facets = null;

var ORDER_BY_BCT = "ORDER_BY_BCT";
var ORDER_ALPHA_ASC = "ORDER_ALPHA_ASC";
var ORDER_BY_HITS = "ORDER_BY_HITS";



function resizePreview() {
    p4.preview.height = $('#PREVIEWIMGCONT').height();
    p4.preview.width = $('#PREVIEWIMGCONT').width();
    setPreview();
}

function getHome(cas, page) {

    if (typeof(page) === 'undefined')
        page = 0;

    switch (cas) {
        case 'QUERY':
            selectedFacets = {};
            newSearch($("#EDIT_query").val());
            break;
        case 'PUBLI':
            publicationModule.fetchPublications(page, answAjax, answAjaxrunning);
            break;
        case 'HELP':
            $.ajax({
                type: "POST",
                url: "/client/home/",
                dataType: 'html',
                data: {
                    type: cas
                },
                beforeSend: function () {
                    if (answAjaxrunning && answAjax.abort)
                        answAjax.abort();
                    clearAnswers();
                    answAjaxrunning = true;
                    $('#answers').addClass('loading');
                },
                error: function () {
                    answAjaxrunning = false;
                    $('#answers').removeClass('loading');
                },
                timeout: function () {
                    answAjaxrunning = false;
                    $('#answers').removeClass('loading');
                },
                success: function (data) {
                    answAjaxrunning = false;
                    $('#answers').append(data);
                    afterSearch();
                    return;
                }

            });
            break;


        default:
            break;
    }
}

function getLanguage() {
    $.ajax({
        type: "GET",
        url: "../prod/language/",
        dataType: 'json',
        success: function (data) {
            language = data;
            return;
        }
    });
}

function is_ctrl_key(event) {
    if (event.altKey)
        return true;
    if (event.ctrlKey)
        return true;
    if (event.metaKey)	// apple key opera
        return true;
    if (event.keyCode === 17)	// apple key opera
        return true;
    if (event.keyCode === 224)	// apple key mozilla
        return true;
    if (event.keyCode === 91)	// apple key safari
        return true;

    return false;
}

function is_shift_key(event) {
    if (event.shiftKey)
        return true;
    return false;
}

/**
 * adv search : check/uncheck all the collections (called by the buttons "all"/"none")
 *
 * @param bool
 */
function checkBases(bool) {
    $('form.phrasea_query .sbas_list').each(function () {

        var sbas_id = $(this).find('input[name=reference]:first').val();
        if (bool)
            $(this).find(':checkbox').prop('checked', true);
        else
            $(this).find(':checkbox').prop('checked', false);
    });

    checkFilters(true);
}

function checkFilters(save) {
    var danger = false;
    var search = {
        bases: {},
        fields: [],
        dates: {},
        status: [],
        elasticSort: {}

    };

    var adv_box = $('form.phrasea_query .adv_options');
    var container = $("#ADVSRCH_OPTIONS_ZONE");
    var fieldsSort = $('#ADVSRCH_SORT_ZONE select[name=sort]', container);
    var fieldsSortOrd = $('#ADVSRCH_SORT_ZONE select[name=ord]', container);
    var fieldsSelectFake = $('#ADVSRCH_FIELDS_ZONE select.term_select_field', container);
    var statusField = $('#ADVSRCH_FIELDS_ZONE .danger_indicator', container);
    var statusFilters = $('#ADVSRCH_SB_ZONE .status-section-title .danger_indicator', container);
    var dateFilterSelect = $('#ADVSRCH_DATE_ZONE select', container);

    // hide all the fields in the "sort by" select, so only the relevant ones will be shown again
    $("option.dbx", fieldsSort).hide().prop("disabled", true);  // dbx is for "field of databases"

    // hide all the fields in the "fields" select, so only the relevant ones will be shown again
    $("option.dbx", fieldsSelectFake).prop("disabled", true);
    // disable the whole select
    $('#ADVSRCH_FIELDS_ZONE .term_select_wrapper select.term_select_field', container).enable(false);
    $('#ADVSRCH_FIELDS_ZONE .term_select_wrapper select.term_select_op', container).enable(false);
    $('#ADVSRCH_FIELDS_ZONE .term_select_wrapper input.term_select_value', container).enable(false);

    // hide all the fields in the "date field" select, so only the relevant ones will be shown again
    $("option.dbx", dateFilterSelect).hide().prop("disabled", true);   // dbx = all "field" entries in the select = all except the firstt

    statusFilters.removeClass('danger');
    $.each($('#ADVSRCH_SB_ZONE .field_switch'), function(index,el){
        if( $(el).prop('checked') === true ) {
            danger = true;
            statusFilters.addClass('danger');
        }
    });

    var nbTotalSelectedColls = 0;
    $.each($('.sbascont', adv_box), function () {
        var $this = $(this);

        var sbas_id = $this.parent().find('input[name="reference"]').val();
        search.bases[sbas_id] = [];

        var nbCols = 0;
        var nbSelectedColls = 0;
        $this.find('.checkbas').each(function (idx, el) {
            nbCols++;
            if($(this).prop('checked')) {
                nbSelectedColls++;
                nbTotalSelectedColls++;
                search.bases[sbas_id].push($(this).val());
            }
        });

        // display the number of selected colls for the databox
        if (nbSelectedColls == nbCols) {
            $('.infos_sbas_' + sbas_id).empty().append(nbCols);
            $(this).siblings(".clksbas").removeClass("danger");
            $(this).siblings(".clksbas").find(".custom_checkbox_label input").prop("checked", "checked");
        }
        else {
            $('.infos_sbas_' + sbas_id).empty().append('<span style="color:#2096F3;font-size: 20px;">' + nbSelectedColls + '</span> / ' + nbCols);
            $(this).siblings(".clksbas").addClass("danger");
            danger = true;
        }

        // if one coll is not checked, show danger
        if(nbSelectedColls != nbCols) {
            $("#ADVSRCH_SBAS_ZONE").addClass("danger");
            danger = true;
        }
        else if (nbSelectedColls == nbCols && danger == false) {
            $("#ADVSRCH_SBAS_ZONE").removeClass("danger");
        }

        if(nbSelectedColls == 0) {
            // no collections checked for this databox
            // hide the status bits
            $("#ADVSRCH_SB_ZONE_"+sbas_id, container).hide();
            // uncheck
            $("#ADVSRCH_SB_ZONE_"+sbas_id+" input:checkbox", container).prop("checked", false);
        }
        else {
            // at least one coll checked for this databox
            // show again the relevant fields in "sort by" select
            $(".db_"+sbas_id, fieldsSort).show().prop("disabled", false);
            // show again the relevant fields in "from fields" select
            $(".db_"+sbas_id, fieldsSelectFake).prop("disabled", false);
            // enable again the whole select, only if the selected option (field name) is relevant for this db
            $("option.db_"+sbas_id+":selected", fieldsSelectFake).closest("select").enable(true);
            // show the sb
            $("#ADVSRCH_SB_ZONE_"+sbas_id, container).show();
            // show again the relevant fields in "date field" select
            $(".db_"+sbas_id, dateFilterSelect).show().prop("disabled", false);
        }
    });

    // enable also the select if the first option ("choose:") was selected
    statusField.removeClass('danger');
    fieldsSelectFake.each(function(e) {
        var $this = $(this);
        var term_ok = $('option:selected:enabled', $this).closest(".term_select_wrapper");
        $("select.term_select_field", term_ok).enable(true);
        if($this.val() !== "") {
            $("select.term_select_op", term_ok).enable(true);
            $("input.term_select_value", term_ok).enable(true);
            danger = true;
            statusField.addClass('danger');
        }
    });

    if (nbTotalSelectedColls == 0) {
        // no collections checked at all
        // hide irrelevant filters
        $("#ADVSRCH_OPTIONS_ZONE").hide();
    }
    else {
        // at least one collection checked
        // show relevant filters
        $("#ADVSRCH_OPTIONS_ZONE").show();
    }

    // --------- sort  --------

    // if no field is selected for sort, select the default option
    if($("option:selected:enabled", fieldsSort).length == 0) {
        $("option.default-selection", fieldsSort).prop("selected", true);
        $("option.default-selection", fieldsSortOrd).prop("selected", true);
    }

    search.elasticSort.by = $("option:selected:enabled", fieldsSort).val();
    search.elasticSort.order = $("option:selected:enabled", fieldsSortOrd).val();

    //--------- status bits filter ---------

    // here only the relevant sb are checked
    for(sbas_id in search.bases) {
        var n_checked = 0;
        var n_unchecked = 0;
        $("#ADVSRCH_SB_ZONE_"+sbas_id+" :checkbox", container).each(function (k, o) {
            var n = $(this).data('sb');
            if($(o).attr('checked')) {
                search.status[n] = $(this).val().split('_');
                n_checked++;
            }
            else {
                n_unchecked++;
            }
        });
        if(n_checked === 0) {
            $("#ADVSRCH_SB_ZONE_"+sbas_id, container).removeClass('danger');
        }
        else {
            $("#ADVSRCH_SB_ZONE_"+sbas_id, container).addClass('danger');
            danger = true;
        }
    }

    //--------- dates filter ---------

    // if no date field is selected for filter, select the first option
    $('#ADVSRCH_DATE_ZONE', adv_box).removeClass('danger');
    if($("option:selected:enabled", dateFilterSelect).length == 0) {
        $("option:eq(0)", dateFilterSelect).prop("selected", true);
    }
    if($("option:selected", dateFilterSelect).val() !== "") {
        $("#ADVSRCH_DATE_SELECTORS", container).show();
        search.dates.minbound = $('#ADVSRCH_DATE_ZONE input[name=date_min]', adv_box).val();
        search.dates.maxbound = $('#ADVSRCH_DATE_ZONE input[name=date_max]', adv_box).val();
        search.dates.field = $('#ADVSRCH_DATE_ZONE select[name=date_field]', adv_box).val();

        if ($.trim(search.dates.minbound) || $.trim(search.dates.maxbound)) {
            danger = true;
            $('#ADVSRCH_DATE_ZONE', adv_box).addClass('danger');
        }
    }
    else {
        $("#ADVSRCH_DATE_SELECTORS", container).hide();
    }

    // if one filter shows danger, show it on the query
    if (danger == true) {
        $('#EDIT_query').addClass('danger');
    }
    else {
        $('#EDIT_query').removeClass('danger');
    }

    if (save === true) {
        setPref('search', JSON.stringify(search));
    }
}

function toggleFilter(filter, ele) {
    var el = $('#' + filter);
    if (el.is(':hidden'))
        $(ele).parent().addClass('open');
    else
        $(ele).parent().removeClass('open');
    el.slideToggle('fast');
}


function setVisible(el) {
    el.style.visibility = 'visible';
}

function resize() {
    var body = $('#mainContainer');
    bodySize.y = body.height();
    bodySize.x = body.width();

    $('.overlay').height(bodySize.y).width(bodySize.x);

    var headBlockH = $('#headBlock').outerHeight();
    var bodyY = bodySize.y - headBlockH - 2;
    var bodyW = bodySize.x - 2;
    //$('#desktop').height(bodyY).width(bodyW);

    if (p4.preview.open)
        resizePreview();

    if ($('#idFrameC').data('ui-resizable')) {
        $('#idFrameC').resizable('option', 'maxWidth', (480));
        $('#idFrameC').resizable('option', 'minWidth', 360);
    }

    answerSizer();
    linearize();
}


function clearAnswers() {
    $('#formAnswerPage').val('');
    $('#searchForm input[name="nba"]').val('');
    $('#answers, #dyn_tool').empty();
}

function reset_adv_search() {
    var jsq = {
        "sort":{
            "field":"created_on",
            "order":"desc"
        },
        "use_truncation":false,
        "phrasea_recordtype":"RECORD",
        "phrasea_mediatype":"",
        "bases":[ ],
        "statuses":[ ],
        "query":{
            "_ux_zone":"PROD",
            "type":"CLAUSES",
            "must_match":"ALL",
            "enabled":true,
            "clauses":[
                {
                    "_ux_zone":"FIELDS",
                    "type":"CLAUSES",
                    "must_match":"ALL",
                    "enabled":false,
                    "clauses":[ ]
                },
                {
                    "_ux_zone":"DATE-FIELD",
                    "type":"DATE-FIELD",
                    "field":"",
                    "from":"",
                    "to":"",
                    "enabled":false
                },
                {
                    "_ux_zone":"AGGREGATES",
                    "type":"CLAUSES",
                    "must_match":"ALL",
                    "enabled":false,
                    "clauses":[ ]
                }
            ]
        },
        "_selectedFacets":{ }
    };

    restoreJsonQuery(jsq, false);
}


function newSearch(query) {
    p4.Results.Selection.empty();

    clearAnswers();
    if(query !== null) {
        // $('#SENT_query').val(query);

        var histo = $('#history-queries ul');

        histo.prepend('<li onclick="doSpecialSearch(\'' + query.replace(/\'/g, "\\'") + '\')">' + query + '</li>');

        var lis = $('li', histo);
        if (lis.length > 25) {
            $('li:last', histo).remove();
        }
    }

    $('#idFrameC li.proposals_WZ').removeClass('active');

    $('#searchForm').submit();
    return false;
}

function beforeSearch() {
    if (answAjaxrunning)
        return;
    answAjaxrunning = true;

    clearAnswers();
    $('#tooltip').css({
        'display': 'none'
    });
    $('#answers').addClass('loading').empty();
    $('#answercontextwrap').remove();
}

function afterSearch() {
    if ($('#answercontextwrap').length === 0)
        $('body').append('<div id="answercontextwrap"></div>');

    $.each($('#answers .contextMenuTrigger'), function () {

        var id = $(this).closest('.IMGT').attr('id').split('_').slice(1, 3).join('_');

        $(this).contextMenu('#IMGT_' + id + ' .answercontextmenu', {
            appendTo: '#answercontextwrap',
            openEvt: 'click',
            dropDown: true,
            theme: 'vista',
            showTransition: 'slideDown',
            hideTransition: 'hide',
            shadow: false
        });
    });

    answAjaxrunning = false;
    $('#answers').removeClass('loading');
    $('.captionTips, .captionRolloverTips').tooltip({
        delay: 0,
        isBrowsable: false,
        extraClass: 'caption-tooltip-container'
    });
    $('.infoTips').tooltip({
        delay: 0
    });
    $('.previewTips').tooltip({
        fixable: true
    });
    $('.captionRolloverTips').tooltip({
        fixable: true
    });
    $('.thumb .rollovable').hover(
        function () {
            $('.rollover-gif-hover', this).show();
            $('.rollover-gif-out', this).hide();
        },
        function () {
            $('.rollover-gif-hover', this).hide();
            $('.rollover-gif-out', this).show();
        }
    );
    viewNbSelect();
    $('#answers div.IMGT').draggable({
        helper: function () {
            $('body').append('<div id="dragDropCursor" style="position:absolute;z-index:9999;background:red;-moz-border-radius:8px;-webkit-border-radius:8px;"><div style="padding:2px 5px;font-weight:bold;">' + p4.Results.Selection.length() + '</div></div>');
            return $('#dragDropCursor');
        },
        scope: "objects",
        distance: 20,
        scroll: false,
        cursorAt: {
            top: -10,
            left: -20
        },
        start: function (event, ui) {
            if (!$(this).hasClass('selected'))
                return false;
        }
    });
    linearize();
}

function initAnswerForm() {

    var searchForm = $('#searchForm');
    $('button[type="submit"]', searchForm).bind('click', function () {
        selectedFacets = {};
        newSearch($("#EDIT_query").val());
        return false;
    });

    searchForm.unbind('submit').bind('submit', function () {
        var $this = $(this),
            method = $this.attr('method') ? $this.attr('method') : 'POST';

        var data = $this.serializeArray();

        // fix bug : if a sb is dual checked, both values are sent with the SAME name
        //     we can remove those since it means we don't care about this sb
        // /!\ silly fixed bug : in sb[] we will test if a key exists using "_undefined()"
        //     BUT sb["sort”] EXISTS ! it is the array.sort() function !
        //     so the side effect in "_filter()" was that data["sort"] was removed.
        //     quick solution : prefix the key with "k_"
        var sb = [];
        _.each(data, function (v) {
            var name = "k_" + v.name;
            if (name.substr(0, 9) === "k_status[") {
                if (_.isUndefined(sb[name])) {
                    sb[name] = 0;
                }
                sb[name]++; // so sb["k_x"] is the number of occurences of sb checkbox named "x"
            }
        });
        // now if a sb checkbox appears 2 times, it is removed from data
        data = _.filter(data, function (e) {
            return (_.isUndefined(sb["k_" + e.name])) || (sb["k_" + e.name] === 1);
        });
        // end of sb fix


        var jsonData = serializeJSON(data, selectedFacets, facets);
        var qry = buildQ(jsonData.query);

        data.push({
                name: 'jsQuery',
                value: JSON.stringify(jsonData)
            },
            {
                name: 'qry',
                value: qry
            });

        answAjax = $.ajax({
            type: method,
            url: $this.attr('action'),
            data: data,
            dataType: 'json',
            beforeSend: function (formData) {
                if (answAjaxrunning && answAjax.abort)
                    answAjax.abort();
                beforeSearch();
            },
            error: function () {
                answAjaxrunning = false;
                $('#answers').removeClass('loading');
            },
            timeout: function () {
                answAjaxrunning = false;
                $('#answers').removeClass('loading');
            },
            success: function (datas) {
                // DEBUG QUERY PARSER
                try {
                    console.info(JSON.parse(datas.parsed_query));
                }
                catch(e) {}

                if(datas.total_answers > 0) {
                    sessionStorage.setItem('search', JSON.stringify(datas.query));
                }

                $('#answers').empty().append(datas.results).removeClass('loading');

                $("#answers img.lazyload").lazyload({
                    container: $('#answers')
                });

                //load last result collected or [] if length == 0
                if(!datas.facets) {
                    datas.facets = [];
                }
                facets = datas.facets;
                loadFacets(facets);

                $('#answers').append('<div id="paginate"><div class="navigation"><div id="tool_navigate"></div></div></div>');

                $('#tool_results').empty().append(datas.infos);
                $('#tool_navigate').empty().append(datas.navigationTpl);

                $.each(p4.Results.Selection.get(), function (i, el) {
                    $('#IMGT_' + el).addClass('selected');
                });

                p4.tot = datas.total_answers;
                p4.tot_options = datas.form;
                p4.tot_query = datas.query;
                p4.navigation = datas.navigation;

                if (datas.next_page) {
                    $("#NEXT_PAGE, #answersNext").bind('click', function () {
                        gotopage(datas.next_page);
                    });
                }
                else {
                    $("#NEXT_PAGE").unbind('click');
                }

                if (datas.prev_page) {
                    $("#PREV_PAGE").bind('click', function () {
                        gotopage(datas.prev_page);
                    });
                }
                else {
                    $("#PREV_PAGE").unbind('click');
                }

                afterSearch();
            }
        });
        return false;
    });

    // if defined, play the first query
    //
    try {
        var jsq = $("#FIRST_QUERY_CONTAINER");
        if(jsq.length > 0) {
            // there is a query to play
            if(jsq.data('format') === "json") {
                // json
                jsq = JSON.parse(jsq.text());
                restoreJsonQuery(jsq, true);
            }
            else {
                // text : do it the old way : restore only fulltext and submit
                searchForm.trigger('submit');
            }
        }
    }
    catch (e) {
        // malformed jsonquery ?
        // no-op
    }
}

var facetStatus = $.parseJSON(sessionStorage.getItem('facetStatus')) || [];
var tokens = [['[',']']];

/**
 *  add missing selected facets fields into "facets", from "selectedFacets"
 *  why : because if we negates all values for a facet field (all red), the facet will disapear from next query->answers
 *        (not in "facets" anymore, not in ux). So we lose the posibility to delete or invert a facet value.
 *  nb : negating all facets values does not mean there will be 0 results, because the field can be empty for some records.
 */
function facetsAddMissingSelected(_selectedFacets, _facets) {
    _.each(_selectedFacets, function (v, k) {
        // todo : "facets" is an array (no key), change to a k->v (same structure as "selectedFacets")
        //        too bad for now, we must lookup...
        var found = _.find(_facets, function (facet) {
            return (facet.field == k);
        });
        if(!found) {
            var i = _facets.push(_.clone(v)); // add a "fake" facet to facets
            _facets[i-1].values = [];      // with no values
        }
    });
}

function loadFacets(facets) {

    //get facets settings
    var hideSingleValueFacet_setting = $('#look_box_settings input[name=filter_facet]').prop('checked');
    var facetOrder = $('#look_box_settings select[name=orderFacet]').val();
    var facetValueOrder = $('#look_box_settings select[name=facetValuesOrder]').val();

    function sortIteration(i) {
        switch(facetValueOrder) {
            case ORDER_ALPHA_ASC:
                return i.value.toString().toLowerCase();
                break;
            case ORDER_BY_HITS:
                return i.count*-1;
                break;
        }
    }

    facetsAddMissingSelected(selectedFacets, facets); // add missing facets fields into "facets" from "selectedFacets"

    // Convert facets data to fancytree source format
    var treeSource = _.map(facets, function(facet) {
        // Values
        var values = _.map(_.sortBy(facet.values, sortIteration), function (value) {
            var type = facet.type;     // todo : define a new phraseanet "color" type for fields. for now we push a "type" for every value, copied from field type
            // patch "color" type values
            var textLimit = 15;     // cut long values (set to 0 to not cut)
            var text = (value.value).toString();
            var label = title = tooltip = text;
            var match = text.match(/^(.*)\[#([0-9a-fA-F]{6})].*$/);
            if(match && match[2] != null) {
                // text looks like a color !
                var colorCode = '#' + match[2];
                // add color circle and remove color code from text;
                var textWithoutColorCode = text.replace('[' + colorCode + ']', '');
                if (textLimit > 0 && textWithoutColorCode.length > textLimit) {
                    textWithoutColorCode = textWithoutColorCode.substring(0, textLimit) + '…';
                }
                // patch
                type = "COLOR-AGGREGATE";
                label = textWithoutColorCode;
                tooltip = _.escape(textWithoutColorCode);
                title = '<span class="color-dot" style="background-color: ' + colorCode + ';"></span> ' + tooltip;
            }
            else {
                // keep text as it is, just cut if too long
                if (textLimit > 0 && text.length > textLimit) {
                    text = text.substring(0, textLimit) + '…';
                }
                label = text;
                title = tooltip = _.escape(text);
            }

            return {
                // custom data
                query:     value.query,
                field:     facet.field,
                raw_value: value.raw_value,
                value:     value.value,
                label:     label,             // displayed when selected (blue/red), escape is done later (render)
                type:      type,              // todo : define a new phraseanet "color" type for fields. for now we push a "type" for every value
                count:     value.count,
                // jquerytree data
                title:     title + ' (' + value.count + ')',
                tooltip:   tooltip + ' (' + value.count + ')'
            };
        });
        // Facet
        return {
            // custom data
            name:     facet.name,
            field:    facet.field,
            label:    facet.label,
            type:     facet.type,
            // jquerytree data
            title:    facet.label,
            folder:   true,
            children: values,
            expanded: !_.some(facetStatus, function(o) { return _.has(o, facet.name)})
        };
    });

    treeSource.sort(sortFacets('title', true, function(a){return a.toUpperCase()}));

    if(facetOrder == ORDER_BY_BCT) {
        treeSource = sortByPredefinedFacets(treeSource, 'name', ['base_aggregate', 'collection_aggregate', 'doctype_aggregate']);
    }

    if(hideSingleValueFacet_setting == true) {
        treeSource = hideSingleValueFacet(treeSource);
    }

    return getFacetsTree().reload(treeSource)
        .done(function () {
            _.each($('#proposals').find('.fancytree-expanded'), function (element, i) {
                $(element).find('.fancytree-title, .fancytree-expander').css('line-height', $(element)[0].offsetHeight + 'px');

                var li_s = $(element).next().children('li');
                var ul = $(element).next();
                if(li_s.length > 5) {
                    _.each(li_s, function(el, i) {
                        if(i > 4) {
                            $(el).hide();
                        }
                    });

                    ul.append('<button class="see_more_btn">See more</button>');
                }
            });
            $('.see_more_btn').on('click', function() {
                $(this).closest('ul').children().show();
                $(this).hide();
                return false;
            });
        });
}

/**
 * hide facets with only one value (experimental)
 *
 * @param source    treesource
 * @returns {*}
 */
function hideSingleValueFacet(source) {
    var filteredSource = [];
    _.forEach(source, function(facet) {
        if(!_.isUndefined(facet.children) && (facet.children.length > 1 || !_.isUndefined(selectedFacets[facet.field]))) {
            filteredSource.push(facet);
        }
    });
    source = filteredSource;

    return source;
}

function sortByPredefinedFacets(source, field, predefinedFieldOrder) {
    var filteredSource = source,
        ordered = [];

    _.forEach(predefinedFieldOrder, function (fieldValue, index) {
        _.forEach(source, function (facet, facetIndex) {
            if (facet[field] !== undefined) {
                if (facet[field] === fieldValue) {
                    ordered.push(facet);
                    // remove from filtered
                    filteredSource.splice(facetIndex, 1);
                }
            }
        });
    });

    var olen = filteredSource.length;
    // fill predefined facets with non predefined facets
    for (var i = 0; i < olen; i++) {
        ordered.push(filteredSource[i]);
    }
    return ordered;
}

// from stackoverflow
// http://stackoverflow.com/questions/979256/sorting-an-array-of-javascript-objects/979325#979325
function sortFacets(field, reverse, primer) {
    var key = function (x) {return primer ? primer(x[field]) : x[field]};

    return function (a,b) {
        var A = key(a), B = key(b);
        return ( (A < B) ? -1 : ((A > B) ? 1 : 0) ) * [-1,1][+!!reverse];
    }
}

function getFacetsTree() {
    var $facetsTree = $('#proposals');
    if (!$facetsTree.data('ui-fancytree')) {
        $facetsTree.fancytree({
            clickFolderMode: 3, // activate and expand
            icons:false,
            source: [],

            activate: function(event, data){
                //var query = data.node.data.query;
                var eventType = event.originalEvent;
                //if user did not click, then no need to perform any query
                if(eventType == null) {
                    return;
                }
                var facet = data.node.parent;
                var facetData = {
                    value:   data.node.data,
                    enabled: true,
                    negated: event.altKey // ,
                    // mode:    event.altKey ? "EXCEPT" : "AND"
                };

                if (selectedFacets[facet.data.field] == null) {
                    selectedFacets[facet.data.field] = facet.data;
                    selectedFacets[facet.data.field].values = [];
                }
                selectedFacets[facet.data.field].values.push(facetData);

                $('#searchForm').submit();
            },

            collapse: function (event, data) {
                var dict = {};
                dict[data.node.data.name] = "collapse";
                if(_.findWhere(facetStatus, dict) !== undefined ) {
                    facetStatus = _.without(facetStatus, _.findWhere(facetStatus, dict))
                }
                facetStatus.push(dict);
                sessionStorage.setItem('facetStatus', JSON.stringify(facetStatus));
            },

            expand: function (event, data) {
                var dict = {};
                dict[data.node.data.name] = "collapse";
                if (_.findWhere(facetStatus, dict) !== undefined) {
                    facetStatus = _.without(facetStatus, _.findWhere(facetStatus, dict))
                }
                sessionStorage.setItem('facetStatus', JSON.stringify(facetStatus));
            },

            renderNode: function(event, data){
                var label = "";
                if(data.node.folder) {
                    // here we render a "fieldname" level
                    if (!_.isUndefined(selectedFacets[data.node.data.field])) {
                        // here the field already contains selected facetvalues (to be rendered blue or red)
                        if ($(".fancytree-folder", data.node.li).find('.dataNode').length == 0) {
                            var dataNode = document.createElement('div');
                            dataNode.setAttribute("class", "dataNode");
                            $(".fancytree-folder", data.node.li).append(
                                dataNode
                            );
                        } else {
                            //remove existing facets
                            $(".dataNode", data.node.li).empty();
                        }
                        _.each(selectedFacets[data.node.data.field].values, function (facetValue) {

                            label = facetValue.value.label;

                            var s_label = document.createElement("SPAN");
                            s_label.setAttribute("class", "facetFilter-label");
                            s_label.setAttribute("title", label);

                            // label is a string ; todo : count may be obsolete when ux is restored ? for now don't print
                            s_label.appendChild(document.createTextNode(label)); // + ' (' + facetValue.value.count + ')');

                            var s_gradient = document.createElement("SPAN");
                            s_gradient.setAttribute("class", "facetFilter-gradient");
                            s_gradient.appendChild(document.createTextNode("\u00A0"));

                            s_label.appendChild(s_gradient);

                            var s_facet = document.createElement("SPAN");
                            s_facet.setAttribute("class", "facetFilter_" + (facetValue.negated ? "EXCEPT" : "AND"));
                            s_facet.appendChild(s_label);

                            var s_closer = document.createElement("A");
                            s_closer.setAttribute("class", "facetFilter-closer");
                            s_closer = $(s_facet.appendChild(s_closer));

                            s_closer.click(
                                function (event) {
                                    event.stopPropagation();
                                    var facetField   = $(this).parent().data("facetField");
                                    var facetLabel   = $(this).parent().data("facetLabel");
                                    var facetNegated = $(this).parent().data("facetNegated");
                                    selectedFacets[facetField].values = _.reject(selectedFacets[facetField].values, function (facetValue) {
                                        return (facetValue.value.label == facetLabel && facetValue.negated == facetNegated);
                                    });

                                    $('#searchForm').submit();
                                    return false;
                                }
                            );

                            var newNode = document.createElement('div');
                            newNode.setAttribute("class", "newNode");
                            s_facet = $(newNode.appendChild(s_facet));
                            s_facet.data("facetField",   data.node.data.field);
                            s_facet.data("facetLabel",   label);
                            s_facet.data("facetNegated", facetValue.negated);

                            $(".fancytree-folder .dataNode", data.node.li).append(
                                newNode
                            );

                            s_facet.click(
                                function (event) {
                                    if (event.altKey) {
                                        event.stopPropagation();
                                        var facetField   = $(this).data("facetField");
                                        var facetLabel   = $(this).data("facetLabel");
                                        var facetNegated = $(this).data("facetNegated");
                                        var found = _.find(selectedFacets[facetField].values, function (facetValue) {
                                            return (facetValue.value.label == facetLabel && facetValue.negated == facetNegated);
                                        });
                                        if (found) {
                                            found.negated = !found.negated;
                                            //replace class attr
                                            $(this).replaceClass($(this).attr('class'), "facetFilter_" + (found.negated ? "EXCEPT" : "AND"));
                                            $('#searchForm').submit();
                                        }
                                    }
                                    return false;
                                }
                            );
                        });
                    }
                    else {
                        // here we render a facet value
                    }
                }
            }
        });

    }
    return $facetsTree.fancytree('getTree');
}

function findClauseBy_ux_zone(clause, ux_zone) {
    if(typeof clause._ux_zone != 'undefined' && clause._ux_zone === ux_zone) {
        return clause;
    }
    if(clause.type === "CLAUSES") {
        for(var i=0; i<clause.clauses.length; i++) {
            var r = findClauseBy_ux_zone(clause.clauses[i], ux_zone);
            if(r != null) {
                return r;
            }
        }
    }
    return null;
}

/**
 * restore the advansearch ux from a json-query
 * elements are restored thank's to custom properties ("_xxx") included in json.
 * nb : for now, _ux_ facets can't be restored _before_sending_the_query_,
 *      but since "selectedFacets" (js) IS restored, sending the query WILL restore facets.
 *
 * @param jsq
 * @param submit
 */
function restoreJsonQuery(jsq, submit) {
    var clause;

    // restore the "fulltext" input-text
    clause = findClauseBy_ux_zone(jsq.query, "FULLTEXT");
    if(clause) {
        $('#EDIT_query').val(clause.value);
    }

    // restore the "records/stories" radios
    if(!_.isUndefined(jsq.phrasea_recordtype)) {
        $('#searchForm INPUT[name=search_type][value="' + ((jsq.phrasea_recordtype == 'STORY') ? '1' : '0') + '"]').prop('checked', true);  // check one radio will uncheck siblings
    }

    // restore the "record type" menu (image, video, audio, ...)
    if(!_.isUndefined(jsq.phrasea_mediatype)) {
        $('#searchForm SELECT[name=record_type] OPTION[value="' + jsq.phrasea_mediatype.toLowerCase() + '"]').prop('selected', true);
    }

    // restore the "use truncation" checkbox
    if(!_.isUndefined(jsq.phrasea_mediatype)) {
        $('#ADVSRCH_USE_TRUNCATION').prop('checked', jsq.phrasea_mediatype);
    }

    // restore the "sort results" menus
    if(!_.isUndefined(jsq.sort)) {
        if(!_.isUndefined(jsq.sort.field)) {
            $('#ADVSRCH_SORT_ZONE SELECT[name=sort] OPTION[value="' + jsq.sort.field + '"]').prop('selected', true);
        }
        if(!_.isUndefined(jsq.sort.order)) {
            $('#ADVSRCH_SORT_ZONE SELECT[name=ord] OPTION[value="' + jsq.sort.order + '"]').prop('selected', true);
        }
    }

    // restore the "bases" checkboxes
    if(!_.isUndefined(jsq.bases)) {
        $('#ADVSRCH_SBAS_ZONE INPUT.checkbas').attr('checked', false);
        if (jsq.bases.length > 0) {
            for (var i = 0; i < jsq.bases.length; i++) {
                $('#ADVSRCH_SBAS_ZONE INPUT.checkbas[value="' + jsq.bases[i] + '"]').attr('checked', true);
            }
        } else {
            // special case : EMPTY array ==> since it's a nonsense, check ALL bases
            $('#ADVSRCH_SBAS_ZONE INPUT.checkbas').attr('checked', true);
        }
    }

    // restore the multiples "fields" (field-menu + op-menu + value-input)
    clause = findClauseBy_ux_zone(jsq.query, "FIELDS");
    if(clause) {
        $('#ADVSRCH_FIELDS_ZONE INPUT[name=must_match][value="' + clause.must_match + '"]').attr('checked', true);
        $('#ADVSRCH_FIELDS_ZONE DIV.term_select_wrapper').remove();
        for (var i = 0; i < clause.clauses.length; i++) {
            var wrapper = AdvSearchAddNewTerm();    // div.term_select_wrapper
            var f = $(".term_select_field", wrapper);
            var o = $(".term_select_op", wrapper);
            var v = $(".term_select_value", wrapper);

            f.data('fieldtype', clause.clauses[i].type);
            $('option[value="' + clause.clauses[i].field + '"]', f).prop('selected', true);
            $('option[value="' + clause.clauses[i].operator + '"]', o).prop('selected', true);
            v.val(clause.clauses[i].value);
        }
        if(i === 0) {
            // if no field, add an empty one to ux
            AdvSearchAddNewTerm();
        }
    }

    // restore the "date field" (field-menu + from + to)
    clause = findClauseBy_ux_zone(jsq.query, "DATE-FIELD");
    if(clause) {
        $("#ADVSRCH_DATE_ZONE SELECT[name=date_field] option[value='" + clause.field + "']").prop('selected', true);
        $("#ADVSRCH_DATE_ZONE INPUT[name=date_min]").val(clause.from);
        $("#ADVSRCH_DATE_ZONE INPUT[name=date_max]").val(clause.to);
    }

    // restore the status-bits (for now dual checked status are restored unchecked)
    if(!_.isUndefined(jsq.statuses)) {
        $("#ADVSRCH_SB_ZONE INPUT:checkbox").prop('checked', false);
        _.each(jsq.statuses, function(db_statuses) {
            var db = db_statuses.databox;
            _.each(db_statuses.status, function(sb) {
                var i = sb.index;
                var v = sb.value ? "1" : "0";
                $("#ADVSRCH_SB_ZONE INPUT[name='status[" + db_statuses.databox + "][" + sb.index + "]'][value=" + v + "]").prop('checked', true);
            });
        });
    }

    // restore the selected facets (whole saved as custom property)
    if(!_.isUndefined(jsq._selectedFacets)) {
        selectedFacets = jsq._selectedFacets;
    }

    // the ux is restored, finish the job (hide unavailable fields/status etc, display "danger" where needed)
    checkFilters(false);
    // loadFacets([]);  // useless, facets will be restored after the query is sent

    if(submit) {
        $('#searchForm').submit();
    }
}

/**
 * advsearch : a status-bit is changed
 * --- NOW BUGGY ---
 * goal : do not allow to uncheck both sb values (1 and 0) since it's nonsense
 * since the "name" is the same for both values, it's buggy to send 2 values in request if both are checked.
 *    so the checkbox are "faked", and the real value is sent by 2 radios
 * @param o
 */
function advSearchChangeStatusBit(o) {
    /*
    * BUGGY : one sb should stay checked, but other js sometimes unselect all.
    * som other js don't know "fakestatus"..
    * todo : to be finished later
    var o_a = $(o);                 // the changed sb (value="0" or value="1")
    var b = o_a.data("sbas_id");
    var i = o_a.data("sb");
    var v = o_a.val();
    var chk = $("#ADVSRCH_SB_"+b+" INPUT[name='fakestatus[" + b + "][" + i + "]']");
    var o_b = chk.filter("[value=" + (v==='1'?'0':'1') + "]");     // the oposite sb (value="1" or value="0")

    // can't uncheck both
    if(!o_a.prop('checked')) {
        o_b.prop('checked', true);  // check the other one
    }

    // transform fakecheckboxes to radio
    var rad = $("#ADVSRCH_SB_"+b+" INPUT[name='status[" + b + "][" + i + "]']");
    rad.prop('checked', false);
    if(!chk.filter("[value=0]").prop('checked') && chk.filter("[value=1]").prop('checked')) {
        rad.filter("[value=1]").prop('checked', true);
    }
    else if (!chk.filter("[value=1]").prop('checked') && chk.filter("[value=0]").prop('checked')) {
        rad.filter("[value=0]").prop('checked', true);
    }
    */
    checkFilters(true);
}

function serializeJSON(data, selectedFacets, facets) {
    var json = {},
        obj = {},
        bases = [],
        statuses = [],
        fields = [],
        aggregates = []
    ;

    $.each(data, function(i, el) {
        obj[el.name] = el.value;

        var col = parseInt(el.value);

        if(el.name === 'bases[]') {
            bases.push(col);
        }

        if(0 && el.name.startsWith('status')) {
            var databoxId = el.name.match(/\d+/g)[0],
                databoxRow = el.name.match(/\d+/g)[1],
                statusMatch = false;

            $.each(statuses, function(i, status) {

                if (status.databox === databoxId) {
                    for (var j = 0; j < status.status.length; j++) {
                        var st = status.status[j].name;
                        var st_id = st.substr(0, st.indexOf(':'));

                        if (st_id === databoxRow) {
                            statusMatch = true;
                        }
                    }
                    statuses.splice((databoxId -1), 1);
                }
            });
            if (!statusMatch) {
                statuses.push({
                    'databox': databoxId,
                    'status': [
                        {
                            'index': databoxRow,
                            'value': !!(parseInt(el.value))
                        }
                    ]
                });
            }
        }
    });

    var _tmpStat = [];
    $("#ADVSRCH_SB_ZONE INPUT[type=checkbox]:checked").each(function(k, o) {
        o = $(o);
        var b = o.data("sbas_id");
        var i = o.data("sb");
        var v = o.val();
        if(_.isUndefined(_tmpStat[b])) {
            _tmpStat[b] = [];
        }
        if(_.isUndefined(_tmpStat[b][i])) {
            // first check
            _tmpStat[b][i] = v;
        }
        else {
            // both checked
            _tmpStat[b][i] = -1;
        }
    });
    _.each(_tmpStat, function(v, sbas_id){
        var status = []
        _.each(v, function(v, sb_index) {
            if( v !== -1) {     // ignore both checked
                status.push({
                    'index': sb_index,
                    'value': (v == "1")
                });
            }
        });
        statuses.push({
            'databox': sbas_id,
            'status': status
        });
    });


    $("#ADVSRCH_FIELDS_ZONE .term_select_wrapper").each(function(i, wrapper) {
        var f = $(".term_select_field option:selected", wrapper);
        var o = $(".term_select_op", wrapper);
        var v = $(".term_select_value", wrapper);
        var type = f.data('fieldtype');
        fields.push({
            'type'    : _.isString(type) ? type.toUpperCase() : "",
            'field'   : f.val(),
            'operator': o.val(),
            'value'   : v.val(),
            "enabled" : !f.prop('disabled') && f.val() !== '' && v.val() !== ''
        });
    });

    _.each(selectedFacets, function(facets) {
        _.each(facets.values, function(facetValue) {
            aggregates.push({
                'type'   : facetValue.value.type,
                'field'  : facetValue.value.field,
                'value'  : facetValue.value.raw_value,
                'negated': facetValue.negated,
                'enabled': facetValue.enabled
            });
        });
    });

    var date_field = $('#ADVSRCH_DATE_ZONE select[name=date_field]', 'form.phrasea_query .adv_options').val();
    var date_from  = $('#ADVSRCH_DATE_ZONE input[name=date_min]', 'form.phrasea_query .adv_options').val();
    var date_to    = $('#ADVSRCH_DATE_ZONE input[name=date_max]', 'form.phrasea_query .adv_options').val();

    json['sort'] = {
        'field': obj.sort,
        'order': obj.ord
    };
    json['perpage'] = parseInt($('#nperpage_value').val());
    json['page'] = obj.pag === "" ? 1 : parseInt(obj.pag);
    json['use_truncation'] = (obj.truncation === "on");
    json['phrasea_recordtype'] = obj.search_type == 0 ? 'RECORD' : 'STORY';
    json['phrasea_mediatype'] = obj.record_type.toUpperCase();
    json['bases'] = bases;
    json['statuses'] = statuses;
    json['query'] = {
        '_ux_zone': $('.menu-bar .selected').text().trim().toUpperCase(),
        'type': 'CLAUSES',
        'must_match': 'ALL',
        'enabled': true,
        'clauses': [
            {
                '_ux_zone': 'FULLTEXT',
                'type': 'FULLTEXT',
                'value': obj.fake_qry,
                'enabled': obj.fake_qry != ""
            },
            {
                '_ux_zone': 'FIELDS',
                'type': 'CLAUSES',
                'must_match': obj.must_match,
                'enabled': fields.length > 0,
                'clauses': fields
            },
            {
                '_ux_zone': "DATE-FIELD",
                "type": "DATE-FIELD",
                "field": date_field,
                "from": date_from,
                "to": date_to,
                "enabled": (date_field != "") && (date_from != "" || date_to != "")
            },
            {
                "_ux_zone": "AGGREGATES",
                "type": "CLAUSES",
                "must_match": "ALL",
                "enabled": aggregates.length > 0,
                "clauses": aggregates
            }
        ]
    };
    json['_selectedFacets'] = selectedFacets;

    return json;
}

var _ALL_Clause_ = "(created_on>1900/01/01)";
function buildQ(clause) {
    if(clause.enabled == false) {
        return "";
    }
    switch(clause.type) {
        case "CLAUSES":
            var t_pos = [];
            var t_neg = [];
            for(var i=0; i<clause.clauses.length; i++) {
                var _clause = clause.clauses[i];
                var _sub_q = buildQ(_clause);
                if(_sub_q !== "()" && _sub_q !== "") {
                    if(_clause.negated == true) {
                        t_neg.push(_sub_q);
                    }
                    else {
                        t_pos.push(_sub_q);
                    }
                }
            }
            if(t_pos.length > 0) {
                // some "yes" clauses
                if(t_neg.length > 0) {
                    // some "yes" and and some "neg" clauses
                    if(clause.must_match=="ONE") {
                        // some "yes" and and some "neg" clauses, one is enough to match
                        var neg = "(" + _ALL_Clause_ + " EXCEPT (" + t_neg.join(" OR ") + "))";
                        t_pos.push(neg);
                        return "(" + t_pos.join(" OR ") + ")";
                    }
                    else {
                        // some "yes" and and some "neg" clauses, all must match
                        return "((" + t_pos.join(" AND ") + ") EXCEPT (" + t_neg.join(" OR ") + "))";
                    }
                }
                else {
                    // only "yes" clauses
                    return "(" + t_pos.join(clause.must_match=="ONE" ? " OR " : " AND ") + ")";
                }
            }
            else {
                // no "yes" clauses
                if(t_neg.length > 0) {
                    // only "neg" clauses
                    return "(" + _ALL_Clause_ + " EXCEPT (" + t_neg.join(clause.must_match == "ALL" ? " OR " : " AND ") + "))";
                }
                else {
                    // no clauses at all
                    return "";
                }
            }

        case "FULLTEXT":
            return clause.value ? ("(" + clause.value + ")") : "";

        case "DATE-FIELD":
            var t="";
            if(clause.from ) {
                t = clause.field + ">=" + clause.from;
            }
            if(clause.to) {
                t += (t?" AND ":"") + clause.field + "<=" + clause.to;
            }
            return t ? ("(" + t + ")") : "";

        case "TEXT-FIELD":
        case "STRING-FIELD":
            return clause.field + clause.operator + "\"" + clause.value + "\"";

        case "GEO-DISTANCE":
            return clause.field + "=\"" + clause.lat + " " + clause.lon + " " + clause.distance + "\"";

        case "STRING-AGGREGATE":
            return clause.field + ":\"" + clause.value + "\"";

        case "COLOR-AGGREGATE":
            return clause.field + ":\"" + clause.value + "\"";

        case "NUMBER-AGGREGATE":
            return clause.field + "=" + clause.value;

        case "BOOLEAN-AGGREGATE":
            return clause.field + "=" + (clause.value ? '1' : '0');

        default :
            console.error("Unknown clause type \"" + clause.type + "\"");
            return null;
    }
}






$(document).ready(function() {
});



function answerSizer() {
    var el = $('#idFrameC').outerWidth();
    if (!$.support.cssFloat) {
       // $('#idFrameC .insidebloc').width(el - 56);
    }
    var widthA = Math.round(bodySize.x - el - 10);
    $('#rightFrame').width(widthA);
    $('#rightFrame').css('left', $('#idFrameC').width());

}

function linearize() {
    var list = $('#answers .list');
    if (list.length > 0) {
        var fllWidth = $('#answers').innerWidth();
        fllWidth -= 16;

        var stdWidth = 460;
        var diff = 28;
        var n = Math.round(fllWidth / (stdWidth));
        var w = Math.floor(fllWidth / n) - diff;
        if (w < 460 && n > 1)
            w = Math.floor(fllWidth / (n - 1)) - diff;
        $('#answers .list').width(w);
    }
    else {
        var minMargin = 5;
        var margin = 0;
        var el = $('#answers .diapo:first');
        var diapoWidth = el.outerWidth() + (minMargin * 2);
        var fllWidth = $('#answers').innerWidth();
        fllWidth -= 26;

        var n = Math.floor(fllWidth / (diapoWidth));

        margin = Math.floor((fllWidth % diapoWidth) / (2 * n));
        margin = margin + minMargin;

        $('#answers .diapo').css('margin', '5px ' + (margin) + 'px');
        
        var answerIcons = $('#answers .bottom_actions_holder .icon-stack');
        var answerIconsHolder = $('.bottom_actions_holder');
        if (el.outerWidth() < 180) {
            answerIcons.css('width', '20px');
        }
        if (el.outerWidth() > 240) {
            answerIcons.css({
                'width': '30px',
                'height': '27px'
            });
            answerIcons.closest('td').css('width', '110px');
            answerIconsHolder.css('height', '36px');
        }
    }

}


function initLook() {
    $('#nperpage_slider').slider({
        value: parseInt($('#nperpage_value').val()),
        min: 10,
        max: 100,
        step: 10,
        slide: function (event, ui) {
            $('#nperpage_value').val(ui.value);
        },
        stop: function (event, ui) {
            setPref('images_per_page', $('#nperpage_value').val());
        }
    });
    $('#sizeAns_slider').slider({
        value: parseInt($('#sizeAns_value').val()),
        min: 90,
        max: 270,
        step: 10,
        slide: function (event, ui) {
            $('#sizeAns_value').val(ui.value);
        },
        stop: function (event, ui) {
            setPref('images_size', $('#sizeAns_value').val());
        }
    });
}

function acceptCgus(name, value) {
    setPref(name, value);
}

function cancelCgus(id) {

    $.ajax({
        type: "POST",
        url: "../prod/TOU/deny/" + id + "/",
        dataType: 'json',
        success: function (data) {
            if (data.success) {
                alert(language.cgusRelog);
                self.location.replace(self.location.href);
            }
            else {
                humane.error(data.message);
            }
        }
    });

}

function activateCgus() {
    var $this = $('.cgu-dialog:first');
    $this.dialog({
        autoOpen: true,
        closeOnEscape: false,
        draggable: false,
        modal: true,
        resizable: false,
        width: 800,
        height: 500,
        open: function () {
            $this.parents(".ui-dialog:first").find(".ui-dialog-titlebar-close").remove();
            $('.cgus-accept', $(this)).bind('click', function () {
                acceptCgus($('.cgus-accept', $this).attr('id'), $('.cgus-accept', $this).attr('date'));
                $this.dialog('close').remove();
                activateCgus();
            });
            $('.cgus-cancel', $(this)).bind('click', function () {
                if (confirm(language.warningDenyCgus)) {
                    cancelCgus($('.cgus-cancel', $this).attr('id').split('_').pop());
                }
            });
        }
    });
}

$(document).ready(function () {
    humane.forceNew = true;
    activateCgus();
});


function triggerShortcuts() {

    $('#keyboard-stop').bind('click', function () {
        var display = $(this).get(0).checked ? '0' : '1';

        setPref('keyboard_infos', display);

    });

    var buttons = {};

    buttons[language.fermer] = function () {
        $("#keyboard-dialog").dialog('close');
    };

    $('#keyboard-dialog').dialog({
        closeOnEscape: false,
        resizable: false,
        draggable: false,
        modal: true,
        width: 600,
        height: 400,
        overlay: {
            backgroundColor: '#000',
            opacity: 0.7
        },
        open: function (event, ui) {
            $(this).dialog("widget").css("z-index", "1400");
        },
        close: function () {
            $(this).dialog("widget").css("z-index", "auto");
            if ($('#keyboard-stop').get(0).checked) {
                var dialog = $('#keyboard-dialog');
                if (dialog.data("ui-dialog")) {
                    dialog.dialog('destroy');
                }
                dialog.remove();
            }
        }
    }).dialog('option', 'buttons', buttons).dialog('open');

    $('#keyboard-dialog').scrollTop(0);

    return false;
}

function activeZoning() {
    $('#idFrameC, #rightFrame').bind('mousedown', function (event) {
        var old_zone = p4.active_zone;
        p4.active_zone = $(this).attr('id');
        if (p4.active_zone !== old_zone && p4.active_zone !== 'headBlock') {
            $('.effectiveZone.activeZone').removeClass('activeZone');
            $('.effectiveZone', this).addClass('activeZone');//.flash('#555555');
        }
        $('#EDIT_query').blur();
    });
    $('#rightFrame').trigger('mousedown');
}

function RGBtoHex(R, G, B) {
    return toHex(R) + toHex(G) + toHex(B);
}
function toHex(N) {
    if (N === null) return "00";
    N = parseInt(N);
    if (N === 0 || isNaN(N)) return "00";
    N = Math.max(0, N);
    N = Math.min(N, 255);
    N = Math.round(N);
    return "0123456789ABCDEF".charAt((N - N % 16) / 16)
        + "0123456789ABCDEF".charAt(N % 16);
}
function hsl2rgb(h, s, l) {
    var m1, m2, hue;
    var r, g, b;
    s /= 100;
    l /= 100;
    if (s === 0)
        r = g = b = (l * 255);
    else {
        if (l <= 0.5)
            m2 = l * (s + 1);
        else
            m2 = l + s - l * s;
        m1 = l * 2 - m2;
        hue = h / 360;
        r = HueToRgb(m1, m2, hue + 1 / 3);
        g = HueToRgb(m1, m2, hue);
        b = HueToRgb(m1, m2, hue - 1 / 3);
    }
    return {
        r: r,
        g: g,
        b: b
    };
}

function HueToRgb(m1, m2, hue) {
    var v;
    if (hue < 0)
        hue += 1;
    else if (hue > 1)
        hue -= 1;

    if (6 * hue < 1)
        v = m1 + (m2 - m1) * hue * 6;
    else if (2 * hue < 1)
        v = m2;
    else if (3 * hue < 2)
        v = m1 + (m2 - m1) * (2 / 3 - hue) * 6;
    else
        v = m1;

    return 255 * v;
}

/**
 * add "field" zone on advsearch
 *
 * @returns {jQuery|HTMLElement}
 * @constructor
 */
function AdvSearchAddNewTerm() {
    var block_template = $('#ADVSRCH_FIELDS_ZONE DIV.term_select_wrapper_template');
    var last_block = $('#ADVSRCH_FIELDS_ZONE DIV.term_select_wrapper:last');
    if(last_block.length === 0) {
        last_block = block_template;
    }
    last_block = block_template.clone(true).insertAfter(last_block);    // true: clone event handlers
    last_block.removeClass('term_select_wrapper_template').addClass('term_select_wrapper').show();
    last_block.css("background-color", "");

    return last_block;
}


$(document).ready(function () {

    var multi_term_select_html = $('.term_select_wrapper').html();

    $('input[name=search_type]').bind('click', function () {
        var $this = $(this);
        var $record_types = $('#recordtype_sel');

        if ($this.hasClass('mode_type_reg')) {
            $record_types.css("visibility", "hidden");  // better than hide because does not change layout
            $record_types.prop("selectedIndex", 0);
        } else {
            $record_types.css("visibility", "visible");
        }
    });

    $(document).on('change', 'select.term_select_field', function () {
        var $this = $(this);
        $this.blur();
        checkFilters(true);
    });

    
    $(document).on('click', '.term_deleter', function (event) {
        event.preventDefault();
        var $this = $(this);
        $this.closest('.term_select_wrapper').remove();
        checkFilters(true);
    });

    $('.add_new_term').on('click', function (event) {
        event.preventDefault();
        AdvSearchAddNewTerm(1);
    });

    $('.adv_search_button').on('click', function () {
        var searchForm = $('#searchForm');
        var parent = searchForm.parent();

        var options = {
            title: 'Advanced search',
            size: (bodySize.x - 120)+'x'+(bodySize.y - 120),
            loading: false,
            closeCallback: function (dialog) {

                var datas = dialog.find('form.phrasea_query').appendTo(parent);//.clone();

                $('.adv_trigger', searchForm).show();
                $('.adv_options', searchForm).hide();
            }
        };

        $dialog = p4.Dialog.Create(options);

        $dialog.getDomElement().closest('.ui-dialog').addClass('advanced_search_dialog_container');
        searchForm.appendTo($dialog.getDomElement());

        $dialog.getDomElement().find('.adv_options').show();
        $dialog.getDomElement().find('.adv_trigger').hide();

        $dialog.getDomElement().find('form').bind('submit.conbo', function () {
            $(this).unbind('submit.conbo');
            $dialog.Close();
            return false;
        });


        return false;
    });

    $(document).bind('contextmenu', function (event) {
        var targ;
        if (event.target)
            targ = event.target;
        else if (event.srcElement)
            targ = event.srcElement;
        if (targ.nodeType === 3)// safari bug
            targ = targ.parentNode;

        var gogo = true;
        var targ_name = targ.nodeName ? targ.nodeName.toLowerCase() : false;

        if (targ_name !== 'input' && targ_name.toLowerCase() !== 'textarea') {
            gogo = false;
        }
        if (targ_name === 'input') {
            if ($(targ).is(':checkbox'))
                gogo = false;
        }

        return gogo;
    });

    $('.basket_refresher').on('click', function () {
        return p4.WorkZone.refresh('current');
        return false;
    });

    $('#loader_bar').stop().animate({
        width: '70%'
    }, 450);
    p4.preview = {
        open: false,
        current: false
    };
    p4.currentViewMode = 'classic';
    p4.nbNoview = 0;
    p4.reg_delete = true;
    p4.sel = [];
    p4.baskSel = [];
    p4.edit = {};
    p4.thesau = {
        tabs: null
    };
    p4.active_zone = false;
    p4.next_bask_scroll = false;


    $('#backcolorpickerHolder').ColorPicker({
        flat: true,
        color: '404040',
        livePreview: false,
        eventName: 'mouseover',
        onSubmit: function (hsb, hex, rgb, el) {
            var back_hex = '';
            var unactive = '';


            if (hsb.b >= 50) {
                back_hex = '000000';

                var sim_b = 0.1 * hsb.b;
            }
            else {
                back_hex = 'FFFFFF';

                var sim_b = 100 - 0.1 * (100 - hsb.b);
            }

            var sim_b = 0.1 * hsb.b;

            var sim_rgb = hsl2rgb(hsb.h, hsb.s, sim_b);
            var sim_hex = RGBtoHex(sim_rgb.r, sim_rgb.g, sim_rgb.b);

            setPref('background-selection', hex);
            setPref('background-selection-disabled', sim_hex);
            setPref('fontcolor-selection', back_hex);

            $('style[title=color_selection]').empty();

            var datas = '.diapo.selected,#reorder_box .diapo.selected, #EDIT_ALL .diapo.selected, .list.selected, .list.selected .diapo' +
                '{' +
                '    COLOR: #' + back_hex + ';' +
                '    BACKGROUND-COLOR: #' + hex + ';' +
                '}';
            $('style[title=color_selection]').empty().text(datas);
        }
    });
    $('#backcolorpickerHolder').find('.colorpicker_submit').append($('#backcolorpickerHolder .submiter')).bind('click', function () {
        $(this).highlight('#CCCCCC');
    });

    $('#search_submit').on('mousedown', function (event) {
        return false;
    });

    $('#history-queries ul li').on('mouseover',function () {
        $(this).addClass('hover');
    }).on('mouseout', function () {
            $(this).removeClass('hover');
        });

    startThesaurus();
    checkFilters();

    activeZoning();

    $('.shortcuts-trigger').bind('click', function () {
        triggerShortcuts();
    });

    $('#idFrameC').resizable({
        handles: 'e',
        resize: function () {
            answerSizer();
            linearize();
        },
        stop: function () {

            var el = $('.SSTT.active').next().find('div:first');
            var w = el.find('div.chim-wrapper:first').outerWidth();
            var iw = el.innerWidth();
            var diff = $('#idFrameC').width() - el.outerWidth();
            var n = Math.floor(iw / w);

            $('#idFrameC').height('auto');

            var nwidth = (n) * w + diff + n;
            if (isNaN(nwidth)) {
                saveWindows();
                return;
            }
            if (nwidth < 265)
                nwidth = 265;
            if (el.find('div.chim-wrapper:first').hasClass('valid') && nwidth < 410)
                nwidth = 410;


            $('#idFrameC').stop().animate({
                    width: nwidth
                },
                360,
                'linear',
                function () {
                    answerSizer();
                    linearize();
                    saveWindows();
                });
        }
    });

    $('#idFrameC .ui-tabs-nav li').on('click', function (event) {
        if($('#idFrameC').attr('data-status') == 'closed'){
            $('#idFrameC').width(360);
            $('#rightFrame').css('left', 360);
            $('#rightFrame').width($(window).width()-360);
            $('#baskets, #proposals, #thesaurus_tab').hide();
            $('.ui-resizable-handle, #basket_menu_trigger').show();
            var IDname = $(this).attr('aria-controls');
            $('#'+IDname).show();
        }

        $('#idFrameC').attr('data-status', 'open');
        $('.WZbasketTab').css('background-position', '9px 21px');
        $('#idFrameC').removeClass('closed');
    });

    var previousTab = "";

    $('#idFrameC #retractableButton').bind('click', function (event) {

        if($('#idFrameC').attr('data-status') != 'closed'){
            $(this).find('i').removeClass('icon-double-angle-left').addClass('icon-double-angle-right')
            $('#idFrameC').width(80);
            $('#rightFrame').css('left', 80);
            $('#rightFrame').width($(window).width()-80);
            $('#idFrameC').attr('data-status', 'closed');
            $('#baskets, #proposals, #thesaurus_tab, .ui-resizable-handle, #basket_menu_trigger').hide();
            $('#idFrameC .ui-tabs-nav li').removeClass('ui-state-active');
            $('.WZbasketTab').css('background-position', '15px 21px');
            $('#idFrameC').addClass('closed');
            previousTab = $('#idFrameC .icon-menu').find('li.ui-tabs-active');
        }else{
            $(this).find('i').removeClass('icon-double-angle-right').addClass('icon-double-angle-left')
            $('#idFrameC').width(360);
            $('#rightFrame').css('left', 360);
            $('#rightFrame').width($(window).width()-360);
            $('#idFrameC').attr('data-status', 'open');
            $('.ui-resizable-handle, #basket_menu_trigger').show();
            $('.WZbasketTab').css('background-position', '9px 21px');
            $('#idFrameC').removeClass('closed');
            $('#idFrameC .icon-menu li').last().find('a').trigger('click');
            $('#idFrameC .icon-menu li').first().find('a').trigger('click');
            $(previousTab).find('a').trigger('click');
        }

        event.stopImmediatePropagation();
        //p4.WorkZone.close();
        return false;
    });

    $('#look_box .tabs').tabs();

    resize();

    $(window).bind('resize', function () {
        resize();
    });
    $('body').append('<iframe id="MODALDL" class="modalbox" src="about:blank;" name="download" style="display:none;border:none;" frameborder="0"></iframe>');

    $('body').append('<iframe id="idHFrameZ" src="about:blank" style="display:none;" name="HFrameZ"></iframe>');

    $('#basket_menu_trigger').contextMenu('#basket_menu', {
        openEvt: 'click',
        dropDown: true,
        theme: 'vista',
        showTransition: 'slideDown',
        hideTransition: 'hide',
        shadow: false
    });

    $('#basket_menu_trigger').trigger("click");
    $('#basket_menu_trigger').trigger("click");

    $('.datepicker').datepicker({
        changeYear: true,
        changeMonth: true,
        dateFormat: 'yy/mm/dd'
    });

    $.ajaxSetup({

        error: function (jqXHR, textStatus, errorThrown) {
            //Request is aborted
            if (errorThrown === 'abort') {
                return false;
            } else {
                showModal('error', {
                    title: language.errorAjaxRequest + ' ' + jqXHR.responseText
                });
            }
        },
        timeout: function () {
            showModal('timeout', {
                title: 'Server not responding'
            });
        }
    });

    $('.tools .answer_selector').bind('click',function () {
        selector($(this));
    }).bind('mouseover',function (event) {
            if (is_ctrl_key(event)) {
                $(this).addClass('add_selector');
            }
            else {
                $(this).removeClass('add_selector');
            }
        }).bind('mouseout', function () {
            $(this).removeClass('add_selector');
        });

    getLanguage();

    activeIcons();

    initAnswerForm();

    initLook();

    setTimeout("pollNotifications();", 10000);

    $(this).bind('keydown', function (event) {
        var cancelKey = false;
        var shortCut = false;

        if ($('#MODALDL').is(':visible')) {
            switch (event.keyCode) {
                case 27:
                    hideDwnl();
                    break;
            }
        }
        else {
            if ($('#EDITWINDOW').is(':visible')) {

                switch (event.keyCode) {
                    case 9:	// tab ou shift-tab
                        edit_chgFld(event, is_shift_key(event) ? -1 : 1);
                        cancelKey = shortCut = true;
                        break;
                    case 27:
                        edit_cancelMultiDesc(event);
                        shortCut = true;
                        break;

                    case 33:	// pg up
                        if (!p4.edit.textareaIsDirty || edit_validField(event, "ask_ok"))
                            skipImage(event, 1);
                        cancelKey = true;
                        break;
                    case 34:	// pg dn
                        if (!p4.edit.textareaIsDirty || edit_validField(event, "ask_ok"))
                            skipImage(event, -1);
                        cancelKey = true;
                        break;
                }

            }
            else {
                if (p4.preview.open) {
                    if (($('#dialog_dwnl:visible').length === 0 && $('#DIALOG1').length === 0 && $('#DIALOG2').length === 0)) {
                        switch (event.keyCode) {
                            case 39:
                                getNext();
                                cancelKey = shortCut = true;
                                break;
                            case 37:
                                getPrevious();
                                cancelKey = shortCut = true;
                                break;
                            case 27://escape
                                closePreview();
                                break;
                            case 32:
                                if (p4.slideShow)
                                    stopSlide();
                                else
                                    startSlide();
                                cancelKey = shortCut = true;
                                break;
                        }
                    }
                }
                else {
                    if ($('#EDIT_query').hasClass('focused'))
                        return true;

                    if ($('.overlay').is(':visible'))
                        return true;

                    if ($('.ui-widget-overlay').is(':visible'))
                        return true;

                    switch (p4.active_zone) {
                        case 'rightFrame':
                            switch (event.keyCode) {
                                case 65:	// a
                                    if (is_ctrl_key(event)) {
                                        $('.tools .answer_selector.all_selector').trigger('click');
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                                case 80://P
                                    if (is_ctrl_key(event)) {
                                        printThis("lst=" + p4.Results.Selection.serialize());
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                                case 69://e
                                    if (is_ctrl_key(event)) {
                                        editThis('IMGT', p4.Results.Selection.serialize());
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                                case 40:	// down arrow
                                    $('#answers').scrollTop($('#answers').scrollTop() + 30);
                                    cancelKey = shortCut = true;
                                    break;
                                case 38:	// down arrow
                                    $('#answers').scrollTop($('#answers').scrollTop() - 30);
                                    cancelKey = shortCut = true;
                                    break;
                                case 37://previous page
                                    $('#PREV_PAGE').trigger('click');
                                    shortCut = true;
                                    break;
                                case 39://previous page
                                    $('#NEXT_PAGE').trigger('click');
                                    shortCut = true;
                                    break;
                                case 9://tab
                                    if (!is_ctrl_key(event) && !$('.ui-widget-overlay').is(':visible') && !$('.overlay_box').is(':visible')) {
                                        document.getElementById('EDIT_query').focus();
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                            }
                            break;


                        case 'idFrameC':
                            switch (event.keyCode) {
                                case 65:	// a
                                    if (is_ctrl_key(event)) {
                                        p4.WorkZone.Selection.selectAll();
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                                case 80://P
                                    if (is_ctrl_key(event)) {
                                        printThis("lst=" + p4.WorkZone.Selection.serialize());
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                                case 69://e
                                    if (is_ctrl_key(event)) {
                                        editThis('IMGT', p4.WorkZone.Selection.serialize());
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                                //						case 46://del
                                //								deleteThis(p4.Results.Selection.serialize());
                                //								cancelKey = true;
                                //							break;
                                case 40:	// down arrow
                                    $('#baskets div.bloc').scrollTop($('#baskets div.bloc').scrollTop() + 30);
                                    cancelKey = shortCut = true;
                                    break;
                                case 38:	// down arrow
                                    $('#baskets div.bloc').scrollTop($('#baskets div.bloc').scrollTop() - 30);
                                    cancelKey = shortCut = true;
                                    break;
                                case 37://previous page
                                    $('#PREV_PAGE').trigger('click');
                                    break;
                                case 39://previous page
                                    $('#NEXT_PAGE').trigger('click');
                                    break;
                                case 9://tab
                                    if (!is_ctrl_key(event) && !$('.ui-widget-overlay').is(':visible') && !$('.overlay_box').is(':visible')) {
                                        document.getElementById('EDIT_query').focus();
                                        cancelKey = shortCut = true;
                                    }
                                    break;
                            }
                            break;


                        case 'mainMenu':
                            break;


                        case 'headBlock':
                            break;

                        default:
                            break;

                    }
                }
            }
        }

        if (!$('#EDIT_query').hasClass('focused') && event.keyCode !== 17) {

            if ($('#keyboard-dialog.auto').length > 0 && shortCut) {
                triggerShortcuts();
            }
        }
        if (cancelKey) {
            event.cancelBubble = true;
            if (event.stopPropagation)
                event.stopPropagation();
            return(false);
        }
        return(true);
    });


    $('#EDIT_query').bind('focus',function () {
        $(this).addClass('focused');
    }).bind('blur', function () {
            $(this).removeClass('focused');
        });

    $('.basketTips').tooltip({
        delay: 200
    });

    $('.basket_title').tooltip({
        extraClass: 'tooltip_flat'
    });

    $('#idFrameC .tabs').tabs({
        activate: function (event, ui) {
            if (ui.newTab.context.hash == "#thesaurus_tab") {
                thesau_show();
            }
            p4.WorkZone.open();
        }
    });

    $('#PREVIEWBOX .gui_vsplitter', p4.edit.editBox).draggable({
        axis: 'x',
        containment: 'parent',
        drag: function (event, ui) {
            var x = $(ui.position.left)[0];
            if (x < 330 || x > (bodySize.x - 400)) {
                return false;
            }
            var v = $(ui.position.left)[0];
            $("#PREVIEWLEFT").width(v);
            $("#PREVIEWRIGHT").css("left", $(ui.position.left)[0]);
            resizePreview();
        }
    });

    $('input.input_select_copy').on('focus', function () {
        $(this).select();
    });
    $('input.input_select_copy').on('blur', function () {
        $(this).deselect();
    });
    $('input.input_select_copy').on('click', function () {
        $(this).select();
    });

    $('#loader_bar').stop().animate({
        width: '100%'
    }, 450, function () {
        $('#loader').parent().fadeOut('slow', function () {
            $(this).remove();
        });
    });

});


function editThis(type, value) {

    $('#idFrameE').empty().addClass('loading');
    showOverlay(2);

    $('#EDITWINDOW').show();

    var options = {
        lst: '',
        ssel: '',
        act: ''
    };

    switch (type) {
        case "IMGT":
            options.lst = value;
            break;

        case "SSTT":
            options.ssel = value;
            break;

        case "STORY":
            options.story = value;
            break;
    }

    $.ajax({
        url: "../prod/records/edit/",
        type: "POST",
        dataType: "html",
        data: options,
        success: function (data) {
            initializeEdit();
            $('#idFrameE').removeClass('loading').empty().html(data);
            $('#tooltip').hide();
            return;
        },
        error: function (XHR, textStatus, errorThrown) {
            if (XHR.status === 0) {
                return false;
            }
        }
    });

    return;
}

(function ($) {
    $.fn.extend({
        highlight: function (color) {
            if ($(this).hasClass('animating')) {
                return;
            }
            color = typeof color !== 'undefined' ? color : 'red';
            var oldColor = $(this).css('backgroundColor');
            return $(this).addClass('animating').stop().animate({
                backgroundColor: color
            }, 50, 'linear', function () {
                $(this).stop().animate({
                    backgroundColor: oldColor
                }, 450, 'linear', function () {
                    $(this).removeClass('animating');
                });
            });
        }
    });
})(jQuery);

(function ($) {
    $.fn.extend({
        flash: function (color) {
            if ($(this).hasClass('animating')) {
                return true;
            }
            color = typeof color !== 'undefined' ? color : 'red';

            var pos = $(this).offset();

            if (!pos) {
                pos = {
                    top: 0,
                    left: 0
                };
            }

            var h = $(this).height();
            var w = $(this).width();
            $('body').append('<div id="flashing" style="border:3px solid ' + color + ';position:absolute;top:' + (pos.top + (h / 2)) + 'px;left:' + (pos.left + (w / 2)) + 'px;width:0px;height:0px"></div>');
            $(this).addClass('animating');
            var el = $(this);

            $('#flashing').stop().animate({
                top: (pos.top + (h / 4)),
                left: (pos.left + (w / 4)),
                opacity: 0,
                width: ($(this).width() / 2),
                height: ($(this).height() / 2)
            }, 700, function () {
                $('#flashing').remove();
                $(el).removeClass('animating');
            });
        }
    });
})(jQuery);


function toggleRemoveReg(el) {
    var state = !el.checked;
    setPref('reg_delete', (state ? '1' : '0'));
    p4.reg_delete = state;
}


function deleteThis(lst) {
    if (lst.split(';').length === 0) {
        alert(language.nodocselected);
        return false;
    }

    var $dialog = p4.Dialog.Create({
        size: '287x153',
        title: language.warning
    });

    $.ajax({
        type: "POST",
        url: "../prod/records/delete/what/",
        dataType: 'html',
        data: {lst: lst},
        success: function (data) {
            var response = JSON.parse(data);
            $dialog.setOption('height', 'auto');

            $dialog.setContent(response.renderView);

            //reset top position of dialog
            $dialog.getDomElement().offsetParent().css('top', ($(window).height() - $dialog.getDomElement()[0].clientHeight)/2);
        }
    });

    return false;
}

function chgCollThis(datas) {
    $dialog = p4.Dialog.Create({
        size: 'Small',
        title: language.move,
        closeButton: true
    });
    $.ajax({
        type: "POST",
        url: "../prod/records/movecollection/",
        data: datas,
        success: function (data) {
            $dialog.setContent(data);
        }
    });
}

function pushThis(sstt_id, lst, story) {
    $dialog = p4.Dialog.Create({
        size: 'Full',
        title: language.push
    });

    $.post("../prod/push/sendform/", {
        lst: lst,
        ssel: sstt_id,
        story: story
    }, function (data) {
        $dialog.setContent(data);
        return;
    });
}

function feedbackThis(sstt_id, lst, story) {
    /* disable push closeonescape as an over dialog may exist (add user) */
    $dialog = p4.Dialog.Create({
        size: 'Full',
        title: language.feedback
    });

    $.post("../prod/push/validateform/", {
        lst: lst,
        ssel: sstt_id,
        story: story
    }, function (data) {
        $dialog.setContent(data);
        return;
    });
}

function toolREFACTOR(datas, activeTab) {

    var dialog = p4.Dialog.Create({
        size: 'Medium',
        title: language.toolbox,
        loading: true
    });

    $.get("../prod/tools/"
        , datas
        , function (data) {
            dialog.setContent(data);
            dialog.setOption('contextArgs', datas);
            var tabs = $('.tabs', dialog.getDomElement()).tabs();

            // activate tab if exists:
            if( activeTab !== undefined ) {
                tabs.tabs('option', 'active', activeTab);
            }
            return;
        }
    );
}

function activeIcons() {
    var $container = $('body');
    $container.on('click', '.TOOL_print_btn', function (e) {
        e.preventDefault();
        var value = "";

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                value = "lst=" + p4.Results.Selection.serialize();
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0)
                    value = "lst=" + p4.WorkZone.Selection.serialize();
                else
                    value = "SSTTID=" + $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    value = "SSTTID=" + $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            value = "lst=" + p4.WorkZone.Selection.serialize();
                        }
                        else {
                            value = "story=" + $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        if (value !== '') {
            printThis(value);
        }
        else {
            alert(language.nodocselected);
        }
    });
    $container.on('click', '.TOOL_bridge_btn', function (e) {
        e.preventDefault();
        var $button = $(this);
        var datas = {};
        var bridgeHref = $button.attr("href");

        if ($button.hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                datas.lst = p4.Results.Selection.serialize();
        }
        else {
            if ($button.hasClass('basket_window')) {
                bridgeHref = $button.attr("data-href");
                if (p4.WorkZone.Selection.length() > 0)
                    datas.lst = p4.WorkZone.Selection.serialize();
                else
                    datas.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
            }
            else {
                if ($button.hasClass('basket_element')) {
                    datas.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($button.hasClass('story_window')) {
                        bridgeHref = $button.attr("data-href");
                        if (p4.WorkZone.Selection.length() > 0) {
                            datas.lst = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            datas.story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        if (datas.ssel || datas.lst || datas.story) {
            init_publicator(bridgeHref, datas);
        }
        else {
            alert(language.nodocselected);
        }
    });

    $container.on('click', '.TOOL_trash_btn', function () {

        var type = "";
        var el = false;

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                type = 'IMGT';
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0)
                    type = 'CHIM';
                else {
                    type = 'SSTT';
                    el = $('.SSTT.active');
                }
            }
            else {
                if ($(this).hasClass('story_window')) {
                    if (p4.WorkZone.Selection.length() > 0) {
                        type = 'CHIM';
                    }
                    else {
                        type = 'STORY';
                        el = $(this).find('input[name=story_key]');
                    }
                }
            }
        }
        if (type !== '') {
            checkDeleteThis(type, el);
        }
        else {
            alert(language.nodocselected);
        }
    });
    $container.on('click', '.TOOL_ppen_btn', function () {

        var value = "";
        var type = "";

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0) {
                type = 'IMGT';
                value = p4.Results.Selection.serialize();
            }
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0) {
                    type = 'IMGT';
                    value = p4.WorkZone.Selection.serialize();
                }
                else {
                    type = 'SSTT';
                    value = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    type = 'SSTT';
                    value = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            type = 'IMGT';
                            value = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            type = 'STORY';
                            value = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        if (value !== '') {
            editThis(type, value);
        }
        else {
            alert(language.nodocselected);
        }
    });
    $container.on('click', '.TOOL_publish_btn', function () {
        var value = "";
        var type = "";

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0) {
                type = 'IMGT';
                value = p4.Results.Selection.serialize();
            }
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0) {
                    type = 'IMGT';
                    value = p4.WorkZone.Selection.serialize();
                }
                else {
                    type = 'SSTT';
                    value = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    type = 'SSTT';
                    value = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            type = 'IMGT';
                            value = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            type = 'STORY';
                            value = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        if (value !== '') {
            publicationModule.publishRecords(type, value);
        }
        else {
            alert(language.nodocselected);
        }
    });


    $container.on('click', '.TOOL_chgcoll_btn', function () {
        var value = {};

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                value.lst = p4.Results.Selection.serialize();
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0)
                    value.lst = p4.WorkZone.Selection.serialize();
                else
                    value.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    value.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            value.lst = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            value.story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        /**
         * if works, then the object is not empty
         */
        for (i in value) {
            return chgCollThis(value);
        }

        alert(language.nodocselected);
    });

    $container.on('click', '.TOOL_chgstatus_btn', function () {
        var params = {};
        var $this = $(this);

        if ($this.hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0) {
                params.lst = p4.Results.Selection.serialize();
            }
        } else {
            if ($this.hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0) {
                    params.lst = p4.WorkZone.Selection.serialize();
                } else {
                    params.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
            } else {
                if ($this.hasClass('basket_element')) {
                    params.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                } else {
                    if ($this.hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            params.lst = p4.WorkZone.Selection.serialize();
                        } else {
                            params.story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        if (false === $.isEmptyObject(params)) {
            var dialog = p4.Dialog.Create();
            dialog.load('../prod/records/property/', 'GET', params);
        } else {
            alert(language.nodocselected);
        }
    });

    $container.on('click', '.TOOL_pushdoc_btn', function () {
        var value = "", type = "", sstt_id = "", story = "";
        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                value = p4.Results.Selection.serialize();
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0)
                    value = p4.WorkZone.Selection.serialize();
                else
                    sstt_id = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    sstt_id = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            value = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }
        if (value !== '' || sstt_id !== '' || story !== '') {
            pushThis(sstt_id, value, story);
        }
        else {
            alert(language.nodocselected);
        }
    });

    $container.on('click', '.TOOL_feedback_btn', function () {
        var value = "", type = "", sstt_id = "", story = '';
        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                value = p4.Results.Selection.serialize();
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0)
                    value = p4.WorkZone.Selection.serialize();
                else
                    sstt_id = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    sstt_id = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            value = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }
        if (value !== '' || sstt_id !== '' || story !== '') {
            feedbackThis(sstt_id, value, story);
        }
        else {
            alert(language.nodocselected);
        }
    });

    $container.on('click', '.TOOL_imgtools_btn', function () {
        var datas = {};

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0)
                datas.lst = p4.Results.Selection.serialize();
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0)
                    datas.lst = p4.WorkZone.Selection.serialize();
                else
                    datas.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    datas.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            datas.lst = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            datas.story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        if (!$.isEmptyObject(datas)) {
            toolREFACTOR(datas);
        }
        else {
            alert(language.nodocselected);
        }
    });

    $container.on('click', '.TOOL_disktt_btn', function () {
        var datas = {};

        if ($(this).hasClass('results_window')) {
            if (p4.Results.Selection.length() > 0) {
                datas.lst = p4.Results.Selection.serialize();
            }
        }
        else {
            if ($(this).hasClass('basket_window')) {
                if (p4.WorkZone.Selection.length() > 0) {
                    datas.lst = p4.WorkZone.Selection.serialize();
                }
                else {
                    datas.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
            }
            else {
                if ($(this).hasClass('basket_element')) {
                    datas.ssel = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                }
                else {
                    if ($(this).hasClass('story_window')) {
                        if (p4.WorkZone.Selection.length() > 0) {
                            datas.lst = p4.WorkZone.Selection.serialize();
                        }
                        else {
                            datas.story = $('.SSTT.active').attr('id').split('_').slice(1, 2).pop();
                        }
                    }
                }
            }
        }

        for (var i in datas) {
            return downloadThis(datas);
        }

        alert(language.nodocselected);
    });


}

function checkDeleteThis(type, el) {
    el = $(el);
    switch (type) {


        case "IMGT":
        case "CHIM":

            var lst = '';

            if (type === 'IMGT')
                lst = p4.Results.Selection.serialize();
            if (type === 'CHIM')
                lst = p4.WorkZone.Selection.serialize();

            deleteThis(lst);

            return;
            break;


        case "SSTT":

            var buttons = {};
            buttons[language.valider] = function (e) {
                deleteBasket(el);
            };

            $('#DIALOG').empty().append(language.confirmDel).attr('title', language.attention).dialog({
                autoOpen: false,
                resizable: false,
                modal: true,
                draggable: false
            }).dialog('open').dialog('option', 'buttons', buttons);
            $('#tooltip').hide();
            return;
            break;
        case "STORY":
            lst = el.val();
            deleteThis(lst);
            break;

    }
}

function shareThis(bas, rec) {
    var dialog = p4.Dialog.Create({
        title: language['share']
    });

    dialog.load("../prod/share/record/" + bas + "/" + rec + "/", "GET");
}

function printThis(value) {
    if ($("#DIALOG").data("ui-dialog")) {
        $("#DIALOG").dialog('destroy');
    }
    $('#DIALOG').attr('title', language.print)
        .empty().addClass('loading')
        .dialog({
            resizable: false,
            closeOnEscape: true,
            modal: true,
            width: '800',
            height: '500',
            open: function (event, ui) {
                $(this).dialog("widget").css("z-index", "1999");
            },
            close: function (event, ui) {
                $(this).dialog("widget").css("z-index", "auto");
            }
        })
        .dialog('open');

    $.ajax({
        type: "POST",
        url: '../prod/printer/?' + value,
        dataType: 'html',
        beforeSend: function () {

        },
        success: function (data) {
            $('#DIALOG').removeClass('loading').empty()
                .append(data);
            return;
        }
    });
}


function downloadThis(datas) {
    var dialog = p4.Dialog.Create({title: language['export']});

    $.post("../prod/export/multi-export/", datas, function (data) {

        dialog.setContent(data);

        $('.tabs', dialog.getDomElement()).tabs();

        $('.close_button', dialog.getDomElement()).bind('click', function () {
            dialog.Close();
        });

        return false;
    });
}


function viewNbSelect() {
    $("#nbrecsel").empty().append(p4.Results.Selection.length());
}

function selector(el) {
    if (el.hasClass('all_selector')) {
        p4.Results.Selection.selectAll();
    }
    else {
        if (el.hasClass('none_selector')) {
            p4.Results.Selection.empty();
        }
        else {
            if (el.hasClass('starred_selector')) {

            }
            else {
                if (el.hasClass('video_selector')) {
                    p4.Results.Selection.empty();
                    p4.Results.Selection.select('.type-video');
                }
                else {
                    if (el.hasClass('image_selector')) {
                        p4.Results.Selection.empty();
                        p4.Results.Selection.select('.type-image');
                    }
                    else {
                        if (el.hasClass('document_selector')) {
                            p4.Results.Selection.empty();
                            p4.Results.Selection.select('.type-document');
                        }
                        else {
                            if (el.hasClass('audio_selector')) {
                                p4.Results.Selection.empty();
                                p4.Results.Selection.select('.type-audio');
                            }
                        }
                    }
                }
            }
        }
    }
}

function evt_dwnl(value) {
    downloadThis("lst=" + value);
}

function evt_print(value) {
    printThis("lst=" + value);
}

function evt_add_in_chutier(sbas_id, record_id, event, singleSelection) {
    var singleSelection = singleSelection || false;
    p4.WorkZone.addElementToBasket(sbas_id, record_id, event, singleSelection);
}

function remove_from_basket(el, confirm) {
    var confirm = confirm || false;
    p4.WorkZone.removeElementFromBasket(el, confirm);
}


function doSpecialSearch(qry, allbase) {
    if (allbase) {
        checkBases(true);
    }
    selectedFacets = {};
    $('#EDIT_query').val(decodeURIComponent(qry).replace(/\+/g, " "));
    newSearch(qry);
}

function clktri(id) {
    var o = $('#TOPIC_UL' + id);
    if ($('#TOPIC_UL' + id).hasClass('closed'))
        $('#TOPIC_TRI' + id + ' ,#TOPIC_UL' + id).removeClass('closed').addClass('opened');
    else
        $('#TOPIC_TRI' + id + ' ,#TOPIC_UL' + id).removeClass('opened').addClass('closed');
}


function archiveBasket(basket_id) {
    $.ajax({
        type: "POST",
        url: "../prod/baskets/" + basket_id + "/archive/?archive=1",
        dataType: 'json',
        beforeSend: function () {

        },
        success: function (data) {
            if (data.success) {
                var basket = $('#SSTT_' + basket_id);
                var next = basket.next();

                if (next.data("ui-droppable")) {
                    next.droppable('destroy');
                }

                next.slideUp().remove();

                if (basket.data("ui-droppable")) {
                    basket.droppable('destroy');
                }

                basket.slideUp().remove();

                if ($('#baskets .SSTT').length === 0) {
                    return p4.WorkZone.refresh(false);
                }
            }
            else {
                alert(data.message);
            }
            return;
        }
    });
}


function deleteBasket(item) {
    if ($("#DIALOG").data("ui-dialog")) {
        $("#DIALOG").dialog('destroy');
    }

    var k = $(item).attr('id').split('_').slice(1, 2).pop();	// id de chutier
    $.ajax({
        type: "POST",
        url: "../prod/baskets/" + k + '/delete/',
        dataType: 'json',
        beforeSend: function () {

        },
        success: function (data) {
            if (data.success) {
                var basket = $('#SSTT_' + k);
                var next = basket.next();

                if (next.data("ui-droppable")) {
                    next.droppable('destroy');
                }

                next.slideUp().remove();

                if (basket.data("ui-droppable")) {
                    basket.droppable('destroy');
                }

                basket.slideUp().remove();

                if ($('#baskets .SSTT').length === 0) {
                    return p4.WorkZone.refresh(false);
                }
            }
            else {
                alert(data.message);
            }
            return;
        }
    });
}

function deploy(deployer, todeploy_selector)
{
    if($(deployer).hasClass("deployer_opened")) {
        $(deployer).removeClass("deployer_opened").addClass("deployer_closed");
        $(todeploy_selector).hide();
    }
    else {
        $(deployer).removeClass("deployer_closed").addClass("deployer_opened");
        $(todeploy_selector).show();
    }
}

function clksbas(el, sbas_id) {
    var bool = $(el).prop('checked');
    $.each($('.sbascont_' + sbas_id + ' :checkbox'), function () {
        this.checked = bool;
    });

    checkFilters(true);
}


function advSearch(event) {
    event.cancelBubble = true;
    //  alternateSearch(false);
    $('#idFrameC .tabs a.adv_search').trigger('click');
}

function start_page_selector() {
    var el = $('#look_box_settings select[name=start_page]');

    switch (el.val()) {
        case "LAST_QUERY":
        case "PUBLI":
        case "HELP":
            $('#look_box_settings input[name=start_page_value]').hide();
            break;
        case "QUERY":
            $('#look_box_settings input[name=start_page_value]').show();
            break;
    }
}

function set_start_page() {
    var el = $('#look_box_settings select[name=start_page]');
    var val = el.val();


    var start_page_query = $('#look_box_settings input[name=start_page_value]').val();

    if (val === 'QUERY') {
        setPref('start_page_query', start_page_query);
    }

    setPref('start_page', val);

}

function basketPrefs() {
    $('#basket_preferences').dialog({
        closeOnEscape: true,
        resizable: false,
        width: 450,
        height: 500,
        modal: true,
        draggable: false,
        overlay: {
            backgroundColor: '#000',
            opacity: 0.7
        }
    }).dialog('open');
}

function lookBox(el, event) {
    $("#look_box").dialog({
        closeOnEscape: true,
        resizable: false,
        width: 450,
        height: 500,
        modal: true,
        draggable: false,
        overlay: {
            backgroundColor: '#000',
            opacity: 0.7
        }
    }).dialog('open');
}

function showAnswer(p) {
    var o;
    if (p === 'Results') {
        // on montre les results
        if (o = document.getElementById("AnswerExplain"))
            o.style.visibility = "hidden";
        if (o = document.getElementById("AnswerResults")) {
            o.style.visibility = "";
            o.style.display = "block";
        }
        // on montre explain
        if (document.getElementById("divpage"))
            document.getElementById("divpage").style.visibility = visibilityDivPage;

        if (document.getElementById("explainResults"))
            document.getElementById("explainResults").style.display = "none";
    }
    else {
        // on montre explain
        if (document.getElementById("divpage")) {
            visibilityDivPage = "visible";
            document.getElementById("divpage").style.visibility = "hidden";
        }
        if (document.getElementById("explainResults"))
            document.getElementById("explainResults").style.display = "block";

        if (o = document.getElementById("AnswerResults")) {
            o.style.visibility = "hidden";
            o.style.display = "none";

        }
        if (o = document.getElementById("AnswerExplain"))
            o.style.visibility = "";
        if (o = document.getElementById("AnswerExplain")) {
            o.style.display = "none";
            setTimeout('document.getElementById("AnswerExplain").style.display = "block";', 200);
        }
    }
}


/**  FROM INDEX.php **/
function saveeditPbar(idesc, ndesc) {
    document.getElementById("saveeditPbarI").innerHTML = idesc;
    document.getElementById("saveeditPbarN").innerHTML = ndesc;
}

function getSelText() {
    var txt = '';
    if (window.getSelection) {
        txt = window.getSelection();
    }
    else if (document.getSelection) {
        txt = document.getSelection();
    }
    else if (document.selection) {
        txt = document.selection.createRange().text;
    }
    else
        return;
    return txt;
}

function getWinPosAsXML() {
    var ret = '<win id="search" ratio="' + ($('#idFrameC').outerWidth() / bodySize.x) + '"/>';

    if ($('#idFrameE').is(':visible') && $('#EDITWINDOW').is(':visible'))
        ret += '<win id="edit" ratio="' + ($('#idFrameE').outerWidth() / $('#EDITWINDOW').innerWidth()) + '"/>';


    return ret;
}

function saveWindows() {
    var key = '';
    var value = '';


    if ($('#idFrameE').is(':visible') && $('#EDITWINDOW').is(':visible')) {
        key = 'edit_window';
        value = $('#idFrameE').outerWidth() / $('#EDITWINDOW').innerWidth();
    }
    else {
        key = 'search_window';
        value = $('#idFrameC').outerWidth() / bodySize.x;
    }
    setPref(key, value);
}

function gotopage(pag) {
    $('#searchForm input[name="sel"]').val(p4.Results.Selection.serialize());
    $('#formAnswerPage').val(pag);
    $('#searchForm').submit();
}

function addFilterMulti(filter, link, sbasid) {
    var clone = $('#filter_multi_' + sbasid + '_' + filter);
    var orig = clone;
    if (!$('#filter_multi_' + sbasid + '_' + filter).is(':visible')) {
        clone = orig.clone(true);
        var par = orig.parent();
        orig.remove();
        par.append(clone);
        clone.slideDown('fast', function () {
            $(this);
        });
        $(link).addClass('filterActive');
    }
    else {
        clone.slideUp();
        $(link).removeClass('filterActive');
    }
    return false;
}

function autoorder() {
    var val = $.trim($('#auto_order').val());

    if (val === '')
        return;

    var sorter = new Array();

    $('#reorder_box .diapo form').each(function (i, n) {

        var id = $('input[name=id]', n).val();

        switch (val) {
            case 'title':
            default:
                var data = $('input[name=title]', n).val();
                break;
            case 'default':
                var data = $('input[name=default]', n).val();
                break;
        }

        sorter[id] = data;
    });

    var data_type = 'string';

    switch (val) {
        case 'default':
            var data_type = 'integer';
            break;
    }

    sorter = arraySortByValue(sorter, data_type);

    var last_moved = false;

    for (i in sorter) {
        var elem = $('#ORDER_' + i);
        if (last_moved) {
            elem.insertAfter(last_moved);
        }
        else {
            $('#reorder_box').prepend(elem);
        }
        last_moved = elem;
    }

}

function setFacet(boolean) {
    setPref("facet", boolean);
    loadFacets(facets);
}

function setFacetOrder(order) {
    setPref("order_facet", order);
    loadFacets(facets);
}

function setFacetValueOrder(valueOrder) {
    setPref("facet_values_order", valueOrder);
    loadFacets(facets);
}


//clear search
$(document).ready(function () {

    $('#thesaurus_tab .input-medium').on('keyup', function(){
        if($('#thesaurus_tab .input-medium').val() != ''){
            $('#thesaurus_tab .th_clear').show();
        }else{
            $('#thesaurus_tab .th_clear').hide();
        }
    });

    $('.th_clear').on('click', function(){
        $('#thesaurus_tab .input-medium').val('');
        $('#thesaurus_tab .gform').submit();
        $('#thesaurus_tab .th_clear').hide();
    });

    $('.treeview>li.expandable>.hitarea').on('click', function(){
        if($(this).css('background-position') == '99% 22px'){
            $(this).css('background-position', '99% -28px');
            $(this).addClass('active');
        }else{
            $(this).css('background-position', '99% 22px');
            $(this).removeClass('active');
        }
    });

});
;var p4 = p4 || {};

(function (p4) {
    function refreshBaskets(baskId, sort, scrolltobottom, type) {
        type = typeof type === 'undefined' ? 'basket' : type;

        var active = $('#baskets .SSTT.ui-state-active');
        if (baskId === 'current' && active.length > 0) {
            baskId = active.attr('id').split('_').slice(1, 2).pop();
        }
        sort = ($.inArray(sort, ['date', 'name']) >= 0) ? sort : '';

        scrolltobottom = typeof scrolltobottom === 'undefined' ? false : scrolltobottom;

        $.ajax({
            type: "GET",
            url: "../prod/WorkZone/",
            data: {
                id: baskId,
                sort: sort,
                type: type
            },
            beforeSend: function () {
                $('#basketcontextwrap').remove();
            },
            success: function (data) {
                var cache = $("#idFrameC #baskets");

                if ($(".SSTT", cache).data("ui-droppable")) {
                    $(".SSTT", cache).droppable('destroy');
                }
                if ($(".bloc", cache).data("ui-droppable")) {
                    $('.bloc', cache).droppable('destroy');
                }
                if (cache.data("ui-accordion")) {
                    cache.accordion('destroy').empty().append(data);
                }

                activeBaskets();

                $('.basketTips').tooltip({
                    delay: 200
                });
                cache.disableSelection();

                if (!scrolltobottom) {
                    return;
                }

                p4.next_bask_scroll = true;
                return;
            }
        });
    }

    function setTemporaryPref(name, value) {
        $.ajax({
            type: "POST",
            url: "/user/preferences/temporary/",
            data: {
                prop: name,
                value: value
            },
            success: function (data) {
                return;
            }
        });
    }

    function WorkZoneElementRemover(el, confirm) {
        var context = el.data('context');

        if (confirm !== true && $(el).hasClass('groupings') && p4.reg_delete) {
            var buttons = {};

            buttons[language.valider] = function () {
                $("#DIALOG-baskets").dialog('close').remove();
                WorkZoneElementRemover(el, true);
            };

            buttons[language.annuler] = function () {
                $("#DIALOG-baskets").dialog('close').remove();
            };

            var texte = '<p>' + language.confirmRemoveReg + '</p><div><input type="checkbox" onchange="toggleRemoveReg(this);"/>' + language.hideMessage + '</div>';
            $('body').append('<div id="DIALOG-baskets"></div>');
            $("#DIALOG-baskets").attr('title', language.removeTitle)
                .empty()
                .append(texte)
                .dialog({
                    autoOpen: false,
                    closeOnEscape: true,
                    resizable: false,
                    draggable: false,
                    modal: true,
                    buttons: buttons,
                    overlay: {
                        backgroundColor: '#000',
                        opacity: 0.7
                    }
                }).dialog('open');
            return;
        }

        var id = $(el).attr('id').split('_').slice(2, 4).join('_');

        $.ajax({
            type: "POST",
            url: $(el).attr('href'),
            dataType: 'json',
            beforeSend: function () {
                $('.wrapCHIM_' + id).find('.CHIM').fadeOut();
            },
            success: function (data) {
                if (data.success) {
                    humane.info(data.message);
                    p4.WorkZone.Selection.remove(id);

                    if ($('.wrapCHIM_' + id).find('.CHIM').data("ui-draggable")) {
                        $('.wrapCHIM_' + id).find('.CHIM').draggable('destroy');
                    }

                    $('.wrapCHIM_' + id).remove();

                    if (context === "reg_train_basket") {
                        var carousel = $("#PREVIEWCURRENTCONT");
                        var carouselItemLength = $('li', carousel).length;
                        var selectedItem = $("li.prevTrainCurrent.selected", carousel);
                        var selectedItemIndex = $('li', carousel).index(selectedItem);
                        var basketId = $('#baskets .content:visible').attr('id').split('_').pop();
                        if(carouselItemLength === 1){
                            closePreview();
                            p4.WorkZone.reloadCurrent();
                            return;
                        }
                        openPreview(this, 'BASK',selectedItemIndex+1,basketId,true);
                        p4.WorkZone.reloadCurrent();

                    } else {
                        return p4.WorkZone.reloadCurrent();
                    }
                } else {
                    humane.error(data.message);
                    $('.wrapCHIM_' + id).find('.CHIM').fadeIn();
                }
                return;
            }
        });
        return false;
    }


    function activeBaskets() {
        var cache = $("#idFrameC #baskets");

        cache.accordion({
            active: 'active',
            heightStyle: "content",
            collapsible: true,
            header: 'div.header',
            activate: function (event, ui) {
                var b_active = $('#baskets .SSTT.active');
                if (p4.next_bask_scroll) {
                    p4.next_bask_scroll = false;

                    if (!b_active.next().is(':visible'))
                        return;

                    var t = $('#baskets .SSTT.active').position().top + b_active.next().height() - 200;

                    t = t < 0 ? 0 : t;

                    $('#baskets .bloc').stop().animate({
                        scrollTop: t
                    });
                }

                var uiactive = $(this).find('.ui-state-active');
                b_active.not('.ui-state-active').removeClass('active');

                if (uiactive.length === 0) {
                    return;
                    /* everything is closed */
                }

                uiactive.addClass('ui-state-focus active');

                p4.WorkZone.Selection.empty();

                getContent(uiactive);

            },
            beforeActivate: function (event, ui) {
                ui.newHeader.addClass('active');
                $('#basketcontextwrap .basketcontextmenu').hide();
            }
        });

        $('.bloc', cache).droppable({
            accept: function (elem) {
                if ($(elem).hasClass('grouping') && !$(elem).hasClass('SSTT'))
                    return true;
                return false;
            },
            scope: 'objects',
            hoverClass: 'groupDrop',
            tolerance: 'pointer',
            drop: function () {
                fix();
            }
        });

        if ($('.SSTT.active', cache).length > 0) {
            var el = $('.SSTT.active', cache)[0];
            $(el).trigger('click');
        }

        $(".SSTT, .content", cache)
            .droppable({
                scope: 'objects',
                hoverClass: 'baskDrop',
                tolerance: 'pointer',
                accept: function (elem) {
                    if ($(elem).hasClass('CHIM')) {
                        if ($(elem).closest('.content').prev()[0] === $(this)[0]) {
                            return false;
                        }
                    }
                    if ($(elem).hasClass('grouping') || $(elem).parent()[0] === $(this)[0])
                        return false;
                    return true;
                },
                drop: function (event, ui) {
                    dropOnBask(event, ui.draggable, $(this));
                }
            });

        if ($('#basketcontextwrap').length === 0)
            $('body').append('<div id="basketcontextwrap"></div>');

        $('.context-menu-item', cache).hover(function () {
            $(this).addClass('context-menu-item-hover');
        }, function () {
            $(this).removeClass('context-menu-item-hover');
        });
        $.each($(".SSTT", cache), function () {
            var el = $(this);
            $(this).find('.contextMenuTrigger').contextMenu('#' + $(this).attr('id') + ' .contextMenu', {
                'appendTo': '#basketcontextwrap',
                openEvt: 'click',
                theme: 'vista',
                dropDown: true,
                showTransition: 'slideDown',
                hideTransition: 'hide',
                shadow: false
            });
        });

    }

    function getContent(header, order) {
        if (window.console) {
            console.log('Reload content for ', header);
        }

        var url = $('a', header).attr('href');

        if (typeof order !== 'undefined') {
            url += '?order=' + order;
        }

        $.ajax({
            type: "GET",
            url: url,
            dataType: 'html',
            beforeSend: function () {
                $('#tooltip').hide();
                header.next().addClass('loading');
            },
            success: function (data) {
                header.removeClass('unread');

                var dest = header.next();
                if (dest.data("ui-droppable")) {
                    dest.droppable('destroy')
                }
                dest.empty().removeClass('loading');

                dest.append(data);

                $('a.WorkZoneElementRemover', dest).bind('mousedown',function (event) {
                    return false;
                }).bind('click', function (event) {
                    return WorkZoneElementRemover($(this), false);
                });
                
                $("#baskets div.content select[name=valid_ord]").on('change', function () {
                    var active = $('#baskets .SSTT.ui-state-active');
                    if (active.length === 0) {
                        return;
                    }
            
                    var order = $(this).val();
            
                    getContent(active, order);
                });

                dest.droppable({
                    accept: function (elem) {
                        if ($(elem).hasClass('CHIM')) {
                            if ($(elem).closest('.content')[0] === $(this)[0]) {
                                return false;
                            }
                        }
                        if ($(elem).hasClass('grouping') || $(elem).parent()[0] === $(this)[0])
                            return false;
                        return true;
                    },
                    hoverClass: 'baskDrop',
                    scope: 'objects',
                    drop: function (event, ui) {
                        dropOnBask(event, ui.draggable, $(this).prev());
                    },
                    tolerance: 'pointer'
                });

                $('.noteTips, .captionRolloverTips', dest).tooltip({
                    extraClass: 'tooltip_flat'
                });

                dest.find('.CHIM').draggable({
                    helper: function () {
                        $('body').append('<div id="dragDropCursor" ' +
                            'style="position:absolute;z-index:9999;background:red;' +
                            '-moz-border-radius:8px;-webkit-border-radius:8px;">' +
                            '<div style="padding:2px 5px;font-weight:bold;">' +
                            p4.WorkZone.Selection.length() + '</div></div>');
                        return $('#dragDropCursor');
                    },
                    scope: "objects",
                    distance: 20,
                    scroll: false,
                    refreshPositions: true,
                    cursorAt: {
                        top: 10,
                        left: -20
                    },
                    start: function (event, ui) {
                        var baskets = $('#baskets');
                        baskets.append('<div class="top-scroller"></div>' +
                            '<div class="bottom-scroller"></div>');
                        $('.bottom-scroller', baskets).bind('mousemove', function () {
                            $('#baskets .bloc').scrollTop($('#baskets .bloc').scrollTop() + 30);
                        });
                        $('.top-scroller', baskets).bind('mousemove', function () {
                            $('#baskets .bloc').scrollTop($('#baskets .bloc').scrollTop() - 30);
                        });
                    },
                    stop: function () {
                        $('#baskets').find('.top-scroller, .bottom-scroller')
                            .unbind()
                            .remove();
                    },
                    drag: function (event, ui) {
                        if (is_ctrl_key(event) || $(this).closest('.content').hasClass('grouping'))
                            $('#dragDropCursor div').empty().append('+ ' + p4.WorkZone.Selection.length());
                        else
                            $('#dragDropCursor div').empty().append(p4.WorkZone.Selection.length());

                    }
                });
                answerSizer();
                return;
            }
        });
    }

    function dropOnBask(event, from, destKey, singleSelection) {
        var action = "", from = $(from), dest_uri = '', lstbr = [], sselcont = [], act = "ADD";

        if (from.hasClass("CHIM")) {
            /* Element(s) come from an open object in the workzone */
            action = $(' #baskets .ui-state-active').hasClass('grouping') ? 'REG2' : 'CHU2';
        } else {
            /* Element(s) come from result */
            action = 'IMGT2';
        }

        action += destKey.hasClass('grouping') ? 'REG' : 'CHU';

        if (destKey.hasClass('content')) {
            /* I dropped on content */
            dest_uri = $('a', destKey.prev()).attr('href');
        } else {
            /* I dropped on Title */
            dest_uri = $('a', destKey).attr('href');
        }

        if (window.console) {
            window.console.log('Requested action is ', action, ' and act on ', dest_uri);
        }

        if (action === "IMGT2CHU" || action === "IMGT2REG") {
            if ($(from).hasClass('.baskAdder')) {
                lstbr = [$(from).attr('id').split('_').slice(2, 4).join('_')];
            } else if (singleSelection) {
                if (from.length === 1) {
                    lstbr = [$(from).attr('id').split('_').slice(1, 3).join('_')];
                } else {
                    lstbr = [$(from).selector.split('_').slice(1, 3).join('_')];
                }
            } else {
                lstbr = p4.Results.Selection.get();
            }
        } else {
            sselcont = $.map(p4.WorkZone.Selection.get(), function (n, i) {
                return $('.CHIM_' + n, $('#baskets .content:visible')).attr('id').split('_').slice(1, 2).pop();
            });
            lstbr = p4.WorkZone.Selection.get();
        }

        switch (action) {
            case "CHU2CHU" :
                if (!is_ctrl_key(event)) act = "MOV";
                break;
            case "IMGT2REG":
            case "CHU2REG" :
            case "REG2REG":
                var sameSbas = true, sbas_reg = destKey.attr('sbas');

                for (var i = 0; i < lstbr.length && sameSbas; i++) {
                    if (lstbr[i].split('_').shift() !== sbas_reg) {
                        sameSbas = false;
                        break;
                    }
                }

                if (sameSbas === false) {
                    return p4.Alerts('', language.reg_wrong_sbas);
                }

                break;
        }

        switch (act + action) {
            case 'MOVCHU2CHU':
                var url = dest_uri + "stealElements/";
                var data = {
                    elements: sselcont
                };
                break;
            case 'ADDCHU2REG':
            case 'ADDREG2REG':
            case 'ADDIMGT2REG':
            case 'ADDCHU2CHU':
            case 'ADDREG2CHU':
            case 'ADDIMGT2CHU':
                var url = dest_uri + "addElements/";
                var data = {
                    lst: lstbr.join(';')
                };
                break;
            default:
                if (window.console) {
                    console.log('Should not happen');
                }
                return;
                break;
        }

        if (window.console) {
            window.console.log('About to execute ajax POST on ', url, ' with datas ', data);
        }

        $.ajax({
            type: "POST",
            url: url,
            data: data,
            dataType: 'json',
            beforeSend: function () {

            },
            success: function (data) {
                if (!data.success) {
                    humane.error(data.message);
                } else {
                    humane.info(data.message);
                }
                if (act === 'MOV' || $(destKey).next().is(':visible') === true || $(destKey).hasClass('content') === true) {
                    $('.CHIM.selected:visible').fadeOut();
                    p4.WorkZone.Selection.empty();
                    return p4.WorkZone.reloadCurrent();
                }

                return;
            }
        });
    }

    function fix() {
        $.ajax({
            type: "POST",
            url: "../prod/WorkZone/attachStories/",
            data: {stories: p4.Results.Selection.get()},
            dataType: "json",
            success: function (data) {
                humane.info(data.message);
                p4.WorkZone.refresh();
            }
        });
    }

    function unfix(link) {
        $.ajax({
            type: "POST",
            url: link,
            dataType: "json",
            success: function (data) {
                humane.info(data.message);
                p4.WorkZone.refresh();
            }
        });
    }

    $(document).ready(function () {
        activeBaskets();
        $('body').on('click', 'a.story_unfix', function () {
            unfix($(this).attr('href'));

            return false;
        });

        p4.WorkZone = {
            'Selection': new Selectable($('#baskets'), {selector: '.CHIM'}),
            'refresh': refreshBaskets,
            'addElementToBasket': function (sbas_id, record_id, event, singleSelection) {
                singleSelection = !!singleSelection || false;

                if ($('#baskets .SSTT.active').length === 1) {
                    return dropOnBask(event, $('#IMGT_' + sbas_id + '_' + record_id), $('#baskets .SSTT.active'), singleSelection);
                } else {
                    humane.info(language.noActiveBasket);
                }
            },
            "removeElementFromBasket": WorkZoneElementRemover,
            'reloadCurrent': function () {
                var sstt = $('#baskets .content:visible');
                if (sstt.length === 0)
                    return;
                getContent(sstt.prev());
            },
            'close': function () {
                var frame = $('#idFrameC'), that = this;

                if (!frame.hasClass('closed')) {
                    // hide tabs content
                    $('#idFrameC .tabs > .ui-tabs-panel').hide();

                    frame.data('openwidth', frame.width());
                    frame.animate({width: 100},
                        300,
                        'linear',
                        function () {
                            answerSizer();
                            linearize();
                            $('#answers').trigger('resize');
                        });
                    frame.addClass('closed');
                    $('.escamote', frame).hide();
                    frame.unbind('click.escamote').bind('click.escamote', function () {
                        that.open();
                    });
                }
            },
            'open': function () {
                var frame = $('#idFrameC');

                if (frame.hasClass('closed')) {
                    var width = frame.data('openwidth') ? frame.data('openwidth') : 300;
                    frame.css({width: width});
                    answerSizer();
                    linearize();
                    frame.removeClass('closed');
                    $('.escamote', frame).show();
                    frame.unbind('click.escamote');
                    // show tabs content
                    var activeTabIdx = $('#idFrameC .tabs').tabs("option", "active");
                    $('#idFrameC .tabs > div:eq(' + activeTabIdx + ')').show();
                }
            }
        };
    });

    return;
}(p4));
;var p4 = p4 || {};

(function (p4) {

    function create_dialog() {
        if ($('#p4_alerts').length === 0) {
            $('body').append('<div id="p4_alerts"></div>');
        }
        return $('#p4_alerts');
    }

    function alert(title, message, callback) {
        var dialog = create_dialog();

        var button = new Object();

        button['Ok'] = function () {
            if (typeof callback === 'function')
                callback();
            else
                dialog.dialog('close');
        };
        if (dialog.data('ui-dialog')) {
            dialog.dialog('destroy');
        }

        dialog.attr('title', title)
            .empty()
            .append(message)
            .dialog({
                autoOpen: false,
                closeOnEscape: true,
                resizable: false,
                draggable: false,
                modal: true,
                buttons: button,
                overlay: {
                    backgroundColor: '#000',
                    opacity: 0.7
                }
            }).dialog('open');

        if (typeof callback === 'function') {
            dialog.bind("dialogclose", function (event, ui) {
                callback();
            });
        }
        else {

        }

        return;
    }

    p4.Alerts = alert;


    return;
}(p4));
;;
var p4 = p4 || {};

;
(function (p4, $) {

    /**
     * UPLOADER MANAGER
     */
    var UploaderManager = function (options) {

        var options = options || {};

        if (false === ("container" in options)) {
            throw "missing container parameter";
        }
        else if (!options.container.jquery) {
            throw "container parameter must be a jquery dom element";
        }

        if (false === ("settingsBox" in options)) {
            throw "missing settingBox parameter";
        }
        else if (!options.settingsBox.jquery) {
            throw "container parameter must be a jquery dom element";
        }

        if (false === ("uploadBox" in options)) {
            throw "missing uploadBox parameter";
        }
        else if (!options.uploadBox.jquery) {
            throw "container parameter must be a jquery dom element";
        }

        if (false === ("downloadBox" in options)) {
            throw "missing downloadBox parameter";
        }
        else if (!options.downloadBox.jquery) {
            throw "container parameter must be a jquery dom element";
        }

        this.recordClass = options.recordClass || 'upload-record';

        this.options = options;

        this.options.uploadBox.wrapInner('<ul class="thumbnails" />');

        this.options.uploadBox = this.options.uploadBox.find('ul:first');

        this.options.downloadBox.wrapInner('<ul class="thumbnails" />');

        this.options.downloadBox = this.options.downloadBox.find('ul:first');

        if ($.isFunction($.fn.sortable)) {
            this.options.uploadBox.sortable();
        }

        this.uploadIndex = 0;

        this.Queue = new Queue();
        this.Formater = new Formater();
        this.Preview = new Preview();
    };

    UploaderManager.prototype = {
        setOptions: function (options) {
            return $.extend(this.options, options);
        },
        getContainer: function () {
            return this.options.container;
        },
        getUploadBox: function () {
            return this.options.uploadBox;
        },
        getSettingsBox: function () {
            return this.options.settingsBox;
        },
        getDownloadBox: function () {
            return this.options.downloadBox;
        },
        clearUploadBox: function () {
            this.getUploadBox().empty();
            this.uploadIndex = 0;
            this.Queue.clear();
        },
        getDatas: function () {
            return this.Queue.all();
        },
        getData: function (index) {
            return this.Queue.get(index);
        },
        addData: function (data) {
            this.uploadIndex++;
            data.uploadIndex = this.uploadIndex;
            this.Queue.set(this.uploadIndex, data);
        },
        removeData: function (index) {
            this.Queue.remove(index);
        },
        addAttributeToData: function (indexOfData, attribute, value) {
            var data = this.getData(indexOfData);
            if ($.type(attribute) === "string") {
                data[attribute] = value;
                this.Queue.set(indexOfData, data);
            }
        },
        getUploadIndex: function () {
            return this.uploadIndex;
        },
        hasData: function () {
            return !this.Queue.isEmpty();
        },
        countData: function () {
            return this.Queue.getLength();
        }
    };


    /**
     * PREVIEW
     *
     * Dependency : loadImage function
     * @see https://github.com/blueimp/JavaScript-Load-Image
     *
     * Options
     *  maxWidth: (int) Max width of preview
     *  maxHeight: (int) Max height of preview
     *  minWidth: (int) Min width of preview
     *  minHeight: (int) Min height of preview
     *  canva: (boolean) render preview as canva if supported by the navigator
     */

    var Preview = function () {
        this.options = {
            fileType: /^image\/(gif|jpeg|png|jpg)$/,
            maxSize: 5242880 // 5MB
        };
    };

    Preview.prototype = {
        setOptions: function (options) {
            this.options = $.extend(this.options, options);
        },
        getOptions: function () {
            return this.options;
        },
        render: function (file, callback) {
            if (typeof loadImage === 'function' && this.options.fileType.test(file.type)) {
                if ($.type(this.options.maxSize) !== 'number' || file.size < this.options.maxSize) {
                    var options = {
                        maxWidth: this.options.maxWidth || 150,
                        maxHeight: this.options.maxHeight || 75,
                        minWidth: this.options.minWidth || 80,
                        minHeight: this.options.minHeight || 40,
                        canvas: this.options.canva || true
                    };
                    loadImage(file, callback, options);
                }
            }
        }
    };


    /**
     * FORMATER
     */

    var Formater = function () {

    };

    Formater.prototype = {
        size: function (bytes) {
            if (typeof bytes !== 'number') {
                throw bytes + ' is not a number';
            }
            if (bytes >= 1073741824) {
                return (bytes / 1073741824).toFixed(2) + ' GB';
            }
            if (bytes >= 1048576) {
                return (bytes / 1048576).toFixed(2) + ' MB';
            }
            return (bytes / 1024).toFixed(2) + ' KB';
        },
        bitrate: function (bits) {
            if (typeof bits !== 'number') {
                throw bits + ' is not a number';
            }
            // 1 byte = 8 bits
            var bytes = (bits >> 3);

            if (bytes >= (1 << 30)) {
                return (bytes / (1 << 30)).toFixed(2) + ' Go/s';
            }
            if (bytes >= (1 << 20)) {
                return (bytes / (1 << 20)).toFixed(2) + ' Mo/s';
            }
            if (bytes >= (1 << 10)) {
                return (bytes / (1 << 10)).toFixed(2) + ' Ko/s';
            }
            return bytes + ' o/s';
        },
        pourcent: function (current, total) {
            return (current / total * 100).toFixed(2);
        }
    };

    /**
     * QUEUE
     */
    var Queue = function () {
        this.list = {};
    };

    Queue.prototype = {
        all: function () {
            return this.list;
        },
        set: function (id, item) {
            this.list[id] = item;
            return this;
        },
        get: function (id) {
            if (!this.list[id]) {
                throw 'Unknown ID' + id;
            }
            return this.list[id];
        },
        remove: function (id) {
            delete this.list[id];
        },
        getLength: function () {
            var count = 0;
            for (var k in this.list) {
                if (this.list.hasOwnProperty(k)) {
                    ++count;
                }
            }
            return count;
        },
        isEmpty: function () {
            return this.getLength() === 0;
        },
        clear: function () {
            var $this = this;
            $.each(this.list, function (k) {
                $this.remove(k);
            });
        }
    };

    p4.UploaderManager = UploaderManager;

})(p4, jQuery);
;;
(function (document) {

    /*****************
     * Canva Object
     *****************/
    var Canva = function (domCanva) {
        this.domCanva = domCanva;
    };

    Canva.prototype = {
        resize: function (elementDomNode, forceWidth) {

            var w = elementDomNode.getWidth();
            var h = null;
            var maxH = elementDomNode.getHeight();
            var ratio = 1;

            if ('' !== elementDomNode.getAspectRatio()) {
                ratio = parseFloat(elementDomNode.getAspectRatio());

                h = Math.round(w * (1 / ratio));

                if (h > maxH) {
                    h = maxH;
                    w = Math.round(h * ratio);
                }
            } else {
                h = maxH;
            }

            if( forceWidth !== undefined ) {
                w = parseInt(forceWidth, 10);

                if (elementDomNode.getAspectRatio() !== '') {
                    h = Math.round(w * (1 / ratio));
                } else {
                    h = maxH;
                }
            }

            this.domCanva.setAttribute("width", w);
            this.domCanva.setAttribute("height", h);

            return this;
        },
        getContext2d: function () {

            if (undefined === this.domCanva.getContext) {
                return G_vmlCanvasManager
                    .initElement(this.domCanva)
                    .getContext("2d");
            }

            return this.domCanva.getContext('2d');
        },
        extractImage: function () {
            return this.domCanva.toDataURL("image/png");
        },
        reset: function () {
            var context = this.getContext2d();
            var w = this.getWidth();
            var h = this.getHeight();

            context.save();
            context.setTransform(1, 0, 0, 1, 0, 0);
            context.clearRect(0, 0, w, h);
            context.restore();

            return this;
        },
        copy: function (elementDomNode) {
            var context = this.getContext2d();

            context.drawImage(
                elementDomNode.getDomElement()
                , 0
                , 0
                , this.getWidth()
                , this.getHeight()
            );

            return this;
        },
        getDomElement: function () {
            return this.domCanva;
        },
        getHeight: function () {
            return this.domCanva.offsetHeight;
        },
        getWidth: function () {
            return this.domCanva.offsetWidth;
        }
    };


    /******************
     *  Image Object
     ******************/
    var Image = function (domElement) {
        this.domElement = domElement;
    };

    Image.prototype = {
        getDomElement: function () {
            return this.domElement;
        },
        getHeight: function () {
            return this.domElement.offsetHeight;
        },
        getWidth: function () {
            return this.domElement.offsetWidth;
        }
    };

    /******************
     *  Video Object inherits from Image object
     ******************/

    var Video = function (domElement) {
        Image.call(this, domElement);
        this.aspectRatio = domElement.getAttribute('data-ratio');
    };

    Video.prototype = new Image();
    Video.prototype.constructor = Video;
    Video.prototype.getCurrentTime = function () {
        return Math.floor(this.domElement.currentTime);
    };
    Video.prototype.getAspectRatio = function () {
        return this.aspectRatio;
    };

    /******************
     *  Cache Object
     ******************/
    var Store = function () {
        this.datas = {};
    };

    Store.prototype = {
        set: function (id, item) {
            this.datas[id] = item;
            return this;
        },
        get: function (id) {
            if (!this.datas[id]) {
                throw 'Unknown ID';
            }
            return this.datas[id];
        },
        remove: function (id) {
            // never reuse same id
            this.datas[id] = null;
        },
        getLength: function () {
            var count = 0;
            for (var k in this.datas) {
                if (this.datas.hasOwnProperty(k)) {
                    ++count;
                }
            }
            return count;
        }
    };

    /******************
     *  Screenshot Object
     ******************/
    var ScreenShot = function (id, canva, video, altCanvas) {

        var date = new Date();
        var options = options || {};
        canva.resize(video);
        canva.copy(video);

        // handle alternative canvas:
        var altCanvas = altCanvas == undefined ? [] : altCanvas;
        this.altScreenShots = [];
        if( altCanvas.length > 0 ) {
            for(var i = 0; i< altCanvas.length; i++) {
                var canvaEl = altCanvas[i].el;
                canvaEl.resize(video, altCanvas[i].width);
                canvaEl.copy(video);

                this.altScreenShots.push({
                    dataURI: canvaEl.extractImage(),
                    name: altCanvas[i].name
                })
            }
        }

        this.id = id;
        this.timestamp = date.getTime();
        this.dataURI = canva.extractImage();
        this.videoTime = video.getCurrentTime();

    };

    ScreenShot.prototype = {
        getId: function () {
            return this.id;
        },
        getDataURI: function () {
            return this.dataURI;
        },
        getTimeStamp: function () {
            return this.timestamp;
        },
        getVideoTime: function () {
            return this.videoTime;
        },
        getAltScreenShots: function() {
            return this.altScreenShots;
        }
    };

    /**
     * THUMB EDITOR
     */
    var ThumbEditor = function (videoId, canvaId, outputOptions) {

        var domElement = document.getElementById(videoId);

        if (null !== domElement) {
            var editorVideo = new Video(domElement);
        }
        var store = new Store();

        function getCanva() {
            return document.getElementById(canvaId);
        }

        var outputOptions = outputOptions || {};

        function setAltCanvas() {
            var domElements = [],
            altCanvas = outputOptions.altCanvas;
            if( altCanvas.length > 0 ) {
                for(var i = 0; i< altCanvas.length; i++) {
                    domElements.push({
                        el: new Canva(altCanvas[i]),
                        width: altCanvas[i].getAttribute('data-width'),
                        name: altCanvas[i].getAttribute('data-name')
                    } );
                }
            }
            return domElements;
        }

        return {
            isSupported: function () {
                var elem = document.createElement('canvas');

                return !!document.getElementById(videoId) && document.getElementById(canvaId)
                    && !!elem.getContext && !!elem.getContext('2d');
            },
            screenshot: function () {
                var screenshot = new ScreenShot(
                    store.getLength() + 1,
                    new Canva(getCanva()),
                    editorVideo,
                    setAltCanvas()
                );

                store.set(screenshot.getId(), screenshot);

                return screenshot;
            },
            store: store,
            copy: function (mainSource, altSources) {

                var elementDomNode = document.createElement('img');
                elementDomNode.src = mainSource;

                var element = new Image(elementDomNode);
                var editorCanva = new Canva(getCanva());
                var altEditorCanva = setAltCanvas();
                editorCanva
                    .reset()
                    .resize(editorVideo)
                    .copy(element);


                // handle alternative canvas:
                if( altEditorCanva.length > 0 ) {
                    for(var i = 0; i< altEditorCanva.length; i++) {

                        var tmpEl = document.createElement('img');
                            tmpEl.src = altSources[i].dataURI;

                        var canvaEl = altEditorCanva[i].el;

                        canvaEl
                            .reset()
                            .resize(editorVideo, altEditorCanva[i].width)
                            .copy(new Image(tmpEl)); // @TODO: should copy the right stored image
                    }
                }
            },
            getCanvaImage: function () {
                var canva = new Canva(getCanva());

                return canva.extractImage();
            },
            resetCanva: function () {
                var editorCanva = new Canva(getCanva());
                editorCanva.reset();
            },
            getNbScreenshot: function () {
                return store.getLength();
            }
        };
    };

    document.THUMB_EDITOR = ThumbEditor;

})(document);

;function publicator_reload_publicator(url) {
    var options = $('#dialog_publicator form[name="current_datas"]').serializeArray();
    var dialog = p4.Dialog.get(1);
    dialog.load(url, 'POST', options);
}

function init_publicator(url, datas) {
    var dialog = p4.Dialog.Create({
        size: 'Full',
        title: 'Bridge',
        loading: false
    });

    dialog.load(url, 'POST', datas);
}
;/**
 * sprintf and vsprintf for jQuery
 * somewhat based on http://jan.moesen.nu/code/javascript/sprintf-and-printf-in-javascript/
 * 
 * Copyright (c) 2008 Sabin Iacob (m0n5t3r) <iacobs@m0n5t3r.info>
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 *
 * @license http://www.gnu.org/licenses/gpl.html 
 * @project jquery.sprintf
 */
(function($){
	var formats = {
		'b': function(val) {return parseInt(val, 10).toString(2);},
		'c': function(val) {return String.fromCharCode(parseInt(val, 10));},
		'd': function(val) {return parseInt(val, 10);},
		'u': function(val) {return Math.abs(val);},
		'f': function(val, p) {
			p = parseInt(p, 10); 
			val = parseFloat(val);
			if(isNaN(p && val)) {
				return NaN;
			}
			return p && val.toFixed(p) || val;
		},
		'o': function(val) {return parseInt(val, 10).toString(8);},
		's': function(val) {return val;},
		'x': function(val) {return ('' + parseInt(val, 10).toString(16)).toLowerCase();},
		'X': function(val) {return ('' + parseInt(val, 10).toString(16)).toUpperCase();}
	};

	var re = /%(?:(\d+)?(?:\.(\d+))?|\(([^)]+)\))([%bcdufosxX])/g;

	var dispatch = function(data){
		if(data.length == 1 && typeof data[0] == 'object') { //python-style printf
			data = data[0];
			return function(match, w, p, lbl, fmt, off, str) {
				return formats[fmt](data[lbl]);
			};
		} else { // regular, somewhat incomplete, printf
			var idx = 0; 
			return function(match, w, p, lbl, fmt, off, str) {
				if(fmt == '%') {
					return '%';
				}
				return formats[fmt](data[idx++], p);
			};
		}
	};

	$.extend({
		sprintf: function(format) {
			var argv = Array.apply(null, arguments).slice(1);
			return format.replace(re, dispatch(argv));
		},
		vsprintf: function(format, data) {
			return format.replace(re, dispatch(data));
		}
	});
})(jQuery);
;var prevAjax, prevAjaxrunning;
prevAjaxrunning = false;
p4.slideShow = false;

$(document).ready(function () {
    $('#PREVIEWIMGDESC').tabs();
});


function getNewVideoToken(lst, obj) {
    $.ajax({
        type: "POST",
        url: "../prod/records/renew-url/",
        dataType: 'json',
        data: {
            lst: lst
        },
        success: function (data) {
            if (!data[lst])
                return;
            obj.unload();
            obj.setClip({url: data[lst]});
            obj.play();
            return;
        }

    });
}

/**
 *
 * @param env
 * @param pos - relative position in current page
 * @param contId
 * @param reload
 */
function openPreview(event, env, pos, contId, reload) {
    if (contId == undefined)
        contId = '';
    var roll = 0;
    var justOpen = false;

    var options_serial = p4.tot_options;
    var query = p4.tot_query;
    var navigation = p4.navigation;
    var navigationContext = '';

    // keep relative position for answer train:
    var relativePos = pos;
    var absolutePos = 0;

    if (!p4.preview.open) {
        showOverlay();

        $('#PREVIEWIMGCONT').disableSelection();

        justOpen = true;

        if (!( navigator.userAgent.match(/msie/i))) {
            $('#PREVIEWBOX').css({
                'display': 'block',
                'opacity': 0
            }).fadeTo(500, 1);
        } else {
            $('#PREVIEWBOX').css({
                'display': 'block',
                'opacity': 1
            });
        }
        p4.preview.open = true;
        p4.preview.nCurrent = 5;
        $('#PREVIEWCURRENT, #PREVIEWOTHERSINNER, #SPANTITLE').empty();
        resizePreview();
        if (env == 'BASK')
            roll = 1;

        // if comes from story and in workzone
        if (env == 'REG') {
            navigationContext = 'storyFromResults';
            var $source = $(event);
            if( $source.hasClass('CHIM')) {
                navigationContext = 'storyFromWorkzone';
            }
        }

    }

    if (reload === true)
        roll = 1;


    $('#tooltip').css({
        'display': 'none'
    });

    $('#PREVIEWIMGCONT').empty();

    if (navigationContext === 'storyFromWorkzone') {
        // if event comes from workzone, set to relative position (CHIM == chutier image)
        absolutePos = relativePos;
    } else if (navigationContext === 'storyFromResults') {
        absolutePos = 0;
    } else {
        // update real absolute position with pagination for records:
        absolutePos = parseInt(navigation.perPage, 10) * (parseInt(navigation.page, 10) - 1) + parseInt(pos, 10);
    }

    prevAjax = $.ajax({
        type: "POST",
        url: "../prod/records/",
        dataType: 'json',
        data: {
            env: env,
            pos: absolutePos,
            cont: contId,
            roll: roll,
            options_serial: options_serial,
            query: query
        },
        beforeSend: function () {
            if (prevAjaxrunning)
                prevAjax.abort();
            if (env == 'RESULT')
                $('#current_result_n').empty().append(parseInt(pos) + 1);
            prevAjaxrunning = true;
            $('#PREVIEWIMGDESC, #PREVIEWOTHERS').addClass('loading');
        },
        error: function (data) {
            prevAjaxrunning = false;
            $('#PREVIEWIMGDESC, #PREVIEWOTHERS').removeClass('loading');
            posAsk = null;
        },
        timeout: function () {
            prevAjaxrunning = false;
            $('#PREVIEWIMGDESC, #PREVIEWOTHERS').removeClass('loading');
            posAsk = null;
        },
        success: function (data) {
            cancelPreview();
            prevAjaxrunning = false;
            posAsk = null;

            if (data.error) {
                $('#PREVIEWIMGDESC, #PREVIEWOTHERS').removeClass('loading');
                alert(data.error);
                if (justOpen)
                    closePreview();
                return;
            }
            posAsk = data.pos - 1;

            $('#PREVIEWIMGCONT').empty().append(data.html_preview);
            $('#PREVIEWIMGCONT .thumb_wrapper')
                .width('100%').height('100%').image_enhance({zoomable: true});

            $('#PREVIEWIMGDESCINNER').empty().append(data.desc);
            $('#HISTORICOPS').empty().append(data.history);
            $('#popularity').empty().append(data.popularity);

            if ($('#popularity .bitly_link').length > 0) {

                BitlyCB.statsResponse = function (data) {
                    var result = data.results;
                    if ($('#popularity .bitly_link_' + result.userHash).length > 0) {
                        $('#popularity .bitly_link_' + result.userHash).append(' (' + result.clicks + ' clicks)');
                    }
                };
                BitlyClient.stats($('#popularity .bitly_link').html(), 'BitlyCB.statsResponse');
            }

            p4.preview.current = {};
            p4.preview.current.width = parseInt($('#PREVIEWIMGCONT input[name=width]').val());
            p4.preview.current.height = parseInt($('#PREVIEWIMGCONT input[name=height]').val());
            p4.preview.current.tot = data.tot;
            p4.preview.current.pos = relativePos;

            if ($('#PREVIEWBOX img.record.zoomable').length > 0) {
                $('#PREVIEWBOX img.record.zoomable').draggable();
            }

            $('#SPANTITLE').empty().append(data.title);
            $("#PREVIEWTITLE_COLLLOGO").empty().append(data.collection_logo);
            $("#PREVIEWTITLE_COLLNAME").empty().append(data.databox_label + ' / ' + data.collection_name);

            setPreview();

            if (env != 'RESULT') {
                if (justOpen || reload) {
                    setCurrent(data.current);
                }
                viewCurrent($('#PREVIEWCURRENT li.selected'));
            }
            else {
                if (!justOpen) {
                    $('#PREVIEWCURRENT li.selected').removeClass('selected');
                    $('#PREVIEWCURRENTCONT li.current' + absolutePos).addClass('selected');
                }
                if (justOpen || ($('#PREVIEWCURRENTCONT li.current' + absolutePos).length === 0) || ($('#PREVIEWCURRENTCONT li:last')[0] == $('#PREVIEWCURRENTCONT li.selected')[0]) || ($('#PREVIEWCURRENTCONT li:first')[0] == $('#PREVIEWCURRENTCONT li.selected')[0])) {
                    getAnswerTrain(pos, data.tools, query, options_serial);
                }

                viewCurrent($('#PREVIEWCURRENT li.selected'));
            }
            if (env == 'REG' && $('#PREVIEWCURRENT').html() === '') {
                getRegTrain(contId, pos, data.tools);
            }
            setOthers(data.others);
            setTools(data.tools);
            $('#tooltip').css({
                'display': 'none'
            });
            $('#PREVIEWIMGDESC, #PREVIEWOTHERS').removeClass('loading');
            if (!justOpen || (p4.preview.mode != env))
                resizePreview();

            p4.preview.mode = env;
            $('#EDIT_query').focus();

            $('#PREVIEWOTHERSINNER .otherBaskToolTip').tooltip();

            return;
        }

    });

}

function zoomPreview(bool) {

    var el = $('#PREVIEWIMGCONT img.record');

    if (el.length === 0)
        return;

    var t1 = parseInt(el.css('top'));
    var l1 = parseInt(el.css('left'));
    var w1 = el.width();
    var h1 = el.height();

    var w2, t2;

    if (bool) {
        if (w1 * 1.08 < 32767)
            w2 = w1 * 1.08;
        else
            w2 = w1;
    }
    else {
        if (w1 / 1.08 > 20)
            w2 = w1 / 1.08;
        else
            w2 = w1;
    }

    var ratio = p4.preview.current.width / p4.preview.current.height;
    h2 = Math.round(w2 / ratio);
    w2 = Math.round(w2);

    t2 = Math.round(t1 - (h2 - h1) / 2) + 'px';
    var l2 = Math.round(l1 - (w2 - w1) / 2) + 'px';

    var wPreview = $('#PREVIEWIMGCONT').width() / 2;
    var hPreview = $('#PREVIEWIMGCONT').height() / 2;

    var nt = Math.round((h2 / h1) * (t1 - hPreview) + hPreview);
    var nl = Math.round(((w2 / w1) * (l1 - wPreview)) + wPreview);

    el.css({
        left: nl,
        top: nt
    }).width(w2).height(h2);
}

function getAnswerTrain(pos, tools, query, options_serial) {
    // keep relative position for answer train:
    var relativePos = pos;
    // update real absolute position with pagination:
    var absolutePos = parseInt(p4.navigation.perPage,10) * (parseInt(p4.navigation.page, 10) - 1) + parseInt(pos,10);

    $('#PREVIEWCURRENTCONT').fadeOut('fast');
    $.ajax({
        type: "POST",
        url: "/prod/query/answer-train/",
        dataType: 'json',
        data: {
            pos: absolutePos,
            options_serial: options_serial,
            query: query
        },
        success: function (data) {
            setCurrent(data.current);
            viewCurrent($('#PREVIEWCURRENT li.selected'));
            setTools(tools);
            return;
        }
    });
}


function getRegTrain(contId, pos, tools) {
    $.ajax({
        type: "POST",
        url: "/prod/query/reg-train/",
        dataType: 'json',
        data: {
            cont: contId,
            pos: pos
        },
        success: function (data) {
            setCurrent(data.current);
            viewCurrent($('#PREVIEWCURRENT li.selected'));
            if (typeof(tools) != 'undefined')
                setTools(tools);
            return;
        }
    });
}

function bounce(sbid, term, field) {
    doThesSearch('T', sbid, term, field);
    closePreview();
}

function setTitle(title) {
    $('#SPANTITLE').empty().append(title);
}

function cancelPreview() {
    $('#PREVIEWIMGDESCINNER').empty();
    $('#PREVIEWIMGCONT').empty();
    p4.preview.current = false;
}


function startSlide() {
    if (!p4.slideShow) {
        p4.slideShow = true;
    }
    if (p4.slideShowCancel) {
        p4.slideShowCancel = false;
        p4.slideShow = false;
        $('#start_slide').show();
        $('#stop_slide').hide();
    }
    if (!p4.preview.open) {
        p4.slideShowCancel = false;
        p4.slideShow = false;
        $('#start_slide').show();
        $('#stop_slide').hide();
    }
    if (p4.slideShow) {
        $('#start_slide').hide();
        $('#stop_slide').show();
        getNext();
        setTimeout("startSlide()", 3000);
    }
}

function stopSlide() {
    p4.slideShowCancel = true;
    $('#start_slide').show();
    $('#stop_slide').hide();
}

//var posAsk = null;

function getNext() {
    if (p4.preview.mode == 'REG' && parseInt(p4.preview.current.pos) === 0)
        $('#PREVIEWCURRENTCONT li img:first').trigger("click");
    else {
        if (p4.preview.mode == 'RESULT') {
            posAsk = parseInt(p4.preview.current.pos) + 1;
            posAsk = (posAsk >= parseInt(p4.tot) || isNaN(posAsk)) ? 0 : posAsk;
            openPreview(false, 'RESULT', posAsk, '', false);
        }
        else {
            if (!$('#PREVIEWCURRENT li.selected').is(':last-child'))
                $('#PREVIEWCURRENT li.selected').next().children('img').trigger("click");
            else
                $('#PREVIEWCURRENT li:first-child').children('img').trigger("click");
        }

    }
}
function reloadPreview() {
    $('#PREVIEWCURRENT li.selected img').trigger("click");
}

function getPrevious() {
    if (p4.preview.mode == 'RESULT') {
        posAsk = parseInt(p4.preview.current.pos) - 1;
        if (p4.navigation.page === 1) {
            // may go to last result
            posAsk = (posAsk < 0) ? ((parseInt(p4.tot) - 1)) : posAsk;
        }
        openPreview(false, 'RESULT', posAsk, '', false);
    }
    else {
        if (!$('#PREVIEWCURRENT li.selected').is(':first-child'))
            $('#PREVIEWCURRENT li.selected').prev().children('img').trigger("click");
        else
            $('#PREVIEWCURRENT li:last-child').children('img').trigger("click");
    }
}

function setOthers(others) {

    $('#PREVIEWOTHERSINNER').empty();
    if (others !== '') {
        $('#PREVIEWOTHERSINNER').append(others);

        $('#PREVIEWOTHERS table.otherRegToolTip').tooltip();
    }
}

function setTools(tools) {
    $('#PREVIEWTOOL').empty().append(tools);
    if (!p4.slideShowCancel && p4.slideShow) {
        $('#start_slide').hide();
        $('#stop_slide').show();
    } else {
        $('#start_slide').show();
        $('#stop_slide').hide();
    }
}

function setCurrent(current) {
    if (current !== '') {
        var el = $('#PREVIEWCURRENT');
        el.removeClass('loading').empty().append(current);

        $('ul', el).width($('li', el).length * 80);
        $('img.prevRegToolTip', el).tooltip();
        $.each($('img.openPreview'), function (i, el) {
            var jsopt = $(el).attr('jsargs').split('|');
            $(el).removeAttr('jsargs');
            $(el).removeClass('openPreview');
            $(el).bind('click', function () {
                viewCurrent($(this).parent());
                // convert abssolute to relative position
                var absolutePos = jsopt[1];
                var relativePos = parseInt(absolutePos, 10) - parseInt(p4.navigation.perPage, 10) * (parseInt(p4.navigation.page, 10) - 1);
                // keep relative position for answer train:
                openPreview(this, jsopt[0], relativePos, jsopt[2],false);
            });
        });
    }
}

function viewCurrent(el) {
    if (el.length === 0) {
        return;
    }
    $('#PREVIEWCURRENT li.selected').removeClass('selected');
    el.addClass('selected');
    $('#PREVIEWCURRENTCONT').animate({'scrollLeft': ($('#PREVIEWCURRENT li.selected').position().left + $('#PREVIEWCURRENT li.selected').width() / 2 - ($('#PREVIEWCURRENTCONT').width() / 2 ))});
    return;
}

function setPreview() {
    if (!p4.preview.current)
        return;

    var zoomable = $('img.record.zoomable');
    if (zoomable.length > 0 && zoomable.hasClass('zoomed'))
        return;

    var h = parseInt(p4.preview.current.height);
    var w = parseInt(p4.preview.current.width);
//	if(p4.preview.current.type == 'video')
//	{
//		var h = parseInt(p4.preview.current.flashcontent.height);
//		var w = parseInt(p4.preview.current.flashcontent.width);
//	}
    var t = 20;
    var de = 0;

    var margX = 0;
    var margY = 0;

    if ($('#PREVIEWIMGCONT .record_audio').length > 0) {
        margY = 100;
        de = 60;
    }


//	if(p4.preview.current.type != 'flash')
//	{
    var ratioP = w / h;
    var ratioD = parseInt(p4.preview.width) / parseInt(p4.preview.height);

    if (ratioD > ratioP) {
        //je regle la hauteur d'abord
        if ((parseInt(h) + margY) > parseInt(p4.preview.height)) {
            h = Math.round(parseInt(p4.preview.height) - margY);
            w = Math.round(h * ratioP);
        }
    }
    else {
        if ((parseInt(w) + margX) > parseInt(p4.preview.width)) {
            w = Math.round(parseInt(p4.preview.width) - margX);
            h = Math.round(w / ratioP);
        }
    }
//	}
//	else
//	{

//		h = Math.round(parseInt(p4.preview.height) - margY);
//		w = Math.round(parseInt(p4.preview.width) - margX);
//	}
    t = Math.round((parseInt(p4.preview.height) - h - de) / 2);
    var l = Math.round((parseInt(p4.preview.width) - w) / 2);
    $('#PREVIEWIMGCONT .record').css({
        width: w,
        height: h,
        top: t,
        left: l
    }).attr('width', w).attr('height', h);
}

function classicMode() {
    $('#PREVIEWCURRENTCONT').animate({'scrollLeft': ($('#PREVIEWCURRENT li.selected').position().left - 160)});
    p4.currentViewMode = 'classic';
}

function closePreview() {
    p4.preview.open = false;
    hideOverlay();

    $('#PREVIEWBOX').fadeTo(500, 0);
    $('#PREVIEWBOX').queue(function () {
        $(this).css({
            'display': 'none'
        });
        cancelPreview();
        $(this).dequeue();
    });

}
;function initializeEdit() {
    p4.edit = {};
    p4.edit.curField = "?";
    p4.edit.editBox = $('#idFrameE');
    p4.edit.textareaIsDirty = false;
    p4.edit.fieldLastValue = "";
    p4.edit.lastClickId = null;
    p4.edit.sbas_id = false;
    p4.edit.what = false;
    p4.edit.regbasprid = false;
    p4.edit.newrepresent = false;
    p4.edit.ssel = false;
}

$(document).ready(function () {
    $(window).bind('resize', function () {
        setPreviewEdit();
        setSizeLimits();
    });
});

function setSizeLimits() {
    if (!$('#EDITWINDOW').is(':visible'))
        return;

    if ($('#EDIT_TOP').data("ui-resizable")) {
        $('#EDIT_TOP').resizable('option', 'maxHeight', ($('#EDIT_ALL').height() - $('#buttonEditing').height() - 10 - 160));
    }
    if ($('#divS_wrapper').data("ui-resizable")) {
        $('#divS_wrapper').resizable('option', 'maxWidth', ($('#EDIT_MID_L').width() - 270));
    }
    if ($('#EDIT_MID_R').data("ui-resizable")) {
        $('#EDIT_MID_R').resizable('option', 'maxWidth', ($('#EDIT_MID_R').width() + $('#idEditZone').width() - 240));
    }
}

function edit_kdwn(evt, src) {
    cancelKey = false;

    switch (evt.keyCode) {
        case 13:
        case 10:
            if (p4.edit.T_fields[p4.edit.curField].type == "date")
                cancelKey = true;
    }

    if (cancelKey) {
        evt.cancelBubble = true;
        if (evt.stopPropagation)
            evt.stopPropagation();
        return(false);
    }
    return(true);
}

// ----------------------------------------------------------------------------------------------
// des events sur le textarea pour tracker la selection (chercher dans le thesaurus...)
// ----------------------------------------------------------------------------------------------
function edit_mdwn_ta(evt) {
    evt.cancelBubble = true;
    return(true);
}

// mouse up textarea
function edit_mup_ta(evt, obj) {

    if (p4.edit.T_fields[p4.edit.curField].tbranch) {
        if (obj.value != "")
            ETHSeeker.search(obj.value);
    }
    return(true);
}

// key up textarea
function edit_kup_ta(evt, obj) {

    var cancelKey = false;
    var o;
    switch (evt.keyCode) {
        case 27:	// esc : on restore la valeur avant editing
            //			$("#btn_cancel", p4.edit.editBox).parent().css("backgroundColor", "#000000");
            edit_validField(evt, "cancel");
            //			self.setTimeout("document.getElementById('btn_cancel').parentNode.style.backgroundColor = '';", 100);
            cancelKey = true;
            break;
    }

    if (cancelKey) {
        evt.cancelBubble = true;
        if (evt.stopPropagation)
            evt.stopPropagation();
        return(false);
    }
    if (!p4.edit.textareaIsDirty && ($("#idEditZTextArea", p4.edit.editBox).val() != p4.edit.fieldLastValue)) {
        p4.edit.textareaIsDirty = true;
    }

    var s = obj.value;
    if (p4.edit.T_fields[p4.edit.curField].tbranch) {
        if (s != "")
            ETHSeeker.search(s);
    }
    return(true);
}

// ---------------------------------------------------------------------------
// on a clique sur le peudo champ 'status'
// ---------------------------------------------------------------------------
function edit_mdwn_status(evt) {
    if (!p4.edit.textareaIsDirty || edit_validField(evt, "ask_ok") == true)
        editStatus(evt);
    evt.cancelBubble = true;
    if (evt.stopPropagation)
        evt.stopPropagation();
    return(false);
}

// ---------------------------------------------------------------------------
// on a clique sur un champ
// ---------------------------------------------------------------------------
function edit_mdwn_fld(evt, meta_struct_id, fieldname) {
    if (!p4.edit.textareaIsDirty || edit_validField(evt, "ask_ok") == true)
        editField(evt, meta_struct_id);
}

// ---------------------------------------------------------------------------
// change de champ (avec les fleches autour du nom champ)
// ---------------------------------------------------------------------------
function edit_chgFld(evt, dir) {
    var current_field = $('#divS .edit_field.active');
    if (current_field.length == 0) {
        current_field = $('#divS .edit_field:first');
        current_field.trigger('click');
    }
    else {
        if (dir >= 0) {
            current_field.next().trigger('click');
        }
        else {
            current_field.prev().trigger('click');
        }
    }
}

// ---------------------------------------------------------------------------
// on active le pseudo champ 'status'
// ---------------------------------------------------------------------------
function editStatus(evt) {
    $(".editDiaButtons", p4.edit.editBox).hide();

    document.getElementById('idEditZTextArea').blur();
    document.getElementById('EditTextMultiValued').blur();

    $("#idFieldNameEdit", p4.edit.editBox).html("[STATUS]");
    $("#idExplain", p4.edit.editBox).html("&nbsp;");

    $("#ZTextMultiValued", p4.edit.editBox).hide();
    $("#ZTextMonoValued", p4.edit.editBox).hide();
    $("#ZTextStatus", p4.edit.editBox).show();

    $("#idEditZone", p4.edit.editBox).show();

    document.getElementById("editFakefocus").focus();
    p4.edit.curField = -1;
    activeField();
}

function activeField() {
    var meta_struct_id = parseInt(p4.edit.curField);

    meta_struct_id = (isNaN(meta_struct_id) || meta_struct_id < 0) ? 'status' : meta_struct_id;

    $('#divS div.active, #divS div.hover').removeClass('active hover');
    $('#EditFieldBox_' + meta_struct_id).addClass('active');

    var cont = $('#divS');
    var calc = $('#EditFieldBox_' + meta_struct_id).offset().top - cont.offset().top;// hauteur relative par rapport au visible

    if (calc > cont.height() || calc < 0) {
        cont.scrollTop(calc + cont.scrollTop());
    }
}

//// ---------------------------------------------------------------------------
//// on change de champ courant
//// ---------------------------------------------------------------------------

function editField(evt, meta_struct_id) {
    document.getElementById('idEditZTextArea').blur();
    document.getElementById('EditTextMultiValued').blur();
    $(".editDiaButtons", p4.edit.editBox).hide();

    $('#idEditZTextArea, #EditTextMultiValued').unbind('keyup.maxLength');

    p4.edit.curField = meta_struct_id;
    if (meta_struct_id >= 0) {
        var name = p4.edit.T_fields[meta_struct_id].label + (p4.edit.T_fields[meta_struct_id].required ? '<span style="font-weight:bold;font-size:16px;"> * </span>' : '');
        $("#idFieldNameEdit", p4.edit.editBox).html(name);

        var vocabType = p4.edit.T_fields[meta_struct_id].vocabularyControl;

        for (f in p4.edit.T_fields)	// tous les champs de la base
            p4.edit.T_fields[f]._sgval = [];
        var t_lsgval = {};
        var t_selcol = {};		// les bases (coll) dont au - une thumb est selectionnee
        var ncolsel = 0;
        var nrecsel = 0;
        for (i in p4.edit.T_records) {
            if (!p4.edit.T_records[i]._selected)
                continue;
            nrecsel++;

            var bid = "b" + p4.edit.T_records[i].bid;
            if (t_selcol[bid])
                continue;

            t_selcol[bid] = 1;
            ncolsel++;
            for (f in p4.edit.T_sgval[bid]) {
                if (!t_lsgval[f])
                    t_lsgval[f] = {};
                for (ivs in p4.edit.T_sgval[bid][f]) {
                    vs = p4.edit.T_sgval[bid][f][ivs];
                    if (!t_lsgval[f][vs])
                        t_lsgval[f][vs] = 0;
                    t_lsgval[f][vs]++;
                }
            }
        }

        var suggestedValuesCollection = t_lsgval;

        if(!$.isEmptyObject(suggestedValuesCollection[meta_struct_id])) {
            var selectElement = $('<select><option selected disabled>Suggested values</option> </select>');
            var selectIdValue = "idSelectSuggestedValues_" + meta_struct_id;
            selectElement.attr('id', selectIdValue);

            $.each(suggestedValuesCollection[meta_struct_id], function(key, value) {
                var optionElement = $("<option></option>");
                optionElement.attr("value",key);
                optionElement.text(key);
                selectElement.append(optionElement);
            });

            selectElement.on('change', function(e) {
                if (p4.edit.T_fields[p4.edit.curField].multi) {
                    $("#EditTextMultiValued", p4.edit.editBox).val($(this).val());
                    $('#EditTextMultiValued').trigger('keyup.maxLength');
                    edit_addmval($('#EditTextMultiValued', p4.edit.editBox).val(), null);
                }
                else {
                    if (is_ctrl_key(e)) {
                        var t = $("#idEditZTextArea", p4.edit.editBox).val();
                        $("#idEditZTextArea", p4.edit.editBox).val(t + (t ? " ; " : "") + $(this).val());
                    }
                    else {
                        $("#idEditZTextArea", p4.edit.editBox).val($(this).val());
                    }
                    $('#idEditZTextArea').trigger('keyup.maxLength');
                    p4.edit.textareaIsDirty = true;
                    if (p4.edit.T_fields[p4.edit.curField]._status != 2)
                        edit_validField(evt, "ask_ok");
                }
            });

            $('#idFieldSuggestedValues', p4.edit.editBox).empty().append(selectElement);

            $('#idFieldSuggestedValues', p4.edit.editBox).css('visibility', 'visible');
            $('.edit-zone-title', p4.edit.editBox).css('height', 80);
            $('#EDIT_EDIT', p4.edit.editBox).css('top', 80);

        }else {
            $('#idFieldSuggestedValues', p4.edit.editBox).css('visibility', 'hidden');
            $('.edit-zone-title', p4.edit.editBox).css('height', 45);
            $('#EDIT_EDIT', p4.edit.editBox).css('top', 45);
        }

        $('#idEditZTextArea, #EditTextMultiValued').autocomplete({
            minLength: 2,
            appendTo: "#idEditZone",
            source: function (request, response) {
                $.ajax({
                    url: '../prod/records/edit/vocabulary/' + vocabType + '/',
                    dataType: "json",
                    data: {
                        sbas_id: p4.edit.sbas_id,
                        query: request.term
                    },
                    success: function (data) {
                        response(data.results);
                    }
                });
            },
            select: function (event, ui) {

                edit_addmval(ui.item.label, ui.item.id);

                return false;
            }
        });


        if (p4.edit.T_fields[meta_struct_id].maxLength > 0) {
            var idexplain = $("#idExplain");
            idexplain.html('');

            $('#idEditZTextArea, #EditTextMultiValued').bind('keyup.maxLength',function () {
                var remaining = Math.max((p4.edit.T_fields[meta_struct_id].maxLength - $(this).val().length), 0);
                idexplain.html("<span class='metadatas_restrictionsTips' tooltipsrc='../prod/tooltip/metas/restrictionsInfos/" + p4.edit.sbas_id + "/" + meta_struct_id + "/'><img src='/assets/common/images/icons/help32.png' /><!--<img src='/assets/common/images/icons/alert.png' />--> Caracteres restants : " + (remaining) + "</span>");
                $('.metadatas_restrictionsTips', idexplain).tooltip();
            }).trigger('keyup.maxLength');
        }
        else {
            $("#idExplain").html("");
        }

        if (!p4.edit.T_fields[meta_struct_id].multi) {
            // champ monovalue : textarea
            $(".editDiaButtons", p4.edit.editBox).hide();

            if (p4.edit.T_fields[meta_struct_id].type == "date") {
                $("#idEditZTextArea", p4.edit.editBox).css("height", "16px");
                $("#idEditDateZone", p4.edit.editBox).show();
            }
            else {
                $("#idEditDateZone", p4.edit.editBox).hide();
                $("#idEditZTextArea", p4.edit.editBox).css("height", "100%");
            }

            $("#ZTextStatus", p4.edit.editBox).hide();
            $("#ZTextMultiValued", p4.edit.editBox).hide();
            $("#ZTextMonoValued", p4.edit.editBox).show();

            if (p4.edit.T_fields[meta_struct_id]._status == 2) {
                // heterogene
                $("#idEditZTextArea", p4.edit.editBox).val(p4.edit.fieldLastValue = "");
                $("#idEditZTextArea", p4.edit.editBox).addClass("hetero");
                $("#idDivButtons", p4.edit.editBox).show();	// valeurs h�t�rog�nes : les 3 boutons remplacer/ajouter/annuler
            }
            else {
                // homogene
                $("#idEditZTextArea", p4.edit.editBox).val(p4.edit.fieldLastValue = p4.edit.T_fields[meta_struct_id]._value);
                $("#idEditZTextArea", p4.edit.editBox).removeClass("hetero");

                $("#idDivButtons", p4.edit.editBox).hide();	// valeurs homog�nes
                if (p4.edit.T_fields[meta_struct_id].type == "date") {
                    var v = p4.edit.T_fields[meta_struct_id]._value.split(' ');
                    d = v[0].split('/');
                    var dateObj = new Date();
                    if (d.length == 3) {
                        dateObj.setYear(d[0]);
                        dateObj.setMonth((d[1] - 1));
                        dateObj.setDate(d[2]);
                    }

                    if ($("#idEditDateZone", p4.edit.editBox).data("ui-datepicker")) {
                        $("#idEditDateZone", p4.edit.editBox).datepicker('setDate', dateObj);
                    }
                }
            }
            p4.edit.textareaIsDirty = false;

            $("#idEditZone", p4.edit.editBox).show();

            $('#idEditZTextArea').trigger('keyup.maxLength');

            self.setTimeout("document.getElementById('idEditZTextArea').focus();", 50);
        }
        else {
            // champ multivalue : liste
            $("#ZTextStatus", p4.edit.editBox).hide();
            $("#ZTextMonoValued", p4.edit.editBox).hide();
            $("#ZTextMultiValued", p4.edit.editBox).show();

            $("#idDivButtons", p4.edit.editBox).hide();	// valeurs homogenes

            updateCurrentMval(meta_struct_id);

            $('#EditTextMultiValued', p4.edit.editBox).val("");
            $('#idEditZone', p4.edit.editBox).show();

            $('#EditTextMultiValued').trigger('keyup.maxLength');

            self.setTimeout("document.getElementById('EditTextMultiValued').focus();", 50);

//      reveal_mval();
        }
    }
    else {
        // pas de champ, masquer la zone du textarea
        $("#idEditZone", p4.edit.editBox).hide();
        $(".editDiaButtons", p4.edit.editBox).hide();

    }
    activeField();
}

function updateCurrentMval(meta_struct_id, HighlightValue, vocabularyId) {

    // on compare toutes les valeurs de chaque fiche selectionnee
    p4.edit.T_mval = [];			// tab des mots, pour trier
    var a = [];		// key : mot ; val : nbr d'occurences distinctes
    var n = 0;					// le nbr de records selectionnes

    for (r in p4.edit.T_records) {
        if (!p4.edit.T_records[r]._selected)
            continue;

        p4.edit.T_records[r].fields[meta_struct_id].sort(SortCompareMetas);

        var values = p4.edit.T_records[r].fields[meta_struct_id].getValues();

        for (v in values) {
            var word = values[v].getValue();
            var key = values[v].getVocabularyId() + '%' + word;

            if (typeof(a[key]) == 'undefined') {
                a[key] = {
                    'n': 0,
                    'f': new Array()
                };	// n:nbr d'occurences DISTINCTES du mot ; f:flag presence mot dans r
                p4.edit.T_mval.push(values[v]);
            }

            if (!a[key].f[r])
                a[key].n++;		// premiere apparition du mot dans le record r
            a[key].f[r] = true;	// on ne recomptera pas le mot s'il apparait a nouveau dans le meme record

        }

        n++;
    }

    p4.edit.T_mval.sort(SortCompareMetas);

    var t = "";
    for (var i in p4.edit.T_mval)	// pour lire le tableau 'a' dans l'ordre trie par 'p4.edit.T_mval'
    {
        var value = p4.edit.T_mval[i];
        var word = value.getValue();
        var key = value.getVocabularyId() + '%' + word;

        var extra = value.getVocabularyId() ? '<img src="/assets/common/images/icons/ressource16.png" /> ' : '';

        if (i > 0) {
            if (value.getVocabularyId() !== null && p4.edit.T_mval[i - 1].getVocabularyId() == value.getVocabularyId()) {
                continue;
            }
            if (value.getVocabularyId() === null && p4.edit.T_mval[i - 1].getVocabularyId() === null) {
                if (p4.edit.T_mval[i - 1].getValue() == value.getValue()) {
                    continue;	// on n'accepte pas les doublons
                }
            }
        }

        t += "<div onclick=\"edit_clkmval(this, " + i + ")\" class='"
            + (((value.getVocabularyId() === null || value.getVocabularyId() == vocabularyId) && HighlightValue == word) ? ' hilighted ' : '')
            + (a[key].n != n ? " hetero " : "") + "'>"
            + '<table><tr><td>'
            + extra
            + '<span class="value" vocabId="' + (value.getVocabularyId() ? value.getVocabularyId() : '') + '">'
            + $('<div/>').text(word).html()
            + "</span></td><td class='options'>"
            + '<a href="#" class="add_all"><img src="/assets/common/images/icons/plus11.png"/></a> '
            + '<a href="#" class="remove_all"><img src="/assets/common/images/icons/minus11.png"/></a>'
            + "</td></tr></table>"
            + "</div>";
    }
    $('#ZTextMultiValued_values', p4.edit.editBox).html(t);

    $('#ZTextMultiValued_values .add_all', p4.edit.editBox).unbind('click').bind('click', function () {
        var container = $(this).closest('div');

        var span = $('span.value', container)

        var value = span.text();
        var vocab_id = span.attr('vocabid');

        edit_addmval(value, vocab_id);
        updateFieldDisplay();
        return false;
    });
    $('#ZTextMultiValued_values .remove_all', p4.edit.editBox).unbind('click').bind('click', function () {
        var container = $(this).closest('div');

        var span = $('span.value', container)

        var value = span.text();
        var vocab_id = span.attr('vocabid');

        edit_delmval(value, vocab_id);
        updateFieldDisplay();
        return false;
    });

    updateFieldDisplay();
}


// ---------------------------------------------------------------------------
// on a clique sur une des multi-valeurs dans la liste
// ---------------------------------------------------------------------------
function edit_clkmval(mvaldiv, ival) {
    $(mvaldiv).parent().find('.hilighted').removeClass('hilighted');
    $(mvaldiv).addClass('hilighted');
    reveal_mval(p4.edit.T_mval[ival].getValue(), p4.edit.T_mval[ival].getVocabularyId());
}


// ---------------------------------------------------------------------------
// highlight la valeur en cours de saisie dans la liste des multi-valeurs
// appele par le onkeyup
// ---------------------------------------------------------------------------
function reveal_mval(value, vocabularyId) {
    if (typeof vocabularyId === 'undefined')
        vocabularyId = null;

    var textZone = $('#EditTextMultiValued');

    if (p4.edit.T_fields[p4.edit.curField].tbranch) {
        if (value != "")
            ETHSeeker.search(value);
    }

    if (value != "") {
        //		var nsel = 0;
        for (rec_i in p4.edit.T_records) {
            if (p4.edit.T_records[rec_i].fields[p4.edit.curField].hasValue(value, vocabularyId)) {
                $("#idEditDiaButtonsP_" + rec_i).hide();
                var talt = $.sprintf(language.editDelSimple, value);
                $("#idEditDiaButtonsM_" + rec_i).show()
                    .attr('alt', talt)
                    .attr('Title', talt)
                    .unbind('click').bind('click', function () {
                        var indice = $(this).attr('id').split('_').pop();
                        edit_diabutton(indice, 'del', value, vocabularyId);
                    });
            }
            else {
                $("#idEditDiaButtonsM_" + rec_i).hide();
                $("#idEditDiaButtonsP_" + rec_i).show();
                var talt = $.sprintf(language.editAddSimple, value);
                $("#idEditDiaButtonsP_" + rec_i).show().attr('alt', talt)
                    .attr('Title', talt)
                    .unbind('click').bind('click', function () {
                        var indice = $(this).attr('id').split('_').pop();
                        edit_diabutton(indice, 'add', value, vocabularyId);
                    });
            }
        }
        $(".editDiaButtons", p4.edit.editBox).show();
    }

    textZone.trigger('focus');
    return(true);
}

function edit_diabutton(record_indice, act, value, vocabularyId) {
    var meta_struct_id = p4.edit.curField;		// le champ en cours d'editing
    if (act == 'del') {
        p4.edit.T_records[record_indice].fields[meta_struct_id].removeValue(value, vocabularyId);
    }

    if (act == 'add') {
        p4.edit.T_records[record_indice].fields[meta_struct_id].addValue(value, false, vocabularyId);
    }
    updateCurrentMval(meta_struct_id, value, vocabularyId);
    reveal_mval(value, vocabularyId);

}

// ---------------------------------------------------------------------------
// on a clique sur le bouton 'ajouter' un mot dans le multi-val
// ---------------------------------------------------------------------------
function edit_addmval(value, VocabularyId) {
    var meta_struct_id = p4.edit.curField;		// le champ en cours d'editing

    // on ajoute le mot dans tous les records selectionnes
    for (var r = 0; r < p4.edit.T_records.length; r++) {
        if (!p4.edit.T_records[r]._selected)
            continue;

        p4.edit.T_records[r].fields[meta_struct_id].addValue(value, false, VocabularyId);
    }

    updateEditSelectedRecords(null);
}

// ---------------------------------------------------------------------------
// on a clique sur le bouton 'supprimer' un mot dans le multi-val
// ---------------------------------------------------------------------------
function edit_delmval(value, VocabularyId) {
    var meta_struct_id = p4.edit.curField;		// le champ en cours d'editing

    for (var r = 0; r < p4.edit.T_records.length; r++) {
        if (!p4.edit.T_records[r]._selected)
            continue;

        p4.edit.T_records[r].fields[meta_struct_id].removeValue(value, VocabularyId);
    }

    updateEditSelectedRecords(null);
}

// ---------------------------------------------------------------------------------------------------------
// en mode textarea, on clique sur ok, cancel ou fusion
// appele egalement quand on essaye de changer de champ ou d'image : si ret=false on interdit le changement
// ---------------------------------------------------------------------------------------------------------
function edit_validField(evt, action) {
    // action : 'ok', 'fusion' ou 'cancel'
    if (p4.edit.curField == "?")
        return(true);

    if (action == "cancel") {
        // on restore le contenu du champ
        $("#idEditZTextArea", p4.edit.editBox).val(p4.edit.fieldLastValue);
        $('#idEditZTextArea').trigger('keyup.maxLength');
        p4.edit.textareaIsDirty = false;
        return(true);
    }

    if (action == "ask_ok" && p4.edit.textareaIsDirty && p4.edit.T_fields[p4.edit.curField]._status == 2) {
        alert(language.edit_hetero);
        return(false);
    }
    var o, newvalue;
    if (o = document.getElementById("idEditField_" + p4.edit.curField)) {
        t = $("#idEditZTextArea", p4.edit.editBox).val();

        status = 0;
        firstvalue = "";
        for (i = 0; i < p4.edit.T_records.length; i++) {
            if (!p4.edit.T_records[i]._selected)
                continue;			// on ne modifie pas les fiches non selectionnees

            if (action == "ok" || action == "ask_ok") {
                p4.edit.T_records[i].fields[p4.edit.curField].addValue(t, false, null);
            }
            else if (action == "fusion" || action == "ask_fusion") {
                p4.edit.T_records[i].fields[p4.edit.curField].addValue(t, true, null);
            }

            check_required(i, p4.edit.curField);
        }
    }

    updateFieldDisplay();

    p4.edit.textareaIsDirty = false;


    editField(evt, p4.edit.curField);
    return(true);
}

function skipImage(evt, step) {
    var cache = $('#EDIT_FILM2');
    var first = $('.diapo.selected:first', cache);
    var last = $('.diapo.selected:last', cache);
    var sel = $('.diapo.selected', cache);

    sel.removeClass('selected');

    var i = step == 1 ? (parseInt(last.attr('pos')) + 1) : (parseInt(first.attr('pos')) - 1);

    if (i < 0)
        i = parseInt($('.diapo:last', cache).attr('pos'));
    else if (i >= $('.diapo', cache).length)
        i = 0;

    edit_clk_editimg(evt, i);
}

function edit_select_all() {
    $('#EDIT_FILM2 .diapo', p4.edit.editBox).addClass('selected');

    for (i in p4.edit.T_records)
        p4.edit.T_records[i]._selected = true;

    p4.edit.lastClickId = 1;

    updateEditSelectedRecords(null);		// null : no evt available
}

// ---------------------------------------------------------------------------
// on a clique sur une thumbnail
// ---------------------------------------------------------------------------
function edit_clk_editimg(evt, i) {
    if (p4.edit.curField >= 0) {
        if (p4.edit.textareaIsDirty && edit_validField(evt, "ask_ok") == false)
            return;
    }

    // guideline : si on mousedown sur une selection, c'est qu'on risque de draguer, donc on ne desectionne pas
    if (evt && evt.type == "mousedown" && p4.edit.T_records[i]._selected)
        return;

    if (evt && is_shift_key(evt) && p4.edit.lastClickId != null) {
        // shift donc on sel du p4.edit.lastClickId a ici
        var pos_from = p4.edit.T_pos[p4.edit.lastClickId];
        var pos_to = p4.edit.T_pos[i];
        if (pos_from > pos_to) {
            var tmp = pos_from;
            pos_from = pos_to;
            pos_to = tmp;
        }

        var pos;
        for (pos = pos_from; pos <= pos_to; pos++) {
            var id = p4.edit.T_id[pos];
            if (!p4.edit.T_records[id]._selected)	// toutes les fiches selectionnees
            {
                p4.edit.T_records[id]._selected = true;
                $("#idEditDiapo_" + id, p4.edit.editBox).addClass('selected');
            }
        }
    }
    else {
        if (!evt || !is_ctrl_key(evt)) {
            // on deselectionne tout avant
            var id;
            for (id in p4.edit.T_records) {
                if (p4.edit.T_records[id]._selected)	// toutes les fiches selectionnees
                {
                    p4.edit.T_records[id]._selected = false;
                    $("#idEditDiapo_" + id, p4.edit.editBox).removeClass('selected');
                }
            }
        }
        if (i >= 0) {
            p4.edit.T_records[i]._selected = !p4.edit.T_records[i]._selected;
            if (p4.edit.T_records[i]._selected)
                $("#idEditDiapo_" + i, p4.edit.editBox).addClass('selected');
            else
                $("#idEditDiapo_" + i, p4.edit.editBox).removeClass('selected');
        }
    }

    $('#TH_Opreview .PNB10').empty();

    var selected = $('#EDIT_FILM2 .diapo.selected');
    if (selected.length == 1) {

        var r = selected.attr('id').split('_').pop();
        previewEdit(r);
    }

    p4.edit.lastClickId = i;
    updateEditSelectedRecords(evt);
}

function toggleGroupSelection() {
    var groupIndex = 0;
    edit_clk_editimg(false, groupIndex);
}

// ---------------------------------------------------------------------------
// on a clique sur une checkbox de status
// ---------------------------------------------------------------------------
function edit_clkstatus(evt, bit, val) {
    var ck0 = $("#idCheckboxStatbit0_" + bit);
    var ck1 = $("#idCheckboxStatbit1_" + bit);
    switch (val) {
        case 0:
            ck0.attr('class', "gui_ckbox_1");
            ck1.attr('class', "gui_ckbox_0");
            break;
        case 1:
            ck0.attr('class', "gui_ckbox_0");
            ck1.attr('class', "gui_ckbox_1");
            break;
    }
    var id;
    for (id in p4.edit.T_records) {
        if (p4.edit.T_records[id]._selected)	// toutes les fiches selectionnees
        {
            if ($('#idEditDiapo_' + id).hasClass('nostatus'))
                continue;

            p4.edit.T_records[id].statbits[bit].value = val;
            p4.edit.T_records[id].statbits[bit].dirty = true;
        }
    }
}

function updateEditSelectedRecords(evt) {
    $(".editDiaButtons", p4.edit.editBox).hide();

    for (n in p4.edit.T_statbits)	// tous les statusbits de la base
    {
        p4.edit.T_statbits[n]._value = "-1";			// val unknown
        for (i in p4.edit.T_records) {
            if (!p4.edit.T_records[i]._selected)
                continue;
            if (p4.edit.T_records[i].statbits.length === 0)
                continue;

            if (p4.edit.T_statbits[n]._value == "-1")
                p4.edit.T_statbits[n]._value = p4.edit.T_records[i].statbits[n].value;
            else if (p4.edit.T_statbits[n]._value != p4.edit.T_records[i].statbits[n].value)
                p4.edit.T_statbits[n]._value = "2";
        }
        var ck0 = $("#idCheckboxStatbit0_" + n);
        var ck1 = $("#idCheckboxStatbit1_" + n);

        switch (p4.edit.T_statbits[n]._value) {
            case "0":
            case 0:
                ck0.removeClass('gui_ckbox_0 gui_ckbox_2').addClass("gui_ckbox_1");
                ck1.removeClass('gui_ckbox_1 gui_ckbox_2').addClass("gui_ckbox_0");
                break;
            case "1":
            case 1:
                ck0.removeClass('gui_ckbox_1 gui_ckbox_2').addClass("gui_ckbox_0");
                ck1.removeClass('gui_ckbox_0 gui_ckbox_2').addClass("gui_ckbox_1");
                break;
            case "2":
                ck0.removeClass('gui_ckbox_0 gui_ckbox_1').addClass("gui_ckbox_2");
                ck1.removeClass('gui_ckbox_0 gui_ckbox_1').addClass("gui_ckbox_2");
                break;
        }
    }


    var nostatus = $('.diapo.selected.nostatus', p4.edit.editBox).length;
    var status_box = $('#ZTextStatus');
    $('.nostatus, .somestatus, .displaystatus', status_box).hide();

    if (nostatus == 0) {
        $('.displaystatus', status_box).show();
    }
    else {
        var yesstatus = $('.diapo.selected', p4.edit.editBox).length;
        if (nostatus == yesstatus) {
            $('.nostatus', status_box).show();
        }
        else {
            $('.somestatus, .displaystatus', status_box).show();
        }
    }

    // calcul des valeurs suggerees COMMUNES aux records (collections) selectionnes //
    for (f in p4.edit.T_fields)	// tous les champs de la base
        p4.edit.T_fields[f]._sgval = [];
    var t_lsgval = {};
    var t_selcol = {};		// les bases (coll) dont au - une thumb est selectionnee
    var ncolsel = 0;
    var nrecsel = 0;
    for (i in p4.edit.T_records) {
        if (!p4.edit.T_records[i]._selected)
            continue;
        nrecsel++;

        var bid = "b" + p4.edit.T_records[i].bid;
        if (t_selcol[bid])
            continue;

        t_selcol[bid] = 1;
        ncolsel++;
        for (f in p4.edit.T_sgval[bid]) {
            if (!t_lsgval[f])
                t_lsgval[f] = {};
            for (ivs in p4.edit.T_sgval[bid][f]) {
                vs = p4.edit.T_sgval[bid][f][ivs];
                if (!t_lsgval[f][vs])
                    t_lsgval[f][vs] = 0;
                t_lsgval[f][vs]++;
            }
        }
    }

    updateFieldDisplay();

    if (p4.edit.curField == -1)
        editStatus(evt);
    else
        editField(evt, p4.edit.curField);
}

function updateFieldDisplay() {
    for (f in p4.edit.T_fields)	// tous les champs de la base
    {
        p4.edit.T_fields[f]._status = 0;			// val unknown
        for (i in p4.edit.T_records) {
            if (!p4.edit.T_records[i]._selected)
                continue;


            if (p4.edit.T_records[i].fields[f].isEmpty()) {
                var v = "";
            }
            else {
                // le champ existe dans la fiche
                if (p4.edit.T_fields[f].multi) {
                    // champ multi : on compare la concat des valeurs
                    var v = p4.edit.T_records[i].fields[f].getSerializedValues()
                }
                else {
                    var v = p4.edit.T_records[i].fields[f].getValue().getValue();
                }
            }

            if (p4.edit.T_fields[f]._status == 0) {
                p4.edit.T_fields[f]._value = v;
                p4.edit.T_fields[f]._status = 1;
            }
            else if (p4.edit.T_fields[f]._status == 1 && p4.edit.T_fields[f]._value != v) {
                p4.edit.T_fields[f]._value = "*****";
                p4.edit.T_fields[f]._status = 2;
                break;	// plus la peine de verifier le champ sur les autres records
            }
        }
        if (o = document.getElementById("idEditField_" + f)) {
            if (p4.edit.T_fields[f]._status == 2)	// mixed
                o.innerHTML = "<span class='hetero'>xxxxx</span>";
            else {
                var v = p4.edit.T_fields[f]._value;
                v = (v instanceof(Array)) ? v.join(";") : v;
                o.innerHTML = cleanTags(v).replace(/\n/gm, "<span style='color:#0080ff'>&para;</span><br/>");
            }
        }
    }
}

function SortCompareMetas(a, b) {
    if (typeof(a) != 'object')
        return(-1);
    if (typeof(b) != 'object')
        return(1);
    var na = a.getValue().toUpperCase();
    var nb = b.getValue().toUpperCase();
    if (na == nb)
        return(0);
    return(na < nb ? -1 : 1);
}

//---------------------------------------------------------------------
//nettoie
//---------------------------------------------------------------------
function cleanTags(string) {
    var chars2replace = [
        {
            f: "&",
            t: "&amp;"
        },
        {
            'f': "<",
            't': "&lt;"
        },
        {
            'f': ">",
            't': "&gt;"
        }
    ];
    for (c in chars2replace)
        string = string.replace(RegExp(chars2replace[c].f, "g"), chars2replace[c].t);
    return string;
}

function check_required(id_r, id_f) {
    var required_fields = false;

    if (typeof id_r == 'undefined')
        id_r = false;
    if (typeof id_f == 'undefined')
        id_f = false;

    for (f in p4.edit.T_fields) {
        if (id_f !== false && f != id_f)
            continue;

        var name = p4.edit.T_fields[f].name;

        if (!p4.edit.T_fields[f].required)
            continue;

        for (r in p4.edit.T_records) {
            if (id_r !== false && r != id_r)
                continue;

            var elem = $('#idEditDiapo_' + r + ' .require_alert');

            elem.hide();

            if (!p4.edit.T_records[r].fields[f]) {
                elem.show();
                required_fields = true;
            }
            else {

                var check_required = '';

                // le champ existe dans la fiche
                if (p4.edit.T_fields[f].multi) {
                    // champ multi : on compare la concat des valeurs
                    check_required = $.trim(p4.edit.T_records[r].fields[f].getSerializedValues())
                }
                else if (p4.edit.T_records[r].fields[f].getValue()) {
                    check_required = $.trim(p4.edit.T_records[r].fields[f].getValue().getValue());
                }


                if (check_required == '') {
                    elem.show();
                    required_fields = true;
                }
            }
        }

    }
    return required_fields;
}

// ----------------------------------------------------------------------------------
// on a clique sur le 'ok' general : save
// ----------------------------------------------------------------------------------
function edit_applyMultiDesc(evt) {
    var sendorder = "";
    var sendChuOrder = "";

    var t = [];

    if (p4.edit.textareaIsDirty && edit_validField(evt, "ask_ok") == false)
        return(false);

    var required_fields = check_required();

    if (required_fields) {
        alert(language.some_required_fields);
        return;
    }

    $("#EDIT_ALL", p4.edit.editBox).hide();

    $("#EDIT_WORKING", p4.edit.editBox).show();

    for (r in p4.edit.T_records) {
        var record_datas = {
            record_id: p4.edit.T_records[r].rid,
            metadatas: [],
            edit: 0,
            status: null
        };

        var editDirty = false;

        for (f in p4.edit.T_records[r].fields) {
            if (!p4.edit.T_records[r].fields[f].isDirty()) {
                continue;
            }

            editDirty = true;
            record_datas.edit = 1;

            record_datas.metadatas = record_datas.metadatas.concat(
                p4.edit.T_records[r].fields[f].exportDatas()
            );
        }

        // les statbits
        var tsb = [];
        for (var n = 0; n < 64; n++)
            tsb[n] = 'x';
        sb_dirty = false;
        for (var n in p4.edit.T_records[r].statbits) {
            if (p4.edit.T_records[r].statbits[n].dirty) {
                tsb[63 - n] = p4.edit.T_records[r].statbits[n].value;
                sb_dirty = true;
            }
        }

        if (sb_dirty || editDirty) {
            if (sb_dirty === true)
                record_datas.status = tsb.join("");

            t.push(record_datas);
        }
    }

    var options = {
        mds: t,
        sbid: p4.edit.sbas_id,
        act: 'WORK',
        lst: $('#edit_lst').val(),
        act_option: 'SAVE' + p4.edit.what,
        regbasprid: p4.edit.regbasprid,
        // newrepresent:p4.edit.newrepresent,
        ssel: p4.edit.ssel
    };
    if (p4.edit.newrepresent != false)
        options.newrepresent = p4.edit.newrepresent;

    $.ajax({
        url: "../prod/records/edit/apply/",
        data: options
        //    ,dataType:'json'
        ,
        type: 'POST',
        success: function (data) {
            if (p4.edit.what == 'GRP' || p4.edit.what == 'SSEL') {
                p4.WorkZone.refresh('current');
            }
            $("#Edit_copyPreset_dlg").remove();
            $('#EDITWINDOW').hide();
            hideOverlay(2);
            if (p4.preview.open)
                reloadPreview();
            return;
        }
    });

}

function edit_cancelMultiDesc(evt) {


    var dirty = false;

    evt.cancelBubble = true;
    if (evt.stopPropagation)
        evt.stopPropagation();

    if (p4.edit.curField >= 0) {
        if (p4.edit.textareaIsDirty && edit_validField(evt, "ask_ok") == false)
            return;
    }

    for (r in p4.edit.T_records) {
        for (f in p4.edit.T_records[r].fields) {
            if ((dirty |= p4.edit.T_records[r].fields[f].isDirty()))
                break;
        }
        for (var n in p4.edit.T_records[r].statbits) {
            if ((dirty |= p4.edit.T_records[r].statbits[n].dirty))
                break;
        }
    }
    if (!dirty || confirm(language.confirm_abandon)) {
        $("#Edit_copyPreset_dlg").remove();
        $('#idFrameE .ww_content', p4.edit.editBox).empty();

        // on reaffiche tous les thesaurus
        for (i in p4.thesau.thlist)	// tous les thesaurus
        {
            var bid = p4.thesau.thlist[i].sbas_id;
            var e = document.getElementById('TH_T.' + bid + '.T');
            if (e)
                e.style.display = "";
        }
        self.setTimeout("$('#EDITWINDOW').fadeOut();hideOverlay(2);", 100);

    }
}

// ======================================================
// ================ gestion du thesaurus ================
// ======================================================

function edit_clickThesaurus(event)	// onclick dans le thesaurus
{
    // on cherche ou on a clique
    for (e = event.srcElement ? event.srcElement : event.target; e && ((!e.tagName) || (!e.id)); e = e.parentNode)
        ;
    if (e) {
        switch (e.id.substr(0, 4)) {
            case "TH_P":	// +/- de deploiement de mot
                js = "edit_thesaurus_ow('" + e.id.substr(5) + "')";
                self.setTimeout(js, 10);
                break;
        }
    }
    return(false);
}

function edit_dblclickThesaurus(event)	// ondblclick dans le thesaurus
{
    for (e = event.srcElement ? event.srcElement : event.target; e && ((!e.tagName) || (!e.id)); e = e.parentNode)
        ;
    if (e) {
        switch (e.id.substr(0, 4)) {
            case "TH_W":
                if (p4.edit.curField >= 0) {
                    var w = $(e).text();
                    if (p4.edit.T_fields[p4.edit.curField].multi) {
                        $("#EditTextMultiValued", p4.edit.editBox).val(w);
                        $('#EditTextMultiValued').trigger('keyup.maxLength');
                        edit_addmval($('#EditTextMultiValued', p4.edit.editBox).val(), null);
                    }
                    else {
                        $("#idEditZTextArea", p4.edit.editBox).val(w);
                        $('#idEditZTextArea').trigger('keyup.maxLength');
                        p4.edit.textareaIsDirty = true;
                    }
                }
                break;
        }
    }
    return(false);
}

function edit_thesaurus_ow(id)	// on ouvre ou ferme une branche de thesaurus
{
    var o = document.getElementById("TH_K." + id);
    if (o.className == "o") {
        // on ferme
        o.className = "c";
        document.getElementById("TH_P." + id).innerHTML = "+";
        document.getElementById("TH_K." + id).innerHTML = language.loading;
    }
    else if (o.className == "c" || o.className == "h") {
        // on ouvre
        o.className = "o";
        document.getElementById("TH_P." + id).innerHTML = "-";

        var t_id = id.split(".");
        var sbas_id = t_id[0];
        t_id.shift();
        var thid = t_id.join(".");
        var url = "/xmlhttp/getterm_prod.x.php";
        var parms = "bid=" + sbas_id;
        parms += "&lng=" + p4.lng;
        parms += "&sortsy=1";
        parms += "&id=" + thid;
        parms += "&typ=TH";

        ETHSeeker.openBranch(id, thid);
    }
    return(false);
}

function EditThesaurusSeeker(sbas_id) {
    this.jq = null;

    this.sbas_id = sbas_id;

    var zid = ("" + sbas_id).replace(new RegExp("\\.", "g"), "\\.") + "\\.T";

    this.TH_P_node = $("#TH_P\\." + zid, p4.edit.editBox);
    this.TH_K_node = $("#TH_K\\." + zid, p4.edit.editBox);

    this._ctimer = null;

    this.search = function (txt) {
        if (this._ctimer)
            clearTimeout(this._ctimer);
        var js = "ETHSeeker.search_delayed('" + txt.replace("'", "\\'") + "');";
        this._ctimer = setTimeout(js, 125);
    };

    this.search_delayed = function (txt) {
        if (this.jq && typeof this.jq.abort == "function") {
            this.jq.abort();
            this.jq = null;
        }
        txt = txt.replace("'", "\\'");
        var url = "/xmlhttp/openbranches_prod.h.php";
        var parms = {
            bid: this.sbas_id,
            lng: p4.lng,
            t: txt,
            mod: "TREE",
            u: Math.random()
        };

        var me = this;

        this.jq = $.ajax({
            url: url,
            data: parms,
            type: 'POST',
            success: function (ret) {
                me.TH_P_node.html("...");
                me.TH_K_node.attr("class", "h").html(ret);
                me.jq = null;
            },
            error: function () {

            },
            timeout: function () {

            }
        });
    };

    this.openBranch = function (id, thid) {
        if (this.jq) {
            this.jq.abort();
            this.jq = null;
        }
        var url = "/xmlhttp/getterm_prod.h.php";
        var parms = {
            bid: this.sbas_id,
            lng: p4.lng,
            sortsy: 1,
            id: thid,
            typ: "TH"
        };
        var me = this;


        this.jq = $.ajax({
            url: url,
            data: parms,
            success: function (ret) {
                var zid = "#TH_K\\." + id.replace(new RegExp("\\.", "g"), "\\.");	// escape les '.' pour jquery
                $(zid, p4.edit.editBox).html(ret);
                me.jq = null;
            },
            error: function () {

            },
            timeout: function () {

            }
        });
    };
}


function replace() {
    var field = $("#EditSRField", p4.edit.editBox).val();
    var search = $("#EditSearch", p4.edit.editBox).val();
    var replace = $("#EditReplace", p4.edit.editBox).val();

    var where = $("[name=EditSR_Where]:checked", p4.edit.editBox).val();
    var commut = "";
    var rgxp = $("#EditSROptionRX", p4.edit.editBox).prop('checked') ? true : false;

    var r_search;
    if (rgxp) {
        r_search = search;
        commut = ($("#EditSR_RXG", p4.edit.editBox).prop('checked') ? "g" : "")
            + ($("#EditSR_RXI", p4.edit.editBox).prop('checked') ? "i" : "");
    }
    else {
        commut = $("#EditSR_case", p4.edit.editBox).prop('checked') ? "g" : "gi";
        r_search = "";
        for (i = 0; i < search.length; i++) {
            var c = search.charAt(i);
            if (("^$[]()|.*+?\\").indexOf(c) != -1)
                r_search += "\\";
            r_search += c;
        }
        if (where == "exact")
            r_search = "^" + r_search + "$";
    }

    search = new RegExp(r_search, commut);

    var r, f, n = 0;
    for (r in p4.edit.T_records) {
        if (!p4.edit.T_records[r]._selected)
            continue;
        for (f in p4.edit.T_records[r].fields) {
            if (field == '' || field == f) {
                n += p4.edit.T_records[r].fields[f].replaceValue(search, replace);
            }
        }
    }

    humane.info($.sprintf(language.nFieldsChanged, n));

    updateEditSelectedRecords(null);
}

function changeReplaceMode(ckRegExp) {


    if (ckRegExp.checked) {
        $("#EditSR_TX", p4.edit.editBox).hide();
        $("#EditSR_RX", p4.edit.editBox).show();
    }
    else {
        $("#EditSR_RX", p4.edit.editBox).hide();
        $("#EditSR_TX", p4.edit.editBox).show();
    }
}

function preset_copy() {
    var html = "";
    for (i in p4.edit.T_fields) {
        if (p4.edit.T_fields[i]._status == 1) {
            if (p4.edit.T_fields[i].readonly) {
                continue;
            }
            var c = p4.edit.T_fields[i]._value === "" ? "" : "checked=\"1\"";
            var v = p4.edit.T_fields[i]._value;
            html += "<div><label class=\"checkbox\" for=\"new_preset_" + p4.edit.T_fields[i].name + "\"><input type=\"checkbox\" class=\"checkbox\" id=\"new_preset_" + p4.edit.T_fields[i].name + "\" value=\"" + i + "\" " + c + "/>" + "<b>" + p4.edit.T_fields[i].label + " : </b></label> ";
            html += cleanTags(p4.edit.T_fields[i]._value) + "</div>";
        }
    }
    $("#Edit_copyPreset_dlg FORM DIV").html(html);
    var $dialog = $("#Edit_copyPreset_dlg");
    if ($dialog.data("ui-dialog")) {
        // to show dialog on top of edit window
        $dialog.dialog("widget").css("z-index", 1300);
        $dialog.dialog("open");
    }
}

function preset_paint(data) {
    $(".EDIT_presets_list", p4.edit.editBox).html(data.html);
    $(".EDIT_presets_list A.triangle").click(
        function () {
            $(this).parent().parent().toggleClass("opened");
            return false;
        }
    );

    $(".EDIT_presets_list A.title").dblclick(
        function () {
            var preset_id = $(this).parent().parent().attr("id");
            if (preset_id.substr(0, 12) == "EDIT_PRESET_")
                preset_load(preset_id.substr(12));
            return false;
        }
    );

    $(".EDIT_presets_list A.delete").click(
        function () {
            var li = $(this).closest("LI");
            var preset_id = li.attr("id");
            var title = $(this).parent().children(".title").html();
            if (preset_id.substr(0, 12) == "EDIT_PRESET_" && confirm("supprimer le preset '" + title + "' ?")) {
                preset_delete(preset_id.substr(12), li);
            }
            return false;
        }
    );
}

function preset_delete(preset_id, li) {
    $.ajax({
        type: 'DELETE',
        url: "../prod/records/edit/presets/" + preset_id,
        data: {},
        dataType: 'json',
        success: function (data, textStatus) {
            li.remove();
        }
    });
}

function preset_load(preset_id) {
    $.ajax({
        type: 'GET',
        url: "../prod/records/edit/presets/" + preset_id,
        data: {},
        dataType: 'json',
        success: function (data, textStatus) {
            if ($("#Edit_copyPreset_dlg").data("ui-dialog")) {
                $("#Edit_copyPreset_dlg").dialog("close");
            }

            for (i in p4.edit.T_fields) {
                p4.edit.T_fields[i].preset = null;
                if (typeof(data.fields[p4.edit.T_fields[i].name]) != "undefined") {
                    p4.edit.T_fields[i].preset = data.fields[p4.edit.T_fields[i].name];
                }
            }
            for (var r = 0; r < p4.edit.T_records.length; r++) {
                if (!p4.edit.T_records[r]._selected)
                    continue;

                for (i in p4.edit.T_fields) {
                    if (p4.edit.T_fields[i].preset != null) {
                        for (val in p4.edit.T_fields[i].preset) {
                            // fix : some (old, malformed) presets values may need trim()
                            p4.edit.T_records[r].fields["" + i].addValue(p4.edit.T_fields[i].preset[val].trim(), false, null);
                        }
                    }
                }
            }
            updateEditSelectedRecords();
        }
    });
}


function hsplit1() {
    var el = $('#EDIT_TOP');
    if (el.length == 0)
        return;
    var h = $(el).outerHeight();
    $(el).height(h);
    var t = $(el).offset().top + h;

    $("#EDIT_MID", p4.edit.editBox).css("top", (t) + "px");
}
function vsplit1() {
    $('#divS_wrapper').height('auto');

    var el = $('#divS_wrapper');
    if (el.length == 0)
        return;
    var a = $(el).width();
    el.width(a);

    $("#idEditZone", p4.edit.editBox).css("left", (a + 20 ));
}
function vsplit2() {
    var el = $('#EDIT_MID_R');
    if (el.length == 0)
        return;
    var a = $(el).width();
    el.width(a);
    var v = $('#EDIT_ALL').width() - a - 35;
    $("#EDIT_MID_L", p4.edit.editBox).width(v);
}

function setPreviewEdit() {
    if (!$('#TH_Opreview').is(':visible'))
        return false;

    var selected = $('#EDIT_FILM2 .diapo.selected');

    if (selected.length != 1) {
        return;
    }

    var id = selected.attr('id').split('_').pop();

    var container = $('#TH_Opreview');
    var zoomable = $('img.record.zoomable', container);

    if (zoomable.length > 0 && zoomable.hasClass('zoomed'))
        return;

    //  var datas = p4.edit.T_records[id].preview;

    var h = parseInt($('input[name=height]', container).val());
    var w = parseInt($('input[name=width]', container).val());

    //  if(datas.doctype == 'video')
    //  {
    //    var h = parseInt(datas.height);
    //    var w = parseInt(datas.width);
    //  }
    var t = 0;
    var de = 0;

    var margX = 0;
    var margY = 0;

    if ($('img.record.record_audio', container).length > 0) {
        var margY = 100;
        de = 60;
    }

    var display_box = $('#TH_Opreview .PNB10');
    var dwidth = display_box.width();
    var dheight = display_box.height();


    //  if(datas.doctype != 'flash')
    //  {
    var ratioP = w / h;
    var ratioD = dwidth / dheight;

    if (ratioD > ratioP) {
        //je regle la hauteur d'abord
        if ((parseInt(h) + margY) > dheight) {
            h = Math.round(dheight - margY);
            w = Math.round(h * ratioP);
        }
    }
    else {
        if ((parseInt(w) + margX) > dwidth) {
            w = Math.round(dwidth - margX);
            h = Math.round(w / ratioP);
        }
    }
    //  }
    //  else
    //  {
    //
    //    h = Math.round(dheight - margY);
    //    w = Math.round(dwidth - margX);
    //  }
    t = Math.round((dheight - h - de) / 2);
    var l = Math.round((dwidth - w) / 2);
    $('.record', container).css({
        width: w,
        height: h,
        top: t,
        left: l
    }).attr('width', w).attr('height', h);

}

function previewEdit(r) {

    $('#TH_Opreview .PNB10').empty().append(p4.edit.T_records[r].preview);

    if ($('img.PREVIEW_PIC.zoomable').length > 0) {
        $('img.PREVIEW_PIC.zoomable').draggable();
    }
    setPreviewEdit();
}

function startThisEditing(sbas_id, what, regbasprid, ssel) {

    p4.edit.sbas_id = sbas_id;
    p4.edit.what = what;
    p4.edit.regbasprid = regbasprid;
    p4.edit.ssel = ssel;

    for (r in p4.edit.T_records) {
        var fields = {};

        for (f in p4.edit.T_records[r].fields) {
            var meta_struct_id = p4.edit.T_records[r].fields[f].meta_struct_id;

            var name = p4.edit.T_fields[meta_struct_id].name;
            var label = p4.edit.T_fields[meta_struct_id].label;

            var multi = p4.edit.T_fields[meta_struct_id].multi;
            var required = p4.edit.T_fields[meta_struct_id].required;
            var readonly = p4.edit.T_fields[meta_struct_id].readonly;
            var maxLength = p4.edit.T_fields[meta_struct_id].maxLength;
            var minLength = p4.edit.T_fields[meta_struct_id].minLength;
            var type = p4.edit.T_fields[meta_struct_id].type;
            var separator = p4.edit.T_fields[meta_struct_id].separator;
            var vocabularyControl = p4.edit.T_fields[meta_struct_id].vocabularyControl;
            var vocabularyRestricted = p4.edit.T_fields[meta_struct_id].vocabularyRestricted;

            var fieldOptions = {
                multi: multi,
                required: required,
                readonly: readonly,
                maxLength: maxLength,
                minLength: minLength,
                type: type,
                separator: separator,
                vocabularyControl: vocabularyControl,
                vocabularyRestricted: vocabularyRestricted
            };

            var databoxField = new p4.databoxField(name, label, meta_struct_id, fieldOptions);

            var values = [];

            for (v in p4.edit.T_records[r].fields[f].values) {
                var meta_id = p4.edit.T_records[r].fields[f].values[v].meta_id;
                var value = p4.edit.T_records[r].fields[f].values[v].value;
                var vocabularyId = p4.edit.T_records[r].fields[f].values[v].vocabularyId;

                values.push(new p4.recordFieldValue(meta_id, value, vocabularyId));
            }

            fields[f] = new p4.recordField(databoxField, values);
        }

        p4.edit.T_records[r].fields = fields;
        p4.edit.fields = fields;

    }

    $('#EditTextMultiValued').bind('keyup', function () {
        reveal_mval($(this).val());
    });

    $('#EDIT_MID_R .tabs').tabs();

    $('#divS div.edit_field:odd').addClass('odd');
    $('#divS div').bind('mouseover',function () {
        $(this).addClass('hover');
    }).bind('mouseout', function () {
            $(this).removeClass('hover');
        });

    $('#editcontextwrap').remove();

    if ($('#editcontextwrap').length == 0)
        $('body').append('<div id="editcontextwrap"></div>');

    // if is a group, only select the group
    if (p4.edit.what === 'GRP') {
        toggleGroupSelection();
    } else {
        self.setTimeout("edit_select_all();", 100);
    }

    $('.previewTips, .DCESTips, .fieldTips', p4.edit.editBox).tooltip({
        fixable: true,
        fixableIndex: 1200
    });
    $('.infoTips', p4.edit.editBox).tooltip();

    if (p4.edit.what == 'GRP') {
        $('#EDIT_FILM2 .reg_opts').show();

        $.each($('#EDIT_FILM2 .contextMenuTrigger'), function () {

            var id = $(this).attr('id').split('_').slice(1, 3).join('_');
            $(this).contextMenu('#editContext_' + id + '', {
                appendTo: '#editcontextwrap',
                openEvt: 'click',
                dropDown: true,
                theme: 'vista',
                dropDown: true,
                showTransition: 'slideDown',
                hideTransition: 'hide',
                shadow: false
            });
        });
    }

    hsplit1();
    vsplit2()
    vsplit1();

    $('#EDIT_TOP', p4.edit.editBox).resizable({
        handles: 's',
        minHeight: 100,
        resize: function () {
            hsplit1();
            setPreviewEdit();
        },
        stop: function () {
            hsplit1();
            setPref('editing_top_box', Math.floor($('#EDIT_TOP').height() * 100 / $('#EDIT_ALL').height()));
            setSizeLimits();
        }
    });

    $('#divS_wrapper', p4.edit.editBox).resizable({
        handles: 'e',
        minWidth: 200,
        resize: function () {
            vsplit1();
            setPreviewEdit();
        },
        stop: function () {
            setPref('editing_right_box', Math.floor($('#divS').width() * 100 / $('#EDIT_MID_L').width()));
            vsplit1();
            setSizeLimits();
        }
    });

    $('#EDIT_MID_R')
        .css('left', $('#EDIT_MID_L').position().left + $('#EDIT_MID_L').width() + 15)
        .resizable({
            handles: 'w',
            minWidth: 200,
            resize: function () {
                vsplit2();
                setPreviewEdit();
            },
            stop: function () {
                setPref('editing_left_box', Math.floor(($('#EDIT_MID_R').width()) * 100 / $('#EDIT_MID').width()));
                vsplit2();
                setSizeLimits();
            }
        });

    $('#EDIT_ZOOMSLIDER', p4.edit.editBox).slider({
        min: 60,
        max: 300,
        value: p4.edit.diapoSize,
        slide: function (event, ui) {
            var v = $(ui.value)[0];
            $('#EDIT_FILM2 .diapo', p4.edit.editBox).width(v).height(v);
        },
        change: function (event, ui) {
            p4.edit.diapoSize = $(ui.value)[0];
            setPref("editing_images_size", p4.edit.diapoSize);
        }
    });

    var buttons = {};
    buttons[language.valider] = function (e) {
        $(this).dialog("close");
        edit_applyMultiDesc(e);
    };
    buttons[language.annuler] = function (e) {
        $(this).dialog("close");
        edit_cancelMultiDesc(e);
    };

    $("#EDIT_CLOSEDIALOG", p4.edit.editBox).dialog({
        autoOpen: false,
        closeOnEscape: true,
        resizable: false,
        draggable: false,
        modal: true,
        buttons: buttons
    });

    var buttons = {};

    buttons[language.valider] = function () {
        var form = $("#Edit_copyPreset_dlg FORM");
        var jtitle = $(".EDIT_presetTitle", form);
        if (jtitle.val() == '') {
            alert(language.needTitle);
            jtitle[0].focus();
            return;
        }

        var fields = [];
        $(":checkbox", form).each(function (idx, elem) {
            var $el = $(elem);
            if ($el.is(":checked")) {
                var val = $el.val();
                var field = {
                    name: p4.edit.T_fields[val].name,
                    value: []
                };
                var tval;
                if (p4.edit.T_fields[val].multi) {
                    field.value = $.map(
                        p4.edit.T_fields[val]._value.split(";"),
                        function(obj, idx){
                            return obj.trim();
                        }
                    );
                } else {
                    field.value = [p4.edit.T_fields[val]._value.trim()];
                }
                fields.push(field);
            }
        });

        $.ajax({
            type: 'POST',
            url: "../prod/records/edit/presets",
            data: {
                sbas_id: p4.edit.sbas_id,
                title: jtitle.val(),
                fields: fields
            },
            dataType: 'json',
            success: function (data, textStatus) {
                preset_paint(data);

                if ($("#Edit_copyPreset_dlg").data("ui-dialog")) {
                    $("#Edit_copyPreset_dlg").dialog("close");
                }
            }
        });
    };

    buttons[language.annuler] = function () {
        $(this).dialog("close");

    };

    $("#Edit_copyPreset_dlg", p4.edit.editBox).dialog({
        stack: true,
        closeOnEscape: true,
        resizable: false,
        draggable: false,
        autoOpen: false,
        modal: true,
        width: 600,
        title: language.newPreset,
        close: function (event, ui) {
            $(this).dialog("widget").css("z-index", "auto");
        },
        open: function (event, ui) {
            $(this).dialog("widget").css("z-index", "5000");
            $(".EDIT_presetTitle")[0].focus();
        },
        buttons: buttons
    });

    $('#idEditDateZone', p4.edit.editBox).datepicker({
        changeYear: true,
        changeMonth: true,
        dateFormat: 'yy/mm/dd',
        onSelect: function (dateText, inst) {
            var lval = $('#idEditZTextArea', p4.edit.editBox).val();
            if (lval != dateText) {
                fieldLastValue = lval;
                $('#idEditZTextArea', p4.edit.editBox).val(dateText);
                $('#idEditZTextArea').trigger('keyup.maxLength');
                textareaIsDirty = true;
                edit_validField(null, 'ok');
            }
        }
    });

    ETHSeeker = new EditThesaurusSeeker(p4.edit.sbas_id);

    setSizeLimits();

    $.ajax({
        type: 'GET',
        url: "../prod/records/edit/presets",
        data: {
            sbas_id: p4.edit.sbas_id,
        },
        dataType: 'json',
        success: function (data, textStatus) {
            preset_paint(data);
        }
    });

    check_required();

    $('#TH_Opresets button.adder').bind('click', function () {
        preset_copy();
    });

    try {
        $('#divS .edit_field:first').trigger('mousedown');
    }
    catch (err) {

    }
}


function setRegDefault(n, record_id) {
    p4.edit.newrepresent = record_id;

    var src = $('#idEditDiapo_' + n).find('img.edit_IMGT').attr('src');
    var style = $('#idEditDiapo_' + n).find('img.edit_IMGT').attr('style');

    $('#EDIT_GRPDIAPO .edit_IMGT').attr('src', src).attr('style', style);
}


;/*
 * jQuery Color Animations
 * Copyright 2007 John Resig
 * Released under the MIT and GPL licenses.
 */

(function(jQuery){

	// We override the animation for all of these color styles
	jQuery.each(['backgroundColor', 'borderBottomColor', 'borderLeftColor', 'borderRightColor', 'borderTopColor', 'color', 'outlineColor'], function(i,attr){
		jQuery.fx.step[attr] = function(fx){
			if ( fx.state == 0 ) {
				fx.start = getColor( fx.elem, attr );
				fx.end = getRGB( fx.end );
			}

			fx.elem.style[attr] = "rgb(" + [
				Math.max(Math.min( parseInt((fx.pos * (fx.end[0] - fx.start[0])) + fx.start[0]), 255), 0),
				Math.max(Math.min( parseInt((fx.pos * (fx.end[1] - fx.start[1])) + fx.start[1]), 255), 0),
				Math.max(Math.min( parseInt((fx.pos * (fx.end[2] - fx.start[2])) + fx.start[2]), 255), 0)
			].join(",") + ")";
		};
	});

	// Color Conversion functions from highlightFade
	// By Blair Mitchelmore
	// http://jquery.offput.ca/highlightFade/

	// Parse strings looking for color tuples [255,255,255]
	function getRGB(color) {
		var result;

		// Check if we're already dealing with an array of colors
		if ( color && color.constructor == Array && color.length == 3 )
			return color;

		// Look for rgb(num,num,num)
		if (result = /rgb\(\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*\)/.exec(color))
			return [parseInt(result[1]), parseInt(result[2]), parseInt(result[3])];

		// Look for rgb(num%,num%,num%)
		if (result = /rgb\(\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\%\s*\)/.exec(color))
			return [parseFloat(result[1])*2.55, parseFloat(result[2])*2.55, parseFloat(result[3])*2.55];

		// Look for #a0b1c2
		if (result = /#([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})/.exec(color))
			return [parseInt(result[1],16), parseInt(result[2],16), parseInt(result[3],16)];

		// Look for #fff
		if (result = /#([a-fA-F0-9])([a-fA-F0-9])([a-fA-F0-9])/.exec(color))
			return [parseInt(result[1]+result[1],16), parseInt(result[2]+result[2],16), parseInt(result[3]+result[3],16)];

		// Otherwise, we're most likely dealing with a named color
		return colors[jQuery.trim(color).toLowerCase()];
	}
	
	function getColor(elem, attr) {
		var color;

		do {
			color = jQuery.curCSS(elem, attr);

			// Keep going until we find an element that has color, or we hit the body
			if ( color != '' && color != 'transparent' || jQuery.nodeName(elem, "body") )
				break; 

			attr = "backgroundColor";
		} while ( elem = elem.parentNode );

		return getRGB(color);
	};
	
	// Some named colors to work with
	// From Interface by Stefan Petre
	// http://interface.eyecon.ro/

	var colors = {
		aqua:[0,255,255],
		azure:[240,255,255],
		beige:[245,245,220],
		black:[0,0,0],
		blue:[0,0,255],
		brown:[165,42,42],
		cyan:[0,255,255],
		darkblue:[0,0,139],
		darkcyan:[0,139,139],
		darkgrey:[169,169,169],
		darkgreen:[0,100,0],
		darkkhaki:[189,183,107],
		darkmagenta:[139,0,139],
		darkolivegreen:[85,107,47],
		darkorange:[255,140,0],
		darkorchid:[153,50,204],
		darkred:[139,0,0],
		darksalmon:[233,150,122],
		darkviolet:[148,0,211],
		fuchsia:[255,0,255],
		gold:[255,215,0],
		green:[0,128,0],
		indigo:[75,0,130],
		khaki:[240,230,140],
		lightblue:[173,216,230],
		lightcyan:[224,255,255],
		lightgreen:[144,238,144],
		lightgrey:[211,211,211],
		lightpink:[255,182,193],
		lightyellow:[255,255,224],
		lime:[0,255,0],
		magenta:[255,0,255],
		maroon:[128,0,0],
		navy:[0,0,128],
		olive:[128,128,0],
		orange:[255,165,0],
		pink:[255,192,203],
		purple:[128,0,128],
		violet:[128,0,128],
		red:[255,0,0],
		silver:[192,192,192],
		white:[255,255,255],
		yellow:[255,255,0]
	};
	
})(jQuery);

;(function ($) {

    var methods = {
        init: function (options) {
            var settings = {
                'zoomable': false,
                'display_full_screen': false
            };
            return this.each(function () {

                var $this = $(this), data = $(this).data('image_enhance');

                if (!data) {
                    if (options) {
                        $.extend(settings, options);
                    }

                    var wrapper = $('.thumb_wrapper', $(this));
                    var $image = $('img', $this);

                    wrapper.css('position', 'relative');

                    reset_position($this);

                    if (settings.display_full_screen) {
                        $image.parent()
                            .append('<div class="image_enhance_titlebar" style="display:none;">\n\
                <div class="image_enhance_title_options"><span class="full"><img src="/assets/common/images/icons/fullscreen.gif" /></span></div>\n\
                <div class="image_enhance_title_bg"></div></div>');

                        var $titlebar = $('.image_enhance_titlebar', $this);

                        $('.image_enhance_title_bg', $titlebar).css('opacity', 0.5);

                        $image.parent()
                            .bind('mouseover.image_enhance', function () {
                                $titlebar.stop().show().animate({
                                    'height': 28
                                }, 150);
                            })
                            .bind('mouseout.image_enhance', function () {
                                $titlebar.stop().animate({
                                    'height': 0
                                }, 150, function () {
                                    $titlebar.hide()
                                });
                            });

                        $('.image_enhance_titlebar .full', wrapper).bind('click.image_enhance', function () {
                            $('body').append('<div class="image_enhance_theatre">\n\
                \n\
                <div class="image_enhance_theatre_closer_wrapper"><span class="closer">close</span></div>\n\
                <img style="width:' + image_width + 'px;height:' + image_height + '" src="' + $image.attr('src') + '"/>\n\
                </div>');

                            var $theatre = $('.image_enhance_theatre');
                            var $theatre_img = $('img', $theatre);
                            $(window).bind('resize.image_enhance dblclick.image_enhance', function (event) {

                                if (event.type == 'dblclick') {
                                    $theatre_img.removeClass('zoomed');
                                }
                                else {
                                    if ($theatre_img.hasClass('zoomed'))
                                        return;
                                }
                                var datas = calculate_sizes($(this).width(), $(this).height(), image_width, image_height, 80);

                                $theatre_img.width(datas.width).height(datas.height).css('top', datas.top).css('left', datas.left);
                            })
                            $(window).trigger('resize.image_enhance');
                            $('.closer', $theatre).bind('click.image_enhance', function () {
                                $theatre.remove();
                            });

                            if (typeof $theatre.disableSelection !== 'function' && window.console)
                                console.error('enhanced image require jquery UI\'s disableSelection');
                            $('img', $theatre).disableSelection();
                        });
                    }


                    if (settings.zoomable) {
                        if (typeof $image.draggable !== 'function' && window.console)
                            console.error('zoomable require jquery UI\'s draggable');

                        if ($image.attr('ondragstart')) {
                            $image.removeAttr('ondragstart');
                        }
                        $image.draggable();
                        $image.css({
                            'max-width': 'none',
                            'max-height': 'none'
                        });

                        var image_width = parseInt($('input[name="width"]', $this).val());
                        var image_height = parseInt($('input[name="height"]', $this).val());
                        var ratio = image_width / image_height;

                        $this.bind('mousewheel',function (event, delta) {
                            $image.addClass('zoomed');
                            if (delta > 0) {
                                event.stopPropagation();
                                zoomPreview(true, ratio, $image, $(this));
                            }
                            else {
                                event.stopPropagation();
                                zoomPreview(false, ratio, $image, $(this));
                            }
                            return false;
                        }).bind('dblclick', function (event) {
                                reset_position($this);
                            });
                    }

                    $(this).data('image_enhance', {
                        width: image_width,
                        height: image_height
                    });
                }

            });
        },
        destroy: function () {
            return this.each(function () {
                $(this).data('image_enhance', null);
                $('.image_enhance_titlebar, .image_enhance_theatre', this).remove();
            });
        }
    };

    function zoomPreview(bool, ratio, $img, $container) {
        if ($img.length === 0)
            return;

        var t1 = parseInt($img.css('top'));
        var l1 = parseInt($img.css('left'));
        var w1 = $img.width();
        var h1 = $img.height();

        var w2, t2;

        if (bool) {
            if ((w1 * 1.08) < 32767) {
                w2 = w1 * 1.08;
            } else {
                w2 = w1;
            }
        }
        else {
            if ((w1 / 1.08) > 20) {
                w2 = w1 / 1.08;
            } else {
                w2 = w1;
            }
        }

        var datas = $(this).data('image_enhance');

        h2 = Math.round(w2 / ratio);
        w2 = Math.round(w2);

        t2 = Math.round(t1 - (h2 - h1) / 2) + 'px';
        var l2 = Math.round(l1 - (w2 - w1) / 2) + 'px';

        var wPreview = $container.width() / 2;
        var hPreview = $container.height() / 2;

        var nt = Math.round((h2 / h1) * (t1 - hPreview) + hPreview);
        var nl = Math.round(((w2 / w1) * (l1 - wPreview)) + wPreview);

        $img.css({
            left: nl,
            top: nt
        }).width(w2).height(h2);
    }

    function calculate_sizes(window_width, window_height, image_width, image_height, border) {
        if (typeof border !== 'number')
            border = 0;

        var width, height;
        var ratio_display = window_width / window_height;
        var ratio_image = image_width / image_height;

        if (ratio_image > ratio_display) {
            width = window_width - border;
            height = Math.round(width / ratio_image);
        }
        else {
            height = window_height - border;
            width = Math.round(height * ratio_image);
        }

        var top = Math.round((window_height - height) / 2);
        var left = Math.round((window_width - width ) / 2);

        return {
            top: top,
            left: left,
            width: width,
            height: height
        };
    }

    function reset_position($this) {
        var display_width = $this.width();
        var display_height = $this.height();
        var image_width = parseInt($('input[name="width"]', $this).val());
        var image_height = parseInt($('input[name="height"]', $this).val());

        var datas = calculate_sizes(display_width, display_height, image_width, image_height);
        var $image = $('img', $this);

        var top = Math.round((display_height - datas.height) / 2) + 'px';
        var left = Math.round((display_width - datas.width) / 2) + 'px';

        $image.width(datas.width).height(datas.height).css({top: top, left: left});
        return;
    }

    $.fn.image_enhance = function (method) {

        if (methods[method]) {
            return methods[method].apply(this, Array.prototype.slice.call(arguments, 1));
        } else if (typeof method === 'object' || !method) {
            return methods.init.apply(this, arguments);
        } else {
            $.error('Method ' + method + ' does not exist on jQuery.image_enhance');
        }


    };
})(jQuery);
;/*
 * Treeview 1.4.2 - jQuery plugin to hide and show branches of a tree
 *
 * http://bassistance.de/jquery-plugins/jquery-plugin-treeview/
 *
 * Copyright Jörn Zaefferer
 * Released under the MIT license:
 *   http://www.opensource.org/licenses/mit-license.php
 */

;(function($) {

	// TODO rewrite as a widget, removing all the extra plugins
	$.extend($.fn, {
		swapClass: function(c1, c2) {
			var c1Elements = this.filter('.' + c1);
			this.filter('.' + c2).removeClass(c2).addClass(c1);
			c1Elements.removeClass(c1).addClass(c2);
			return this;
		},
		replaceClass: function(c1, c2) {
			return this.filter('.' + c1).removeClass(c1).addClass(c2).end();
		},
		hoverClass: function(className) {
			className = className || "hover";
			return this.hover(function() {
				$(this).addClass(className);
			}, function() {
				$(this).removeClass(className);
			});
		},
		heightToggle: function(animated, callback) {
			animated ?
				this.animate({ height: "toggle" }, animated, callback) :
				this.each(function(){
					jQuery(this)[ jQuery(this).is(":hidden") ? "show" : "hide" ]();
					if(callback)
						callback.apply(this, arguments);
				});
		},
		heightHide: function(animated, callback) {
			if (animated) {
				this.animate({ height: "hide" }, animated, callback);
			} else {
				this.hide();
				if (callback)
					this.each(callback);
			}
		},
		prepareBranches: function(settings) {
			if (!settings.prerendered) {
				// mark last tree items
				this.filter(":last-child:not(ul)").addClass(CLASSES.last);
				// collapse whole tree, or only those marked as closed, anyway except those marked as open
				this.filter((settings.collapsed ? "" : "." + CLASSES.closed) + ":not(." + CLASSES.open + ")").find(">ul").hide();
			}
			// return all items with sublists
			return this.filter(":has(>ul)");
		},
		applyClasses: function(settings, toggler) {
			// TODO use event delegation
			this.filter(":has(>ul):not(:has(>a))").find(">span").unbind("click.treeview").bind("click.treeview", function(event) {
				// don't handle click events on children, eg. checkboxes
				if ( this == event.target )
					toggler.apply($(this).next());
			}).add( $("a", this) ).hoverClass();

			if (!settings.prerendered) {
				// handle closed ones first
				this.filter(":has(>ul:hidden)")
						.addClass(CLASSES.expandable)
						.replaceClass(CLASSES.last, CLASSES.lastExpandable);

				// handle open ones
				this.not(":has(>ul:hidden)")
						.addClass(CLASSES.collapsable)
						.replaceClass(CLASSES.last, CLASSES.lastCollapsable);

	            // create hitarea if not present
				var hitarea = this.find("div." + CLASSES.hitarea);
				if (!hitarea.length)
					hitarea = this.prepend("<div class=\"" + CLASSES.hitarea + "\"/>").find("div." + CLASSES.hitarea);
				hitarea.removeClass().addClass(CLASSES.hitarea).each(function() {
					var classes = "";
					$.each($(this).parent().attr("class").split(" "), function() {
						classes += this + "-hitarea ";
					});
					$(this).addClass( classes );
				})
			}

			// apply event to hitarea
			this.find("div." + CLASSES.hitarea).click( toggler );
		},
		treeview: function(settings) {

			settings = $.extend({
				cookieId: "treeview"
			}, settings);

			if ( settings.toggle ) {
				var callback = settings.toggle;
				settings.toggle = function() {
					return callback.apply($(this).parent()[0], arguments);
				};
			}

			// factory for treecontroller
			function treeController(tree, control) {
				// factory for click handlers
				function handler(filter) {
					return function() {
						// reuse toggle event handler, applying the elements to toggle
						// start searching for all hitareas
						toggler.apply( $("div." + CLASSES.hitarea, tree).filter(function() {
							// for plain toggle, no filter is provided, otherwise we need to check the parent element
							return filter ? $(this).parent("." + filter).length : true;
						}) );
						return false;
					};
				}
				// click on first element to collapse tree
				$("a:eq(0)", control).click( handler(CLASSES.collapsable) );
				// click on second to expand tree
				$("a:eq(1)", control).click( handler(CLASSES.expandable) );
				// click on third to toggle tree
				$("a:eq(2)", control).click( handler() );
			}

			// handle toggle event
			function toggler() {
				$(this)
					.parent()
					// swap classes for hitarea
					.find(">.hitarea")
						.swapClass( CLASSES.collapsableHitarea, CLASSES.expandableHitarea )
						.swapClass( CLASSES.lastCollapsableHitarea, CLASSES.lastExpandableHitarea )
					.end()
					// swap classes for parent li
					.swapClass( CLASSES.collapsable, CLASSES.expandable )
					.swapClass( CLASSES.lastCollapsable, CLASSES.lastExpandable )
					// find child lists
					.find( ">ul" )
					// toggle them
					.heightToggle( settings.animated, settings.toggle );
				if ( settings.unique ) {
					$(this).parent()
						.siblings()
						// swap classes for hitarea
						.find(">.hitarea")
							.replaceClass( CLASSES.collapsableHitarea, CLASSES.expandableHitarea )
							.replaceClass( CLASSES.lastCollapsableHitarea, CLASSES.lastExpandableHitarea )
						.end()
						.replaceClass( CLASSES.collapsable, CLASSES.expandable )
						.replaceClass( CLASSES.lastCollapsable, CLASSES.lastExpandable )
						.find( ">ul" )
						.heightHide( settings.animated, settings.toggle );
				}
			}
			this.data("toggler", toggler);

			function serialize() {
				function binary(arg) {
					return arg ? 1 : 0;
				}
				var data = [];
				branches.each(function(i, e) {
					data[i] = $(e).is(":has(>ul:visible)") ? 1 : 0;
				});
				$.cookie(settings.cookieId, data.join(""), settings.cookieOptions );
			}

			function deserialize() {
				var stored = $.cookie(settings.cookieId);
				if ( stored ) {
					var data = stored.split("");
					branches.each(function(i, e) {
						$(e).find(">ul")[ parseInt(data[i]) ? "show" : "hide" ]();
					});
				}
			}

			// add treeview class to activate styles
			this.addClass("treeview");

			// prepare branches and find all tree items with child lists
			var branches = this.find("li").prepareBranches(settings);

			switch(settings.persist) {
			case "cookie":
				var toggleCallback = settings.toggle;
				settings.toggle = function() {
					serialize();
					if (toggleCallback) {
						toggleCallback.apply(this, arguments);
					}
				};
				deserialize();
				break;
			case "location":
				var current = this.find("a").filter(function() {
					return location.href.toLowerCase().indexOf(this.href.toLowerCase()) == 0;
				});
				if ( current.length ) {
					// TODO update the open/closed classes
					var items = current.addClass("selected").parents("ul, li").add( current.next() ).show();
					if (settings.prerendered) {
						// if prerendered is on, replicate the basic class swapping
						items.filter("li")
							.swapClass( CLASSES.collapsable, CLASSES.expandable )
							.swapClass( CLASSES.lastCollapsable, CLASSES.lastExpandable )
							.find(">.hitarea")
								.swapClass( CLASSES.collapsableHitarea, CLASSES.expandableHitarea )
								.swapClass( CLASSES.lastCollapsableHitarea, CLASSES.lastExpandableHitarea );
					}
				}
				break;
			}

			branches.applyClasses(settings, toggler);

			// if control option is set, create the treecontroller and show it
			if ( settings.control ) {
				treeController(this, settings.control);
				$(settings.control).show();
			}

			return this;
		}
	});

	// classes used by the plugin
	// need to be styled via external stylesheet, see first example
	$.treeview = {};
	var CLASSES = ($.treeview.classes = {
		open: "open",
		closed: "closed",
		expandable: "expandable",
		expandableHitarea: "expandable-hitarea",
		lastExpandableHitarea: "lastExpandable-hitarea",
		collapsable: "collapsable",
		collapsableHitarea: "collapsable-hitarea",
		lastCollapsableHitarea: "lastCollapsable-hitarea",
		lastCollapsable: "lastCollapsable",
		lastExpandable: "lastExpandable",
		last: "last",
		hitarea: "hitarea"
	});

})(jQuery);
;/*
 * Async Treeview 0.1 - Lazy-loading extension for Treeview
 *
 * http://bassistance.de/jquery-plugins/jquery-plugin-treeview/
 *
 * Copyright Jörn Zaefferer
 * Released under the MIT license:
 *   http://www.opensource.org/licenses/mit-license.php
 */

;(function($) {

function load(settings, root, child, container) {
	function createNode(parent) {
		var current = $("<li/>").attr("id", this.id || "").html("<span>" + this.text + "</span>").appendTo(parent);
		if (this.classes) {
			current.children("span").addClass(this.classes);
		}
		if (this.expanded) {
			current.addClass("open");
		}
		if (this.hasChildren || this.children && this.children.length) {
			var branch = $("<ul/>").appendTo(current);
			if (this.hasChildren) {
				current.addClass("hasChildren");
				createNode.call({
					classes: "placeholder",
					text: "&nbsp;",
					children:[]
				}, branch);
			}
			if (this.children && this.children.length) {
				$.each(this.children, createNode, [branch])
			}
		}
	}
	$.ajax($.extend(true, {
		url: settings.url,
		dataType: "json",
		data: {
			root: root
		},
		success: function(response) {
			child.empty();
			$.each(response, createNode, [child]);
	        $(container).treeview({add: child});
	    }
	}, settings.ajax));
	/*
	$.getJSON(settings.url, {root: root}, function(response) {
		function createNode(parent) {
			var current = $("<li/>").attr("id", this.id || "").html("<span>" + this.text + "</span>").appendTo(parent);
			if (this.classes) {
				current.children("span").addClass(this.classes);
			}
			if (this.expanded) {
				current.addClass("open");
			}
			if (this.hasChildren || this.children && this.children.length) {
				var branch = $("<ul/>").appendTo(current);
				if (this.hasChildren) {
					current.addClass("hasChildren");
					createNode.call({
						classes: "placeholder",
						text: "&nbsp;",
						children:[]
					}, branch);
				}
				if (this.children && this.children.length) {
					$.each(this.children, createNode, [branch])
				}
			}
		}
		child.empty();
		$.each(response, createNode, [child]);
        $(container).treeview({add: child});
    });
    */
}

var proxied = $.fn.treeview;
$.fn.treeview = function(settings) {
	if (!settings.url) {
		return proxied.apply(this, arguments);
	}
	if (!settings.root) {
		settings.root = "source";
	}
	var container = this;
	if (!container.children().size())
		load(settings, settings.root, this, container);
	var userToggle = settings.toggle;
	return proxied.call(this, $.extend({}, settings, {
		collapsed: true,
		toggle: function() {
			var $this = $(this);
			if ($this.hasClass("hasChildren")) {
				var childList = $this.removeClass("hasChildren").find("ul");
				load(settings, this.id, childList, container);
			}
			if (userToggle) {
				userToggle.apply(this, arguments);
			}
		}
	}));
};

})(jQuery);
;/*! jQuery Fancytree Plugin - 2.7.0 - 2014-12-21T15:57
  * https://github.com/mar10/fancytree
  * Copyright (c) 2014 Martin Wendt; Licensed MIT */
(function( factory ) {
	if ( typeof define === "function" && define.amd ) {
		define( [ "jquery" ], factory );
	} else {
		factory( jQuery );
	}
}(function( $ ) {

!function(a,b,c,d){"use strict";function e(b,c){b||(c=c?": "+c:"",a.error("Fancytree assertion failed"+c))}function f(a,c){var d,e,f=b.console?b.console[a]:null;if(f)try{f.apply(b.console,c)}catch(g){for(e="",d=0;d<c.length;d++)e+=c[d];f(e)}}function g(a){return!(!a.tree||a.statusNodeType===d)}function h(b){var c,d,e,f=a.map(a.trim(b).split("."),function(a){return parseInt(a,10)}),g=a.map(Array.prototype.slice.call(arguments,1),function(a){return parseInt(a,10)});for(c=0;c<g.length;c++)if(d=f[c]||0,e=g[c]||0,d!==e)return d>e;return!0}function i(a,b,c,d,e){var f=function(){var c=b[a],f=d[a],g=b.ext[e],h=function(){return c.apply(b,arguments)};return function(){var a=b._local,c=b._super;try{return b._local=g,b._super=h,f.apply(b,arguments)}finally{b._local=a,b._super=c}}}();return f}function j(b,c,d,e){for(var f in d)"function"==typeof d[f]?"function"==typeof b[f]?b[f]=i(f,b,c,d,e):"_"===f.charAt(0)?b.ext[e][f]=i(f,b,c,d,e):a.error("Could not override tree."+f+". Use prefix '_' to create tree."+e+"._"+f):"options"!==f&&(b.ext[e][f]=d[f])}function k(b,c){return b===d?a.Deferred(function(){this.resolve()}).promise():a.Deferred(function(){this.resolveWith(b,c)}).promise()}function l(b,c){return b===d?a.Deferred(function(){this.reject()}).promise():a.Deferred(function(){this.rejectWith(b,c)}).promise()}function m(a,b){return function(){a.resolveWith(b)}}function n(b){var c=a.extend({},b.data()),d=c.json;return delete c.fancytree,d&&(delete c.json,c=a.extend(c,d)),c}function o(a){return a=a.toLowerCase(),function(b){return b.title.toLowerCase().indexOf(a)>=0}}function p(a){var b=new RegExp("^"+a,"i");return function(a){return b.test(a.title)}}function q(b,c){var d,f,g,h;for(this.parent=b,this.tree=b.tree,this.ul=null,this.li=null,this.statusNodeType=null,this._isLoading=!1,this._error=null,this.data={},d=0,f=z.length;f>d;d++)g=z[d],this[g]=c[g];c.data&&a.extend(this.data,c.data);for(g in c)A[g]||a.isFunction(c[g])||B[g]||(this.data[g]=c[g]);null==this.key?this.tree.options.defaultKey?(this.key=this.tree.options.defaultKey(this),e(this.key,"defaultKey() must return a unique key")):this.key="_"+t._nextNodeKey++:this.key=""+this.key,c.active&&(e(null===this.tree.activeNode,"only one active node allowed"),this.tree.activeNode=this),c.selected&&(this.tree.lastSelectedNode=this),this.children=null,h=c.children,h&&h.length&&this._setChildren(h),this.tree._callHook("treeRegisterNode",this.tree,!0,this)}function r(b){this.widget=b,this.$div=b.element,this.options=b.options,this.options&&(a.isFunction(this.options.lazyload)&&!a.isFunction(this.options.lazyLoad)&&(this.options.lazyLoad=function(){return t.warn("The 'lazyload' event is deprecated since 2014-02-25. Use 'lazyLoad' (with uppercase L) instead."),b.options.lazyload.apply(this,arguments)}),a.isFunction(this.options.loaderror)&&a.error("The 'loaderror' event was renamed since 2014-07-03. Use 'loadError' (with uppercase E) instead."),this.options.fx!==d&&t.warn("The 'fx' options was replaced by 'toggleEffect' since 2014-11-30.")),this.ext={},this.data=n(this.$div),this._id=a.ui.fancytree._nextId++,this._ns=".fancytree-"+this._id,this.activeNode=null,this.focusNode=null,this._hasFocus=null,this.lastSelectedNode=null,this.systemFocusElement=null,this.lastQuicksearchTerm="",this.lastQuicksearchTime=0,this.statusClassPropName="span",this.ariaPropName="li",this.nodeContainerAttrName="li",this.$div.find(">ul.fancytree-container").remove();var c,e={tree:this};this.rootNode=new q(e,{title:"root",key:"root_"+this._id,children:null,expanded:!0}),this.rootNode.parent=null,c=a("<ul>",{"class":"ui-fancytree fancytree-container"}).appendTo(this.$div),this.$container=c,this.rootNode.ul=c[0],null==this.options.debugLevel&&(this.options.debugLevel=t.debugLevel),this.$container.attr("tabindex",this.options.tabbable?"0":"-1"),this.options.aria&&this.$container.attr("role","tree").attr("aria-multiselectable",!0)}if(a.ui&&a.ui.fancytree)return void a.ui.fancytree.warn("Fancytree: ignored duplicate include");e(a.ui,"Fancytree requires jQuery UI (http://jqueryui.com)");var s,t=null,u={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;","/":"&#x2F;"},v={16:!0,17:!0,18:!0},w={8:"backspace",9:"tab",10:"return",13:"return",19:"pause",20:"capslock",27:"esc",32:"space",33:"pageup",34:"pagedown",35:"end",36:"home",37:"left",38:"up",39:"right",40:"down",45:"insert",46:"del",59:";",61:"=",96:"0",97:"1",98:"2",99:"3",100:"4",101:"5",102:"6",103:"7",104:"8",105:"9",106:"*",107:"+",109:"-",110:".",111:"/",112:"f1",113:"f2",114:"f3",115:"f4",116:"f5",117:"f6",118:"f7",119:"f8",120:"f9",121:"f10",122:"f11",123:"f12",144:"numlock",145:"scroll",173:"-",186:";",187:"=",188:",",189:"-",190:".",191:"/",192:"`",219:"[",220:"\\",221:"]",222:"'"},x="active expanded focus folder hideCheckbox lazy selected unselectable".split(" "),y={},z="expanded extraClasses folder hideCheckbox key lazy refKey selected title tooltip unselectable".split(" "),A={},B={active:!0,children:!0,data:!0,focus:!0};for(s=0;s<x.length;s++)y[x[s]]=!0;for(s=0;s<z.length;s++)A[z[s]]=!0;q.prototype={_findDirectChild:function(a){var b,c,d=this.children;if(d)if("string"==typeof a){for(b=0,c=d.length;c>b;b++)if(d[b].key===a)return d[b]}else{if("number"==typeof a)return this.children[a];if(a.parent===this)return a}return null},_setChildren:function(a){e(a&&(!this.children||0===this.children.length),"only init supported"),this.children=[];for(var b=0,c=a.length;c>b;b++)this.children.push(new q(this,a[b]))},addChildren:function(b,c){var d,f,g,h=null,i=[];for(a.isPlainObject(b)&&(b=[b]),this.children||(this.children=[]),d=0,f=b.length;f>d;d++)i.push(new q(this,b[d]));return h=i[0],null==c?this.children=this.children.concat(i):(c=this._findDirectChild(c),g=a.inArray(c,this.children),e(g>=0,"insertBefore must be an existing child"),this.children.splice.apply(this.children,[g,0].concat(i))),(!this.parent||this.parent.ul||this.tr)&&this.render(),3===this.tree.options.selectMode&&this.fixSelection3FromEndNodes(),h},addNode:function(a,b){switch((b===d||"over"===b)&&(b="child"),b){case"after":return this.getParent().addChildren(a,this.getNextSibling());case"before":return this.getParent().addChildren(a,this);case"firstChild":var c=this.children?this.children[0]:null;return this.addChildren(a,c);case"child":case"over":return this.addChildren(a)}e(!1,"Invalid mode: "+b)},appendSibling:function(a){return this.addNode(a,"after")},applyPatch:function(b){if(null===b)return this.remove(),k(this);var c,d,e,f={children:!0,expanded:!0,parent:!0};for(c in b)e=b[c],f[c]||a.isFunction(e)||(A[c]?this[c]=e:this.data[c]=e);return b.hasOwnProperty("children")&&(this.removeChildren(),b.children&&this._setChildren(b.children)),this.isVisible()&&(this.renderTitle(),this.renderStatus()),d=b.hasOwnProperty("expanded")?this.setExpanded(b.expanded):k(this)},collapseSiblings:function(){return this.tree._callHook("nodeCollapseSiblings",this)},copyTo:function(a,b,c){return a.addNode(this.toDict(!0,c),b)},countChildren:function(a){var b,c,d,e=this.children;if(!e)return 0;if(d=e.length,a!==!1)for(b=0,c=d;c>b;b++)d+=e[b].countChildren();return d},debug:function(){this.tree.options.debugLevel>=2&&(Array.prototype.unshift.call(arguments,this.toString()),f("log",arguments))},discard:function(){return this.warn("FancytreeNode.discard() is deprecated since 2014-02-16. Use .resetLazy() instead."),this.resetLazy()},findAll:function(b){b=a.isFunction(b)?b:o(b);var c=[];return this.visit(function(a){b(a)&&c.push(a)}),c},findFirst:function(b){b=a.isFunction(b)?b:o(b);var c=null;return this.visit(function(a){return b(a)?(c=a,!1):void 0}),c},_changeSelectStatusAttrs:function(a){var b=!1;switch(a){case!1:b=this.selected||this.partsel,this.selected=!1,this.partsel=!1;break;case!0:b=!this.selected||!this.partsel,this.selected=!0,this.partsel=!0;break;case d:b=this.selected||!this.partsel,this.selected=!1,this.partsel=!0;break;default:e(!1,"invalid state: "+a)}return b&&this.renderStatus(),b},fixSelection3AfterClick:function(){var a=this.isSelected();this.visit(function(b){b._changeSelectStatusAttrs(a)}),this.fixSelection3FromEndNodes()},fixSelection3FromEndNodes:function(){function a(b){var c,e,f,g,h,i,j,k=b.children;if(k&&k.length){for(i=!0,j=!1,c=0,e=k.length;e>c;c++)f=k[c],g=a(f),g!==!1&&(j=!0),g!==!0&&(i=!1);h=i?!0:j?d:!1}else h=!!b.selected;return b._changeSelectStatusAttrs(h),h}e(3===this.tree.options.selectMode,"expected selectMode 3"),a(this),this.visitParents(function(a){var b,c,e,f,g=a.children,h=!0,i=!1;for(b=0,c=g.length;c>b;b++)e=g[b],(e.selected||e.partsel)&&(i=!0),e.unselectable||e.selected||(h=!1);f=h?!0:i?d:!1,a._changeSelectStatusAttrs(f)})},fromDict:function(b){for(var c in b)A[c]?this[c]=b[c]:"data"===c?a.extend(this.data,b.data):a.isFunction(b[c])||B[c]||(this.data[c]=b[c]);b.children&&(this.removeChildren(),this.addChildren(b.children)),this.renderTitle()},getChildren:function(){return this.hasChildren()===d?d:this.children},getFirstChild:function(){return this.children?this.children[0]:null},getIndex:function(){return a.inArray(this,this.parent.children)},getIndexHier:function(b){b=b||".";var c=[];return a.each(this.getParentList(!1,!0),function(a,b){c.push(b.getIndex()+1)}),c.join(b)},getKeyPath:function(a){var b=[],c=this.tree.options.keyPathSeparator;return this.visitParents(function(a){a.parent&&b.unshift(a.key)},!a),c+b.join(c)},getLastChild:function(){return this.children?this.children[this.children.length-1]:null},getLevel:function(){for(var a=0,b=this.parent;b;)a++,b=b.parent;return a},getNextSibling:function(){if(this.parent){var a,b,c=this.parent.children;for(a=0,b=c.length-1;b>a;a++)if(c[a]===this)return c[a+1]}return null},getParent:function(){return this.parent},getParentList:function(a,b){for(var c=[],d=b?this:this.parent;d;)(a||d.parent)&&c.unshift(d),d=d.parent;return c},getPrevSibling:function(){if(this.parent){var a,b,c=this.parent.children;for(a=1,b=c.length;b>a;a++)if(c[a]===this)return c[a-1]}return null},hasChildren:function(){return this.lazy?null==this.children?d:0===this.children.length?!1:1===this.children.length&&this.children[0].isStatusNode()?d:!0:!(!this.children||!this.children.length)},hasFocus:function(){return this.tree.hasFocus()&&this.tree.focusNode===this},info:function(){this.tree.options.debugLevel>=1&&(Array.prototype.unshift.call(arguments,this.toString()),f("info",arguments))},isActive:function(){return this.tree.activeNode===this},isChildOf:function(a){return this.parent&&this.parent===a},isDescendantOf:function(a){if(!a||a.tree!==this.tree)return!1;for(var b=this.parent;b;){if(b===a)return!0;b=b.parent}return!1},isExpanded:function(){return!!this.expanded},isFirstSibling:function(){var a=this.parent;return!a||a.children[0]===this},isFolder:function(){return!!this.folder},isLastSibling:function(){var a=this.parent;return!a||a.children[a.children.length-1]===this},isLazy:function(){return!!this.lazy},isLoaded:function(){return!this.lazy||this.hasChildren()!==d},isLoading:function(){return!!this._isLoading},isRoot:function(){return this.isRootNode()},isRootNode:function(){return this.tree.rootNode===this},isSelected:function(){return!!this.selected},isStatusNode:function(){return!!this.statusNodeType},isTopLevel:function(){return this.tree.rootNode===this.parent},isUndefined:function(){return this.hasChildren()===d},isVisible:function(){var a,b,c=this.getParentList(!1,!1);for(a=0,b=c.length;b>a;a++)if(!c[a].expanded)return!1;return!0},lazyLoad:function(a){return this.warn("FancytreeNode.lazyLoad() is deprecated since 2014-02-16. Use .load() instead."),this.load(a)},load:function(a){var b,c,d=this;return e(this.isLazy(),"load() requires a lazy node"),a||this.isUndefined()?(this.isLoaded()&&this.resetLazy(),c=this.tree._triggerNodeEvent("lazyLoad",this),c===!1?k(this):(e("boolean"!=typeof c,"lazyLoad event must return source in data.result"),b=this.tree._callHook("nodeLoadChildren",this,c),this.expanded&&b.always(function(){d.render()}),b)):k(this)},makeVisible:function(b){var c,d=this,e=[],f=new a.Deferred,g=this.getParentList(!1,!1),h=g.length,i=!(b&&b.noAnimation===!0),j=!(b&&b.scrollIntoView===!1);for(c=h-1;c>=0;c--)e.push(g[c].setExpanded(!0,b));return a.when.apply(a,e).done(function(){j?d.scrollIntoView(i).done(function(){f.resolve()}):f.resolve()}),f.promise()},moveTo:function(b,c,f){(c===d||"over"===c)&&(c="child");var g,h=this.parent,i="child"===c?b:b.parent;if(this!==b){if(!this.parent)throw"Cannot move system root";if(i.isDescendantOf(this))throw"Cannot move a node to its own descendant";if(1===this.parent.children.length){if(this.parent===i)return;this.parent.children=this.parent.lazy?[]:null,this.parent.expanded=!1}else g=a.inArray(this,this.parent.children),e(g>=0),this.parent.children.splice(g,1);if(this.parent=i,i.hasChildren())switch(c){case"child":i.children.push(this);break;case"before":g=a.inArray(b,i.children),e(g>=0),i.children.splice(g,0,this);break;case"after":g=a.inArray(b,i.children),e(g>=0),i.children.splice(g+1,0,this);break;default:throw"Invalid mode "+c}else i.children=[this];f&&b.visit(f,!0),this.tree!==b.tree&&(this.warn("Cross-tree moveTo is experimantal!"),this.visit(function(a){a.tree=b.tree},!0)),h.isDescendantOf(i)||h.render(),i.isDescendantOf(h)||i===h||i.render()}},navigate:function(b,c){function d(d){if(d){try{d.makeVisible()}catch(e){}return a(d.span).is(":visible")?c===!1?d.setFocus():d.setActive():(d.debug("Navigate: skipping hidden node"),void d.navigate(b,c))}}var e,f,g=!0,h=a.ui.keyCode,i=null;switch(b){case h.BACKSPACE:this.parent&&this.parent.parent&&d(this.parent);break;case h.LEFT:this.expanded?(this.setExpanded(!1),d(this)):this.parent&&this.parent.parent&&d(this.parent);break;case h.RIGHT:this.expanded||!this.children&&!this.lazy?this.children&&this.children.length&&d(this.children[0]):(this.setExpanded(),d(this));break;case h.UP:for(i=this.getPrevSibling();i&&!a(i.span).is(":visible");)i=i.getPrevSibling();for(;i&&i.expanded&&i.children&&i.children.length;)i=i.children[i.children.length-1];!i&&this.parent&&this.parent.parent&&(i=this.parent),d(i);break;case h.DOWN:if(this.expanded&&this.children&&this.children.length)i=this.children[0];else for(f=this.getParentList(!1,!0),e=f.length-1;e>=0;e--){for(i=f[e].getNextSibling();i&&!a(i.span).is(":visible");)i=i.getNextSibling();if(i)break}d(i);break;default:g=!1}},remove:function(){return this.parent.removeChild(this)},removeChild:function(a){return this.tree._callHook("nodeRemoveChild",this,a)},removeChildren:function(){return this.tree._callHook("nodeRemoveChildren",this)},render:function(a,b){return this.tree._callHook("nodeRender",this,a,b)},renderTitle:function(){return this.tree._callHook("nodeRenderTitle",this)},renderStatus:function(){return this.tree._callHook("nodeRenderStatus",this)},resetLazy:function(){this.removeChildren(),this.expanded=!1,this.lazy=!0,this.children=d,this.renderStatus()},scheduleAction:function(a,b){this.tree.timer&&clearTimeout(this.tree.timer),this.tree.timer=null;var c=this;switch(a){case"cancel":break;case"expand":this.tree.timer=setTimeout(function(){c.tree.debug("setTimeout: trigger expand"),c.setExpanded(!0)},b);break;case"activate":this.tree.timer=setTimeout(function(){c.tree.debug("setTimeout: trigger activate"),c.setActive(!0)},b);break;default:throw"Invalid mode "+a}},scrollIntoView:function(f,h){h!==d&&g(h)&&(this.warn("scrollIntoView() with 'topNode' option is deprecated since 2014-05-08. Use 'options.topNode' instead."),h={topNode:h});var i,j,k,l,m=a.extend({effects:f===!0?{duration:200,queue:!1}:f,scrollOfs:this.tree.options.scrollOfs,scrollParent:this.tree.options.scrollParent||this.tree.$container,topNode:null},h),n=new a.Deferred,o=this,p=a(this.span).height(),q=a(m.scrollParent),r=m.scrollOfs.top||0,s=m.scrollOfs.bottom||0,t=q.height(),u=q.scrollTop(),v=q,w=q[0]===b,x=m.topNode||null,y=null;return e(a(this.span).is(":visible"),"scrollIntoView node is invisible"),w?(j=a(this.span).offset().top,i=x&&x.span?a(x.span).offset().top:0,v=a("html,body")):(e(q[0]!==c&&q[0]!==c.body,"scrollParent should be an simple element or `window`, not document or body."),l=q.offset().top,j=a(this.span).offset().top-l+u,i=x?a(x.span).offset().top-l+u:0,k=Math.max(0,q.innerHeight()-q[0].clientHeight),t-=k),u+r>j?y=j-r:j+p>u+t-s&&(y=j+p-t+s,x&&(e(x.isRoot()||a(x.span).is(":visible"),"topNode must be visible"),y>i&&(y=i-r))),null!==y?m.effects?(m.effects.complete=function(){n.resolveWith(o)},v.stop(!0).animate({scrollTop:y},m.effects)):(v[0].scrollTop=y,n.resolveWith(this)):n.resolveWith(this),n.promise()},setActive:function(a,b){return this.tree._callHook("nodeSetActive",this,a,b)},setExpanded:function(a,b){return this.tree._callHook("nodeSetExpanded",this,a,b)},setFocus:function(a){return this.tree._callHook("nodeSetFocus",this,a)},setSelected:function(a){return this.tree._callHook("nodeSetSelected",this,a)},setStatus:function(a,b,c){return this.tree._callHook("nodeSetStatus",this,a,b,c)},setTitle:function(a){this.title=a,this.renderTitle()},sortChildren:function(a,b){var c,d,e=this.children;if(e){if(a=a||function(a,b){var c=a.title.toLowerCase(),d=b.title.toLowerCase();return c===d?0:c>d?1:-1},e.sort(a),b)for(c=0,d=e.length;d>c;c++)e[c].children&&e[c].sortChildren(a,"$norender$");"$norender$"!==b&&this.render()}},toDict:function(b,c){var d,e,f,g={},h=this;if(a.each(z,function(a,b){(h[b]||h[b]===!1)&&(g[b]=h[b])}),a.isEmptyObject(this.data)||(g.data=a.extend({},this.data),a.isEmptyObject(g.data)&&delete g.data),c&&c(g),b&&this.hasChildren())for(g.children=[],d=0,e=this.children.length;e>d;d++)f=this.children[d],f.isStatusNode()||g.children.push(f.toDict(!0,c));return g},toggleExpanded:function(){return this.tree._callHook("nodeToggleExpanded",this)},toggleSelected:function(){return this.tree._callHook("nodeToggleSelected",this)},toString:function(){return"<FancytreeNode(#"+this.key+", '"+this.title+"')>"},visit:function(a,b){var c,d,e=!0,f=this.children;if(b===!0&&(e=a(this),e===!1||"skip"===e))return e;if(f)for(c=0,d=f.length;d>c&&(e=f[c].visit(a,!0),e!==!1);c++);return e},visitAndLoad:function(b,c,d){var e,f,g,h=this;return b&&c===!0&&(f=b(h),f===!1||"skip"===f)?d?f:k():h.children||h.lazy?(e=new a.Deferred,g=[],h.load().done(function(){for(var c=0,d=h.children.length;d>c;c++){if(f=h.children[c].visitAndLoad(b,!0,!0),f===!1){e.reject();break}"skip"!==f&&g.push(f)}a.when.apply(this,g).then(function(){e.resolve()})}),e.promise()):k()},visitParents:function(a,b){if(b&&a(this)===!1)return!1;for(var c=this.parent;c;){if(a(c)===!1)return!1;c=c.parent}return!0},warn:function(){Array.prototype.unshift.call(arguments,this.toString()),f("warn",arguments)}},r.prototype={_makeHookContext:function(b,c,e){var f,g;return b.node!==d?(c&&b.originalEvent!==c&&a.error("invalid args"),f=b):b.tree?(g=b.tree,f={node:b,tree:g,widget:g.widget,options:g.widget.options,originalEvent:c}):b.widget?f={node:null,tree:b,widget:b.widget,options:b.widget.options,originalEvent:c}:a.error("invalid args"),e&&a.extend(f,e),f},_callHook:function(b,c){var d=this._makeHookContext(c),e=this[b],f=Array.prototype.slice.call(arguments,2);return a.isFunction(e)||a.error("_callHook('"+b+"') is not a function"),f.unshift(d),e.apply(this,f)},_requireExtension:function(b,c,d,f){d=!!d;var g=this._local.name,h=this.options.extensions,i=a.inArray(b,h)<a.inArray(g,h),j=c&&null==this.ext[b],k=!j&&null!=d&&d!==i;return e(g&&g!==b),j||k?(f||(j||c?(f="'"+g+"' extension requires '"+b+"'",k&&(f+=" to be registered "+(d?"before":"after")+" itself")):f="If used together, `"+b+"` must be registered "+(d?"before":"after")+" `"+g+"`"),a.error(f),!1):!0},activateKey:function(a){var b=this.getNodeByKey(a);return b?b.setActive():this.activeNode&&this.activeNode.setActive(!1),b},applyPatch:function(b){var c,d,f,g,h,i,j=b.length,k=[];for(d=0;j>d;d++)f=b[d],e(2===f.length,"patchList must be an array of length-2-arrays"),g=f[0],h=f[1],i=null===g?this.rootNode:this.getNodeByKey(g),i?(c=new a.Deferred,k.push(c),i.applyPatch(h).always(m(c,i))):this.warn("could not find node with key '"+g+"'");return a.when.apply(a,k).promise()},count:function(){return this.rootNode.countChildren()},debug:function(){this.options.debugLevel>=2&&(Array.prototype.unshift.call(arguments,this.toString()),f("log",arguments))},findNextNode:function(b,c){var d=null,e=c.parent.children,f=null,g=function(a,b,c){var d,e,f=a.children,h=f.length,i=f[b];if(i&&c(i)===!1)return!1;if(i&&i.children&&i.expanded&&g(i,0,c)===!1)return!1;for(d=b+1;h>d;d++)if(g(a,d,c)===!1)return!1;return e=a.parent,e?g(e,e.children.indexOf(a)+1,c):g(a,0,c)};return b="string"==typeof b?p(b):b,c=c||this.getFirstChild(),g(c.parent,e.indexOf(c),function(e){return e===d?!1:(d=d||e,a(e.span).is(":visible")?b(e)&&(f=e,f!==c)?!1:void 0:void e.debug("quicksearch: skipping hidden node"))}),f},generateFormElements:function(b,c){var d,e=b!==!1?"ft_"+this._id+"[]":b,f=c!==!1?"ft_"+this._id+"_active":c,g="fancytree_result_"+this._id,h=a("#"+g);h.length?h.empty():h=a("<div>",{id:g}).hide().insertAfter(this.$container),e&&(d=this.getSelectedNodes(3===this.options.selectMode),a.each(d,function(b,c){h.append(a("<input>",{type:"checkbox",name:e,value:c.key,checked:!0}))})),f&&this.activeNode&&h.append(a("<input>",{type:"radio",name:f,value:this.activeNode.key,checked:!0}))},getActiveNode:function(){return this.activeNode},getFirstChild:function(){return this.rootNode.getFirstChild()},getFocusNode:function(){return this.focusNode},getNodeByKey:function(a,b){var d,e;return!b&&(d=c.getElementById(this.options.idPrefix+a))?d.ftnode?d.ftnode:null:(b=b||this.rootNode,e=null,b.visit(function(b){return b.key===a?(e=b,!1):void 0},!0),e)},getRootNode:function(){return this.rootNode},getSelectedNodes:function(a){var b=[];return this.rootNode.visit(function(c){return c.selected&&(b.push(c),a===!0)?"skip":void 0}),b},hasFocus:function(){return!!this._hasFocus},info:function(){this.options.debugLevel>=1&&(Array.prototype.unshift.call(arguments,this.toString()),f("info",arguments))},loadKeyPath:function(b,c,e){function f(a,b,d){c.call(r,b,"loading"),b.load().done(function(){r.loadKeyPath.call(r,l[a],c,b).always(m(d,r))}).fail(function(){r.warn("loadKeyPath: error loading: "+a+" (parent: "+o+")"),c.call(r,b,"error"),d.reject()})}var g,h,i,j,k,l,n,o,p,q=this.options.keyPathSeparator,r=this;for(a.isArray(b)||(b=[b]),l={},i=0;i<b.length;i++)for(o=e||this.rootNode,j=b[i],j.charAt(0)===q&&(j=j.substr(1)),p=j.split(q);p.length;){if(k=p.shift(),n=o._findDirectChild(k),!n){this.warn("loadKeyPath: key not found: "+k+" (parent: "+o+")"),c.call(this,k,"error");break}if(0===p.length){c.call(this,n,"ok");break}if(n.lazy&&n.hasChildren()===d){c.call(this,n,"loaded"),l[k]?l[k].push(p.join(q)):l[k]=[p.join(q)];break}c.call(this,n,"loaded"),o=n}g=[];for(k in l)n=o._findDirectChild(k),h=new a.Deferred,g.push(h),f(k,n,h);return a.when.apply(a,g).promise()},reactivate:function(a){var b,c=this.activeNode;return c?(this.activeNode=null,b=c.setActive(),a&&c.setFocus(),b):k()},reload:function(a){return this._callHook("treeClear",this),this._callHook("treeLoad",this,a)},render:function(a,b){return this.rootNode.render(a,b)},setFocus:function(a){return this._callHook("treeSetFocus",this,a)},toDict:function(a,b){var c=this.rootNode.toDict(!0,b);return a?c:c.children},toString:function(){return"<Fancytree(#"+this._id+")>"},_triggerNodeEvent:function(a,b,c,e){var f=this._makeHookContext(b,c,e),g=this.widget._trigger(a,c,f);return g!==!1&&f.result!==d?f.result:g},_triggerTreeEvent:function(a,b,c){var e=this._makeHookContext(this,b,c),f=this.widget._trigger(a,b,e);return f!==!1&&e.result!==d?e.result:f},visit:function(a){return this.rootNode.visit(a,!1)},warn:function(){Array.prototype.unshift.call(arguments,this.toString()),f("warn",arguments)}},a.extend(r.prototype,{nodeClick:function(a){var b,c,d=a.targetType,e=a.node;if("expander"===d)this._callHook("nodeToggleExpanded",a);else if("checkbox"===d)this._callHook("nodeToggleSelected",a),a.options.focusOnSelect&&this._callHook("nodeSetFocus",a,!0);else{if(c=!1,b=!0,e.folder)switch(a.options.clickFolderMode){case 2:c=!0,b=!1;break;case 3:b=!0,c=!0}b&&(this.nodeSetFocus(a),this._callHook("nodeSetActive",a,!0)),c&&this._callHook("nodeToggleExpanded",a)}},nodeCollapseSiblings:function(a,b){var c,d,e,f=a.node;if(f.parent)for(c=f.parent.children,d=0,e=c.length;e>d;d++)c[d]!==f&&c[d].expanded&&this._callHook("nodeSetExpanded",c[d],!1,b)},nodeDblclick:function(a){"title"===a.targetType&&4===a.options.clickFolderMode&&this._callHook("nodeToggleExpanded",a),"title"===a.targetType&&a.originalEvent.preventDefault()},nodeKeydown:function(b){var c,d,e,f=b.originalEvent,g=b.node,h=b.tree,i=b.options,j=f.which,k=String.fromCharCode(j),l=!(f.altKey||f.ctrlKey||f.metaKey||f.shiftKey),m=a(f.target),n=!0,o=!(f.ctrlKey||!i.autoActivate);if(g||(this.getFirstChild().setFocus(),g=b.node=this.focusNode,g.debug("Keydown force focus on first node")),i.quicksearch&&l&&/\w/.test(k)&&!m.is(":input:enabled"))return d=(new Date).getTime(),d-h.lastQuicksearchTime>500&&(h.lastQuicksearchTerm=""),h.lastQuicksearchTime=d,h.lastQuicksearchTerm+=k,c=h.findNextNode(h.lastQuicksearchTerm,h.getActiveNode()),c&&c.setActive(),void f.preventDefault();switch(t.keyEventToString(f)){case"+":case"=":h.nodeSetExpanded(b,!0);break;case"-":h.nodeSetExpanded(b,!1);break;case"space":i.checkbox?h.nodeToggleSelected(b):h.nodeSetActive(b,!0);break;case"enter":h.nodeSetActive(b,!0);break;case"backspace":case"left":case"right":case"up":case"down":e=g.navigate(f.which,o);break;default:n=!1}n&&f.preventDefault()},nodeLoadChildren:function(b,c){var d,f,g,h=b.tree,i=b.node;return a.isFunction(c)&&(c=c()),c.url&&(d=a.extend({},b.options.ajax,c),d.debugDelay?(f=d.debugDelay,a.isArray(f)&&(f=f[0]+Math.random()*(f[1]-f[0])),i.debug("nodeLoadChildren waiting debug delay "+Math.round(f)+"ms"),d.debugDelay=!1,g=a.Deferred(function(b){setTimeout(function(){a.ajax(d).done(function(){b.resolveWith(this,arguments)}).fail(function(){b.rejectWith(this,arguments)})},f)})):g=a.ajax(d),c=new a.Deferred,g.done(function(d){var e,f;if("string"==typeof d&&a.error("Ajax request returned a string (did you get the JSON dataType wrong?)."),b.options.postProcess){if(f=h._triggerNodeEvent("postProcess",b,b.originalEvent,{response:d,error:null,dataType:this.dataType}),f.error)return e=a.isPlainObject(f.error)?f.error:{message:f.error},e=h._makeHookContext(i,null,e),void c.rejectWith(this,[e]);d=a.isArray(f)?f:d}else d&&d.hasOwnProperty("d")&&b.options.enableAspx&&(d="string"==typeof d.d?a.parseJSON(d.d):d.d);c.resolveWith(this,[d])}).fail(function(a,b,d){var e=h._makeHookContext(i,null,{error:a,args:Array.prototype.slice.call(arguments),message:d,details:a.status+": "+d});c.rejectWith(this,[e])})),a.isFunction(c.promise)&&(e(!i.isLoading()),h.nodeSetStatus(b,"loading"),c.done(function(){h.nodeSetStatus(b,"ok")}).fail(function(a){var c;c=a.node&&a.error&&a.message?a:h._makeHookContext(i,null,{error:a,args:Array.prototype.slice.call(arguments),message:a?a.message||a.toString():""}),h._triggerNodeEvent("loadError",c,null)!==!1&&h.nodeSetStatus(b,"error",c.message,c.details)})),a.when(c).done(function(b){var c;a.isPlainObject(b)&&(e(a.isArray(b.children),"source must contain (or be) an array of children"),e(i.isRoot(),"source may only be an object for root nodes"),c=b,b=b.children,delete c.children,a.extend(h.data,c)),e(a.isArray(b),"expected array of children"),i._setChildren(b),h._triggerNodeEvent("loadChildren",i)})},nodeLoadKeyPath:function(){},nodeRemoveChild:function(b,c){var d,f=b.node,g=b.options,h=a.extend({},b,{node:c}),i=f.children;return 1===i.length?(e(c===i[0]),this.nodeRemoveChildren(b)):(this.activeNode&&(c===this.activeNode||this.activeNode.isDescendantOf(c))&&this.activeNode.setActive(!1),this.focusNode&&(c===this.focusNode||this.focusNode.isDescendantOf(c))&&(this.focusNode=null),this.nodeRemoveMarkup(h),this.nodeRemoveChildren(h),d=a.inArray(c,i),e(d>=0),c.visit(function(a){a.parent=null},!0),this._callHook("treeRegisterNode",this,!1,c),g.removeNode&&g.removeNode.call(b.tree,{type:"removeNode"},h),void i.splice(d,1))},nodeRemoveChildMarkup:function(b){var c=b.node;c.ul&&(c.isRoot()?a(c.ul).empty():(a(c.ul).remove(),c.ul=null),c.visit(function(a){a.li=a.ul=null}))},nodeRemoveChildren:function(b){var c,d=b.tree,e=b.node,f=e.children,g=b.options;f&&(this.activeNode&&this.activeNode.isDescendantOf(e)&&this.activeNode.setActive(!1),this.focusNode&&this.focusNode.isDescendantOf(e)&&(this.focusNode=null),this.nodeRemoveChildMarkup(b),c=a.extend({},b),e.visit(function(a){a.parent=null,d._callHook("treeRegisterNode",d,!1,a),g.removeNode&&(c.node=a,g.removeNode.call(b.tree,{type:"removeNode"},c))}),e.children=e.lazy?[]:null,this.nodeRenderStatus(b))},nodeRemoveMarkup:function(b){var c=b.node;c.li&&(a(c.li).remove(),c.li=null),this.nodeRemoveChildMarkup(b)},nodeRender:function(b,d,f,g,h){var i,j,k,l,m,n,o,p=b.node,q=b.tree,r=b.options,s=r.aria,t=!1,u=p.parent,v=!u,w=p.children;if(v||u.ul){if(e(v||u.ul,"parent UL must exist"),v||(p.li&&(d||p.li.parentNode!==p.parent.ul)&&(p.li.parentNode!==p.parent.ul&&this.debug("Unlinking "+p+" (must be child of "+p.parent+")"),this.nodeRemoveMarkup(b)),p.li?this.nodeRenderStatus(b):(t=!0,p.li=c.createElement("li"),p.li.ftnode=p,p.key&&r.generateIds&&(p.li.id=r.idPrefix+p.key),p.span=c.createElement("span"),p.span.className="fancytree-node",s&&a(p.span).attr("aria-labelledby","ftal_"+p.key),p.li.appendChild(p.span),this.nodeRenderTitle(b),r.createNode&&r.createNode.call(q,{type:"createNode"},b)),r.renderNode&&r.renderNode.call(q,{type:"renderNode"},b)),w){if(v||p.expanded||f===!0){for(p.ul||(p.ul=c.createElement("ul"),(g===!0&&!h||!p.expanded)&&(p.ul.style.display="none"),s&&a(p.ul).attr("role","group"),p.li?p.li.appendChild(p.ul):p.tree.$div.append(p.ul)),l=0,m=w.length;m>l;l++)o=a.extend({},b,{node:w[l]}),this.nodeRender(o,d,f,!1,!0);for(i=p.ul.firstChild;i;)k=i.ftnode,k&&k.parent!==p?(p.debug("_fixParent: remove missing "+k,i),n=i.nextSibling,i.parentNode.removeChild(i),i=n):i=i.nextSibling;for(i=p.ul.firstChild,l=0,m=w.length-1;m>l;l++)j=w[l],k=i.ftnode,j!==k?p.ul.insertBefore(j.li,k.li):i=i.nextSibling}}else p.ul&&(this.warn("remove child markup for "+p),this.nodeRemoveChildMarkup(b));v||t&&u.ul.appendChild(p.li)}},nodeRenderTitle:function(a,b){var c,e,f,g,h,i,j=a.node,k=a.tree,l=a.options,m=l.aria,n=j.getLevel(),o=[],p=j.data.icon;b!==d&&(j.title=b),j.span&&(n<l.minExpandLevel?(j.lazy||(j.expanded=!0),n>1&&o.push(m?"<span role='button' class='fancytree-expander fancytree-expander-fixed'></span>":"<span class='fancytree-expander fancytree-expander-fixed''></span>")):o.push(m?"<span role='button' class='fancytree-expander'></span>":"<span class='fancytree-expander'></span>"),l.checkbox&&j.hideCheckbox!==!0&&!j.isStatusNode()&&o.push(m?"<span role='checkbox' class='fancytree-checkbox'></span>":"<span class='fancytree-checkbox'></span>"),g=m?" role='img'":"",(p===!0||p!==!1&&l.icons!==!1)&&(p&&"string"==typeof p?(p="/"===p.charAt(0)?p:(l.imagePath||"")+p,o.push("<img src='"+p+"' class='fancytree-icon' alt='' />")):(e=l.iconClass&&l.iconClass.call(k,j,a)||j.data.iconclass||null,o.push(e?"<span "+g+" class='fancytree-custom-icon "+e+"'></span>":"<span "+g+" class='fancytree-icon'></span>"))),f="",l.renderTitle&&(f=l.renderTitle.call(k,{type:"renderTitle"},a)||""),f||(i=j.tooltip?" title='"+t.escapeHtml(j.tooltip)+"'":"",c=m?" id='ftal_"+j.key+"'":"",g=m?" role='treeitem'":"",h=l.titlesTabbable?" tabindex='0'":"",f="<span "+g+" class='fancytree-title'"+c+i+h+">"+j.title+"</span>"),o.push(f),j.span.innerHTML=o.join(""),this.nodeRenderStatus(a))},nodeRenderStatus:function(b){var c=b.node,d=b.tree,e=b.options,f=c.hasChildren(),g=c.isLastSibling(),h=e.aria,i=a(c.span).find(".fancytree-title"),j=e._classNames,k=[],l=c[d.statusClassPropName];l&&(k.push(j.node),d.activeNode===c&&k.push(j.active),d.focusNode===c?(k.push(j.focused),h&&i.attr("aria-activedescendant",!0)):h&&i.removeAttr("aria-activedescendant"),c.expanded?(k.push(j.expanded),h&&i.attr("aria-expanded",!0)):h&&i.removeAttr("aria-expanded"),c.folder&&k.push(j.folder),f!==!1&&k.push(j.hasChildren),g&&k.push(j.lastsib),c.lazy&&null==c.children&&k.push(j.lazy),c.partsel&&k.push(j.partsel),c.unselectable&&k.push(j.unselectable),c._isLoading&&k.push(j.loading),c._error&&k.push(j.error),c.selected?(k.push(j.selected),h&&i.attr("aria-selected",!0)):h&&i.attr("aria-selected",!1),c.extraClasses&&k.push(c.extraClasses),k.push(f===!1?j.combinedExpanderPrefix+"n"+(g?"l":""):j.combinedExpanderPrefix+(c.expanded?"e":"c")+(c.lazy&&null==c.children?"d":"")+(g?"l":"")),k.push(j.combinedIconPrefix+(c.expanded?"e":"c")+(c.folder?"f":"")),l.className=k.join(" "),c.li&&(c.li.className=g?j.lastsib:""))
},nodeSetActive:function(b,c,d){d=d||{};var f,g=b.node,h=b.tree,i=b.options,j=d.noEvents===!0,m=g===h.activeNode;return c=c!==!1,m===c?k(g):c&&!j&&this._triggerNodeEvent("beforeActivate",g,b.originalEvent)===!1?l(g,["rejected"]):void(c?(h.activeNode&&(e(h.activeNode!==g,"node was active (inconsistency)"),f=a.extend({},b,{node:h.activeNode}),h.nodeSetActive(f,!1),e(null===h.activeNode,"deactivate was out of sync?")),i.activeVisible&&g.makeVisible({scrollIntoView:!1}),h.activeNode=g,h.nodeRenderStatus(b),h.nodeSetFocus(b),j||h._triggerNodeEvent("activate",g,b.originalEvent)):(e(h.activeNode===g,"node was not active (inconsistency)"),h.activeNode=null,this.nodeRenderStatus(b),j||b.tree._triggerNodeEvent("deactivate",g,b.originalEvent)))},nodeSetExpanded:function(b,c,e){e=e||{};var f,g,h,i,j,m,n=b.node,o=b.tree,p=b.options,q=e.noAnimation===!0,r=e.noEvents===!0;if(c=c!==!1,n.expanded&&c||!n.expanded&&!c)return k(n);if(c&&!n.lazy&&!n.hasChildren())return k(n);if(!c&&n.getLevel()<p.minExpandLevel)return l(n,["locked"]);if(!r&&this._triggerNodeEvent("beforeExpand",n,b.originalEvent)===!1)return l(n,["rejected"]);if(q||n.isVisible()||(q=e.noAnimation=!0),g=new a.Deferred,c&&!n.expanded&&p.autoCollapse){j=n.getParentList(!1,!0),m=p.autoCollapse;try{for(p.autoCollapse=!1,h=0,i=j.length;i>h;h++)this._callHook("nodeCollapseSiblings",j[h],e)}finally{p.autoCollapse=m}}return g.done(function(){c&&p.autoScroll&&!q?n.getLastChild().scrollIntoView(!0,{topNode:n}).always(function(){r||b.tree._triggerNodeEvent(c?"expand":"collapse",b)}):r||b.tree._triggerNodeEvent(c?"expand":"collapse",b)}),f=function(d){var e,f,g=p.toggleEffect;if(n.expanded=c,o._callHook("nodeRender",b,!1,!1,!0),n.ul)if(e="none"!==n.ul.style.display,f=!!n.expanded,e===f)n.warn("nodeSetExpanded: UL.style.display already set");else{if(g&&!q)return void a(n.ul).toggle(g.effect,g.options,g.duration,function(){d()});n.ul.style.display=n.expanded||!parent?"":"none"}d()},c&&n.lazy&&n.hasChildren()===d?n.load().done(function(){g.notifyWith&&g.notifyWith(n,["loaded"]),f(function(){g.resolveWith(n)})}).fail(function(a){f(function(){g.rejectWith(n,["load failed ("+a+")"])})}):f(function(){g.resolveWith(n)}),g.promise()},nodeSetFocus:function(b,c){var d,e=b.tree,f=b.node;if(c=c!==!1,e.focusNode){if(e.focusNode===f&&c)return;d=a.extend({},b,{node:e.focusNode}),e.focusNode=null,this._triggerNodeEvent("blur",d),this._callHook("nodeRenderStatus",d)}c&&(this.hasFocus()||(f.debug("nodeSetFocus: forcing container focus"),this._callHook("treeSetFocus",b,!0,!0)),f.makeVisible({scrollIntoView:!1}),e.focusNode=f,this._triggerNodeEvent("focus",b),b.options.autoScroll&&f.scrollIntoView(),this._callHook("nodeRenderStatus",b))},nodeSetSelected:function(a,b){var c=a.node,d=a.tree,e=a.options;if(b=b!==!1,c.debug("nodeSetSelected("+b+")",a),!c.unselectable){if(c.selected&&b||!c.selected&&!b)return!!c.selected;if(this._triggerNodeEvent("beforeSelect",c,a.originalEvent)===!1)return!!c.selected;b&&1===e.selectMode?d.lastSelectedNode&&d.lastSelectedNode.setSelected(!1):3===e.selectMode&&(c.selected=b,c.fixSelection3AfterClick()),c.selected=b,this.nodeRenderStatus(a),d.lastSelectedNode=b?c:null,d._triggerNodeEvent("select",a)}},nodeSetStatus:function(b,c,d,e){function f(){var a=h.children?h.children[0]:null;if(a&&a.isStatusNode()){try{h.ul&&(h.ul.removeChild(a.li),a.li=null)}catch(b){}1===h.children.length?h.children=[]:h.children.shift()}}function g(b,c){var d=h.children?h.children[0]:null;return d&&d.isStatusNode()?(a.extend(d,b),i._callHook("nodeRenderTitle",d)):(b.key="_statusNode",h._setChildren([b]),h.children[0].statusNodeType=c,i.render()),h.children[0]}var h=b.node,i=b.tree;switch(c){case"ok":f(),h._isLoading=!1,h._error=null,h.renderStatus();break;case"loading":h.parent||g({title:i.options.strings.loading+(d?" ("+d+") ":""),tooltip:e,extraClasses:"fancytree-statusnode-wait"},c),h._isLoading=!0,h._error=null,h.renderStatus();break;case"error":g({title:i.options.strings.loadError+(d?" ("+d+") ":""),tooltip:e,extraClasses:"fancytree-statusnode-error"},c),h._isLoading=!1,h._error={message:d,details:e},h.renderStatus();break;default:a.error("invalid node status "+c)}},nodeToggleExpanded:function(a){return this.nodeSetExpanded(a,!a.node.expanded)},nodeToggleSelected:function(a){return this.nodeSetSelected(a,!a.node.selected)},treeClear:function(a){var b=a.tree;b.activeNode=null,b.focusNode=null,b.$div.find(">ul.fancytree-container").empty(),b.rootNode.children=null},treeCreate:function(){},treeDestroy:function(){},treeInit:function(a){this.treeLoad(a)},treeLoad:function(b,c){var d,e,f,g=b.tree,h=b.widget.element,i=a.extend({},b,{node:this.rootNode});if(g.rootNode.children&&this.treeClear(b),c=c||this.options.source)"string"==typeof c&&a.error("Not implemented");else switch(d=h.data("type")||"html"){case"html":e=h.find(">ul:first"),e.addClass("ui-fancytree-source ui-helper-hidden"),c=a.ui.fancytree.parseHtml(e),this.data=a.extend(this.data,n(e));break;case"json":c=a.parseJSON(h.text()),c.children&&(c.title&&(g.title=c.title),c=c.children);break;default:a.error("Invalid data-type: "+d)}return f=this.nodeLoadChildren(i,c).done(function(){g.render(),3===b.options.selectMode&&g.rootNode.fixSelection3FromEndNodes(),g._triggerTreeEvent("init",null,{status:!0})}).fail(function(){g.render(),g._triggerTreeEvent("init",null,{status:!1})})},treeRegisterNode:function(){},treeSetFocus:function(a,b){b=b!==!1,b!==this.hasFocus()&&(this._hasFocus=b,this.$container.toggleClass("fancytree-treefocus",b),this._triggerTreeEvent(b?"focusTree":"blurTree"))}}),a.widget("ui.fancytree",{options:{activeVisible:!0,ajax:{type:"GET",cache:!1,dataType:"json"},aria:!1,autoActivate:!0,autoCollapse:!1,autoScroll:!1,checkbox:!1,clickFolderMode:4,debugLevel:null,disabled:!1,enableAspx:!0,extensions:[],toggleEffect:{effect:"blind",options:{direction:"vertical",scale:"box"},duration:200},generateIds:!1,icons:!0,idPrefix:"ft_",focusOnSelect:!1,keyboard:!0,keyPathSeparator:"/",minExpandLevel:1,quicksearch:!1,scrollOfs:{top:0,bottom:0},scrollParent:null,selectMode:2,strings:{loading:"Loading&#8230;",loadError:"Load error!"},tabbable:!0,titlesTabbable:!1,_classNames:{node:"fancytree-node",folder:"fancytree-folder",combinedExpanderPrefix:"fancytree-exp-",combinedIconPrefix:"fancytree-ico-",hasChildren:"fancytree-has-children",active:"fancytree-active",selected:"fancytree-selected",expanded:"fancytree-expanded",lazy:"fancytree-lazy",focused:"fancytree-focused",partsel:"fancytree-partsel",unselectable:"fancytree-unselectable",lastsib:"fancytree-lastsib",loading:"fancytree-loading",error:"fancytree-error"},lazyLoad:null,postProcess:null},_create:function(){this.tree=new r(this),this.$source=this.source||"json"===this.element.data("type")?this.element:this.element.find(">ul:first");var b,c,f,g=this.options.extensions,h=this.tree;for(f=0;f<g.length;f++)c=g[f],b=a.ui.fancytree._extensions[c],b||a.error("Could not apply extension '"+c+"' (it is not registered, did you forget to include it?)"),this.tree.options[c]=a.extend(!0,{},b.options,this.tree.options[c]),e(this.tree.ext[c]===d,"Extension name must not exist as Fancytree.ext attribute: '"+c+"'"),this.tree.ext[c]={},j(this.tree,h,b,c),h=b;this.tree._callHook("treeCreate",this.tree)},_init:function(){this.tree._callHook("treeInit",this.tree),this._bind()},_setOption:function(b,c){var d=!0,e=!1;switch(b){case"aria":case"checkbox":case"icons":case"minExpandLevel":case"tabbable":this.tree._callHook("treeCreate",this.tree),e=!0;break;case"source":d=!1,this.tree._callHook("treeLoad",this.tree,c)}this.tree.debug("set option "+b+"="+c+" <"+typeof c+">"),d&&a.Widget.prototype._setOption.apply(this,arguments),e&&this.tree.render(!0,!1)},destroy:function(){this._unbind(),this.tree._callHook("treeDestroy",this.tree),this.tree.$div.find(">ul.fancytree-container").remove(),this.$source&&this.$source.removeClass("ui-helper-hidden"),a.Widget.prototype.destroy.call(this)},_unbind:function(){var b=this.tree._ns;this.element.unbind(b),this.tree.$container.unbind(b),a(c).unbind(b)},_bind:function(){var a=this,b=this.options,c=this.tree,d=c._ns;this._unbind(),c.$container.on("focusin"+d+" focusout"+d,function(a){var b=t.getNode(a),d="focusin"===a.type;b?c._callHook("nodeSetFocus",b,d):c._callHook("treeSetFocus",c,d)}).on("selectstart"+d,"span.fancytree-title",function(a){a.preventDefault()}).on("keydown"+d,function(a){if(b.disabled||b.keyboard===!1)return!0;var d,e=c.focusNode,f=c._makeHookContext(e||c,a),g=c.phase;try{return c.phase="userEvent",d=e?c._triggerNodeEvent("keydown",e,a):c._triggerTreeEvent("keydown",a),"preventNav"===d?d=!0:d!==!1&&(d=c._callHook("nodeKeydown",f)),d}finally{c.phase=g}}).on("click"+d+" dblclick"+d,function(c){if(b.disabled)return!0;var d,e=t.getEventTarget(c),f=e.node,g=a.tree,h=g.phase;if(!f)return!0;d=g._makeHookContext(f,c);try{switch(g.phase="userEvent",c.type){case"click":return d.targetType=e.type,g._triggerNodeEvent("click",d,c)===!1?!1:g._callHook("nodeClick",d);case"dblclick":return d.targetType=e.type,g._triggerNodeEvent("dblclick",d,c)===!1?!1:g._callHook("nodeDblclick",d)}}finally{g.phase=h}})},getActiveNode:function(){return this.tree.activeNode},getNodeByKey:function(a){return this.tree.getNodeByKey(a)},getRootNode:function(){return this.tree.rootNode},getTree:function(){return this.tree}}),t=a.ui.fancytree,a.extend(a.ui.fancytree,{version:"2.7.0",buildType: "production",debugLevel: 1,_nextId:1,_nextNodeKey:1,_extensions:{},_FancytreeClass:r,_FancytreeNodeClass:q,jquerySupports:{positionMyOfs:h(a.ui.version,1,9)},assert:function(a,b){return e(a,b)},debounce:function(a,b,c,d){var e;return 3===arguments.length&&"boolean"!=typeof c&&(d=c,c=!1),function(){var f=arguments;d=d||this,c&&!e&&b.apply(d,f),clearTimeout(e),e=setTimeout(function(){c||b.apply(d,f),e=null},a)}},debug:function(){a.ui.fancytree.debugLevel>=2&&f("log",arguments)},error:function(){f("error",arguments)},escapeHtml:function(a){return(""+a).replace(/[&<>"'\/]/g,function(a){return u[a]})},getEventTargetType:function(a){return this.getEventTarget(a).type},getEventTarget:function(b){var c=b&&b.target?b.target.className:"",e={node:this.getNode(b.target),type:d};return/\bfancytree-title\b/.test(c)?e.type="title":/\bfancytree-expander\b/.test(c)?e.type=e.node.hasChildren()===!1?"prefix":"expander":/\bfancytree-checkbox\b/.test(c)||/\bfancytree-radio\b/.test(c)?e.type="checkbox":/\bfancytree-icon\b/.test(c)?e.type="icon":/\bfancytree-node\b/.test(c)?e.type="title":b&&b.target&&a(b.target).closest(".fancytree-title").length&&(e.type="title"),e},getNode:function(a){if(a instanceof q)return a;for(a.selector!==d?a=a[0]:a.originalEvent!==d&&(a=a.target);a;){if(a.ftnode)return a.ftnode;a=a.parentNode}return null},info:function(){a.ui.fancytree.debugLevel>=1&&f("info",arguments)},keyEventToString:function(a){var b=a.which,c=[];return a.altKey&&c.push("alt"),a.ctrlKey&&c.push("ctrl"),a.metaKey&&c.push("meta"),a.shiftKey&&c.push("shift"),v[b]||c.push(w[b]||String.fromCharCode(b).toLowerCase()),c.join("+")},parseHtml:function(b){var c,e,f,g,h,i,j,k,l=b.find(">li"),m=[];return l.each(function(){var l,o=a(this),p=o.find(">span:first",this),q=p.length?null:o.find(">a:first"),r={tooltip:null,data:{}};for(p.length?r.title=p.html():q&&q.length?(r.title=q.html(),r.data.href=q.attr("href"),r.data.target=q.attr("target"),r.tooltip=q.attr("title")):(r.title=o.html(),g=r.title.search(/<ul/i),g>=0&&(r.title=r.title.substring(0,g))),r.title=a.trim(r.title),e=0,f=x.length;f>e;e++)r[x[e]]=d;for(j=this.className.split(" "),c=[],e=0,f=j.length;f>e;e++)k=j[e],y[k]?r[k]=!0:c.push(k);if(r.extraClasses=c.join(" "),h=o.attr("title"),h&&(r.tooltip=h),h=o.attr("id"),h&&(r.key=h),l=n(o),l&&!a.isEmptyObject(l)){for(e=0,f=z.length;f>e;e++)h=z[e],i=l[h],null!=i&&(delete l[h],r[h]=i);a.extend(r.data,l)}b=o.find(">ul:first"),r.children=b.length?a.ui.fancytree.parseHtml(b):r.lazy?d:null,m.push(r)}),m},registerExtension:function(b){e(null!=b.name,"extensions must have a `name` property."),e(null!=b.version,"extensions must have a `version` property."),a.ui.fancytree._extensions[b.name]=b},unescapeHtml:function(a){var b=c.createElement("div");return b.innerHTML=a,0===b.childNodes.length?"":b.childNodes[0].nodeValue},warn:function(){f("warn",arguments)}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.childcounter.min.js' */
!function(a){"use strict";a.ui.fancytree._FancytreeClass.prototype.countSelected=function(a){{var b=this;b.options}return b.getSelectedNodes(a).length},a.ui.fancytree._FancytreeNodeClass.prototype.toUpper=function(){var a=this;return a.setTitle(a.title.toUpperCase())},a.ui.fancytree.prototype.widgetMethod1=function(a){this.tree;return a},a.ui.fancytree.registerExtension({name:"childcounter",version:"1.0.0",options:{deep:!0,hideZeros:!0,hideExpanded:!1},foo:42,_appendCounter:function(){},treeInit:function(a){a.options,a.options.childcounter;this._super(a),this.$container.addClass("fancytree-ext-childcounter")},treeDestroy:function(a){this._super(a)},nodeRenderTitle:function(b,c){var d=b.node,e=b.options.childcounter,f=null==d.data.childCounter?d.countChildren(e.deep):+d.data.childCounter;this._super(b,c),!f&&e.hideZeros||d.isExpanded()&&e.hideExpanded||a("span.fancytree-icon",d.span).append(a("<span class='fancytree-childcounter'/>").text(f))},nodeSetExpanded:function(a,b,c){{var d=a.tree;a.node}return this._super(a,b,c).always(function(){d.nodeRenderTitle(a)})}})}(jQuery);

/*! Extension 'jquery.fancytree.clones.min.js' */
!function(a){"use strict";function b(b,c){b||(c=c?": "+c:"",a.error("Assertion failed"+c))}function c(a,b){var c;for(c=a.length-1;c>=0;c--)if(a[c]===b)return a.splice(c,1),!0;return!1}function d(a,b,c){for(var d,e,f=3&a.length,g=a.length-f,h=c,i=3432918353,j=461845907,k=0;g>k;)e=255&a.charCodeAt(k)|(255&a.charCodeAt(++k))<<8|(255&a.charCodeAt(++k))<<16|(255&a.charCodeAt(++k))<<24,++k,e=(65535&e)*i+(((e>>>16)*i&65535)<<16)&4294967295,e=e<<15|e>>>17,e=(65535&e)*j+(((e>>>16)*j&65535)<<16)&4294967295,h^=e,h=h<<13|h>>>19,d=5*(65535&h)+((5*(h>>>16)&65535)<<16)&4294967295,h=(65535&d)+27492+(((d>>>16)+58964&65535)<<16);switch(e=0,f){case 3:e^=(255&a.charCodeAt(k+2))<<16;case 2:e^=(255&a.charCodeAt(k+1))<<8;case 1:e^=255&a.charCodeAt(k),e=(65535&e)*i+(((e>>>16)*i&65535)<<16)&4294967295,e=e<<15|e>>>17,e=(65535&e)*j+(((e>>>16)*j&65535)<<16)&4294967295,h^=e}return h^=a.length,h^=h>>>16,h=2246822507*(65535&h)+((2246822507*(h>>>16)&65535)<<16)&4294967295,h^=h>>>13,h=3266489909*(65535&h)+((3266489909*(h>>>16)&65535)<<16)&4294967295,h^=h>>>16,b?("0000000"+(h>>>0).toString(16)).substr(-8):h>>>0}function e(b){var c,e=a.map(b.getParentList(!1,!0),function(a){return a.refKey||a.key});return e=e.join("/"),c="id_"+d(e,!0)}a.ui.fancytree._FancytreeNodeClass.prototype.getCloneList=function(b){var c,d=this.tree,e=d.refMap[this.refKey]||null,f=d.keyMap;return e&&(c=this.key,b?e=a.map(e,function(a){return f[a]}):(e=a.map(e,function(a){return a===c?null:f[a]}),e.length<1&&(e=null))),e},a.ui.fancytree._FancytreeNodeClass.prototype.isClone=function(){var a=this.refKey||null,b=a&&this.tree.refMap[a]||null;return!!(b&&b.length>1)},a.ui.fancytree._FancytreeNodeClass.prototype.reRegister=function(b,c){b=null==b?null:""+b,c=null==c?null:""+c;var d=this.tree,e=this.key,f=this.refKey,g=d.keyMap,h=d.refMap,i=h[f]||null,j=!1;return null!=b&&b!==this.key&&(g[b]&&a.error("[ext-clones] reRegister("+b+"): already exists: "+this),delete g[e],g[b]=this,i&&(h[f]=a.map(i,function(a){return a===e?b:a})),this.key=b,j=!0),null!=c&&c!==this.refKey&&(i&&(1===i.length?delete h[f]:h[f]=a.map(i,function(a){return a===e?null:a})),h[c]?h[c].append(b):h[c]=[this.key],this.refKey=c,j=!0),j},a.ui.fancytree._FancytreeClass.prototype.getNodesByRef=function(b,c){var d=this.keyMap,e=this.refMap[b]||null;return e&&(e=c?a.map(e,function(a){var b=d[a];return b.isDescendantOf(c)?b:null}):a.map(e,function(a){return d[a]}),e.length<1&&(e=null)),e},a.ui.fancytree._FancytreeClass.prototype.changeRefKey=function(a,b){var c,d,e=this.keyMap,f=this.refMap[a]||null;if(f){for(c=0;c<f.length;c++)d=e[f[c]],d.refKey=b;delete this.refMap[a],this.refMap[b]=f}},a.ui.fancytree.registerExtension({name:"clones",version:"0.0.3",options:{highlightActiveClones:!0,highlightClones:!1},treeCreate:function(a){this._super(a),a.tree.refMap={},a.tree.keyMap={}},treeInit:function(a){this.$container.addClass("fancytree-ext-clones"),b(null==a.options.defaultKey),a.options.defaultKey=function(a){return e(a)},this._super(a)},treeClear:function(a){return a.tree.refMap={},a.tree.keyMap={},this._super(a)},treeRegisterNode:function(d,e,f){var g,h,i=d.tree,j=i.keyMap,k=i.refMap,l=f.key,m=f&&null!=f.refKey?""+f.refKey:null;return"_statusNode"===l?this._super(d,e,f):(e?(null!=j[f.key]&&a.error("clones.treeRegisterNode: node.key already exists: "+f),j[l]=f,m&&(g=k[m],g?(g.push(l),2===g.length&&d.options.clones.highlightClones&&j[g[0]].renderStatus()):k[m]=[l])):(null==j[l]&&a.error("clones.treeRegisterNode: node.key not registered: "+f.key),delete j[l],m&&(g=k[m],g&&(h=g.length,1>=h?(b(1===h),b(g[0]===l),delete k[m]):(c(g,l),2===h&&d.options.clones.highlightClones&&j[g[0]].renderStatus())))),this._super(d,e,f))},nodeRenderStatus:function(b){var c,d,e=b.node;return d=this._super(b),b.options.clones.highlightClones&&(c=a(e[b.tree.statusClassPropName]),c.length&&e.isClone()&&c.addClass("fancytree-clone")),d},nodeSetActive:function(b,c){var d,e=b.tree.statusClassPropName,f=b.node;return d=this._super(b,c),b.options.clones.highlightActiveClones&&f.isClone()&&a.each(f.getCloneList(!0),function(b,d){a(d[e]).toggleClass("fancytree-active-clone",c!==!1)}),d}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.dnd.min.js' */
!function(a,b,c,d){"use strict";function e(a){return 0===a?"":a>0?"+"+a:""+a}function f(b){var c=b.options.dnd||null;c&&g(),c&&c.dragStart&&b.widget.element.draggable(a.extend({addClasses:!1,appendTo:b.$container,containment:!1,delay:0,distance:4,revert:!1,scroll:!0,scrollSpeed:7,scrollSensitivity:10,connectToFancytree:!0,helper:function(b){var c,d=a.ui.fancytree.getNode(b.target),e=a(d.span);return d?(c=a("<div class='fancytree-drag-helper'><span class='fancytree-drag-helper-img' /></div>").css({zIndex:3,position:"relative"}).append(e.find("span.fancytree-title").clone()),c.data("ftSourceNode",d),c):"<div>ERROR?: helper requested but sourceNode not found</div>"},start:function(a,b){var c=b.helper.data("ftSourceNode");return!!c}},b.options.dnd.draggable)),c&&c.dragDrop&&b.widget.element.droppable(a.extend({addClasses:!1,tolerance:"intersect",greedy:!1},b.options.dnd.droppable))}function g(){h||(a.ui.plugin.add("draggable","connectToFancytree",{start:function(b,c){var d=a(this).data("ui-draggable")||a(this).data("draggable"),e=c.helper.data("ftSourceNode")||null;return e?(d.offset.click.top=-2,d.offset.click.left=16,e.tree.ext.dnd._onDragEvent("start",e,null,b,c,d)):void 0},drag:function(b,c){var d,e,f=a(this).data("ui-draggable")||a(this).data("draggable"),g=c.helper.data("ftSourceNode")||null,h=c.helper.data("ftTargetNode")||null,i=a.ui.fancytree.getNode(b.target);return b.target&&!i&&(d=a(b.target).closest("div.fancytree-drag-helper,#fancytree-drop-marker").length>0)?(e=g||h||a.ui.fancytree,void e.debug("Drag event over helper: ignored.")):(c.helper.data("ftTargetNode",i),h&&h!==i&&h.tree.ext.dnd._onDragEvent("leave",h,g,b,c,f),void(i&&i.tree.options.dnd.dragDrop&&(i===h?i.tree.ext.dnd._onDragEvent("over",i,g,b,c,f):i.tree.ext.dnd._onDragEvent("enter",i,g,b,c,f))))},stop:function(b,c){var d,e=a(this).data("ui-draggable")||a(this).data("draggable"),f=c.helper.data("ftSourceNode")||null,g=c.helper.data("ftTargetNode")||null,h=b.type,i="mouseup"===h&&1===b.which;i||(d=f||g||a.ui.fancytree,d.debug("Drag was cancelled")),g&&(i&&g.tree.ext.dnd._onDragEvent("drop",g,f,b,c,e),g.tree.ext.dnd._onDragEvent("leave",g,f,b,c,e)),f&&f.tree.ext.dnd._onDragEvent("stop",f,null,b,c,e)}}),h=!0)}var h=!1;a.ui.fancytree.registerExtension({name:"dnd",version:"0.1.0",options:{autoExpandMS:1e3,draggable:null,droppable:null,focusOnClick:!1,preventVoidMoves:!0,preventRecursiveMoves:!0,dragStart:null,dragStop:null,dragEnter:null,dragOver:null,dragDrop:null,dragLeave:null},treeInit:function(b){var c=b.tree;this._super(b),c.options.dnd.dragStart&&c.$container.on("mousedown",function(d){if(!c.hasFocus()&&b.options.dnd.focusOnClick){var e=a.ui.fancytree.getNode(d);e.debug("Re-enable focus that was prevented by jQuery UI draggable."),setTimeout(function(){a(d.target).closest(":tabbable").focus()},10)}}),f(c)},nodeKeydown:function(b){var c=b.originalEvent;return c.which===a.ui.keyCode.ESCAPE&&this._local._cancelDrag(),this._super(b)},nodeClick:function(a){return this._super(a)},_setDndStatus:function(b,c,d,f,g){var h,i=0,j="center",k=this._local,l=b?a(b.span):null,m=a(c.span);if(k.$dropMarker||(k.$dropMarker=a("<div id='fancytree-drop-marker'></div>").hide().css({"z-index":1e3}).prependTo(a(this.$div).parent())),"after"===f||"before"===f||"over"===f){switch(f){case"before":k.$dropMarker.removeClass("fancytree-drop-after fancytree-drop-over").addClass("fancytree-drop-before"),j="top";break;case"after":k.$dropMarker.removeClass("fancytree-drop-before fancytree-drop-over").addClass("fancytree-drop-after"),j="bottom";break;default:k.$dropMarker.removeClass("fancytree-drop-after fancytree-drop-before").addClass("fancytree-drop-over"),m.addClass("fancytree-drop-target"),i=8}h=a.ui.fancytree.jquerySupports.positionMyOfs?{my:"left"+e(i)+" center",at:"left "+j,of:m}:{my:"left center",at:"left "+j,of:m,offset:""+i+" 0"},k.$dropMarker.show().position(h)}else m.removeClass("fancytree-drop-target"),k.$dropMarker.hide();"after"===f?m.addClass("fancytree-drop-after"):m.removeClass("fancytree-drop-after"),"before"===f?m.addClass("fancytree-drop-before"):m.removeClass("fancytree-drop-before"),g===!0?(l&&l.addClass("fancytree-drop-accept"),m.addClass("fancytree-drop-accept"),d.addClass("fancytree-drop-accept")):(l&&l.removeClass("fancytree-drop-accept"),m.removeClass("fancytree-drop-accept"),d.removeClass("fancytree-drop-accept")),g===!1?(l&&l.addClass("fancytree-drop-reject"),m.addClass("fancytree-drop-reject"),d.addClass("fancytree-drop-reject")):(l&&l.removeClass("fancytree-drop-reject"),m.removeClass("fancytree-drop-reject"),d.removeClass("fancytree-drop-reject"))},_onDragEvent:function(b,c,e,f,g,h){"over"!==b&&this.debug("tree.ext.dnd._onDragEvent(%s, %o, %o) - %o",b,c,e,this);var i,j,k,l,m,n,o=this.options,p=o.dnd,q=this._makeHookContext(c,f,{otherNode:e,ui:g,draggable:h}),r=null,s=a(c.span);switch(b){case"start":c.isStatusNode()?r=!1:p.dragStart&&(r=p.dragStart(c,q)),r===!1?(this.debug("tree.dragStart() cancelled"),g.helper.trigger("mouseup").hide()):s.addClass("fancytree-drag-source");break;case"enter":n=p.preventRecursiveMoves&&c.isDescendantOf(e)?!1:p.dragEnter?p.dragEnter(c,q):null,r=n?a.isArray(n)?{over:a.inArray("over",n)>=0,before:a.inArray("before",n)>=0,after:a.inArray("after",n)>=0}:{over:n===!0||"over"===n,before:n===!0||"before"===n,after:n===!0||"after"===n}:!1,g.helper.data("enterResponse",r),this.debug("helper.enterResponse: %o",r);break;case"over":l=g.helper.data("enterResponse"),m=null,l===!1||("string"==typeof l?m=l:(i=s.offset(),j={x:f.pageX-i.left,y:f.pageY-i.top},k={x:j.x/s.width(),y:j.y/s.height()},l.after&&k.y>.75?m="after":!l.over&&l.after&&k.y>.5?m="after":l.before&&k.y<=.25?m="before":!l.over&&l.before&&k.y<=.5?m="before":l.over&&(m="over"),p.preventVoidMoves&&(c===e?(this.debug("    drop over source node prevented"),m=null):"before"===m&&e&&c===e.getNextSibling()?(this.debug("    drop after source node prevented"),m=null):"after"===m&&e&&c===e.getPrevSibling()?(this.debug("    drop before source node prevented"),m=null):"over"===m&&e&&e.parent===c&&e.isLastSibling()&&(this.debug("    drop last child over own parent prevented"),m=null)),g.helper.data("hitMode",m))),"over"===m&&p.autoExpandMS&&c.hasChildren()!==!1&&!c.expanded&&c.scheduleAction("expand",p.autoExpandMS),m&&p.dragOver&&(q.hitMode=m,r=p.dragOver(c,q)),this._local._setDndStatus(e,c,g.helper,m,r!==!1&&null!==m);break;case"drop":m=g.helper.data("hitMode"),m&&p.dragDrop&&(q.hitMode=m,p.dragDrop(c,q));break;case"leave":c.scheduleAction("cancel"),g.helper.data("enterResponse",null),g.helper.data("hitMode",null),this._local._setDndStatus(e,c,g.helper,"out",d),p.dragLeave&&p.dragLeave(c,q);break;case"stop":s.removeClass("fancytree-drag-source"),p.dragStop&&p.dragStop(c,q);break;default:a.error("Unsupported drag event: "+b)}return r},_cancelDrag:function(){var b=a.ui.ddmanager.current;b&&b.cancel()}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.edit.min.js' */
!function(a,b,c){"use strict";var d=/Mac/.test(navigator.platform),e=a.ui.fancytree.escapeHtml,f=a.ui.fancytree.unescapeHtml;a.ui.fancytree._FancytreeNodeClass.prototype.editStart=function(){var b,d=this,e=this.tree,g=e.ext.edit,h=e.options.edit,i=a(".fancytree-title",d.span),j={node:d,tree:e,options:e.options,isNew:a(d.span).hasClass("fancytree-edit-new"),orgTitle:d.title,input:null,dirty:!1};return h.beforeEdit.call(d,{type:"beforeEdit"},j)===!1?!1:(a.ui.fancytree.assert(!g.currentNode,"recursive edit"),g.currentNode=this,g.eventData=j,e.widget._unbind(),a(c).on("mousedown.fancytree-edit",function(b){a(b.target).hasClass("fancytree-edit-input")||d.editEnd(!0,b)}),b=a("<input />",{"class":"fancytree-edit-input",type:"text",value:f(j.orgTitle)}),g.eventData.input=b,null!=h.adjustWidthOfs&&b.width(i.width()+h.adjustWidthOfs),null!=h.inputCss&&b.css(h.inputCss),i.html(b),b.focus().change(function(){b.addClass("fancytree-edit-dirty")}).keydown(function(b){switch(b.which){case a.ui.keyCode.ESCAPE:d.editEnd(!1,b);break;case a.ui.keyCode.ENTER:return d.editEnd(!0,b),!1}b.stopPropagation()}).blur(function(a){return d.editEnd(!0,a)}),void h.edit.call(d,{type:"edit"},j))},a.ui.fancytree._FancytreeNodeClass.prototype.editEnd=function(b){var d,f=this,g=this.tree,h=g.ext.edit,i=h.eventData,j=g.options.edit,k=a(".fancytree-title",f.span),l=k.find("input.fancytree-edit-input");return j.trim&&l.val(a.trim(l.val())),d=l.val(),i.dirty=d!==f.title,i.save=b===!1?!1:i.isNew?""!==d:i.dirty&&""!==d,j.beforeClose.call(f,{type:"beforeClose"},i)===!1?!1:i.save&&j.save.call(f,{type:"save"},i)===!1?!1:(l.removeClass("fancytree-edit-dirty").unbind(),a(c).off(".fancytree-edit"),i.save?(f.setTitle(e(d)),f.setFocus()):i.isNew?(f.remove(),f=i.node=null,h.relatedNode.setFocus()):(f.renderTitle(),f.setFocus()),h.eventData=null,h.currentNode=null,h.relatedNode=null,g.widget._bind(),a(g.$container).focus(),i.input=null,j.close.call(f,{type:"close"},i),!0)},a.ui.fancytree._FancytreeNodeClass.prototype.editCreateNode=function(b,c){var d,e=this;return b=b||"child",null==c?c={title:""}:"string"==typeof c?c={title:c}:a.ui.fancytree.assert(a.isPlainObject(c)),"child"!==b||this.isExpanded()||this.hasChildren()===!1?(d=this.addNode(c,b),d.makeVisible(),a(d.span).addClass("fancytree-edit-new"),this.tree.ext.edit.relatedNode=this,void d.editStart()):void this.setExpanded().done(function(){e.editCreateNode(b,c)})},a.ui.fancytree._FancytreeClass.prototype.isEditing=function(){return this.ext.edit.currentNode},a.ui.fancytree._FancytreeNodeClass.prototype.isEditing=function(){return this.tree.ext.edit.currentNode===this},a.ui.fancytree.registerExtension({name:"edit",version:"0.2.0",options:{adjustWidthOfs:4,allowEmpty:!1,inputCss:{minWidth:"3em"},triggerCancel:["esc","tab","click"],triggerStart:["f2","shift+click","mac+enter"],trim:!0,beforeClose:a.noop,beforeEdit:a.noop,close:a.noop,edit:a.noop,save:a.noop},currentNode:null,treeInit:function(a){this._super(a),this.$container.addClass("fancytree-ext-edit")},nodeClick:function(b){return a.inArray("shift+click",b.options.edit.triggerStart)>=0&&b.originalEvent.shiftKey?(b.node.editStart(),!1):this._super(b)},nodeDblclick:function(b){return a.inArray("dblclick",b.options.edit.triggerStart)>=0?(b.node.editStart(),!1):this._super(b)},nodeKeydown:function(b){switch(b.originalEvent.which){case 113:if(a.inArray("f2",b.options.edit.triggerStart)>=0)return b.node.editStart(),!1;break;case a.ui.keyCode.ENTER:if(a.inArray("mac+enter",b.options.edit.triggerStart)>=0&&d)return b.node.editStart(),!1}return this._super(b)}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.filter.min.js' */
!function(a){"use strict";function b(a){return(a+"").replace(/([.?*+\^\$\[\]\\(){}|-])/g,"\\$1")}a.ui.fancytree._FancytreeClass.prototype._applyFilterImpl=function(a,c,d){var e,f,g=0,h="hide"===this.options.filter.mode;return d=!!d&&!c,"string"==typeof a&&(e=b(a),f=new RegExp(".*"+e+".*","i"),a=function(a){return!!f.exec(a.title)}),this.enableFilter=!0,this.lastFilterArgs=arguments,this.$div.addClass("fancytree-ext-filter"),this.$div.addClass(h?"fancytree-ext-filter-hide":"fancytree-ext-filter-dimm"),this.visit(function(a){delete a.match,delete a.subMatch}),this.visit(function(b){return d&&null!=b.children||!a(b)||(g++,b.match=!0,b.visitParents(function(a){a.subMatch=!0}),!c)?void 0:(b.visit(function(a){a.match=!0}),"skip")}),this.render(),g},a.ui.fancytree._FancytreeClass.prototype.filterNodes=function(a,b){return this._applyFilterImpl(a,!1,b)},a.ui.fancytree._FancytreeClass.prototype.applyFilter=function(){return this.warn("Fancytree.applyFilter() is deprecated since 2014-05-10. Use .filterNodes() instead."),this.filterNodes.apply(this,arguments)},a.ui.fancytree._FancytreeClass.prototype.filterBranches=function(a){return this._applyFilterImpl(a,!0,null)},a.ui.fancytree._FancytreeClass.prototype.clearFilter=function(){this.visit(function(a){delete a.match,delete a.subMatch}),this.enableFilter=!1,this.lastFilterArgs=null,this.$div.removeClass("fancytree-ext-filter fancytree-ext-filter-dimm fancytree-ext-filter-hide"),this.render()},a.ui.fancytree.registerExtension({name:"filter",version:"0.3.0",options:{autoApply:!0,mode:"dimm"},treeInit:function(a){this._super(a)},nodeLoadChildren:function(a,b){return this._super(a,b).done(function(){a.tree.enableFilter&&a.tree.lastFilterArgs&&a.options.filter.autoApply&&a.tree._applyFilterImpl.apply(a.tree,a.tree.lastFilterArgs)})},nodeRenderStatus:function(b){var c,d=b.node,e=b.tree,f=a(d[e.statusClassPropName]);return c=this._super(b),f.length&&e.enableFilter?(f.toggleClass("fancytree-match",!!d.match).toggleClass("fancytree-submatch",!!d.subMatch).toggleClass("fancytree-hide",!(d.match||d.subMatch)),c):c}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.glyph.min.js' */
!function(a){"use strict";function b(a,b){return a.map[b]}a.ui.fancytree.registerExtension({name:"glyph",version:"0.2.0",options:{map:{checkbox:"icon-check-empty",checkboxSelected:"icon-check",checkboxUnknown:"icon-check icon-muted",error:"icon-exclamation-sign",expanderClosed:"icon-caret-right",expanderLazy:"icon-angle-right",expanderOpen:"icon-caret-down",doc:"icon-file-alt",noExpander:"",docOpen:"icon-file-alt",loading:"icon-refresh icon-spin",folder:"icon-folder-close-alt",folderOpen:"icon-folder-open-alt"}},treeInit:function(a){var b=a.tree;this._super(a),b.$container.addClass("fancytree-ext-glyph")},nodeRenderStatus:function(c){var d,e,f=c.node,g=a(f.span),h=c.options.glyph,i=h.map;this._super(c),f.isRoot()||(e=g.children("span.fancytree-expander").get(0),e&&(d=f.isLoading()?"loading":f.expanded?"expanderOpen":f.isUndefined()?"expanderLazy":f.hasChildren()?"expanderClosed":"noExpander",e.className="fancytree-expander "+i[d]),e=f.tr?a("td",f.tr).children("span.fancytree-checkbox").get(0):g.children("span.fancytree-checkbox").get(0),e&&(d=f.selected?"checkboxSelected":f.partsel?"checkboxUnknown":"checkbox",e.className="fancytree-checkbox "+i[d]),e=g.children("span.fancytree-icon").get(0),e&&(d=f.folder?f.expanded?b(h,"folderOpen"):b(h,"folder"):f.expanded?b(h,"docOpen"):b(h,"doc"),e.className="fancytree-icon "+d))},nodeSetStatus:function(c,d,e,f){var g,h=c.options.glyph,i=c.node;this._super(c,d,e,f),g=i.parent?a("span.fancytree-expander",i.span).get(0):a(".fancytree-statusnode-wait, .fancytree-statusnode-error",i[this.nodeContainerAttrName]).find("span.fancytree-expander").get(0),"loading"===d?g.className="fancytree-expander "+b(h,"loading"):"error"===d&&(g.className="fancytree-expander "+b(h,"error"))}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.gridnav.min.js' */
!function(a){"use strict";function b(b,c){var d,e=c.get(0),f=0;return b.children().each(function(){return this===e?!1:(d=a(this).prop("colspan"),void(f+=d?d:1))}),f}function c(b,c){var d,e=null,f=0;return b.children().each(function(){return f>=c?(e=a(this),!1):(d=a(this).prop("colspan"),void(f+=d?d:1))}),e}function d(a,d){var f,g,h=a.closest("td"),i=null;switch(d){case e.LEFT:i=h.prev();break;case e.RIGHT:i=h.next();break;case e.UP:case e.DOWN:for(f=h.parent(),g=b(f,h);;){if(f=d===e.UP?f.prev():f.next(),!f.length)break;if(!f.is(":hidden")&&(i=c(f,g),i&&i.find(":input").length))break}}return i}var e=a.ui.keyCode,f={text:[e.UP,e.DOWN],checkbox:[e.UP,e.DOWN,e.LEFT,e.RIGHT],radiobutton:[e.UP,e.DOWN,e.LEFT,e.RIGHT],"select-one":[e.LEFT,e.RIGHT],"select-multiple":[e.LEFT,e.RIGHT]};a.ui.fancytree.registerExtension({name:"gridnav",version:"0.0.1",options:{autofocusInput:!1,handleCursorKeys:!0},treeInit:function(b){this._requireExtension("table",!0,!0),this._super(b),this.$container.addClass("fancytree-ext-gridnav"),this.$container.on("focusin",function(c){var d,e=a.ui.fancytree.getNode(c.target);e&&!e.isActive()&&(d=b.tree._makeHookContext(e,c),b.tree._callHook("nodeSetActive",d,!0))})},nodeSetActive:function(b,c){var d,e=b.options.gridnav,f=b.node,g=b.originalEvent||{},h=a(g.target).is(":input");c=c!==!1,this._super(b,c),c&&(b.options.titlesTabbable?(h||(a(f.span).find("span.fancytree-title").focus(),f.setFocus()),b.tree.$container.attr("tabindex","-1")):e.autofocusInput&&!h&&(d=a(f.tr||f.span),d.find(":input:enabled:first").focus()))},nodeKeydown:function(b){var c,e,g,h=b.options.gridnav,i=b.originalEvent,j=a(i.target);return c=j.is(":input:enabled")?j.prop("type"):null,c&&h.handleCursorKeys?(e=f[c],e&&a.inArray(i.which,e)>=0&&(g=d(j,i.which),g&&g.length)?(g.find(":input:enabled").focus(),!1):!0):(b.tree.debug("ext-gridnav NOT HANDLED",i,c),this._super(b))}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.persist.min.js' */
!function(a,b,c,d){"use strict";function e(b,c,d,f,g){var i,j,k,l,m=!1,n=[],o=[];for(d=d||[],g=g||a.Deferred(),i=0,k=d.length;k>i;i++)j=d[i],l=b.getNodeByKey(j),l?f&&l.isUndefined()?(m=!0,b.debug("_loadLazyNodes: "+l+" is lazy: loading..."),n.push("expand"===f?l.setExpanded():l.load())):(b.debug("_loadLazyNodes: "+l+" already loaded."),l.setExpanded()):(o.push(j),b.debug("_loadLazyNodes: "+l+" was not yet found."));return a.when.apply(a,n).always(function(){if(m&&o.length>0)e(b,c,o,f,g);else{if(o.length)for(b.warn("_loadLazyNodes: could not load those keys: ",o),i=0,k=o.length;k>i;i++)j=d[i],c._appendKey(h,d[i],!1);g.resolve()}}),g}var f=a.ui.fancytree.assert,g="active",h="expanded",i="focus",j="selected";a.ui.fancytree._FancytreeClass.prototype.clearCookies=function(a){var b=this.ext.persist,c=b.cookiePrefix;a=a||"active expanded focus selected",a.indexOf(g)>=0&&b._data(c+g,null),a.indexOf(h)>=0&&b._data(c+h,null),a.indexOf(i)>=0&&b._data(c+i,null),a.indexOf(j)>=0&&b._data(c+j,null)},a.ui.fancytree._FancytreeClass.prototype.getPersistData=function(){var a=this.ext.persist,b=a.cookiePrefix,c=a.cookieDelimiter,d={};return d[g]=a._data(b+g),d[h]=(a._data(b+h)||"").split(c),d[j]=(a._data(b+j)||"").split(c),d[i]=a._data(b+i),d},a.ui.fancytree.registerExtension({name:"persist",version:"0.3.0",options:{cookieDelimiter:"~",cookiePrefix:d,cookie:{raw:!1,expires:"",path:"",domain:"",secure:!1},expandLazy:!1,overrideSource:!0,store:"auto",types:"active expanded focus selected"},_data:function(b,c){var e=this._local.localStorage;return c===d?e?e.getItem(b):a.cookie(b):void(null===c?e?e.removeItem(b):a.removeCookie(b):e?e.setItem(b,c):a.cookie(b,c,this.options.persist.cookie))},_appendKey:function(b,c,d){c=""+c;var e=this._local,f=this.options.persist,g=f.cookieDelimiter,h=e.cookiePrefix+b,i=e._data(h),j=i?i.split(g):[],k=a.inArray(c,j);k>=0&&j.splice(k,1),d&&j.push(c),e._data(h,j.join(g))},treeInit:function(c){var k=c.tree,l=c.options,m=this._local,n=this.options.persist;return f("localStore"===n.store||a.cookie,"Missing required plugin for 'persist' extension: jquery.cookie.js"),m.cookiePrefix=n.cookiePrefix||"fancytree-"+k._id+"-",m.storeActive=n.types.indexOf(g)>=0,m.storeExpanded=n.types.indexOf(h)>=0,m.storeSelected=n.types.indexOf(j)>=0,m.storeFocus=n.types.indexOf(i)>=0,m.localStorage="cookie"!==n.store&&b.localStorage?"local"===n.store?b.localStorage:b.sessionStorage:null,k.$div.bind("fancytreeinit",function(){var b,c,f,o,p,q=a.cookie(m.cookiePrefix+i);b=m._data(m.cookiePrefix+h),o=b&&b.split(n.cookieDelimiter),c=m.storeExpanded?e(k,m,o,n.expandLazy?"expand":!1,null):(new a.Deferred).resolve(),c.done(function(){if(m.storeSelected){if(b=m._data(m.cookiePrefix+j))for(o=b.split(n.cookieDelimiter),f=0;f<o.length;f++)p=k.getNodeByKey(o[f]),p?(p.selected===d||n.overrideSource&&p.selected===!1)&&(p.selected=!0,p.renderStatus()):m._appendKey(j,o[f],!1);3===k.options.selectMode&&k.visit(function(a){return a.selected?(a.fixSelection3AfterClick(),"skip"):void 0})}m.storeActive&&(b=m._data(m.cookiePrefix+g),!b||!l.persist.overrideSource&&k.activeNode||(p=k.getNodeByKey(b),p&&p.setActive())),m.storeFocus&&q&&(p=k.getNodeByKey(q),p&&p.setFocus()),k._triggerTreeEvent("restore",null,{})})}),this._super(c)},nodeSetActive:function(a,b,c){var d,e=this._local;return b=b!==!1,d=this._super(a,b,c),e.storeActive&&e._data(e.cookiePrefix+g,this.activeNode?this.activeNode.key:null),d},nodeSetExpanded:function(a,b,c){var d,e=a.node,f=this._local;return b=b!==!1,d=this._super(a,b,c),f.storeExpanded&&f._appendKey(h,e.key,b),d},nodeSetFocus:function(a,b){var c,d=this._local;return b=b!==!1,c=this._super(a,b),b&&d.storeFocus&&d._data(d.cookiePrefix+i,this.focusNode?this.focusNode.key:null),c},nodeSetSelected:function(b,c){var d,e,f=b.tree,g=b.node,h=this._local;return c=c!==!1,d=this._super(b,c),h.storeSelected&&(3===f.options.selectMode?(e=a.map(f.getSelectedNodes(!0),function(a){return a.key}),e=e.join(b.options.persist.cookieDelimiter),h._data(h.cookiePrefix+j,e)):h._appendKey(j,g.key,c)),d}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.table.min.js' */
!function(a,b,c){"use strict";function d(b,c){c=c||"",b||a.error("Assertion failed "+c)}function e(a,b){a.parentNode.insertBefore(b,a.nextSibling)}function f(a,b){a.visit(function(a){var c=a.tr;return c&&(c.style.display=a.hide||!b?"none":""),a.expanded?void 0:"skip"})}function g(b){var c,e,f,g=b.parent,h=g?g.children:null;if(h&&h.length>1&&h[0]!==b)for(c=a.inArray(b,h),f=h[c-1],d(f.tr);f.children&&(e=f.children[f.children.length-1],e.tr);)f=e;else f=g;return f}a.ui.fancytree.registerExtension({name:"table",version:"0.2.0",options:{checkboxColumnIdx:null,customStatus:!1,indentation:16,nodeColumnIdx:0},treeInit:function(b){var d,e,f,g=b.tree,h=g.widget.element;for(h.addClass("fancytree-container fancytree-ext-table"),g.tbody=h.find("> tbody")[0],g.columnCount=a("thead >tr >th",h).length,a(g.tbody).empty(),g.rowFragment=c.createDocumentFragment(),e=a("<tr />"),f="",b.options.aria&&(e.attr("role","row"),f=" role='gridcell'"),d=0;d<g.columnCount;d++)e.append(b.options.table.nodeColumnIdx===d?"<td"+f+"><span class='fancytree-node' /></td>":"<td"+f+" />");g.rowFragment.appendChild(e.get(0)),g.statusClassPropName="tr",g.ariaPropName="tr",this.nodeContainerAttrName="tr",this._super(b),a(g.rootNode.ul).remove(),g.rootNode.ul=null,g.$container=h,this.$container.attr("tabindex",this.options.tabbable?"0":"-1"),this.options.aria&&g.$container.attr("role","treegrid").attr("aria-readonly",!0)},nodeRemoveChildMarkup:function(b){var c=b.node;c.visit(function(b){b.tr&&(a(b.tr).remove(),b.tr=null)})},nodeRemoveMarkup:function(b){var c=b.node;c.tr&&(a(c.tr).remove(),c.tr=null),this.nodeRemoveChildMarkup(b)},nodeRender:function(b,c,h,i,j){var k,l,m,n,o,p,q,r,s=b.tree,t=b.node,u=b.options,v=!t.parent;if(j||(b.hasCollapsedParents=t.parent&&!t.parent.expanded),!v)if(t.tr)c?this.nodeRenderTitle(b):this.nodeRenderStatus(b);else{if(b.hasCollapsedParents)return void t.debug("nodeRender ignored due to unrendered parent");o=s.rowFragment.firstChild.cloneNode(!0),p=g(t),d(p),i===!0&&j?o.style.display="none":h&&b.hasCollapsedParents&&(o.style.display="none"),p.tr?e(p.tr,o):(d(!p.parent,"prev. row must have a tr, or is system root"),s.tbody.appendChild(o)),t.tr=o,t.key&&u.generateIds&&(t.tr.id=u.idPrefix+t.key),t.tr.ftnode=t,u.aria&&a(t.tr).attr("aria-labelledby","ftal_"+t.key),t.span=a("span.fancytree-node",t.tr).get(0),this.nodeRenderTitle(b),u.createNode&&u.createNode.call(s,{type:"createNode"},b)}if(u.renderNode&&u.renderNode.call(s,{type:"renderNode"},b),k=t.children,k&&(v||h||t.expanded))for(m=0,n=k.length;n>m;m++)r=a.extend({},b,{node:k[m]}),r.hasCollapsedParents=r.hasCollapsedParents||!t.expanded,this.nodeRender(r,c,h,i,!0);k&&!j&&(q=t.tr||null,l=s.tbody.firstChild,t.visit(function(a){if(a.tr){if(a.parent.expanded||"none"===a.tr.style.display||(a.tr.style.display="none",f(a,!1)),a.tr.previousSibling!==q){t.debug("_fixOrder: mismatch at node: "+a);var b=q?q.nextSibling:l;s.tbody.insertBefore(a.tr,b)}q=a.tr}}))},nodeRenderTitle:function(b){var c,d=b.node,e=b.options;this._super(b),e.checkbox&&null!=e.table.checkboxColumnIdx&&(c=a("span.fancytree-checkbox",d.span).detach(),a(d.tr).find("td:first").html(c)),d.isRoot()||this.nodeRenderStatus(b),!e.table.customStatus&&d.isStatusNode()||e.renderColumns&&e.renderColumns.call(b.tree,{type:"renderColumns"},b)},nodeRenderStatus:function(b){var c,d=b.node,e=b.options;this._super(b),a(d.tr).removeClass("fancytree-node"),c=(d.getLevel()-1)*e.table.indentation,a(d.span).css({marginLeft:c+"px"})},nodeSetExpanded:function(b,c,d){function e(a){c=c!==!1,f(b.node,c),a?c&&b.options.autoScroll&&!d.noAnimation&&b.node.hasChildren()?b.node.getLastChild().scrollIntoView(!0,{topNode:b.node}).always(function(){d.noEvents||b.tree._triggerNodeEvent(c?"expand":"collapse",b),g.resolveWith(b.node)}):(d.noEvents||b.tree._triggerNodeEvent(c?"expand":"collapse",b),g.resolveWith(b.node)):(d.noEvents||b.tree._triggerNodeEvent(c?"expand":"collapse",b),g.rejectWith(b.node))}var g=new a.Deferred,h=a.extend({},d,{noEvents:!0,noAnimation:!0});return d=d||{},this._super(b,c,h).done(function(){e(!0)}).fail(function(){e(!1)}),g.promise()},nodeSetStatus:function(b,c,d,e){if("ok"===c){var f=b.node,g=f.children?f.children[0]:null;g&&g.isStatusNode()&&a(g.tr).remove()}this._super(b,c,d,e)},treeClear:function(a){return this.nodeRemoveChildMarkup(this._makeHookContext(this.rootNode)),this._super(a)}})}(jQuery,window,document);

/*! Extension 'jquery.fancytree.themeroller.min.js' */
!function(a){"use strict";a.ui.fancytree.registerExtension({name:"themeroller",version:"0.0.1",options:{activeClass:"ui-state-active",foccusClass:"ui-state-focus",hoverClass:"ui-state-hover",selectedClass:"ui-state-highlight"},treeInit:function(b){this._super(b);var c=b.widget.element;"TABLE"===c[0].nodeName?(c.addClass("ui-widget ui-corner-all"),c.find(">thead tr").addClass("ui-widget-header"),c.find(">tbody").addClass("ui-widget-conent")):c.addClass("ui-widget ui-widget-content ui-corner-all"),c.delegate(".fancytree-node","mouseenter mouseleave",function(b){var c=a.ui.fancytree.getNode(b.target),d="mouseenter"===b.type;c.debug("hover: "+d),a(c.span).toggleClass("ui-state-hover ui-corner-all",d)})},treeDestroy:function(a){this._super(a),a.widget.element.removeClass("ui-widget ui-widget-content ui-corner-all")},nodeRenderStatus:function(b){var c=b.node,d=a(c.span);this._super(b),d.toggleClass("ui-state-active",c.isActive()),d.toggleClass("ui-state-focus",c.hasFocus()),d.toggleClass("ui-state-highlight",c.isSelected())}})}(jQuery,window,document);
}));
